/* Generated by Cython 0.29.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "include/busclique/bundle_cache.hpp",
            "include/busclique/cell_cache.hpp",
            "include/busclique/clique_cache.hpp",
            "include/busclique/coordinate_types.hpp",
            "include/busclique/find_biclique.hpp",
            "include/busclique/find_clique.hpp",
            "include/busclique/topo_cache.hpp",
            "include/busclique/util.hpp"
        ],
        "include_dirs": [
            "minorminer",
            "",
            "./include/",
            ".include/busclique"
        ],
        "language": "c++",
        "name": "minorminer.busclique",
        "sources": [
            "./minorminer/busclique.pyx"
        ]
    },
    "module_name": "minorminer.busclique"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__minorminer__busclique
#define __PYX_HAVE_API__minorminer__busclique
/* Early includes */
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <vector>
#include <utility>

    #if __cplusplus > 199711L
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <stdint.h>
#include "../include/busclique/util.hpp"
#include "../include/busclique/cell_cache.hpp"
#include "../include/busclique/bundle_cache.hpp"
#include "../include/busclique/clique_cache.hpp"
#include "../include/busclique/topo_cache.hpp"
#include "../include/busclique/find_clique.hpp"
#include "../include/busclique/find_biclique.hpp"
#include "../include/busclique/coordinate_types.hpp"
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "minorminer/busclique.pyx",
  "stringsource",
  "env/lib/python3.9/site-packages/Cython/Includes/cpython/type.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph;
struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph;
struct __pyx_obj_10minorminer_9busclique__chimera_busgraph;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding;
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;

/* "minorminer/busclique_h.pxi":18
 * from libcpp.pair cimport pair
 * from libc.stdint cimport uint8_t, uint32_t, uint64_t
 * ctypedef vector[size_t] nodes_t             # <<<<<<<<<<<<<<
 * ctypedef vector[vector[size_t]] embedding_t
 * ctypedef vector[pair[size_t,size_t]] edges_t
 */
typedef std::vector<size_t>  __pyx_t_10minorminer_9busclique_nodes_t;

/* "minorminer/busclique_h.pxi":19
 * from libc.stdint cimport uint8_t, uint32_t, uint64_t
 * ctypedef vector[size_t] nodes_t
 * ctypedef vector[vector[size_t]] embedding_t             # <<<<<<<<<<<<<<
 * ctypedef vector[pair[size_t,size_t]] edges_t
 * 
 */
typedef std::vector<std::vector<size_t> >  __pyx_t_10minorminer_9busclique_embedding_t;

/* "minorminer/busclique_h.pxi":20
 * ctypedef vector[size_t] nodes_t
 * ctypedef vector[vector[size_t]] embedding_t
 * ctypedef vector[pair[size_t,size_t]] edges_t             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "../include/busclique/util.hpp" namespace "busclique":
 */
typedef std::vector<std::pair<size_t,size_t> >  __pyx_t_10minorminer_9busclique_edges_t;
struct __pyx_defaults {
  PyObject *__pyx_arg_version;
};

/* "minorminer/busclique.pyx":793
 *     return ret, shortcode
 * 
 * cdef class _zephyr_busgraph:             # <<<<<<<<<<<<<<
 *     cdef topo_cache[zephyr_spec] *topo
 *     cdef nodes_t nodes
 */
struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph {
  PyObject_HEAD
  busclique::topo_cache<busclique::zephyr_spec>  *topo;
  __pyx_t_10minorminer_9busclique_nodes_t nodes;
  __pyx_t_10minorminer_9busclique_embedding_t emb_1;
  PyObject *relabel;
  PyObject *delabel;
  PyObject *identifier;
  PyObject *short_identifier;
};


/* "minorminer/busclique.pyx":904
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * cdef class _pegasus_busgraph:             # <<<<<<<<<<<<<<
 *     cdef topo_cache[pegasus_spec] *topo
 *     cdef nodes_t nodes
 */
struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph {
  PyObject_HEAD
  busclique::topo_cache<busclique::pegasus_spec>  *topo;
  __pyx_t_10minorminer_9busclique_nodes_t nodes;
  __pyx_t_10minorminer_9busclique_embedding_t emb_1;
  PyObject *relabel;
  PyObject *delabel;
  PyObject *identifier;
  PyObject *short_identifier;
};


/* "minorminer/busclique.pyx":1017
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * cdef class _chimera_busgraph:             # <<<<<<<<<<<<<<
 *     """Class for managing a single Chimera graph, and dispatches various
 *     structure-aware C++ embedding functions on it.
 */
struct __pyx_obj_10minorminer_9busclique__chimera_busgraph {
  PyObject_HEAD
  busclique::topo_cache<busclique::chimera_spec>  *topo;
  __pyx_t_10minorminer_9busclique_embedding_t emb_1;
  __pyx_t_10minorminer_9busclique_nodes_t nodes;
  PyObject *relabel;
  PyObject *delabel;
  PyObject *identifier;
  PyObject *short_identifier;
};


/* "minorminer/busclique.pyx":254
 *         return cache
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):             # <<<<<<<<<<<<<<
 *         """Replaces or sets the cliques stored by this cache with the provided
 *         cache object.
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "minorminer/busclique.pyx":287
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):             # <<<<<<<<<<<<<<
 *         """Inserts a clique embedding into the cache.
 * 
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "minorminer/busclique.pyx":341
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):             # <<<<<<<<<<<<<<
 *         """Merges the clique cache of another ``busgraph_cache`` object
 * 
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "minorminer/busclique.pyx":752
 *     return {v: tuple(chain) for v, chain in emb.items()}
 * 
 * def _make_relabeler(f):             # <<<<<<<<<<<<<<
 *     """
 *     This returns an embedding-relabeling function, which applies `f` to each
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler {
  PyObject_HEAD
  PyObject *__pyx_v_f;
};


/* "minorminer/busclique.pyx":1135
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * def _default_quality_function(emb):             # <<<<<<<<<<<<<<
 *     """A function that returns a tuple corresponds to the "quality" of the embedding.
 * 
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function {
  PyObject_HEAD
  PyObject *__pyx_v_emb;
  PyObject *__pyx_v_maxlen;
};


/* "minorminer/busclique.pyx":1153
 *     """
 *     maxlen = max(map(len, emb))
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)             # <<<<<<<<<<<<<<
 * 
 * def _regularize_embedding(g, emb):
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *__pyx_outer_scope;
  PyObject *__pyx_v_chain;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "minorminer/busclique.pyx":1155
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 * def _regularize_embedding(g, emb):             # <<<<<<<<<<<<<<
 *     """Best-effort augmented embedding, with hopefully uniform chainlegnths.
 * 
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding {
  PyObject_HEAD
  PyObject *__pyx_v_boundary;
  PyObject *__pyx_v_chain_node;
};


/* "minorminer/busclique.pyx":1187
 *                 chain_node = (i, 'chain', j)
 *                 top_nodes.append(chain_node)
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)             # <<<<<<<<<<<<<<
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 *         for chain_node in top_nodes:
 */
struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *__pyx_outer_scope;
  PyObject *__pyx_v_q;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_FloorDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceFloorDivide(op1, op2) : PyNumber_FloorDivide(op1, op2))
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_TrueDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'libcpp.utility' */

/* Module declarations from 'libcpp.pair' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'minorminer.busclique' */
static PyTypeObject *__pyx_ptype_10minorminer_9busclique__zephyr_busgraph = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique__pegasus_busgraph = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique__chimera_busgraph = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct__update_clique_cache = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_5_genexpr = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding = 0;
static PyTypeObject *__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_7_genexpr = 0;
static int __pyx_v_10minorminer_9busclique___cache_version;
static int __pyx_v_10minorminer_9busclique___lru_size;
static PyObject *__pyx_v_10minorminer_9busclique___global_locks = 0;
static PyObject *__pyx_f_10minorminer_9busclique__make_clique_cache(std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  &); /*proto*/
static PyObject *__pyx_f_10minorminer_9busclique__keep_biclique_key(PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_10minorminer_9busclique__make_biclique_cache(std::vector<std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t> >  &); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_10minorminer_9busclique__serialize(busclique::topo_cache<busclique::zephyr_spec>  *, PyObject *); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_10minorminer_9busclique__serialize(busclique::topo_cache<busclique::pegasus_spec>  *, PyObject *); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_10minorminer_9busclique__serialize(busclique::topo_cache<busclique::chimera_spec>  *, PyObject *); /*proto*/
static std::vector<size_t>  __pyx_convert_vector_from_py_size_t(PyObject *); /*proto*/
static std::vector<std::vector<size_t> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(PyObject *); /*proto*/
static std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(PyObject *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_size_t(const std::vector<size_t>  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(const std::vector<std::vector<size_t> >  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t(const std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  &); /*proto*/
static PyObject *__pyx_convert_pair_to_py_size_t____size_t(std::pair<size_t,size_t>  const &); /*proto*/
static PyObject *__pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t(std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t>  const &); /*proto*/
static std::pair<size_t,size_t>  __pyx_convert_pair_from_py_size_t__and_size_t(PyObject *); /*proto*/
static std::vector<std::pair<size_t,size_t> >  __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(PyObject *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(const std::vector<std::pair<size_t,size_t> >  &); /*proto*/
static std::vector<uint8_t>  __pyx_convert_vector_from_py_uint8_t(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "minorminer.busclique"
extern int __pyx_module_is_main_minorminer__busclique;
int __pyx_module_is_main_minorminer__busclique = 0;

/* Implementation of 'minorminer.busclique' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_RuntimeError;
static const char __pyx_k_M[] = "M";
static const char __pyx_k_N[] = "N";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_hm[] = "hm";
static const char __pyx_k_mm[] = "mm";
static const char __pyx_k_nn[] = "nn";
static const char __pyx_k_nx[] = "nx";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_s0[] = "s0";
static const char __pyx_k_s1[] = "s1";
static const char __pyx_k_wb[] = "wb";
static const char __pyx_k_16x[] = "{}_{:16x}";
static const char __pyx_k_LRU[] = "LRU";
static const char __pyx_k__21[] = "~";
static const char __pyx_k_bgc[] = "bgc";
static const char __pyx_k_dnx[] = "dnx";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_emb[] = "emb";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_len[] = "len";
static const char __pyx_k_lru[] = ".lru";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_num[] = "num";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_raw[] = "raw";
static const char __pyx_k_row[] = "row";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_top[] = "top";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_Lock[] = "Lock";
static const char __pyx_k_Path[] = "Path";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_cols[] = "cols";
static const char __pyx_k_dump[] = "dump";
static const char __pyx_k_emb0[] = "emb0";
static const char __pyx_k_emb1[] = "emb1";
static const char __pyx_k_embs[] = "embs";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_gzip[] = "gzip";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_item[] = "item";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_json[] = "json";
static const char __pyx_k_key0[] = "key0";
static const char __pyx_k_key1[] = "key1";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_load[] = "load";
static const char __pyx_k_lock[] = ".lock";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mmax[] = "mmax";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_nice[] = "nice";
static const char __pyx_k_nmax[] = "nmax";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_raw0[] = "raw0";
static const char __pyx_k_raw1[] = "raw1";
static const char __pyx_k_rows[] = "rows";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tile[] = "tile";
static const char __pyx_k_used[] = "used";
static const char __pyx_k_utf8[] = "utf8";
static const char __pyx_k_Graph[] = "Graph";
static const char __pyx_k_bgc_i[] = "bgc_i";
static const char __pyx_k_blank[] = "blank";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_chain[] = "chain";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_dwave[] = "dwave";
static const char __pyx_k_edges[] = "edges";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_f_emb[] = "f_emb";
static const char __pyx_k_files[] = "files";
static const char __pyx_k_force[] = "force";
static const char __pyx_k_graph[] = "graph";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_loads[] = "loads";
static const char __pyx_k_lru_2[] = "lru";
static const char __pyx_k_miner[] = "miner";
static const char __pyx_k_mkdir[] = "mkdir";
static const char __pyx_k_nodes[] = "nodes";
static const char __pyx_k_qual0[] = "qual0";
static const char __pyx_k_qual1[] = "qual1";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_rmdir[] = "rmdir";
static const char __pyx_k_s_min[] = "s_min";
static const char __pyx_k_size0[] = "size0";
static const char __pyx_k_size1[] = "size1";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_tries[] = "tries";
static const char __pyx_k_union[] = "union";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_clique[] = "clique";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_direct[] = "direct";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_exists[] = "exists";
static const char __pyx_k_family[] = "family";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_is_dir[] = "is_dir";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_labels[] = "labels";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_maxlen[] = "maxlen";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_newLRU[] = "newLRU";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_oldkey[] = "oldkey";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_sparse[] = "sparse";
static const char __pyx_k_unlink[] = "unlink";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_zephyr[] = "zephyr";
static const char __pyx_k_basedir[] = "basedir";
static const char __pyx_k_biggest[] = "biggest";
static const char __pyx_k_by_size[] = "by_size";
static const char __pyx_k_chimera[] = "chimera";
static const char __pyx_k_cliques[] = "_cliques";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_compute[] = "compute";
static const char __pyx_k_delabel[] = "delabel";
static const char __pyx_k_dirname[] = "dirname";
static const char __pyx_k_discard[] = "discard";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_graph_2[] = "_graph";
static const char __pyx_k_iterdir[] = "iterdir";
static const char __pyx_k_length0[] = "length0";
static const char __pyx_k_length1[] = "length1";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_lrufile[] = "lrufile";
static const char __pyx_k_matched[] = "matched";
static const char __pyx_k_no_seed[] = "_no_seed";
static const char __pyx_k_parents[] = "parents";
static const char __pyx_k_pathlib[] = "pathlib";
static const char __pyx_k_pegasus[] = "pegasus";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_randint[] = "randint";
static const char __pyx_k_raw_emb[] = "raw_emb";
static const char __pyx_k_relabel[] = "relabel";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_rootdir[] = "rootdir";
static const char __pyx_k_urandom[] = "urandom";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_biclique[] = "biclique";
static const char __pyx_k_boundary[] = "boundary";
static const char __pyx_k_busgraph[] = "busgraph";
static const char __pyx_k_compress[] = "compress";
static const char __pyx_k_dirstack[] = "dirstack";
static const char __pyx_k_emb_list[] = "emb_list";
static const char __pyx_k_exist_ok[] = "exist_ok";
static const char __pyx_k_family_2[] = "_family";
static const char __pyx_k_file_obj[] = "file_obj";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_homebase[] = "homebase";
static const char __pyx_k_lockfile[] = "lockfile";
static const char __pyx_k_matching[] = "matching";
static const char __pyx_k_max_beta[] = "max_beta";
static const char __pyx_k_max_side[] = "max_side";
static const char __pyx_k_networkx[] = "networkx";
static const char __pyx_k_progress[] = "progress";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_raw_emb0[] = "raw_emb0";
static const char __pyx_k_raw_emb1[] = "raw_emb1";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_substack[] = "substack";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_bicliques[] = "_bicliques";
static const char __pyx_k_bipartite[] = "bipartite";
static const char __pyx_k_busclique[] = "busclique";
static const char __pyx_k_by_length[] = "by_length";
static const char __pyx_k_cachefile[] = "cachefile";
static const char __pyx_k_cliques_2[] = "cliques";
static const char __pyx_k_currcache[] = "currcache";
static const char __pyx_k_edge_list[] = "edge_list";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_fasteners[] = "fasteners";
static const char __pyx_k_file_name[] = "file_name";
static const char __pyx_k_filecache[] = "filecache";
static const char __pyx_k_getLogger[] = "getLogger";
static const char __pyx_k_graphdata[] = "graphdata";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_node_list[] = "node_list";
static const char __pyx_k_num_nodes[] = "_num_nodes";
static const char __pyx_k_num_seeds[] = "num_seeds";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_relabeler[] = "_relabeler";
static const char __pyx_k_remainder[] = "remainder";
static const char __pyx_k_shortcode[] = "shortcode";
static const char __pyx_k_temp_name[] = "temp_name";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_top_nodes[] = "top_nodes";
static const char __pyx_k_use_cache[] = "use_cache";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_chain_node[] = "chain_node";
static const char __pyx_k_coordinate[] = "coordinate";
static const char __pyx_k_decompress[] = "decompress";
static const char __pyx_k_graphclass[] = "graphclass";
static const char __pyx_k_identifier[] = "identifier";
static const char __pyx_k_mask_bound[] = "mask_bound";
static const char __pyx_k_max_length[] = "max_length";
static const char __pyx_k_minorminer[] = "minorminer";
static const char __pyx_k_regularize[] = "regularize";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_bicliques_2[] = "bicliques";
static const char __pyx_k_chainlength[] = "chainlength";
static const char __pyx_k_fetch_cache[] = "_fetch_cache";
static const char __pyx_k_force_write[] = "force_write";
static const char __pyx_k_zip_longest[] = "zip_longest";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_file_context[] = "file_context";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_cache_rootdir[] = "cache_rootdir";
static const char __pyx_k_chimera_graph[] = "chimera_graph";
static const char __pyx_k_node_boundary[] = "node_boundary";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_user_data_dir[] = "user_data_dir";
static const char __pyx_k_write_to_disk[] = "write_to_disk";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_add_edges_from[] = "add_edges_from";
static const char __pyx_k_busgraph_cache[] = "busgraph_cache";
static const char __pyx_k_complete_graph[] = "complete_graph";
static const char __pyx_k_dwave_networkx[] = "dwave_networkx";
static const char __pyx_k_fragment_nodes[] = "fragment_nodes";
static const char __pyx_k_heuristic_args[] = "heuristic_args";
static const char __pyx_k_heuristic_runs[] = "heuristic_runs";
static const char __pyx_k_largest_clique[] = "largest_clique";
static const char __pyx_k_make_relabeler[] = "_make_relabeler";
static const char __pyx_k_matching_graph[] = "matching_graph";
static const char __pyx_k_set_mask_bound[] = "set_mask_bound";
static const char __pyx_k_find_embeddings[] = "find_embeddings";
static const char __pyx_k_heuristic_bound[] = "heuristic_bound";
static const char __pyx_k_independent_set[] = "independent_set";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_zephyr_busgraph[] = "_zephyr_busgraph";
static const char __pyx_k_InterProcessLock[] = "InterProcessLock";
static const char __pyx_k_chimera_busgraph[] = "_chimera_busgraph";
static const char __pyx_k_clear_all_caches[] = "clear_all_caches";
static const char __pyx_k_get_clique_cache[] = "get_clique_cache";
static const char __pyx_k_maximum_matching[] = "maximum_matching";
static const char __pyx_k_pegasus_busgraph[] = "_pegasus_busgraph";
static const char __pyx_k_quality_function[] = "quality_function";
static const char __pyx_k_short_identifier[] = "short_identifier";
static const char __pyx_k_vertical_offsets[] = "vertical_offsets";
static const char __pyx_k_FileNotFoundError[] = "FileNotFoundError";
static const char __pyx_k_trivial_relabeler[] = "_trivial_relabeler";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_compute_identifier[] = "compute_identifier";
static const char __pyx_k_horizontal_offsets[] = "horizontal_offsets";
static const char __pyx_k_merge_clique_cache[] = "merge_clique_cache";
static const char __pyx_k_zephyr_coordinates[] = "zephyr_coordinates";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_chimera_coordinates[] = "chimera_coordinates";
static const char __pyx_k_ensure_clique_cache[] = "_ensure_clique_cache";
static const char __pyx_k_fragment_graph_spec[] = "fragment_graph_spec";
static const char __pyx_k_iter_linear_to_nice[] = "iter_linear_to_nice";
static const char __pyx_k_iter_nice_to_linear[] = "iter_nice_to_linear";
static const char __pyx_k_pegasus_coordinates[] = "pegasus_coordinates";
static const char __pyx_k_update_clique_cache[] = "update_clique_cache";
static const char __pyx_k_chainlength_patience[] = "chainlength_patience";
static const char __pyx_k_minorminer_busclique[] = "minorminer.busclique";
static const char __pyx_k_regularize_embedding[] = "_regularize_embedding";
static const char __pyx_k_busgraph_cache___init[] = "busgraph_cache.__init__";
static const char __pyx_k_copy_on_close_context[] = "copy_on_close_context";
static const char __pyx_k_ensure_biclique_cache[] = "_ensure_biclique_cache";
static const char __pyx_k_find_clique_embedding[] = "find_clique_embedding";
static const char __pyx_k_iter_linear_to_zephyr[] = "iter_linear_to_zephyr";
static const char __pyx_k_iter_zephyr_to_linear[] = "iter_zephyr_to_linear";
static const char __pyx_k_draw_chimera_embedding[] = "draw_chimera_embedding";
static const char __pyx_k_iter_chimera_to_linear[] = "iter_chimera_to_linear";
static const char __pyx_k_iter_linear_to_chimera[] = "iter_linear_to_chimera";
static const char __pyx_k_iter_linear_to_pegasus[] = "iter_linear_to_pegasus";
static const char __pyx_k_iter_pegasus_to_linear[] = "iter_pegasus_to_linear";
static const char __pyx_k_mine_clique_embeddings[] = "mine_clique_embeddings";
static const char __pyx_k_draw_fragment_embedding[] = "draw_fragment_embedding";
static const char __pyx_k_find_biclique_embedding[] = "find_biclique_embedding";
static const char __pyx_k_insert_clique_embedding[] = "insert_clique_embedding";
static const char __pyx_k_default_quality_function[] = "_default_quality_function";
static const char __pyx_k_minorminer_busclique_pyx[] = "minorminer/busclique.pyx";
static const char __pyx_k_iter_nice_to_linear_pairs[] = "iter_nice_to_linear_pairs";
static const char __pyx_k_largest_balanced_biclique[] = "largest_balanced_biclique";
static const char __pyx_k_copy_on_close_context_open[] = "copy_on_close_context.open";
static const char __pyx_k_graph_family_not_supported[] = "graph family not supported";
static const char __pyx_k_busgraph_cache__fetch_cache[] = "busgraph_cache._fetch_cache";
static const char __pyx_k_copy_on_close_context_close[] = "copy_on_close_context.close";
static const char __pyx_k_iter_zephyr_to_linear_pairs[] = "iter_zephyr_to_linear_pairs";
static const char __pyx_k_unrecognized_graph_labeling[] = "unrecognized graph labeling";
static const char __pyx_k_busgraph_cache_cache_rootdir[] = "busgraph_cache.cache_rootdir";
static const char __pyx_k_copy_on_close_context___init[] = "copy_on_close_context.__init__";
static const char __pyx_k_iter_chimera_to_linear_pairs[] = "iter_chimera_to_linear_pairs";
static const char __pyx_k_iter_pegasus_to_linear_pairs[] = "iter_pegasus_to_linear_pairs";
static const char __pyx_k_busgraph_cache_largest_clique[] = "busgraph_cache.largest_clique";
static const char __pyx_k_largest_clique_by_chainlength[] = "largest_clique_by_chainlength";
static const char __pyx_k_polynomial_embedder_run_d_of_d[] = "polynomial embedder run %d of %d";
static const char __pyx_k_busgraph_cache__ensure_biclique[] = "busgraph_cache._ensure_biclique_cache";
static const char __pyx_k_busgraph_cache_clear_all_caches[] = "busgraph_cache.clear_all_caches";
static const char __pyx_k_busgraph_cache_get_clique_cache[] = "busgraph_cache.get_clique_cache";
static const char __pyx_k_busgraph_cache_largest_balanced[] = "busgraph_cache.largest_balanced_biclique.<locals>.<lambda>";
static const char __pyx_k_default_quality_function_locals[] = "_default_quality_function.<locals>.genexpr";
static const char __pyx_k_make_relabeler_locals__relabele[] = "_make_relabeler.<locals>._relabeler";
static const char __pyx_k_quality_function_is_ignored_for[] = "quality_function is ignored for direct insertion";
static const char __pyx_k_regularize_embedding_locals_gen[] = "_regularize_embedding.<locals>.genexpr";
static const char __pyx_k_writing_embedding_cache_to_disk[] = "writing embedding cache to disk";
static const char __pyx_k_A_cache_class_for_Chimera_Pegasu[] = "A cache class for Chimera, Pegasus and Zephyr graphs, and their\n    associated cliques and bicliques.\n\n    The cache files are stored in a directory determined by `homebase` (use\n    :meth:`busgraph_cache.cache_rootdir` to retrieve the path to this directory).\n    Subdirectories named `cliques` and `bicliques` are then created to store the\n    respective caches in each.\n\n    Args:\n        g (NetworkX Graph):\n            A :func:`dwave_networkx.pegasus_graph` or :func:`dwave_networkx.chimera_graph`.\n                or :func:`dwave_networkx.zephyr_graph`.\n    Note:\n        Due to internal optimizations, not all Chimera graphs are supported by\n        this code. Specifically, the graphs :func:`dwave_networkx.chimera_graph(m, n, t)`\n        are only supported for :math:`t<=8`. The code currently supports D-Wave\n        products, which have :math:`t=4`, but not all graphs. For graphs with\n        :math:`t>8`, use the legacy chimera-embedding package.\n\n    ";
static const char __pyx_k_busgraph_cache__ensure_clique_ca[] = "busgraph_cache._ensure_clique_cache";
static const char __pyx_k_busgraph_cache_draw_fragment_emb[] = "busgraph_cache.draw_fragment_embedding";
static const char __pyx_k_busgraph_cache_find_biclique_emb[] = "busgraph_cache.find_biclique_embedding";
static const char __pyx_k_busgraph_cache_find_clique_embed[] = "busgraph_cache.find_clique_embedding";
static const char __pyx_k_busgraph_cache_insert_clique_emb[] = "busgraph_cache.insert_clique_embedding.<locals>.<lambda>";
static const char __pyx_k_busgraph_cache_largest_clique_by[] = "busgraph_cache.largest_clique_by_chainlength";
static const char __pyx_k_busgraph_cache_merge_clique_cach[] = "busgraph_cache.merge_clique_cache.<locals>.<lambda>";
static const char __pyx_k_busgraph_cache_update_clique_cac[] = "busgraph_cache.update_clique_cache.<locals>.<lambda>";
static const char __pyx_k_copy_on_close_context_can_only_o[] = "copy_on_close_context can only open files once.  Discarding all files.";
static const char __pyx_k_could_not_allocate_memory_for_to[] = "could not allocate memory for topology serialization";
static const char __pyx_k_input_graph_must_either_be_a_dwa[] = "input graph must either be a dwave_networkx.pegasus_graph, dwave_networkx.chimera_graph or dwave_networkx.zephyr_graph";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_running_heuristic_embeddings_for[] = "running heuristic embeddings for size %d";
static const char __pyx_k_this_clique_embedder_supports_ch[] = "this clique embedder supports chimera graphs with a tile size of 8 or less";
static const char __pyx_k_this_clique_embedder_supports_ze[] = "this clique embedder supports zephyr graphs with a tile size of 4 or less";
static const char __pyx_k_busgraph_cache_largest_balanced_2[] = "busgraph_cache.largest_balanced_biclique";
static const char __pyx_k_busgraph_cache_insert_clique_emb_2[] = "busgraph_cache.insert_clique_embedding";
static const char __pyx_k_busgraph_cache_merge_clique_cach_2[] = "busgraph_cache.merge_clique_cache";
static const char __pyx_k_busgraph_cache_update_clique_cac_2[] = "busgraph_cache.update_clique_cache";
static PyObject *__pyx_kp_u_16x;
static PyObject *__pyx_kp_s_A_cache_class_for_Chimera_Pegasu;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_FileNotFoundError;
static PyObject *__pyx_n_s_Graph;
static PyObject *__pyx_n_s_InterProcessLock;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_LRU;
static PyObject *__pyx_n_s_Lock;
static PyObject *__pyx_n_s_M;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_N;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_Path;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u__21;
static PyObject *__pyx_n_s_add_edges_from;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_basedir;
static PyObject *__pyx_n_s_bgc;
static PyObject *__pyx_n_s_bgc_i;
static PyObject *__pyx_n_u_biclique;
static PyObject *__pyx_n_s_bicliques;
static PyObject *__pyx_n_s_bicliques_2;
static PyObject *__pyx_n_s_biggest;
static PyObject *__pyx_n_s_bipartite;
static PyObject *__pyx_n_s_blank;
static PyObject *__pyx_n_s_boundary;
static PyObject *__pyx_n_u_busclique;
static PyObject *__pyx_n_s_busgraph;
static PyObject *__pyx_n_s_busgraph_cache;
static PyObject *__pyx_n_s_busgraph_cache___init;
static PyObject *__pyx_n_s_busgraph_cache__ensure_biclique;
static PyObject *__pyx_n_s_busgraph_cache__ensure_clique_ca;
static PyObject *__pyx_n_s_busgraph_cache__fetch_cache;
static PyObject *__pyx_n_s_busgraph_cache_cache_rootdir;
static PyObject *__pyx_n_s_busgraph_cache_clear_all_caches;
static PyObject *__pyx_n_s_busgraph_cache_draw_fragment_emb;
static PyObject *__pyx_n_s_busgraph_cache_find_biclique_emb;
static PyObject *__pyx_n_s_busgraph_cache_find_clique_embed;
static PyObject *__pyx_n_s_busgraph_cache_get_clique_cache;
static PyObject *__pyx_n_s_busgraph_cache_insert_clique_emb;
static PyObject *__pyx_n_s_busgraph_cache_insert_clique_emb_2;
static PyObject *__pyx_n_s_busgraph_cache_largest_balanced;
static PyObject *__pyx_n_s_busgraph_cache_largest_balanced_2;
static PyObject *__pyx_n_s_busgraph_cache_largest_clique;
static PyObject *__pyx_n_s_busgraph_cache_largest_clique_by;
static PyObject *__pyx_n_s_busgraph_cache_merge_clique_cach;
static PyObject *__pyx_n_s_busgraph_cache_merge_clique_cach_2;
static PyObject *__pyx_n_s_busgraph_cache_update_clique_cac;
static PyObject *__pyx_n_s_busgraph_cache_update_clique_cac_2;
static PyObject *__pyx_n_s_by_length;
static PyObject *__pyx_n_s_by_size;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_cache;
static PyObject *__pyx_n_s_cache_rootdir;
static PyObject *__pyx_n_s_cachefile;
static PyObject *__pyx_n_s_chain;
static PyObject *__pyx_n_u_chain;
static PyObject *__pyx_n_s_chain_node;
static PyObject *__pyx_n_s_chainlength;
static PyObject *__pyx_n_s_chainlength_patience;
static PyObject *__pyx_n_u_chimera;
static PyObject *__pyx_n_s_chimera_busgraph;
static PyObject *__pyx_n_s_chimera_coordinates;
static PyObject *__pyx_n_s_chimera_graph;
static PyObject *__pyx_n_s_clear;
static PyObject *__pyx_n_s_clear_all_caches;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_u_clique;
static PyObject *__pyx_n_s_cliques;
static PyObject *__pyx_n_s_cliques_2;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cols;
static PyObject *__pyx_n_u_cols;
static PyObject *__pyx_n_u_columns;
static PyObject *__pyx_n_s_complete_graph;
static PyObject *__pyx_n_s_compress;
static PyObject *__pyx_n_s_compute;
static PyObject *__pyx_n_s_compute_identifier;
static PyObject *__pyx_n_u_coordinate;
static PyObject *__pyx_n_s_copy_on_close_context;
static PyObject *__pyx_n_s_copy_on_close_context___init;
static PyObject *__pyx_kp_u_copy_on_close_context_can_only_o;
static PyObject *__pyx_n_s_copy_on_close_context_close;
static PyObject *__pyx_n_s_copy_on_close_context_open;
static PyObject *__pyx_kp_u_could_not_allocate_memory_for_to;
static PyObject *__pyx_n_s_currcache;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_decompress;
static PyObject *__pyx_n_s_default_quality_function;
static PyObject *__pyx_n_s_default_quality_function_locals;
static PyObject *__pyx_n_s_delabel;
static PyObject *__pyx_n_s_direct;
static PyObject *__pyx_n_s_dirname;
static PyObject *__pyx_n_s_dirstack;
static PyObject *__pyx_n_s_discard;
static PyObject *__pyx_n_s_dnx;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_draw_chimera_embedding;
static PyObject *__pyx_n_s_draw_fragment_embedding;
static PyObject *__pyx_n_s_dump;
static PyObject *__pyx_n_s_dumps;
static PyObject *__pyx_n_u_dwave;
static PyObject *__pyx_n_s_dwave_networkx;
static PyObject *__pyx_n_s_edge_list;
static PyObject *__pyx_n_s_edges;
static PyObject *__pyx_n_s_emb;
static PyObject *__pyx_n_s_emb0;
static PyObject *__pyx_n_s_emb1;
static PyObject *__pyx_n_s_emb_list;
static PyObject *__pyx_n_s_embs;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_ensure_biclique_cache;
static PyObject *__pyx_n_s_ensure_clique_cache;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_exist_ok;
static PyObject *__pyx_n_s_exists;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_f_emb;
static PyObject *__pyx_n_s_family;
static PyObject *__pyx_n_u_family;
static PyObject *__pyx_n_s_family_2;
static PyObject *__pyx_n_s_fasteners;
static PyObject *__pyx_n_s_fetch_cache;
static PyObject *__pyx_n_s_file_context;
static PyObject *__pyx_n_s_file_name;
static PyObject *__pyx_n_s_file_obj;
static PyObject *__pyx_n_s_filecache;
static PyObject *__pyx_n_s_files;
static PyObject *__pyx_n_s_find_biclique_embedding;
static PyObject *__pyx_n_s_find_clique_embedding;
static PyObject *__pyx_n_s_find_embeddings;
static PyObject *__pyx_n_s_force;
static PyObject *__pyx_n_s_force_write;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fragment_graph_spec;
static PyObject *__pyx_n_s_fragment_nodes;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getLogger;
static PyObject *__pyx_n_s_get_clique_cache;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_graph;
static PyObject *__pyx_n_s_graph_2;
static PyObject *__pyx_kp_u_graph_family_not_supported;
static PyObject *__pyx_n_s_graphclass;
static PyObject *__pyx_n_s_graphdata;
static PyObject *__pyx_n_s_gzip;
static PyObject *__pyx_n_s_heuristic_args;
static PyObject *__pyx_n_s_heuristic_bound;
static PyObject *__pyx_n_s_heuristic_runs;
static PyObject *__pyx_n_s_hm;
static PyObject *__pyx_n_s_homebase;
static PyObject *__pyx_n_u_horizontal_offsets;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_identifier;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_independent_set;
static PyObject *__pyx_n_s_info;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_kp_u_input_graph_must_either_be_a_dwa;
static PyObject *__pyx_n_s_insert_clique_embedding;
static PyObject *__pyx_n_u_int;
static PyObject *__pyx_n_s_is_dir;
static PyObject *__pyx_n_s_item;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_iter_chimera_to_linear;
static PyObject *__pyx_n_s_iter_chimera_to_linear_pairs;
static PyObject *__pyx_n_s_iter_linear_to_chimera;
static PyObject *__pyx_n_s_iter_linear_to_nice;
static PyObject *__pyx_n_s_iter_linear_to_pegasus;
static PyObject *__pyx_n_s_iter_linear_to_zephyr;
static PyObject *__pyx_n_s_iter_nice_to_linear;
static PyObject *__pyx_n_s_iter_nice_to_linear_pairs;
static PyObject *__pyx_n_s_iter_pegasus_to_linear;
static PyObject *__pyx_n_s_iter_pegasus_to_linear_pairs;
static PyObject *__pyx_n_s_iter_zephyr_to_linear;
static PyObject *__pyx_n_s_iter_zephyr_to_linear_pairs;
static PyObject *__pyx_n_s_iterdir;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_json;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key0;
static PyObject *__pyx_n_s_key1;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_u_labels;
static PyObject *__pyx_n_s_largest_balanced_biclique;
static PyObject *__pyx_n_s_largest_clique;
static PyObject *__pyx_n_s_largest_clique_by_chainlength;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_u_length;
static PyObject *__pyx_n_s_length0;
static PyObject *__pyx_n_s_length1;
static PyObject *__pyx_n_s_load;
static PyObject *__pyx_n_s_loads;
static PyObject *__pyx_kp_u_lock;
static PyObject *__pyx_n_s_lockfile;
static PyObject *__pyx_n_s_logger;
static PyObject *__pyx_n_s_logging;
static PyObject *__pyx_kp_u_lru;
static PyObject *__pyx_n_s_lru_2;
static PyObject *__pyx_n_s_lrufile;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_relabeler;
static PyObject *__pyx_n_s_make_relabeler_locals__relabele;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_mask_bound;
static PyObject *__pyx_n_s_matched;
static PyObject *__pyx_n_s_matching;
static PyObject *__pyx_n_s_matching_graph;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max_beta;
static PyObject *__pyx_n_s_max_length;
static PyObject *__pyx_n_u_max_side;
static PyObject *__pyx_n_s_maximum_matching;
static PyObject *__pyx_n_s_maxlen;
static PyObject *__pyx_n_s_merge_clique_cache;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_mine_clique_embeddings;
static PyObject *__pyx_n_s_miner;
static PyObject *__pyx_n_s_minorminer;
static PyObject *__pyx_n_s_minorminer_busclique;
static PyObject *__pyx_kp_s_minorminer_busclique_pyx;
static PyObject *__pyx_n_s_mkdir;
static PyObject *__pyx_n_s_mm;
static PyObject *__pyx_n_s_mmax;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_networkx;
static PyObject *__pyx_n_s_newLRU;
static PyObject *__pyx_n_u_nice;
static PyObject *__pyx_n_s_nmax;
static PyObject *__pyx_n_s_nn;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_no_seed;
static PyObject *__pyx_n_s_node_boundary;
static PyObject *__pyx_n_s_node_list;
static PyObject *__pyx_n_s_nodes;
static PyObject *__pyx_n_s_num;
static PyObject *__pyx_n_s_num_nodes;
static PyObject *__pyx_n_s_num_seeds;
static PyObject *__pyx_n_s_nx;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_oldkey;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_parents;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_pathlib;
static PyObject *__pyx_n_u_pegasus;
static PyObject *__pyx_n_s_pegasus_busgraph;
static PyObject *__pyx_n_s_pegasus_coordinates;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_kp_u_polynomial_embedder_run_d_of_d;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_progress;
static PyObject *__pyx_n_s_qual0;
static PyObject *__pyx_n_s_qual1;
static PyObject *__pyx_n_s_quality_function;
static PyObject *__pyx_kp_u_quality_function_is_ignored_for;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_randint;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_raw;
static PyObject *__pyx_n_u_raw;
static PyObject *__pyx_n_s_raw0;
static PyObject *__pyx_n_s_raw1;
static PyObject *__pyx_n_s_raw_emb;
static PyObject *__pyx_n_s_raw_emb0;
static PyObject *__pyx_n_s_raw_emb1;
static PyObject *__pyx_n_u_rb;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_regularize;
static PyObject *__pyx_n_s_regularize_embedding;
static PyObject *__pyx_n_s_regularize_embedding_locals_gen;
static PyObject *__pyx_n_s_relabel;
static PyObject *__pyx_n_s_relabeler;
static PyObject *__pyx_n_s_remainder;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_s_rmdir;
static PyObject *__pyx_n_s_rootdir;
static PyObject *__pyx_n_s_row;
static PyObject *__pyx_n_s_rows;
static PyObject *__pyx_n_u_rows;
static PyObject *__pyx_kp_u_running_heuristic_embeddings_for;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_s0;
static PyObject *__pyx_n_s_s1;
static PyObject *__pyx_n_s_s_min;
static PyObject *__pyx_n_s_seed;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set_mask_bound;
static PyObject *__pyx_n_s_setdefault;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_short_identifier;
static PyObject *__pyx_n_s_shortcode;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_u_size;
static PyObject *__pyx_n_s_size0;
static PyObject *__pyx_n_s_size1;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_sparse;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_substack;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_temp_name;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_u_this_clique_embedder_supports_ch;
static PyObject *__pyx_kp_u_this_clique_embedder_supports_ze;
static PyObject *__pyx_n_s_threading;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tile;
static PyObject *__pyx_n_u_tile;
static PyObject *__pyx_n_s_top;
static PyObject *__pyx_n_s_top_nodes;
static PyObject *__pyx_n_s_tries;
static PyObject *__pyx_n_s_trivial_relabeler;
static PyObject *__pyx_n_s_union;
static PyObject *__pyx_n_s_unlink;
static PyObject *__pyx_kp_u_unrecognized_graph_labeling;
static PyObject *__pyx_n_s_update_clique_cache;
static PyObject *__pyx_n_s_urandom;
static PyObject *__pyx_n_s_use_cache;
static PyObject *__pyx_n_s_used;
static PyObject *__pyx_n_s_user_data_dir;
static PyObject *__pyx_n_u_utf8;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_version;
static PyObject *__pyx_n_u_vertical_offsets;
static PyObject *__pyx_n_u_wb;
static PyObject *__pyx_n_s_write_to_disk;
static PyObject *__pyx_kp_u_writing_embedding_cache_to_disk;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_u_zephyr;
static PyObject *__pyx_n_s_zephyr_busgraph;
static PyObject *__pyx_n_s_zephyr_coordinates;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_n_s_zip_longest;
static PyObject *__pyx_pf_10minorminer_9busclique__num_nodes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nn, PyObject *__pyx_v_offset); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_2find_clique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nodes, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_use_cache); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_2_ensure_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_blank); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_4_ensure_biclique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_6cache_rootdir(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_version); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_8clear_all_caches(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_10get_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sparse, PyObject *__pyx_v_compress); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_12update_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_cache, PyObject *__pyx_v_write_to_disk); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda1(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_14insert_clique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_emb, PyObject *__pyx_v_write_to_disk, PyObject *__pyx_v_quality_function, PyObject *__pyx_v_direct); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda2(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_16merge_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_cache, PyObject *__pyx_v_quality_function, PyObject *__pyx_v_write_to_disk); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_18_fetch_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_dirname, PyObject *__pyx_v_compute, PyObject *__pyx_v_force_write); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_20largest_clique(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_22largest_clique_by_chainlength(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_chainlength); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_24find_clique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nn); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_26largest_balanced_biclique(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_28find_biclique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nn, PyObject *__pyx_v_mm); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_30draw_fragment_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_emb, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_4_trivial_relabeler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_emb); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_15_make_relabeler__relabeler(PyObject *__pyx_self, PyObject *__pyx_v_emb); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_6_make_relabeler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static int __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph___cinit__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_compute_identifier); /* proto */
static void __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_2__dealloc__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_4set_mask_bound(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_num_masks); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_6bicliques(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_8cliques(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_10independent_set(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_12find_clique_embedding(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_nn); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_14fragment_graph_spec(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_16fragment_nodes(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_nodes); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_7relabel___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_7delabel___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_10identifier___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_16short_identifier___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph___cinit__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_compute_identifier); /* proto */
static void __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_2__dealloc__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_4set_mask_bound(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_num_masks); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_6bicliques(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_8cliques(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_10independent_set(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_12find_clique_embedding(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_nn); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_14fragment_graph_spec(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_16fragment_nodes(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_nodes); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_7relabel___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_7delabel___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_10identifier___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_16short_identifier___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10minorminer_9busclique_17_chimera_busgraph___cinit__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_compute_identifier); /* proto */
static void __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_2__dealloc__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_4set_mask_bound(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_num_masks); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_6bicliques(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_8cliques(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_10independent_set(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_12find_clique_embedding(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_nn); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_14fragment_graph_spec(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_16fragment_nodes(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_nodes); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_7relabel___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_7delabel___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_10identifier___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_16short_identifier___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_25_default_quality_function_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_8_default_quality_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_emb); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_21_regularize_embedding_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_10_regularize_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_g, PyObject *__pyx_v_emb); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_12mine_clique_embeddings(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_g, PyObject *__pyx_v_quality_function, PyObject *__pyx_v_regularize, PyObject *__pyx_v_num_seeds, PyObject *__pyx_v_mask_bound, PyObject *__pyx_v_heuristic_bound, PyObject *__pyx_v_heuristic_args, PyObject *__pyx_v_heuristic_runs); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_21copy_on_close_context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_21copy_on_close_context_2open(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_file_name, PyObject *__pyx_v_mode); /* proto */
static PyObject *__pyx_pf_10minorminer_9busclique_21copy_on_close_context_4close(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_10minorminer_9busclique__zephyr_busgraph(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique__pegasus_busgraph(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique__chimera_busgraph(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct__update_clique_cache(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_7_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_setdefault = {0, &__pyx_n_s_setdefault, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_64;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_4294967295;
static PyObject *__pyx_k_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__12;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_codeobj__6;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__80;
/* Late includes */

/* "minorminer/busclique.pyx":46
 *                             'biclique': threading.Lock()}
 * 
 * def _num_nodes(nn, offset = 0):             # <<<<<<<<<<<<<<
 *     """Internal-use function to normalize inputs.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_1_num_nodes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique__num_nodes[] = "Internal-use function to normalize inputs.\n\n    Args:\n        nn: int or iterable\n\n    Returns:\n        (n, nodes)\n            n: nn if n is an int, otherwise len(nodes)\n            nodes: tuple(offset, offset + nn) if nn is an iterable, otherwise\n            tuple(range(offset, offset + n))\n    ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_1_num_nodes = {"_num_nodes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_1_num_nodes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique__num_nodes};
static PyObject *__pyx_pw_10minorminer_9busclique_1_num_nodes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nn = 0;
  PyObject *__pyx_v_offset = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_num_nodes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nn,&__pyx_n_s_offset,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nn)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_num_nodes") < 0)) __PYX_ERR(0, 46, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nn = values[0];
    __pyx_v_offset = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_num_nodes", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 46, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._num_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique__num_nodes(__pyx_self, __pyx_v_nn, __pyx_v_offset);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique__num_nodes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nn, PyObject *__pyx_v_offset) {
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_num_nodes", 0);

  /* "minorminer/busclique.pyx":58
 *             tuple(range(offset, offset + n))
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         nodes = tuple(range(offset, offset + nn))
 *         num = nn
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "minorminer/busclique.pyx":59
 *     """
 *     try:
 *         nodes = tuple(range(offset, offset + nn))             # <<<<<<<<<<<<<<
 *         num = nn
 *     except TypeError:
 */
      __pyx_t_4 = PyNumber_Add(__pyx_v_offset, __pyx_v_nn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 59, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 59, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_offset);
      __Pyx_GIVEREF(__pyx_v_offset);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_offset);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 59, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 59, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_nodes = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "minorminer/busclique.pyx":60
 *     try:
 *         nodes = tuple(range(offset, offset + nn))
 *         num = nn             # <<<<<<<<<<<<<<
 *     except TypeError:
 *         nodes = tuple(nn)
 */
      __Pyx_INCREF(__pyx_v_nn);
      __pyx_v_num = __pyx_v_nn;

      /* "minorminer/busclique.pyx":58
 *             tuple(range(offset, offset + n))
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         nodes = tuple(range(offset, offset + nn))
 *         num = nn
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":61
 *         nodes = tuple(range(offset, offset + nn))
 *         num = nn
 *     except TypeError:             # <<<<<<<<<<<<<<
 *         nodes = tuple(nn)
 *         num = len(nodes)
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("minorminer.busclique._num_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 61, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);

      /* "minorminer/busclique.pyx":62
 *         num = nn
 *     except TypeError:
 *         nodes = tuple(nn)             # <<<<<<<<<<<<<<
 *         num = len(nodes)
 *     return num, nodes
 */
      __pyx_t_8 = __Pyx_PySequence_Tuple(__pyx_v_nn); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 62, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_nodes, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "minorminer/busclique.pyx":63
 *     except TypeError:
 *         nodes = tuple(nn)
 *         num = len(nodes)             # <<<<<<<<<<<<<<
 *     return num, nodes
 * 
 */
      __pyx_t_9 = PyTuple_GET_SIZE(__pyx_v_nodes); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 63, __pyx_L5_except_error)
      __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 63, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_num, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "minorminer/busclique.pyx":58
 *             tuple(range(offset, offset + n))
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         nodes = tuple(range(offset, offset + nn))
 *         num = nn
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "minorminer/busclique.pyx":64
 *         nodes = tuple(nn)
 *         num = len(nodes)
 *     return num, nodes             # <<<<<<<<<<<<<<
 * 
 * _no_seed = object()
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_num);
  __Pyx_GIVEREF(__pyx_v_num);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_num);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_nodes);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":46
 *                             'biclique': threading.Lock()}
 * 
 * def _num_nodes(nn, offset = 0):             # <<<<<<<<<<<<<<
 *     """Internal-use function to normalize inputs.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("minorminer.busclique._num_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":67
 * 
 * _no_seed = object()
 * def find_clique_embedding(nodes, g, seed = _no_seed, use_cache = True):             # <<<<<<<<<<<<<<
 *     """Finds a clique embedding in the graph ``g`` using a polynomial-time
 *     algorithm.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_3find_clique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_2find_clique_embedding[] = "Finds a clique embedding in the graph ``g`` using a polynomial-time\n    algorithm.\n\n    Args:\n        nodes (int/iterable):\n            A number (indicating the size of the desired clique) or an\n            iterable (specifying the node labels of the desired clique).\n\n        g (NetworkX Graph):\n            The target graph that is either a :func:`dwave_networkx.chimera_graph`\n            or :func:`dwave_networkx.pegasus_graph`.\n\n        use_cache (bool, optional, default=True):\n            Whether or not to compute/restore a cache of clique embeddings for\n            ``g``. Note that this function only uses the filesystem cache, and\n            does not maintain the cache in memory. If many (or even several)\n            embeddings are desired in a single session, it is recommended to use\n            :class:`.busgraph_cache`.\n\n        seed (int, optional):\n            A seed for an internal random number generator.  If ``use_cache`` is\n            True, then the seed defaults to an internally-defined value which\n            is consistent between runs.  Otherwise, a seed is generated from the\n            current python random state.\n\n    Returns:\n        dict: An embedding of node labels (either nodes, or range(nodes)) mapped\n        to chains of a clique embedding.\n\n    Note:\n        Due to internal optimizations, not all Chimera graphs are supported by\n        this code. Specifically, the graphs :func:`dwave_networkx.chimera_graph(m, n, t)`\n        are only supported for :math:`t<=8`. The code currently supports D-Wave\n        products, which have :math:`t=4`, but not all graphs. For graphs with\n        :math:`t>8`, use the legacy chimera-embedding package.\n\n    Note:\n        When the cache is used, clique embeddings of all sizes are computed\n        and cached. This takes somewhat longer than a single embedding, but tends\n        to pay off after a fairly small number of calls. An exceptional use case\n        is when there a""re a large number of missing internal couplers, where the\n        result is nondeterministic -- avoiding the cache in this case may be\n        preferable.\n\n    ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_3find_clique_embedding = {"find_clique_embedding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_3find_clique_embedding, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_2find_clique_embedding};
static PyObject *__pyx_pw_10minorminer_9busclique_3find_clique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0;
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_use_cache = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_clique_embedding (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,&__pyx_n_s_g,&__pyx_n_s_seed,&__pyx_n_s_use_cache,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = __pyx_k_;
    values[3] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_clique_embedding", 0, 2, 4, 1); __PYX_ERR(0, 67, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_cache);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_clique_embedding") < 0)) __PYX_ERR(0, 67, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nodes = values[0];
    __pyx_v_g = values[1];
    __pyx_v_seed = values[2];
    __pyx_v_use_cache = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_clique_embedding", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_2find_clique_embedding(__pyx_self, __pyx_v_nodes, __pyx_v_g, __pyx_v_seed, __pyx_v_use_cache);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_2find_clique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nodes, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_use_cache) {
  PyObject *__pyx_v_graphdata = NULL;
  PyObject *__pyx_v_family = NULL;
  PyObject *__pyx_v_busgraph = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_clique_embedding", 0);
  __Pyx_INCREF(__pyx_v_seed);

  /* "minorminer/busclique.pyx":113
 * 
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         graphdata = g.graph
 *         family = graphdata['family']
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "minorminer/busclique.pyx":114
 *     """
 *     try:
 *         graphdata = g.graph             # <<<<<<<<<<<<<<
 *         family = graphdata['family']
 *         busgraph = {'pegasus': _pegasus_busgraph,
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 114, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_graphdata = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":115
 *     try:
 *         graphdata = g.graph
 *         family = graphdata['family']             # <<<<<<<<<<<<<<
 *         busgraph = {'pegasus': _pegasus_busgraph,
 *                     'zephyr': _zephyr_busgraph,
 */
      __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_graphdata, __pyx_n_u_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 115, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_family = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":116
 *         graphdata = g.graph
 *         family = graphdata['family']
 *         busgraph = {'pegasus': _pegasus_busgraph,             # <<<<<<<<<<<<<<
 *                     'zephyr': _zephyr_busgraph,
 *                     'chimera': _chimera_busgraph}[family]
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_pegasus, ((PyObject *)__pyx_ptype_10minorminer_9busclique__pegasus_busgraph)) < 0) __PYX_ERR(0, 116, __pyx_L3_error)

      /* "minorminer/busclique.pyx":117
 *         family = graphdata['family']
 *         busgraph = {'pegasus': _pegasus_busgraph,
 *                     'zephyr': _zephyr_busgraph,             # <<<<<<<<<<<<<<
 *                     'chimera': _chimera_busgraph}[family]
 *     except (AttributeError, KeyError):
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_zephyr, ((PyObject *)__pyx_ptype_10minorminer_9busclique__zephyr_busgraph)) < 0) __PYX_ERR(0, 116, __pyx_L3_error)

      /* "minorminer/busclique.pyx":118
 *         busgraph = {'pegasus': _pegasus_busgraph,
 *                     'zephyr': _zephyr_busgraph,
 *                     'chimera': _chimera_busgraph}[family]             # <<<<<<<<<<<<<<
 *     except (AttributeError, KeyError):
 *         raise ValueError(("input graph must either be a "
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_chimera, ((PyObject *)__pyx_ptype_10minorminer_9busclique__chimera_busgraph)) < 0) __PYX_ERR(0, 116, __pyx_L3_error)
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_t_4, __pyx_v_family); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 118, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_busgraph = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "minorminer/busclique.pyx":113
 * 
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         graphdata = g.graph
 *         family = graphdata['family']
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":119
 *                     'zephyr': _zephyr_busgraph,
 *                     'chimera': _chimera_busgraph}[family]
 *     except (AttributeError, KeyError):             # <<<<<<<<<<<<<<
 *         raise ValueError(("input graph must either be a "
 *                           "dwave_networkx.pegasus_graph, "
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("minorminer.busclique.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 119, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);

      /* "minorminer/busclique.pyx":120
 *                     'chimera': _chimera_busgraph}[family]
 *     except (AttributeError, KeyError):
 *         raise ValueError(("input graph must either be a "             # <<<<<<<<<<<<<<
 *                           "dwave_networkx.pegasus_graph, "
 *                           "dwave_networkx.chimera_graph or "
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 120, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 120, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "minorminer/busclique.pyx":113
 * 
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         graphdata = g.graph
 *         family = graphdata['family']
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "minorminer/busclique.pyx":125
 *                           "dwave_networkx.zephyr_graph"))
 * 
 *     if use_cache:             # <<<<<<<<<<<<<<
 *         if seed is _no_seed:
 *             seed = 0
 */
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_use_cache); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "minorminer/busclique.pyx":126
 * 
 *     if use_cache:
 *         if seed is _no_seed:             # <<<<<<<<<<<<<<
 *             seed = 0
 *         return busgraph_cache(g, seed=seed).find_clique_embedding(nodes)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_no_seed); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = (__pyx_v_seed == __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (__pyx_t_10) {

      /* "minorminer/busclique.pyx":127
 *     if use_cache:
 *         if seed is _no_seed:
 *             seed = 0             # <<<<<<<<<<<<<<
 *         return busgraph_cache(g, seed=seed).find_clique_embedding(nodes)
 *     else:
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_DECREF_SET(__pyx_v_seed, __pyx_int_0);

      /* "minorminer/busclique.pyx":126
 * 
 *     if use_cache:
 *         if seed is _no_seed:             # <<<<<<<<<<<<<<
 *             seed = 0
 *         return busgraph_cache(g, seed=seed).find_clique_embedding(nodes)
 */
    }

    /* "minorminer/busclique.pyx":128
 *         if seed is _no_seed:
 *             seed = 0
 *         return busgraph_cache(g, seed=seed).find_clique_embedding(nodes)             # <<<<<<<<<<<<<<
 *     else:
 *         if seed is _no_seed:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_busgraph_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_g);
    __Pyx_GIVEREF(__pyx_v_g);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_g);
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_seed, __pyx_v_seed) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_find_clique_embedding); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_11, __pyx_v_nodes) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_nodes);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":125
 *                           "dwave_networkx.zephyr_graph"))
 * 
 *     if use_cache:             # <<<<<<<<<<<<<<
 *         if seed is _no_seed:
 *             seed = 0
 */
  }

  /* "minorminer/busclique.pyx":130
 *         return busgraph_cache(g, seed=seed).find_clique_embedding(nodes)
 *     else:
 *         if seed is _no_seed:             # <<<<<<<<<<<<<<
 *             seed = None
 *         return busgraph(g, seed=seed).find_clique_embedding(nodes)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_no_seed); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = (__pyx_v_seed == __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = (__pyx_t_10 != 0);
    if (__pyx_t_9) {

      /* "minorminer/busclique.pyx":131
 *     else:
 *         if seed is _no_seed:
 *             seed = None             # <<<<<<<<<<<<<<
 *         return busgraph(g, seed=seed).find_clique_embedding(nodes)
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_seed, Py_None);

      /* "minorminer/busclique.pyx":130
 *         return busgraph_cache(g, seed=seed).find_clique_embedding(nodes)
 *     else:
 *         if seed is _no_seed:             # <<<<<<<<<<<<<<
 *             seed = None
 *         return busgraph(g, seed=seed).find_clique_embedding(nodes)
 */
    }

    /* "minorminer/busclique.pyx":132
 *         if seed is _no_seed:
 *             seed = None
 *         return busgraph(g, seed=seed).find_clique_embedding(nodes)             # <<<<<<<<<<<<<<
 * 
 * class busgraph_cache:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_g);
    __Pyx_GIVEREF(__pyx_v_g);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_g);
    __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_seed, __pyx_v_seed) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v_busgraph, __pyx_t_8, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_find_clique_embedding); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_5, __pyx_v_nodes) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_v_nodes);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":67
 * 
 * _no_seed = object()
 * def find_clique_embedding(nodes, g, seed = _no_seed, use_cache = True):             # <<<<<<<<<<<<<<
 *     """Finds a clique embedding in the graph ``g`` using a polynomial-time
 *     algorithm.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("minorminer.busclique.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_graphdata);
  __Pyx_XDECREF(__pyx_v_family);
  __Pyx_XDECREF(__pyx_v_busgraph);
  __Pyx_XDECREF(__pyx_v_seed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":155
 * 
 *     """
 *     def __init__(self, g, seed = 0):             # <<<<<<<<<<<<<<
 *         self._family = g.graph['family']
 *         graphclass = {'pegasus': _pegasus_busgraph,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_seed = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_g,&__pyx_n_s_seed,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_0));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 155, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_g = values[1];
    __pyx_v_seed = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 155, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache___init__(__pyx_self, __pyx_v_self, __pyx_v_g, __pyx_v_seed);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed) {
  PyObject *__pyx_v_graphclass = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "minorminer/busclique.pyx":156
 *     """
 *     def __init__(self, g, seed = 0):
 *         self._family = g.graph['family']             # <<<<<<<<<<<<<<
 *         graphclass = {'pegasus': _pegasus_busgraph,
 *                       'zephyr': _zephyr_busgraph,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_family); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_family_2, __pyx_t_2) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":157
 *     def __init__(self, g, seed = 0):
 *         self._family = g.graph['family']
 *         graphclass = {'pegasus': _pegasus_busgraph,             # <<<<<<<<<<<<<<
 *                       'zephyr': _zephyr_busgraph,
 *                       'chimera': _chimera_busgraph}.get(self._family)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_pegasus, ((PyObject *)__pyx_ptype_10minorminer_9busclique__pegasus_busgraph)) < 0) __PYX_ERR(0, 157, __pyx_L1_error)

  /* "minorminer/busclique.pyx":158
 *         self._family = g.graph['family']
 *         graphclass = {'pegasus': _pegasus_busgraph,
 *                       'zephyr': _zephyr_busgraph,             # <<<<<<<<<<<<<<
 *                       'chimera': _chimera_busgraph}.get(self._family)
 *         if graphclass is None:
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_zephyr, ((PyObject *)__pyx_ptype_10minorminer_9busclique__zephyr_busgraph)) < 0) __PYX_ERR(0, 157, __pyx_L1_error)

  /* "minorminer/busclique.pyx":159
 *         graphclass = {'pegasus': _pegasus_busgraph,
 *                       'zephyr': _zephyr_busgraph,
 *                       'chimera': _chimera_busgraph}.get(self._family)             # <<<<<<<<<<<<<<
 *         if graphclass is None:
 *             raise ValueError(("input graph must either be a "
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_chimera, ((PyObject *)__pyx_ptype_10minorminer_9busclique__chimera_busgraph)) < 0) __PYX_ERR(0, 157, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_family_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyDict_GetItemDefault(__pyx_t_2, __pyx_t_1, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_graphclass = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":160
 *                       'zephyr': _zephyr_busgraph,
 *                       'chimera': _chimera_busgraph}.get(self._family)
 *         if graphclass is None:             # <<<<<<<<<<<<<<
 *             raise ValueError(("input graph must either be a "
 *                               "dwave_networkx.pegasus_graph, "
 */
  __pyx_t_4 = (__pyx_v_graphclass == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (unlikely(__pyx_t_5)) {

    /* "minorminer/busclique.pyx":161
 *                       'chimera': _chimera_busgraph}.get(self._family)
 *         if graphclass is None:
 *             raise ValueError(("input graph must either be a "             # <<<<<<<<<<<<<<
 *                               "dwave_networkx.pegasus_graph, "
 *                               "dwave_networkx.chimera_graph or "
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 161, __pyx_L1_error)

    /* "minorminer/busclique.pyx":160
 *                       'zephyr': _zephyr_busgraph,
 *                       'chimera': _chimera_busgraph}.get(self._family)
 *         if graphclass is None:             # <<<<<<<<<<<<<<
 *             raise ValueError(("input graph must either be a "
 *                               "dwave_networkx.pegasus_graph, "
 */
  }

  /* "minorminer/busclique.pyx":165
 *                               "dwave_networkx.chimera_graph or "
 *                               "dwave_networkx.zephyr_graph"))
 *         self._graph = graphclass(g, seed=seed, compute_identifier=True)             # <<<<<<<<<<<<<<
 *         self._cliques = None
 *         self._bicliques = None
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_g);
  __Pyx_GIVEREF(__pyx_v_g);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_g);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_seed, __pyx_v_seed) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_compute_identifier, Py_True) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_graphclass, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_graph_2, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":166
 *                               "dwave_networkx.zephyr_graph"))
 *         self._graph = graphclass(g, seed=seed, compute_identifier=True)
 *         self._cliques = None             # <<<<<<<<<<<<<<
 *         self._bicliques = None
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cliques, Py_None) < 0) __PYX_ERR(0, 166, __pyx_L1_error)

  /* "minorminer/busclique.pyx":167
 *         self._graph = graphclass(g, seed=seed, compute_identifier=True)
 *         self._cliques = None
 *         self._bicliques = None             # <<<<<<<<<<<<<<
 * 
 *     def _ensure_clique_cache(self, blank=False):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bicliques, Py_None) < 0) __PYX_ERR(0, 167, __pyx_L1_error)

  /* "minorminer/busclique.pyx":155
 * 
 *     """
 *     def __init__(self, g, seed = 0):             # <<<<<<<<<<<<<<
 *         self._family = g.graph['family']
 *         graphclass = {'pegasus': _pegasus_busgraph,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_graphclass);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":169
 *         self._bicliques = None
 * 
 *     def _ensure_clique_cache(self, blank=False):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_3_ensure_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_2_ensure_clique_cache[] = "Fetch/compute the clique cache, if it's not already in memory.";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_3_ensure_clique_cache = {"_ensure_clique_cache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_3_ensure_clique_cache, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_2_ensure_clique_cache};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_3_ensure_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_blank = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_ensure_clique_cache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_blank,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_blank);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_ensure_clique_cache") < 0)) __PYX_ERR(0, 169, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_blank = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_ensure_clique_cache", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 169, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._ensure_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_2_ensure_clique_cache(__pyx_self, __pyx_v_self, __pyx_v_blank);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_2_ensure_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_blank) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_clique_cache", 0);

  /* "minorminer/busclique.pyx":171
 *     def _ensure_clique_cache(self, blank=False):
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:             # <<<<<<<<<<<<<<
 *             if blank:
 *                 self._cliques = _make_clique_cache([[]])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "minorminer/busclique.pyx":172
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:
 *             if blank:             # <<<<<<<<<<<<<<
 *                 self._cliques = _make_clique_cache([[]])
 *             else:
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_blank); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 172, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "minorminer/busclique.pyx":173
 *         if self._cliques is None:
 *             if blank:
 *                 self._cliques = _make_clique_cache([[]])             # <<<<<<<<<<<<<<
 *             else:
 *                 self._cliques = self._fetch_cache(
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_5 = __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __pyx_f_10minorminer_9busclique__make_clique_cache(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cliques, __pyx_t_4) < 0) __PYX_ERR(0, 173, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":172
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:
 *             if blank:             # <<<<<<<<<<<<<<
 *                 self._cliques = _make_clique_cache([[]])
 *             else:
 */
      goto __pyx_L4;
    }

    /* "minorminer/busclique.pyx":175
 *                 self._cliques = _make_clique_cache([[]])
 *             else:
 *                 self._cliques = self._fetch_cache(             # <<<<<<<<<<<<<<
 *                     'clique',
 *                     self._graph.cliques
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_fetch_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "minorminer/busclique.pyx":177
 *                 self._cliques = self._fetch_cache(
 *                     'clique',
 *                     self._graph.cliques             # <<<<<<<<<<<<<<
 *                 )
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_cliques_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_u_clique, __pyx_t_7};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_u_clique, __pyx_t_7};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_n_u_clique);
        __Pyx_GIVEREF(__pyx_n_u_clique);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_n_u_clique);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":175
 *                 self._cliques = _make_clique_cache([[]])
 *             else:
 *                 self._cliques = self._fetch_cache(             # <<<<<<<<<<<<<<
 *                     'clique',
 *                     self._graph.cliques
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cliques, __pyx_t_4) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L4:;

    /* "minorminer/busclique.pyx":171
 *     def _ensure_clique_cache(self, blank=False):
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:             # <<<<<<<<<<<<<<
 *             if blank:
 *                 self._cliques = _make_clique_cache([[]])
 */
  }

  /* "minorminer/busclique.pyx":169
 *         self._bicliques = None
 * 
 *     def _ensure_clique_cache(self, blank=False):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._ensure_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":180
 *                 )
 * 
 *     def _ensure_biclique_cache(self):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_5_ensure_biclique_cache(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_4_ensure_biclique_cache[] = "Fetch/compute the clique cache, if it's not already in memory.";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_5_ensure_biclique_cache = {"_ensure_biclique_cache", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_5_ensure_biclique_cache, METH_O, __pyx_doc_10minorminer_9busclique_14busgraph_cache_4_ensure_biclique_cache};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_5_ensure_biclique_cache(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_ensure_biclique_cache (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_4_ensure_biclique_cache(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_4_ensure_biclique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_biclique_cache", 0);

  /* "minorminer/busclique.pyx":182
 *     def _ensure_biclique_cache(self):
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:             # <<<<<<<<<<<<<<
 *             self._bicliques = self._fetch_cache('biclique', self._graph.bicliques)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bicliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "minorminer/busclique.pyx":183
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:
 *             self._bicliques = self._fetch_cache('biclique', self._graph.bicliques)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_fetch_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_bicliques_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_u_biclique, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_u_biclique, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_n_u_biclique);
      __Pyx_GIVEREF(__pyx_n_u_biclique);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_n_u_biclique);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bicliques, __pyx_t_1) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":182
 *     def _ensure_biclique_cache(self):
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:             # <<<<<<<<<<<<<<
 *             self._bicliques = self._fetch_cache('biclique', self._graph.bicliques)
 * 
 */
  }

  /* "minorminer/busclique.pyx":180
 *                 )
 * 
 *     def _ensure_biclique_cache(self):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._ensure_biclique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":186
 * 
 *     @staticmethod
 *     def cache_rootdir(version=__cache_version):             # <<<<<<<<<<<<<<
 *         """Returns the directory corresponding to the provided cache version.
 * 
 */

static PyObject *__pyx_pf_10minorminer_9busclique_14__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_version);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_version);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_version);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("minorminer.busclique.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_7cache_rootdir(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_6cache_rootdir[] = "Returns the directory corresponding to the provided cache version.\n\n        Args:\n            version (int, optional, default=current cache version):\n                Cache version.\n\n        Returns:\n            str\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_7cache_rootdir = {"cache_rootdir", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_7cache_rootdir, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_6cache_rootdir};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_7cache_rootdir(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_version = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cache_rootdir (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_version,0};
    PyObject* values[1] = {0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[0] = __pyx_dynamic_args->__pyx_arg_version;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_version);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cache_rootdir") < 0)) __PYX_ERR(0, 186, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_version = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cache_rootdir", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 186, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.cache_rootdir", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_6cache_rootdir(__pyx_self, __pyx_v_version);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_6cache_rootdir(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_version) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cache_rootdir", 0);

  /* "minorminer/busclique.pyx":196
 *             str
 *         """
 *         return homebase.user_data_dir('busclique', 'dwave', version)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_homebase); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_user_data_dir); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_n_u_busclique, __pyx_n_u_dwave, __pyx_v_version};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_n_u_busclique, __pyx_n_u_dwave, __pyx_v_version};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_n_u_busclique);
    __Pyx_GIVEREF(__pyx_n_u_busclique);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_n_u_busclique);
    __Pyx_INCREF(__pyx_n_u_dwave);
    __Pyx_GIVEREF(__pyx_n_u_dwave);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_u_dwave);
    __Pyx_INCREF(__pyx_v_version);
    __Pyx_GIVEREF(__pyx_v_version);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_version);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":186
 * 
 *     @staticmethod
 *     def cache_rootdir(version=__cache_version):             # <<<<<<<<<<<<<<
 *         """Returns the directory corresponding to the provided cache version.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.cache_rootdir", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":199
 * 
 *     @staticmethod
 *     def clear_all_caches():             # <<<<<<<<<<<<<<
 *         """Removes all caches created by this class, up to and including the
 *         current version.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_9clear_all_caches(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_8clear_all_caches[] = "Removes all caches created by this class, up to and including the\n        current version.\n\n        Returns:\n            None\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_9clear_all_caches = {"clear_all_caches", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_9clear_all_caches, METH_NOARGS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_8clear_all_caches};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_9clear_all_caches(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_all_caches (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_8clear_all_caches(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_8clear_all_caches(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_dirstack = NULL;
  long __pyx_v_i;
  PyObject *__pyx_v_rootdir = NULL;
  PyObject *__pyx_v_top = NULL;
  PyObject *__pyx_v_substack = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear_all_caches", 0);

  /* "minorminer/busclique.pyx":206
 *             None
 *         """
 *         dirstack = []             # <<<<<<<<<<<<<<
 *         for i in range(__cache_version + 1):
 *             rootdir = pathlib.Path(busgraph_cache.cache_rootdir(i))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dirstack = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":207
 *         """
 *         dirstack = []
 *         for i in range(__cache_version + 1):             # <<<<<<<<<<<<<<
 *             rootdir = pathlib.Path(busgraph_cache.cache_rootdir(i))
 *             if rootdir.exists():
 */
  __pyx_t_2 = (__pyx_v_10minorminer_9busclique___cache_version + 1);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "minorminer/busclique.pyx":208
 *         dirstack = []
 *         for i in range(__cache_version + 1):
 *             rootdir = pathlib.Path(busgraph_cache.cache_rootdir(i))             # <<<<<<<<<<<<<<
 *             if rootdir.exists():
 *                 dirstack.append(rootdir)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pathlib); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Path); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_busgraph_cache); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_cache_rootdir); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_rootdir, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":209
 *         for i in range(__cache_version + 1):
 *             rootdir = pathlib.Path(busgraph_cache.cache_rootdir(i))
 *             if rootdir.exists():             # <<<<<<<<<<<<<<
 *                 dirstack.append(rootdir)
 *         while dirstack:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_rootdir, __pyx_n_s_exists); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {

      /* "minorminer/busclique.pyx":210
 *             rootdir = pathlib.Path(busgraph_cache.cache_rootdir(i))
 *             if rootdir.exists():
 *                 dirstack.append(rootdir)             # <<<<<<<<<<<<<<
 *         while dirstack:
 *             top = dirstack[-1]
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_dirstack, __pyx_v_rootdir); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 210, __pyx_L1_error)

      /* "minorminer/busclique.pyx":209
 *         for i in range(__cache_version + 1):
 *             rootdir = pathlib.Path(busgraph_cache.cache_rootdir(i))
 *             if rootdir.exists():             # <<<<<<<<<<<<<<
 *                 dirstack.append(rootdir)
 *         while dirstack:
 */
    }
  }

  /* "minorminer/busclique.pyx":211
 *             if rootdir.exists():
 *                 dirstack.append(rootdir)
 *         while dirstack:             # <<<<<<<<<<<<<<
 *             top = dirstack[-1]
 *             substack = []
 */
  while (1) {
    __pyx_t_10 = (PyList_GET_SIZE(__pyx_v_dirstack) != 0);
    if (!__pyx_t_10) break;

    /* "minorminer/busclique.pyx":212
 *                 dirstack.append(rootdir)
 *         while dirstack:
 *             top = dirstack[-1]             # <<<<<<<<<<<<<<
 *             substack = []
 *             for item in top.iterdir():
 */
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dirstack, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_top, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":213
 *         while dirstack:
 *             top = dirstack[-1]
 *             substack = []             # <<<<<<<<<<<<<<
 *             for item in top.iterdir():
 *                 if item.is_dir():
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_substack, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":214
 *             top = dirstack[-1]
 *             substack = []
 *             for item in top.iterdir():             # <<<<<<<<<<<<<<
 *                 if item.is_dir():
 *                     substack.append(item)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_top, __pyx_n_s_iterdir); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_6 = __pyx_t_1; __Pyx_INCREF(__pyx_t_6); __pyx_t_12 = 0;
      __pyx_t_13 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 214, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_13)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 214, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 214, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_13(__pyx_t_6);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 214, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":215
 *             substack = []
 *             for item in top.iterdir():
 *                 if item.is_dir():             # <<<<<<<<<<<<<<
 *                     substack.append(item)
 *                 else:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_is_dir); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_10) {

        /* "minorminer/busclique.pyx":216
 *             for item in top.iterdir():
 *                 if item.is_dir():
 *                     substack.append(item)             # <<<<<<<<<<<<<<
 *                 else:
 *                     item.unlink()
 */
        __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_substack, __pyx_v_item); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

        /* "minorminer/busclique.pyx":215
 *             substack = []
 *             for item in top.iterdir():
 *                 if item.is_dir():             # <<<<<<<<<<<<<<
 *                     substack.append(item)
 *                 else:
 */
        goto __pyx_L10;
      }

      /* "minorminer/busclique.pyx":218
 *                     substack.append(item)
 *                 else:
 *                     item.unlink()             # <<<<<<<<<<<<<<
 *             if substack:
 *                 dirstack.extend(substack)
 */
      /*else*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_unlink); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L10:;

      /* "minorminer/busclique.pyx":214
 *             top = dirstack[-1]
 *             substack = []
 *             for item in top.iterdir():             # <<<<<<<<<<<<<<
 *                 if item.is_dir():
 *                     substack.append(item)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "minorminer/busclique.pyx":219
 *                 else:
 *                     item.unlink()
 *             if substack:             # <<<<<<<<<<<<<<
 *                 dirstack.extend(substack)
 *             else:
 */
    __pyx_t_10 = (PyList_GET_SIZE(__pyx_v_substack) != 0);
    if (__pyx_t_10) {

      /* "minorminer/busclique.pyx":220
 *                     item.unlink()
 *             if substack:
 *                 dirstack.extend(substack)             # <<<<<<<<<<<<<<
 *             else:
 *                 top.rmdir()
 */
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_dirstack, __pyx_v_substack); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 220, __pyx_L1_error)

      /* "minorminer/busclique.pyx":219
 *                 else:
 *                     item.unlink()
 *             if substack:             # <<<<<<<<<<<<<<
 *                 dirstack.extend(substack)
 *             else:
 */
      goto __pyx_L11;
    }

    /* "minorminer/busclique.pyx":222
 *                 dirstack.extend(substack)
 *             else:
 *                 top.rmdir()             # <<<<<<<<<<<<<<
 *                 dirstack.pop()
 * 
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_top, __pyx_n_s_rmdir); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "minorminer/busclique.pyx":223
 *             else:
 *                 top.rmdir()
 *                 dirstack.pop()             # <<<<<<<<<<<<<<
 * 
 *     def get_clique_cache(self, sparse=False, compress=True):
 */
      __pyx_t_6 = __Pyx_PyList_Pop(__pyx_v_dirstack); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_L11:;
  }

  /* "minorminer/busclique.pyx":199
 * 
 *     @staticmethod
 *     def clear_all_caches():             # <<<<<<<<<<<<<<
 *         """Removes all caches created by this class, up to and including the
 *         current version.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.clear_all_caches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dirstack);
  __Pyx_XDECREF(__pyx_v_rootdir);
  __Pyx_XDECREF(__pyx_v_top);
  __Pyx_XDECREF(__pyx_v_substack);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":225
 *                 dirstack.pop()
 * 
 *     def get_clique_cache(self, sparse=False, compress=True):             # <<<<<<<<<<<<<<
 *         """Returns an object representing the cliques stored in this cache.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_11get_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_10get_clique_cache[] = "Returns an object representing the cliques stored in this cache.\n\n        Args:\n            sparse: (bool, optional, default=False)\n                If True, the cache will be a dictionary mapping embedding sizes\n                to lists of chains.  Otherwise, the cache will be a list of\n                lists of chains; the index of the outermost list corresponding\n                to embedding size.\n\n            compress: (bool, optional, default=True)\n                if True, the cache will be json-utf8-gzipped.  Otherwise, the\n                cache will be a python object whose type is determined by the\n                ``sparse`` argument.\n        Returns:\n            a cache object\n\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_11get_clique_cache = {"get_clique_cache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_11get_clique_cache, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_10get_clique_cache};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_11get_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_sparse = 0;
  PyObject *__pyx_v_compress = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_clique_cache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_sparse,&__pyx_n_s_compress,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)((PyObject *)Py_False));
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sparse);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compress);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_clique_cache") < 0)) __PYX_ERR(0, 225, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_sparse = values[1];
    __pyx_v_compress = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_clique_cache", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 225, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.get_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_10get_clique_cache(__pyx_self, __pyx_v_self, __pyx_v_sparse, __pyx_v_compress);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_10get_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sparse, PyObject *__pyx_v_compress) {
  PyObject *__pyx_v_raw = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_cache = NULL;
  PyObject *__pyx_7genexpr__pyx_v_i = NULL;
  PyObject *__pyx_7genexpr__pyx_v_s = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_clique_cache", 0);

  /* "minorminer/busclique.pyx":243
 * 
 *         """
 *         self._ensure_clique_cache()             # <<<<<<<<<<<<<<
 *         raw = self._cliques['raw']
 *         size = self._cliques['size']
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":244
 *         """
 *         self._ensure_clique_cache()
 *         raw = self._cliques['raw']             # <<<<<<<<<<<<<<
 *         size = self._cliques['size']
 *         if sparse:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_raw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_raw = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":245
 *         self._ensure_clique_cache()
 *         raw = self._cliques['raw']
 *         size = self._cliques['size']             # <<<<<<<<<<<<<<
 *         if sparse:
 *             cache = {i: raw[s][:i] for i, s in size.items()}
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":246
 *         raw = self._cliques['raw']
 *         size = self._cliques['size']
 *         if sparse:             # <<<<<<<<<<<<<<
 *             cache = {i: raw[s][:i] for i, s in size.items()}
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_sparse); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 246, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":247
 *         size = self._cliques['size']
 *         if sparse:
 *             cache = {i: raw[s][:i] for i, s in size.items()}             # <<<<<<<<<<<<<<
 *         else:
 *             cache = [raw[size[i]][:i] if i in size else () for i in range(max(size))]
 */
    { /* enter inner scope */
      __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = 0;
      if (unlikely(__pyx_v_size == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 247, __pyx_L6_error)
      }
      __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_size, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      while (1) {
        __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_6, &__pyx_t_5, &__pyx_t_3, &__pyx_t_8, NULL, __pyx_t_7);
        if (unlikely(__pyx_t_9 == 0)) break;
        if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 247, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_i, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_s, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_raw, __pyx_7genexpr__pyx_v_s); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 247, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_8, 0, 0, NULL, &__pyx_7genexpr__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_7genexpr__pyx_v_i, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 247, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_s); __pyx_7genexpr__pyx_v_s = 0;
      goto __pyx_L9_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_s); __pyx_7genexpr__pyx_v_s = 0;
      goto __pyx_L1_error;
      __pyx_L9_exit_scope:;
    } /* exit inner scope */
    __pyx_v_cache = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":246
 *         raw = self._cliques['raw']
 *         size = self._cliques['size']
 *         if sparse:             # <<<<<<<<<<<<<<
 *             cache = {i: raw[s][:i] for i, s in size.items()}
 *         else:
 */
    goto __pyx_L3;
  }

  /* "minorminer/busclique.pyx":249
 *             cache = {i: raw[s][:i] for i, s in size.items()}
 *         else:
 *             cache = [raw[size[i]][:i] if i in size else () for i in range(max(size))]             # <<<<<<<<<<<<<<
 *         if compress:
 *             cache = gzip.compress(dumps(cache).encode('utf8'))
 */
  /*else*/ {
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
        __pyx_t_10 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 249, __pyx_L12_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_10)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 249, __pyx_L12_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 249, __pyx_L12_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_10(__pyx_t_2);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 249, __pyx_L12_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_i, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr1__pyx_v_i, __pyx_v_size, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 249, __pyx_L12_error)
        if ((__pyx_t_4 != 0)) {
          __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_size, __pyx_8genexpr1__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 249, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_raw, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 249, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_GetSlice(__pyx_t_11, 0, 0, NULL, &__pyx_8genexpr1__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 249, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_3 = __pyx_t_8;
          __pyx_t_8 = 0;
        } else {
          __Pyx_INCREF(__pyx_empty_tuple);
          __pyx_t_3 = __pyx_empty_tuple;
        }
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 249, __pyx_L12_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i); __pyx_8genexpr1__pyx_v_i = 0;
      goto __pyx_L15_exit_scope;
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i); __pyx_8genexpr1__pyx_v_i = 0;
      goto __pyx_L1_error;
      __pyx_L15_exit_scope:;
    } /* exit inner scope */
    __pyx_v_cache = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "minorminer/busclique.pyx":250
 *         else:
 *             cache = [raw[size[i]][:i] if i in size else () for i in range(max(size))]
 *         if compress:             # <<<<<<<<<<<<<<
 *             cache = gzip.compress(dumps(cache).encode('utf8'))
 *         return cache
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compress); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 250, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":251
 *             cache = [raw[size[i]][:i] if i in size else () for i in range(max(size))]
 *         if compress:
 *             cache = gzip.compress(dumps(cache).encode('utf8'))             # <<<<<<<<<<<<<<
 *         return cache
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_gzip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compress); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_dumps); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_v_cache) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_v_cache);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_8, __pyx_n_u_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_n_u_utf8);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_11, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":250
 *         else:
 *             cache = [raw[size[i]][:i] if i in size else () for i in range(max(size))]
 *         if compress:             # <<<<<<<<<<<<<<
 *             cache = gzip.compress(dumps(cache).encode('utf8'))
 *         return cache
 */
  }

  /* "minorminer/busclique.pyx":252
 *         if compress:
 *             cache = gzip.compress(dumps(cache).encode('utf8'))
 *         return cache             # <<<<<<<<<<<<<<
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cache);
  __pyx_r = __pyx_v_cache;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":225
 *                 dirstack.pop()
 * 
 *     def get_clique_cache(self, sparse=False, compress=True):             # <<<<<<<<<<<<<<
 *         """Returns an object representing the cliques stored in this cache.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.get_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_cache);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_s);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":254
 *         return cache
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):             # <<<<<<<<<<<<<<
 *         """Replaces or sets the cliques stored by this cache with the provided
 *         cache object.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_13update_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_12update_clique_cache[] = "Replaces or sets the cliques stored by this cache with the provided\n        cache object.\n\n        Args:\n            cache: (bytes, str, dict or iterable)\n                The cache object to replace.  Should be an object returned by\n                :func:`self.get_clique_cache`.\n\n            write_to_disk: (bool, optional, default=True)\n                If True, the updated cache will be written to disk.\n                Otherwise, the cache will only be updated in memory.\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_13update_clique_cache = {"update_clique_cache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_13update_clique_cache, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_12update_clique_cache};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_13update_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_cache = 0;
  PyObject *__pyx_v_write_to_disk = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_clique_cache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_cache,&__pyx_n_s_write_to_disk,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cache)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_clique_cache", 0, 2, 3, 1); __PYX_ERR(0, 254, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write_to_disk);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "update_clique_cache") < 0)) __PYX_ERR(0, 254, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_cache = values[1];
    __pyx_v_write_to_disk = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update_clique_cache", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 254, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.update_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_12update_clique_cache(__pyx_self, __pyx_v_self, __pyx_v_cache, __pyx_v_write_to_disk);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":285
 * 
 *         if write_to_disk:
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)             # <<<<<<<<<<<<<<
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_19update_clique_cache_lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_19update_clique_cache_lambda = {"lambda", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_19update_clique_cache_lambda, METH_NOARGS, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_19update_clique_cache_lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *__pyx_cur_scope;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 0);
  __pyx_outer_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 285, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.update_clique_cache.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":254
 *         return cache
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):             # <<<<<<<<<<<<<<
 *         """Replaces or sets the cliques stored by this cache with the provided
 *         cache object.
 */

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_12update_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_cache, PyObject *__pyx_v_write_to_disk) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *__pyx_cur_scope;
  int __pyx_v_empty;
  CYTHON_UNUSED PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_emb = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_clique_cache", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct__update_clique_cache(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct__update_clique_cache, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 254, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_cache);

  /* "minorminer/busclique.pyx":267
 *                 Otherwise, the cache will only be updated in memory.
 *         """
 *         if isinstance(cache, bytes):             # <<<<<<<<<<<<<<
 *             cache = gzip.decompress(cache).decode('utf8')
 *         if isinstance(cache, str):
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_cache); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":268
 *         """
 *         if isinstance(cache, bytes):
 *             cache = gzip.decompress(cache).decode('utf8')             # <<<<<<<<<<<<<<
 *         if isinstance(cache, str):
 *             cache = loads(cache)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_gzip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_decompress); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_v_cache) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_cache);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_decode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_n_u_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_utf8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":267
 *                 Otherwise, the cache will only be updated in memory.
 *         """
 *         if isinstance(cache, bytes):             # <<<<<<<<<<<<<<
 *             cache = gzip.decompress(cache).decode('utf8')
 *         if isinstance(cache, str):
 */
  }

  /* "minorminer/busclique.pyx":269
 *         if isinstance(cache, bytes):
 *             cache = gzip.decompress(cache).decode('utf8')
 *         if isinstance(cache, str):             # <<<<<<<<<<<<<<
 *             cache = loads(cache)
 *         if isinstance(cache, dict):
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_cache); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "minorminer/busclique.pyx":270
 *             cache = gzip.decompress(cache).decode('utf8')
 *         if isinstance(cache, str):
 *             cache = loads(cache)             # <<<<<<<<<<<<<<
 *         if isinstance(cache, dict):
 *             cache = cache.items()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_loads); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_cache) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_cache);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":269
 *         if isinstance(cache, bytes):
 *             cache = gzip.decompress(cache).decode('utf8')
 *         if isinstance(cache, str):             # <<<<<<<<<<<<<<
 *             cache = loads(cache)
 *         if isinstance(cache, dict):
 */
  }

  /* "minorminer/busclique.pyx":271
 *         if isinstance(cache, str):
 *             cache = loads(cache)
 *         if isinstance(cache, dict):             # <<<<<<<<<<<<<<
 *             cache = cache.items()
 *         else:
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_cache); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":272
 *             cache = loads(cache)
 *         if isinstance(cache, dict):
 *             cache = cache.items()             # <<<<<<<<<<<<<<
 *         else:
 *             cache = enumerate(cache)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_items); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":271
 *         if isinstance(cache, str):
 *             cache = loads(cache)
 *         if isinstance(cache, dict):             # <<<<<<<<<<<<<<
 *             cache = cache.items()
 *         else:
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":274
 *             cache = cache.items()
 *         else:
 *             cache = enumerate(cache)             # <<<<<<<<<<<<<<
 * 
 *         empty = True
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_v_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L5:;

  /* "minorminer/busclique.pyx":276
 *             cache = enumerate(cache)
 * 
 *         empty = True             # <<<<<<<<<<<<<<
 *         for size, emb in cache:
 *             empty = False
 */
  __pyx_v_empty = 1;

  /* "minorminer/busclique.pyx":277
 * 
 *         empty = True
 *         for size, emb in cache:             # <<<<<<<<<<<<<<
 *             empty = False
 *             self.insert_clique_embedding(emb, write_to_disk=False, direct=True)
 */
  if (likely(PyList_CheckExact(__pyx_v_cache)) || PyTuple_CheckExact(__pyx_v_cache)) {
    __pyx_t_3 = __pyx_v_cache; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 277, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 277, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 277, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 277, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 277, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_emb, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":278
 *         empty = True
 *         for size, emb in cache:
 *             empty = False             # <<<<<<<<<<<<<<
 *             self.insert_clique_embedding(emb, write_to_disk=False, direct=True)
 * 
 */
    __pyx_v_empty = 0;

    /* "minorminer/busclique.pyx":279
 *         for size, emb in cache:
 *             empty = False
 *             self.insert_clique_embedding(emb, write_to_disk=False, direct=True)             # <<<<<<<<<<<<<<
 * 
 *         if empty:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_insert_clique_embedding); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_emb);
    __Pyx_GIVEREF(__pyx_v_emb);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_emb);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_write_to_disk, Py_False) < 0) __PYX_ERR(0, 279, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_direct, Py_True) < 0) __PYX_ERR(0, 279, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "minorminer/busclique.pyx":277
 * 
 *         empty = True
 *         for size, emb in cache:             # <<<<<<<<<<<<<<
 *             empty = False
 *             self.insert_clique_embedding(emb, write_to_disk=False, direct=True)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":281
 *             self.insert_clique_embedding(emb, write_to_disk=False, direct=True)
 * 
 *         if empty:             # <<<<<<<<<<<<<<
 *             self._ensure_clique_cache(blank=True)
 * 
 */
  __pyx_t_2 = (__pyx_v_empty != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":282
 * 
 *         if empty:
 *             self._ensure_clique_cache(blank=True)             # <<<<<<<<<<<<<<
 * 
 *         if write_to_disk:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_blank, Py_True) < 0) __PYX_ERR(0, 282, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":281
 *             self.insert_clique_embedding(emb, write_to_disk=False, direct=True)
 * 
 *         if empty:             # <<<<<<<<<<<<<<
 *             self._ensure_clique_cache(blank=True)
 * 
 */
  }

  /* "minorminer/busclique.pyx":284
 *             self._ensure_clique_cache(blank=True)
 * 
 *         if write_to_disk:             # <<<<<<<<<<<<<<
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_write_to_disk); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":285
 * 
 *         if write_to_disk:
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)             # <<<<<<<<<<<<<<
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fetch_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_19update_clique_cache_lambda, 0, __pyx_n_s_busgraph_cache_update_clique_cac, ((PyObject*)__pyx_cur_scope), __pyx_n_s_minorminer_busclique, __pyx_d, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_u_clique);
    __Pyx_GIVEREF(__pyx_n_u_clique);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_clique);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_force_write, Py_True) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":284
 *             self._ensure_clique_cache(blank=True)
 * 
 *         if write_to_disk:             # <<<<<<<<<<<<<<
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 */
  }

  /* "minorminer/busclique.pyx":254
 *         return cache
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):             # <<<<<<<<<<<<<<
 *         """Replaces or sets the cliques stored by this cache with the provided
 *         cache object.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.update_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XDECREF(__pyx_v_cache);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":287
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):             # <<<<<<<<<<<<<<
 *         """Inserts a clique embedding into the cache.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_15insert_clique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_14insert_clique_embedding[] = "Inserts a clique embedding into the cache.\n\n        Args:\n            emb: (dict or list)\n                The embedding to insert into the cache.\n\n            write_to_disk: (bool, optional, default=True)\n                If True, the updated cache will be written to disk.\n                Otherwise, the cache will only be updated in memory.\n\n            quality_function: (callable or None)\n                A function used to evaluate the quality of an embedding before\n                incorporating it into the cache.  If ``quality_function`` is\n                None, preference is given to the embedding with shortest maximum\n                chain length.  The embeddings evaluated by ``quality_function``\n                will be provided as a list of list of integers, corresponding\n                to the integer-labeled graph underlying ``self``.\n\n            direct: (bool, optional, default=False)\n                If True, the embedding will be stored directly into the cache,\n                overwriting the current clique of that size if it exists.  In\n                this case, ``quality_function`` must be ``None``.\n                Otherwise, the embedding and embeddings derived by removing\n                chains in decreasing order of length may be stored into the\n                cache, if ``quality function`` evaluated on the new embedding is\n                greater than the old.  If ``quality_function`` is None, new\n                embeddings are retained when their chainlength is shorter than\n                the old.\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_15insert_clique_embedding = {"insert_clique_embedding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_15insert_clique_embedding, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_14insert_clique_embedding};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_15insert_clique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_emb = 0;
  PyObject *__pyx_v_write_to_disk = 0;
  PyObject *__pyx_v_quality_function = 0;
  PyObject *__pyx_v_direct = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert_clique_embedding (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_emb,&__pyx_n_s_write_to_disk,&__pyx_n_s_quality_function,&__pyx_n_s_direct,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_emb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert_clique_embedding", 0, 2, 5, 1); __PYX_ERR(0, 287, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write_to_disk);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quality_function);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_direct);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert_clique_embedding") < 0)) __PYX_ERR(0, 287, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_emb = values[1];
    __pyx_v_write_to_disk = values[2];
    __pyx_v_quality_function = values[3];
    __pyx_v_direct = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert_clique_embedding", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 287, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.insert_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_14insert_clique_embedding(__pyx_self, __pyx_v_self, __pyx_v_emb, __pyx_v_write_to_disk, __pyx_v_quality_function, __pyx_v_direct);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":334
 *                 by_length[length] = key
 *             if write_to_disk:
 *                 self._fetch_cache('clique', lambda: self._cliques, force_write=True)             # <<<<<<<<<<<<<<
 *         else:
 *             max_length = max(map(len, emb)) if emb else 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_23insert_clique_embedding_lambda1(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_23insert_clique_embedding_lambda1 = {"lambda1", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_23insert_clique_embedding_lambda1, METH_NOARGS, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_23insert_clique_embedding_lambda1(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda1(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda1(PyObject *__pyx_self) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *__pyx_cur_scope;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __pyx_outer_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 334, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.insert_clique_embedding.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":287
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):             # <<<<<<<<<<<<<<
 *         """Inserts a clique embedding into the cache.
 * 
 */

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_14insert_clique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_emb, PyObject *__pyx_v_write_to_disk, PyObject *__pyx_v_quality_function, PyObject *__pyx_v_direct) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *__pyx_cur_scope;
  PyObject *__pyx_v_raw = NULL;
  Py_ssize_t __pyx_v_key;
  PyObject *__pyx_v_by_length = NULL;
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_v_max_length = NULL;
  PyObject *__pyx_v_emb_list = NULL;
  PyObject *__pyx_v_cache = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_i = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_t_16;
  int __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insert_clique_embedding", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 287, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_emb);

  /* "minorminer/busclique.pyx":317
 *                 the old.
 *         """
 *         if not isinstance(emb, dict):             # <<<<<<<<<<<<<<
 *             emb = dict(enumerate(emb))
 *         emb = self._graph.delabel(emb)
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_emb); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":318
 *         """
 *         if not isinstance(emb, dict):
 *             emb = dict(enumerate(emb))             # <<<<<<<<<<<<<<
 *         emb = self._graph.delabel(emb)
 *         emb = sorted(emb.values(), key=len)
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_v_emb); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_emb, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":317
 *                 the old.
 *         """
 *         if not isinstance(emb, dict):             # <<<<<<<<<<<<<<
 *             emb = dict(enumerate(emb))
 *         emb = self._graph.delabel(emb)
 */
  }

  /* "minorminer/busclique.pyx":319
 *         if not isinstance(emb, dict):
 *             emb = dict(enumerate(emb))
 *         emb = self._graph.delabel(emb)             # <<<<<<<<<<<<<<
 *         emb = sorted(emb.values(), key=len)
 *         if direct:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_delabel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_emb) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_emb);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_emb, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "minorminer/busclique.pyx":320
 *             emb = dict(enumerate(emb))
 *         emb = self._graph.delabel(emb)
 *         emb = sorted(emb.values(), key=len)             # <<<<<<<<<<<<<<
 *         if direct:
 *             if quality_function is not None:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_emb, __pyx_n_s_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_emb, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":321
 *         emb = self._graph.delabel(emb)
 *         emb = sorted(emb.values(), key=len)
 *         if direct:             # <<<<<<<<<<<<<<
 *             if quality_function is not None:
 *                 raise ValueError("quality_function is ignored for direct insertion")
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_direct); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 321, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":322
 *         emb = sorted(emb.values(), key=len)
 *         if direct:
 *             if quality_function is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError("quality_function is ignored for direct insertion")
 *             self._ensure_clique_cache(blank=True)
 */
    __pyx_t_2 = (__pyx_v_quality_function != Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (unlikely(__pyx_t_1)) {

      /* "minorminer/busclique.pyx":323
 *         if direct:
 *             if quality_function is not None:
 *                 raise ValueError("quality_function is ignored for direct insertion")             # <<<<<<<<<<<<<<
 *             self._ensure_clique_cache(blank=True)
 *             raw = self._cliques['raw']
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 323, __pyx_L1_error)

      /* "minorminer/busclique.pyx":322
 *         emb = sorted(emb.values(), key=len)
 *         if direct:
 *             if quality_function is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError("quality_function is ignored for direct insertion")
 *             self._ensure_clique_cache(blank=True)
 */
    }

    /* "minorminer/busclique.pyx":324
 *             if quality_function is not None:
 *                 raise ValueError("quality_function is ignored for direct insertion")
 *             self._ensure_clique_cache(blank=True)             # <<<<<<<<<<<<<<
 *             raw = self._cliques['raw']
 *             key = len(raw)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_blank, Py_True) < 0) __PYX_ERR(0, 324, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":325
 *                 raise ValueError("quality_function is ignored for direct insertion")
 *             self._ensure_clique_cache(blank=True)
 *             raw = self._cliques['raw']             # <<<<<<<<<<<<<<
 *             key = len(raw)
 *             raw.append(emb)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_u_raw); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_raw = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":326
 *             self._ensure_clique_cache(blank=True)
 *             raw = self._cliques['raw']
 *             key = len(raw)             # <<<<<<<<<<<<<<
 *             raw.append(emb)
 *             by_length = self._cliques['length']
 */
    __pyx_t_6 = PyObject_Length(__pyx_v_raw); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 326, __pyx_L1_error)
    __pyx_v_key = __pyx_t_6;

    /* "minorminer/busclique.pyx":327
 *             raw = self._cliques['raw']
 *             key = len(raw)
 *             raw.append(emb)             # <<<<<<<<<<<<<<
 *             by_length = self._cliques['length']
 *             self._cliques['size'][len(emb)] = key
 */
    __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_raw, __pyx_v_emb); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 327, __pyx_L1_error)

    /* "minorminer/busclique.pyx":328
 *             key = len(raw)
 *             raw.append(emb)
 *             by_length = self._cliques['length']             # <<<<<<<<<<<<<<
 *             self._cliques['size'][len(emb)] = key
 *             length = len(emb[-1]) if emb else 0
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_u_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_by_length = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":329
 *             raw.append(emb)
 *             by_length = self._cliques['length']
 *             self._cliques['size'][len(emb)] = key             # <<<<<<<<<<<<<<
 *             length = len(emb[-1]) if emb else 0
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):
 */
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_u_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = PyObject_Length(__pyx_v_emb); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 329, __pyx_L1_error)
    if (unlikely(__Pyx_SetItemInt(__pyx_t_3, __pyx_t_6, __pyx_t_5, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":330
 *             by_length = self._cliques['length']
 *             self._cliques['size'][len(emb)] = key
 *             length = len(emb[-1]) if emb else 0             # <<<<<<<<<<<<<<
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):
 *                 by_length[length] = key
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_emb); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 330, __pyx_L1_error)
    if (__pyx_t_1) {
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_emb, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __pyx_t_8;
    } else {
      __pyx_t_6 = 0;
    }
    __pyx_v_length = __pyx_t_6;

    /* "minorminer/busclique.pyx":331
 *             self._cliques['size'][len(emb)] = key
 *             length = len(emb[-1]) if emb else 0
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):             # <<<<<<<<<<<<<<
 *                 by_length[length] = key
 *             if write_to_disk:
 */
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_v_by_length, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_9 = (__pyx_t_2 != 0);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_1 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_by_length, __pyx_v_length, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_raw, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = PyObject_Length(__pyx_v_emb); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 331, __pyx_L1_error)
    __pyx_t_9 = ((__pyx_t_6 < __pyx_t_8) != 0);
    __pyx_t_1 = __pyx_t_9;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "minorminer/busclique.pyx":332
 *             length = len(emb[-1]) if emb else 0
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):
 *                 by_length[length] = key             # <<<<<<<<<<<<<<
 *             if write_to_disk:
 *                 self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_SetItemInt(__pyx_v_by_length, __pyx_v_length, __pyx_t_3, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "minorminer/busclique.pyx":331
 *             self._cliques['size'][len(emb)] = key
 *             length = len(emb[-1]) if emb else 0
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):             # <<<<<<<<<<<<<<
 *                 by_length[length] = key
 *             if write_to_disk:
 */
    }

    /* "minorminer/busclique.pyx":333
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):
 *                 by_length[length] = key
 *             if write_to_disk:             # <<<<<<<<<<<<<<
 *                 self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_write_to_disk); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 333, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "minorminer/busclique.pyx":334
 *                 by_length[length] = key
 *             if write_to_disk:
 *                 self._fetch_cache('clique', lambda: self._cliques, force_write=True)             # <<<<<<<<<<<<<<
 *         else:
 *             max_length = max(map(len, emb)) if emb else 0
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fetch_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_23insert_clique_embedding_lambda1, 0, __pyx_n_s_busgraph_cache_insert_clique_emb, ((PyObject*)__pyx_cur_scope), __pyx_n_s_minorminer_busclique, __pyx_d, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_u_clique);
      __Pyx_GIVEREF(__pyx_n_u_clique);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_clique);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_force_write, Py_True) < 0) __PYX_ERR(0, 334, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "minorminer/busclique.pyx":333
 *             if length not in by_length or len(raw[by_length[length]]) < len(emb):
 *                 by_length[length] = key
 *             if write_to_disk:             # <<<<<<<<<<<<<<
 *                 self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 *         else:
 */
    }

    /* "minorminer/busclique.pyx":321
 *         emb = self._graph.delabel(emb)
 *         emb = sorted(emb.values(), key=len)
 *         if direct:             # <<<<<<<<<<<<<<
 *             if quality_function is not None:
 *                 raise ValueError("quality_function is ignored for direct insertion")
 */
    goto __pyx_L4;
  }

  /* "minorminer/busclique.pyx":336
 *                 self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 *         else:
 *             max_length = max(map(len, emb)) if emb else 0             # <<<<<<<<<<<<<<
 *             emb_list = [[c for c in emb if len(c) <= i] for i in range(max_length+1)]
 *             cache = _make_clique_cache(emb_list)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_emb); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
    if (__pyx_t_1) {
      __pyx_t_5 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_emb);
      __Pyx_GIVEREF(__pyx_v_emb);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_emb);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = __pyx_t_4;
      __pyx_t_4 = 0;
    } else {
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_10 = __pyx_int_0;
    }
    __pyx_v_max_length = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":337
 *         else:
 *             max_length = max(map(len, emb)) if emb else 0
 *             emb_list = [[c for c in emb if len(c) <= i] for i in range(max_length+1)]             # <<<<<<<<<<<<<<
 *             cache = _make_clique_cache(emb_list)
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 */
    { /* enter inner scope */
      __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 337, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_max_length, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
        __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4); __pyx_t_8 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 337, __pyx_L12_error)
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 337, __pyx_L12_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 337, __pyx_L12_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_11(__pyx_t_4);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 337, __pyx_L12_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_i, __pyx_t_5);
        __pyx_t_5 = 0;
        { /* enter inner scope */
          __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (likely(PyList_CheckExact(__pyx_v_emb)) || PyTuple_CheckExact(__pyx_v_emb)) {
            __pyx_t_3 = __pyx_v_emb; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
            __pyx_t_12 = NULL;
          } else {
            __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_emb); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_12 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 337, __pyx_L17_error)
          }
          for (;;) {
            if (likely(!__pyx_t_12)) {
              if (likely(PyList_CheckExact(__pyx_t_3))) {
                if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_13 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_13); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 337, __pyx_L17_error)
                #else
                __pyx_t_13 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 337, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_13);
                #endif
              } else {
                if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_13); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 337, __pyx_L17_error)
                #else
                __pyx_t_13 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 337, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_13);
                #endif
              }
            } else {
              __pyx_t_13 = __pyx_t_12(__pyx_t_3);
              if (unlikely(!__pyx_t_13)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 337, __pyx_L17_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_13);
            }
            __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_c, __pyx_t_13);
            __pyx_t_13 = 0;
            __pyx_t_14 = PyObject_Length(__pyx_8genexpr3__pyx_v_c); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 337, __pyx_L17_error)
            __pyx_t_13 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 337, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_15 = PyObject_RichCompare(__pyx_t_13, __pyx_8genexpr2__pyx_v_i, Py_LE); __Pyx_XGOTREF(__pyx_t_15); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 337, __pyx_L17_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_15); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 337, __pyx_L17_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (__pyx_t_1) {
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_8genexpr3__pyx_v_c))) __PYX_ERR(0, 337, __pyx_L17_error)
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_c); __pyx_8genexpr3__pyx_v_c = 0;
          goto __pyx_L21_exit_scope;
          __pyx_L17_error:;
          __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_c); __pyx_8genexpr3__pyx_v_c = 0;
          goto __pyx_L12_error;
          __pyx_L21_exit_scope:;
        } /* exit inner scope */
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_10, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 337, __pyx_L12_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_i); __pyx_8genexpr2__pyx_v_i = 0;
      goto __pyx_L22_exit_scope;
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_i); __pyx_8genexpr2__pyx_v_i = 0;
      goto __pyx_L1_error;
      __pyx_L22_exit_scope:;
    } /* exit inner scope */
    __pyx_v_emb_list = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":338
 *             max_length = max(map(len, emb)) if emb else 0
 *             emb_list = [[c for c in emb if len(c) <= i] for i in range(max_length+1)]
 *             cache = _make_clique_cache(emb_list)             # <<<<<<<<<<<<<<
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 */
    __pyx_t_16 = __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(__pyx_v_emb_list); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 338, __pyx_L1_error)
    __pyx_t_10 = __pyx_f_10minorminer_9busclique__make_clique_cache(__pyx_t_16); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_v_cache = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":339
 *             emb_list = [[c for c in emb if len(c) <= i] for i in range(max_length+1)]
 *             cache = _make_clique_cache(emb_list)
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)             # <<<<<<<<<<<<<<
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_merge_clique_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_17 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_17 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_cache, __pyx_v_quality_function, __pyx_v_write_to_disk};
      __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_10);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_cache, __pyx_v_quality_function, __pyx_v_write_to_disk};
      __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_10);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_cache);
      __Pyx_GIVEREF(__pyx_v_cache);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_17, __pyx_v_cache);
      __Pyx_INCREF(__pyx_v_quality_function);
      __Pyx_GIVEREF(__pyx_v_quality_function);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_17, __pyx_v_quality_function);
      __Pyx_INCREF(__pyx_v_write_to_disk);
      __Pyx_GIVEREF(__pyx_v_write_to_disk);
      PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_17, __pyx_v_write_to_disk);
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __pyx_L4:;

  /* "minorminer/busclique.pyx":287
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):             # <<<<<<<<<<<<<<
 *         """Inserts a clique embedding into the cache.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.insert_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_by_length);
  __Pyx_XDECREF(__pyx_v_max_length);
  __Pyx_XDECREF(__pyx_v_emb_list);
  __Pyx_XDECREF(__pyx_v_cache);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_i);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":341
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):             # <<<<<<<<<<<<<<
 *         """Merges the clique cache of another ``busgraph_cache`` object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_17merge_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_16merge_clique_cache[] = "Merges the clique cache of another ``busgraph_cache`` object\n\n        Args:\n            cache: (busgraph_cache or object)\n                If ``cache`` is a ``busgraph_cache``, its clique cache will be\n                incorporated that of ``self``.  Otherwise, the cache object is\n                an undocumented, internal cache representation (do not use).\n\n            quality_function: (callable or None)\n                A function used to evaluate the quality of an embedding before\n                incorporating it into the cache.  If ``quality_function`` is\n                None, preference is given to the embedding with shortest maximum\n                chain length.  The embeddings evaluated by ``quality_function``\n                will be provided as a list of list of integers, corresponding\n                to the integer-labeled graph underlying ``self``.\n\n            write_to_disk: (bool, optional, default=True)\n                If True, the merged cache will be written to disk.\n                Otherwise, the cache will only be updated in memory.\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_17merge_clique_cache = {"merge_clique_cache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_17merge_clique_cache, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_16merge_clique_cache};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_17merge_clique_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_cache = 0;
  PyObject *__pyx_v_quality_function = 0;
  PyObject *__pyx_v_write_to_disk = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("merge_clique_cache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_cache,&__pyx_n_s_quality_function,&__pyx_n_s_write_to_disk,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cache)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("merge_clique_cache", 0, 2, 4, 1); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quality_function);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write_to_disk);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "merge_clique_cache") < 0)) __PYX_ERR(0, 341, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_cache = values[1];
    __pyx_v_quality_function = values[2];
    __pyx_v_write_to_disk = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("merge_clique_cache", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 341, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.merge_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_16merge_clique_cache(__pyx_self, __pyx_v_self, __pyx_v_cache, __pyx_v_quality_function, __pyx_v_write_to_disk);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":419
 *         self._cliques = {"raw": raw, "size": size, "length": length}
 *         if write_to_disk:
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)             # <<<<<<<<<<<<<<
 * 
 *     def _fetch_cache(self, dirname, compute, force_write=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_18merge_clique_cache_lambda2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_18merge_clique_cache_lambda2 = {"lambda2", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_18merge_clique_cache_lambda2, METH_NOARGS, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_18merge_clique_cache_lambda2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda2 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda2(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda2(PyObject *__pyx_self) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *__pyx_cur_scope;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda2", 0);
  __pyx_outer_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 419, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.merge_clique_cache.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":341
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):             # <<<<<<<<<<<<<<
 *         """Merges the clique cache of another ``busgraph_cache`` object
 * 
 */

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_16merge_clique_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_cache, PyObject *__pyx_v_quality_function, PyObject *__pyx_v_write_to_disk) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *__pyx_cur_scope;
  PyObject *__pyx_v_raw0 = NULL;
  PyObject *__pyx_v_raw1 = NULL;
  PyObject *__pyx_v_used = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_size0 = NULL;
  PyObject *__pyx_v_size1 = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_key0 = NULL;
  PyObject *__pyx_v_key1 = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_qual0 = NULL;
  PyObject *__pyx_v_qual1 = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_v_length0 = NULL;
  PyObject *__pyx_v_length1 = NULL;
  PyObject *__pyx_v_raw = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_key = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("merge_clique_cache", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 341, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_cache);

  /* "minorminer/busclique.pyx":363
 *         """
 *         cdef vector[embedding_t] embs
 *         self._ensure_clique_cache(blank=True)             # <<<<<<<<<<<<<<
 *         if isinstance(cache, busgraph_cache):
 *             cache._ensure_clique_cache()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_blank, Py_True) < 0) __PYX_ERR(0, 363, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":364
 *         cdef vector[embedding_t] embs
 *         self._ensure_clique_cache(blank=True)
 *         if isinstance(cache, busgraph_cache):             # <<<<<<<<<<<<<<
 *             cache._ensure_clique_cache()
 *             cache = cache._cliques
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_busgraph_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_cache, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "minorminer/busclique.pyx":365
 *         self._ensure_clique_cache(blank=True)
 *         if isinstance(cache, busgraph_cache):
 *             cache._ensure_clique_cache()             # <<<<<<<<<<<<<<
 *             cache = cache._cliques
 *         raw0 = cache['raw']
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":366
 *         if isinstance(cache, busgraph_cache):
 *             cache._ensure_clique_cache()
 *             cache = cache._cliques             # <<<<<<<<<<<<<<
 *         raw0 = cache['raw']
 *         raw1 = self._cliques['raw']
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_cliques); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":364
 *         cdef vector[embedding_t] embs
 *         self._ensure_clique_cache(blank=True)
 *         if isinstance(cache, busgraph_cache):             # <<<<<<<<<<<<<<
 *             cache._ensure_clique_cache()
 *             cache = cache._cliques
 */
  }

  /* "minorminer/busclique.pyx":367
 *             cache._ensure_clique_cache()
 *             cache = cache._cliques
 *         raw0 = cache['raw']             # <<<<<<<<<<<<<<
 *         raw1 = self._cliques['raw']
 *         used = {}
 */
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_cache, __pyx_n_u_raw); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_raw0 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":368
 *             cache = cache._cliques
 *         raw0 = cache['raw']
 *         raw1 = self._cliques['raw']             # <<<<<<<<<<<<<<
 *         used = {}
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_raw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_raw1 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":369
 *         raw0 = cache['raw']
 *         raw1 = self._cliques['raw']
 *         used = {}             # <<<<<<<<<<<<<<
 * 
 *         size = {}
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_used = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":371
 *         used = {}
 * 
 *         size = {}             # <<<<<<<<<<<<<<
 *         size0 = cache['size']
 *         size1 = self._cliques['size']
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_size = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":372
 * 
 *         size = {}
 *         size0 = cache['size']             # <<<<<<<<<<<<<<
 *         size1 = self._cliques['size']
 *         for i in sorted(set(size0).union(size1)):
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_cache, __pyx_n_u_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_size0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":373
 *         size = {}
 *         size0 = cache['size']
 *         size1 = self._cliques['size']             # <<<<<<<<<<<<<<
 *         for i in sorted(set(size0).union(size1)):
 *             key0 = size0.get(i)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size1 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":374
 *         size0 = cache['size']
 *         size1 = self._cliques['size']
 *         for i in sorted(set(size0).union(size1)):             # <<<<<<<<<<<<<<
 *             key0 = size0.get(i)
 *             key1 = size1.get(i)
 */
  __pyx_t_1 = PySet_New(__pyx_v_size0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_union); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_v_size1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_size1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_7 = PyList_Sort(__pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 374, __pyx_L1_error)
  if (unlikely(__pyx_t_3 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 374, __pyx_L1_error)
  }
  __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_6)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 374, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":375
 *         size1 = self._cliques['size']
 *         for i in sorted(set(size0).union(size1)):
 *             key0 = size0.get(i)             # <<<<<<<<<<<<<<
 *             key1 = size1.get(i)
 *             if key1 is None:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_size0, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_i);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key0, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":376
 *         for i in sorted(set(size0).union(size1)):
 *             key0 = size0.get(i)
 *             key1 = size1.get(i)             # <<<<<<<<<<<<<<
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_size1, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_i);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":377
 *             key0 = size0.get(i)
 *             key1 = size1.get(i)
 *             if key1 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:
 */
    __pyx_t_5 = (__pyx_v_key1 == Py_None);
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "minorminer/busclique.pyx":378
 *             key1 = size1.get(i)
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))             # <<<<<<<<<<<<<<
 *             elif key0 is None:
 *                 key = used.setdefault((key1, 1), len(used))
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_key0);
      __Pyx_GIVEREF(__pyx_v_key0);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_key0);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
      __pyx_t_9 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 378, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_3, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":377
 *             key0 = size0.get(i)
 *             key1 = size1.get(i)
 *             if key1 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:
 */
      goto __pyx_L6;
    }

    /* "minorminer/busclique.pyx":379
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key1, 1), len(used))
 *             else:
 */
    __pyx_t_4 = (__pyx_v_key0 == Py_None);
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "minorminer/busclique.pyx":380
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:
 *                 key = used.setdefault((key1, 1), len(used))             # <<<<<<<<<<<<<<
 *             else:
 *                 if i == 0:
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_key1);
      __Pyx_GIVEREF(__pyx_v_key1);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_key1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_1);
      __pyx_t_9 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 380, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_1, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "minorminer/busclique.pyx":379
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key1, 1), len(used))
 *             else:
 */
      goto __pyx_L6;
    }

    /* "minorminer/busclique.pyx":382
 *                 key = used.setdefault((key1, 1), len(used))
 *             else:
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     qual0 = qual1 = 0
 *                 elif quality_function is None:
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_i, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_5) {

        /* "minorminer/busclique.pyx":383
 *             else:
 *                 if i == 0:
 *                     qual0 = qual1 = 0             # <<<<<<<<<<<<<<
 *                 elif quality_function is None:
 *                     # by default, shortest chainlength is highest quality
 */
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_XDECREF_SET(__pyx_v_qual0, __pyx_int_0);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_XDECREF_SET(__pyx_v_qual1, __pyx_int_0);

        /* "minorminer/busclique.pyx":382
 *                 key = used.setdefault((key1, 1), len(used))
 *             else:
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     qual0 = qual1 = 0
 *                 elif quality_function is None:
 */
        goto __pyx_L7;
      }

      /* "minorminer/busclique.pyx":384
 *                 if i == 0:
 *                     qual0 = qual1 = 0
 *                 elif quality_function is None:             # <<<<<<<<<<<<<<
 *                     # by default, shortest chainlength is highest quality
 *                     qual0 = -len(raw0[key0][i-1])
 */
      __pyx_t_5 = (__pyx_v_quality_function == Py_None);
      __pyx_t_4 = (__pyx_t_5 != 0);
      if (__pyx_t_4) {

        /* "minorminer/busclique.pyx":386
 *                 elif quality_function is None:
 *                     # by default, shortest chainlength is highest quality
 *                     qual0 = -len(raw0[key0][i-1])             # <<<<<<<<<<<<<<
 *                     qual1 = -len(raw1[key1][i-1])
 *                 else:
 */
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_raw0, __pyx_v_key0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_9 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyInt_FromSsize_t((-__pyx_t_9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_qual0, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "minorminer/busclique.pyx":387
 *                     # by default, shortest chainlength is highest quality
 *                     qual0 = -len(raw0[key0][i-1])
 *                     qual1 = -len(raw1[key1][i-1])             # <<<<<<<<<<<<<<
 *                 else:
 *                     qual0 = quality_function(raw0[key0][:i])
 */
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_raw1, __pyx_v_key1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_9 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyInt_FromSsize_t((-__pyx_t_9)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_qual1, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "minorminer/busclique.pyx":384
 *                 if i == 0:
 *                     qual0 = qual1 = 0
 *                 elif quality_function is None:             # <<<<<<<<<<<<<<
 *                     # by default, shortest chainlength is highest quality
 *                     qual0 = -len(raw0[key0][i-1])
 */
        goto __pyx_L7;
      }

      /* "minorminer/busclique.pyx":389
 *                     qual1 = -len(raw1[key1][i-1])
 *                 else:
 *                     qual0 = quality_function(raw0[key0][:i])             # <<<<<<<<<<<<<<
 *                     qual1 = quality_function(raw1[key1][:i])
 *                 if qual0 > qual1:
 */
      /*else*/ {
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_raw0, __pyx_v_key0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_v_quality_function);
        __pyx_t_2 = __pyx_v_quality_function; __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_qual0, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "minorminer/busclique.pyx":390
 *                 else:
 *                     qual0 = quality_function(raw0[key0][:i])
 *                     qual1 = quality_function(raw1[key1][:i])             # <<<<<<<<<<<<<<
 *                 if qual0 > qual1:
 *                     key = used.setdefault((key0, 0), len(used))
 */
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_raw1, __pyx_v_key1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_v_quality_function);
        __pyx_t_2 = __pyx_v_quality_function; __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_qual1, __pyx_t_3);
        __pyx_t_3 = 0;
      }
      __pyx_L7:;

      /* "minorminer/busclique.pyx":391
 *                     qual0 = quality_function(raw0[key0][:i])
 *                     qual1 = quality_function(raw1[key1][:i])
 *                 if qual0 > qual1:             # <<<<<<<<<<<<<<
 *                     key = used.setdefault((key0, 0), len(used))
 *                 else:
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_qual0, __pyx_v_qual1, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_4) {

        /* "minorminer/busclique.pyx":392
 *                     qual1 = quality_function(raw1[key1][:i])
 *                 if qual0 > qual1:
 *                     key = used.setdefault((key0, 0), len(used))             # <<<<<<<<<<<<<<
 *                 else:
 *                     # if the quality is the same, retain the old embedding
 */
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_key0);
        __Pyx_GIVEREF(__pyx_v_key0);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_key0);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
        __pyx_t_9 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 392, __pyx_L1_error)
        __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_3, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "minorminer/busclique.pyx":391
 *                     qual0 = quality_function(raw0[key0][:i])
 *                     qual1 = quality_function(raw1[key1][:i])
 *                 if qual0 > qual1:             # <<<<<<<<<<<<<<
 *                     key = used.setdefault((key0, 0), len(used))
 *                 else:
 */
        goto __pyx_L8;
      }

      /* "minorminer/busclique.pyx":395
 *                 else:
 *                     # if the quality is the same, retain the old embedding
 *                     key = used.setdefault((key1, 1), len(used))             # <<<<<<<<<<<<<<
 *             size[i] = key
 * 
 */
      /*else*/ {
        __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_key1);
        __Pyx_GIVEREF(__pyx_v_key1);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_key1);
        __Pyx_INCREF(__pyx_int_1);
        __Pyx_GIVEREF(__pyx_int_1);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_1);
        __pyx_t_9 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 395, __pyx_L1_error)
        __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_1, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
        __pyx_t_3 = 0;
      }
      __pyx_L8:;
    }
    __pyx_L6:;

    /* "minorminer/busclique.pyx":396
 *                     # if the quality is the same, retain the old embedding
 *                     key = used.setdefault((key1, 1), len(used))
 *             size[i] = key             # <<<<<<<<<<<<<<
 * 
 *         length = {}
 */
    if (unlikely(PyDict_SetItem(__pyx_v_size, __pyx_v_i, __pyx_v_key) < 0)) __PYX_ERR(0, 396, __pyx_L1_error)

    /* "minorminer/busclique.pyx":374
 *         size0 = cache['size']
 *         size1 = self._cliques['size']
 *         for i in sorted(set(size0).union(size1)):             # <<<<<<<<<<<<<<
 *             key0 = size0.get(i)
 *             key1 = size1.get(i)
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "minorminer/busclique.pyx":398
 *             size[i] = key
 * 
 *         length = {}             # <<<<<<<<<<<<<<
 *         length0 = cache['length']
 *         length1 = self._cliques['length']
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_length = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "minorminer/busclique.pyx":399
 * 
 *         length = {}
 *         length0 = cache['length']             # <<<<<<<<<<<<<<
 *         length1 = self._cliques['length']
 *         for i in sorted(set(length0).union(length1)):
 */
  __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_cache, __pyx_n_u_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_length0 = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "minorminer/busclique.pyx":400
 *         length = {}
 *         length0 = cache['length']
 *         length1 = self._cliques['length']             # <<<<<<<<<<<<<<
 *         for i in sorted(set(length0).union(length1)):
 *             key0 = length0.get(i)
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_length1 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":401
 *         length0 = cache['length']
 *         length1 = self._cliques['length']
 *         for i in sorted(set(length0).union(length1)):             # <<<<<<<<<<<<<<
 *             key0 = length0.get(i)
 *             key1 = length1.get(i)
 */
  __pyx_t_2 = PySet_New(__pyx_v_length0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_union); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_length1) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_length1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_7 = PyList_Sort(__pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 401, __pyx_L1_error)
  if (unlikely(__pyx_t_3 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 401, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 401, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":402
 *         length1 = self._cliques['length']
 *         for i in sorted(set(length0).union(length1)):
 *             key0 = length0.get(i)             # <<<<<<<<<<<<<<
 *             key1 = length1.get(i)
 *             if key1 is None:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_length0, __pyx_n_s_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_i);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key0, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":403
 *         for i in sorted(set(length0).union(length1)):
 *             key0 = length0.get(i)
 *             key1 = length1.get(i)             # <<<<<<<<<<<<<<
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_length1, __pyx_n_s_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_i);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":404
 *             key0 = length0.get(i)
 *             key1 = length1.get(i)
 *             if key1 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:
 */
    __pyx_t_4 = (__pyx_v_key1 == Py_None);
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "minorminer/busclique.pyx":405
 *             key1 = length1.get(i)
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))             # <<<<<<<<<<<<<<
 *             elif key0 is None:
 *                 key = used.setdefault((key1, 1), len(used))
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_key0);
      __Pyx_GIVEREF(__pyx_v_key0);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_key0);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
      __pyx_t_9 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 405, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_3, __pyx_t_6, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "minorminer/busclique.pyx":404
 *             key0 = length0.get(i)
 *             key1 = length1.get(i)
 *             if key1 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:
 */
      goto __pyx_L11;
    }

    /* "minorminer/busclique.pyx":406
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key1, 1), len(used))
 *             elif len(raw0[key0]) > len(raw1[key1]):
 */
    __pyx_t_5 = (__pyx_v_key0 == Py_None);
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "minorminer/busclique.pyx":407
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:
 *                 key = used.setdefault((key1, 1), len(used))             # <<<<<<<<<<<<<<
 *             elif len(raw0[key0]) > len(raw1[key1]):
 *                 # pick the larger embedding
 */
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_key1);
      __Pyx_GIVEREF(__pyx_v_key1);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_key1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
      __pyx_t_9 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 407, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_2, __pyx_t_6, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "minorminer/busclique.pyx":406
 *             if key1 is None:
 *                 key = used.setdefault((key0, 0), len(used))
 *             elif key0 is None:             # <<<<<<<<<<<<<<
 *                 key = used.setdefault((key1, 1), len(used))
 *             elif len(raw0[key0]) > len(raw1[key1]):
 */
      goto __pyx_L11;
    }

    /* "minorminer/busclique.pyx":408
 *             elif key0 is None:
 *                 key = used.setdefault((key1, 1), len(used))
 *             elif len(raw0[key0]) > len(raw1[key1]):             # <<<<<<<<<<<<<<
 *                 # pick the larger embedding
 *                 key = used.setdefault((key0, 0), len(used))
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_raw0, __pyx_v_key0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_raw1, __pyx_v_key1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = ((__pyx_t_9 > __pyx_t_11) != 0);
    if (__pyx_t_4) {

      /* "minorminer/busclique.pyx":410
 *             elif len(raw0[key0]) > len(raw1[key1]):
 *                 # pick the larger embedding
 *                 key = used.setdefault((key0, 0), len(used))             # <<<<<<<<<<<<<<
 *             else:
 *                 # if the embeddings have the same size, retain the old one
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_key0);
      __Pyx_GIVEREF(__pyx_v_key0);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_key0);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
      __pyx_t_11 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_3, __pyx_t_6, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "minorminer/busclique.pyx":408
 *             elif key0 is None:
 *                 key = used.setdefault((key1, 1), len(used))
 *             elif len(raw0[key0]) > len(raw1[key1]):             # <<<<<<<<<<<<<<
 *                 # pick the larger embedding
 *                 key = used.setdefault((key0, 0), len(used))
 */
      goto __pyx_L11;
    }

    /* "minorminer/busclique.pyx":413
 *             else:
 *                 # if the embeddings have the same size, retain the old one
 *                 key = used.setdefault((key1, 1), len(used))             # <<<<<<<<<<<<<<
 *             length[i] = key
 * 
 */
    /*else*/ {
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_key1);
      __Pyx_GIVEREF(__pyx_v_key1);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_key1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
      __pyx_t_11 = PyDict_Size(__pyx_v_used); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 413, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyDict_SetDefault(__pyx_v_used, __pyx_t_2, __pyx_t_6, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;
    }
    __pyx_L11:;

    /* "minorminer/busclique.pyx":414
 *                 # if the embeddings have the same size, retain the old one
 *                 key = used.setdefault((key1, 1), len(used))
 *             length[i] = key             # <<<<<<<<<<<<<<
 * 
 *         raw = [raw0[key] if i == 0 else raw1[key] for (key, i) in used]
 */
    if (unlikely(PyDict_SetItem(__pyx_v_length, __pyx_v_i, __pyx_v_key) < 0)) __PYX_ERR(0, 414, __pyx_L1_error)

    /* "minorminer/busclique.pyx":401
 *         length0 = cache['length']
 *         length1 = self._cliques['length']
 *         for i in sorted(set(length0).union(length1)):             # <<<<<<<<<<<<<<
 *             key0 = length0.get(i)
 *             key1 = length1.get(i)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":416
 *             length[i] = key
 * 
 *         raw = [raw0[key] if i == 0 else raw1[key] for (key, i) in used]             # <<<<<<<<<<<<<<
 *         self._cliques = {"raw": raw, "size": size, "length": length}
 *         if write_to_disk:
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = 0;
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_used, 1, ((PyObject *)NULL), (&__pyx_t_11), (&__pyx_t_12)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_13 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_11, &__pyx_t_8, &__pyx_t_6, NULL, NULL, __pyx_t_12);
      if (unlikely(__pyx_t_13 == 0)) break;
      if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 416, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_6);
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 416, __pyx_L14_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_10);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 416, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_14 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 416, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_14)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_15(__pyx_t_14); if (unlikely(!__pyx_t_2)) goto __pyx_L17_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_10 = __pyx_t_15(__pyx_t_14); if (unlikely(!__pyx_t_10)) goto __pyx_L17_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_10);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_14), 2) < 0) __PYX_ERR(0, 416, __pyx_L14_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L18_unpacking_done;
        __pyx_L17_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 416, __pyx_L14_error)
        __pyx_L18_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_i, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyInt_EqObjC(__pyx_8genexpr4__pyx_v_i, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 416, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 416, __pyx_L14_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_4) {
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_raw0, __pyx_8genexpr4__pyx_v_key); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 416, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_6 = __pyx_t_10;
        __pyx_t_10 = 0;
      } else {
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_raw1, __pyx_8genexpr4__pyx_v_key); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 416, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_6 = __pyx_t_10;
        __pyx_t_10 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 416, __pyx_L14_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_i); __pyx_8genexpr4__pyx_v_i = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_key); __pyx_8genexpr4__pyx_v_key = 0;
    goto __pyx_L19_exit_scope;
    __pyx_L14_error:;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_i); __pyx_8genexpr4__pyx_v_i = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_key); __pyx_8genexpr4__pyx_v_key = 0;
    goto __pyx_L1_error;
    __pyx_L19_exit_scope:;
  } /* exit inner scope */
  __pyx_v_raw = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":417
 * 
 *         raw = [raw0[key] if i == 0 else raw1[key] for (key, i) in used]
 *         self._cliques = {"raw": raw, "size": size, "length": length}             # <<<<<<<<<<<<<<
 *         if write_to_disk:
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_raw, __pyx_v_raw) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_size, __pyx_v_size) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_length, __pyx_v_length) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_cliques, __pyx_t_1) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":418
 *         raw = [raw0[key] if i == 0 else raw1[key] for (key, i) in used]
 *         self._cliques = {"raw": raw, "size": size, "length": length}
 *         if write_to_disk:             # <<<<<<<<<<<<<<
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_write_to_disk); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 418, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":419
 *         self._cliques = {"raw": raw, "size": size, "length": length}
 *         if write_to_disk:
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)             # <<<<<<<<<<<<<<
 * 
 *     def _fetch_cache(self, dirname, compute, force_write=False):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fetch_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_18merge_clique_cache_lambda2, 0, __pyx_n_s_busgraph_cache_merge_clique_cach, ((PyObject*)__pyx_cur_scope), __pyx_n_s_minorminer_busclique, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_n_u_clique);
    __Pyx_GIVEREF(__pyx_n_u_clique);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_clique);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_force_write, Py_True) < 0) __PYX_ERR(0, 419, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":418
 *         raw = [raw0[key] if i == 0 else raw1[key] for (key, i) in used]
 *         self._cliques = {"raw": raw, "size": size, "length": length}
 *         if write_to_disk:             # <<<<<<<<<<<<<<
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 */
  }

  /* "minorminer/busclique.pyx":341
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):             # <<<<<<<<<<<<<<
 *         """Merges the clique cache of another ``busgraph_cache`` object
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.merge_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_raw0);
  __Pyx_XDECREF(__pyx_v_raw1);
  __Pyx_XDECREF(__pyx_v_used);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_size0);
  __Pyx_XDECREF(__pyx_v_size1);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_key0);
  __Pyx_XDECREF(__pyx_v_key1);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_qual0);
  __Pyx_XDECREF(__pyx_v_qual1);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XDECREF(__pyx_v_length0);
  __Pyx_XDECREF(__pyx_v_length1);
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_key);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_cache);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":421
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def _fetch_cache(self, dirname, compute, force_write=False):             # <<<<<<<<<<<<<<
 *         """This is an ad-hoc implementation of a file-cache using a LRU strategy.
 *         It's intended to be platform independent, thread- and multiprocess-safe,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_19_fetch_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_18_fetch_cache[] = "This is an ad-hoc implementation of a file-cache using a LRU strategy.\n        It's intended to be platform independent, thread- and multiprocess-safe,\n        and reasonably performant -- I couldn't find a ready-made solution that\n        satisfies those requirements, so I had to roll my own.  TODO: keep\n        looking for something cleaner.\n\n        We use `homebase` to locate a viable directory to store our caches.\n        We create two subdirectories (for now): `cliques` and `bicliques`, and\n        store respective caches in each.\n\n        We derive a unique identifier from `self._graph`, and hash that\n        identifier to produce a `shortcode`.  The cache associated with\n        `self.graph` has the filename `str(shortcode)` -- this is preferable to\n        using a single file to store the entire cache.  In the case of hash\n        collisions, the databases of multiple graphs will be stored in a single\n        cache file -- so that file contains a `pickle`'d dict mapping\n        indentifiers (not shortcodes) to the corresponding clique/biclique\n        caches.\n\n        Inside a cache directory, we have two special files and perhaps many\n        cache files.  The `.lock` file is used by `fasteners.InterProcessLock`\n        to provide process-level locking.  The `.lru` file is a list of up to\n        `__lru_size` different cache filenames, sorted descending in the\n        recentness of the last access of a given filename.  This enables us to\n        automatically clean up the cache before it gets too large.\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_19_fetch_cache = {"_fetch_cache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_19_fetch_cache, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_18_fetch_cache};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_19_fetch_cache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_dirname = 0;
  PyObject *__pyx_v_compute = 0;
  PyObject *__pyx_v_force_write = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fetch_cache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_dirname,&__pyx_n_s_compute,&__pyx_n_s_force_write,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dirname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fetch_cache", 0, 3, 4, 1); __PYX_ERR(0, 421, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_fetch_cache", 0, 3, 4, 2); __PYX_ERR(0, 421, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_force_write);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fetch_cache") < 0)) __PYX_ERR(0, 421, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_dirname = values[1];
    __pyx_v_compute = values[2];
    __pyx_v_force_write = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fetch_cache", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 421, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_18_fetch_cache(__pyx_self, __pyx_v_self, __pyx_v_dirname, __pyx_v_compute, __pyx_v_force_write);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_18_fetch_cache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_dirname, PyObject *__pyx_v_compute, PyObject *__pyx_v_force_write) {
  PyObject *__pyx_v_rootdir = NULL;
  PyObject *__pyx_v_basedir = NULL;
  PyObject *__pyx_v_lockfile = NULL;
  PyObject *__pyx_v_lrufile = NULL;
  PyObject *__pyx_v_identifier = NULL;
  PyObject *__pyx_v_shortcode = NULL;
  PyObject *__pyx_v_file_context = NULL;
  PyObject *__pyx_v_cachefile = NULL;
  PyObject *__pyx_v_filecache = NULL;
  PyObject *__pyx_v_currcache = NULL;
  PyObject *__pyx_v_cache = NULL;
  PyObject *__pyx_v_lru = NULL;
  PyObject *__pyx_v_LRU = NULL;
  PyObject *__pyx_v_newLRU = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_oldkey = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  Py_ssize_t __pyx_t_25;
  PyObject *(*__pyx_t_26)(PyObject *);
  int __pyx_t_27;
  PyObject *__pyx_t_28 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fetch_cache", 0);

  /* "minorminer/busclique.pyx":448
 *         automatically clean up the cache before it gets too large.
 *         """
 *         rootdir = busgraph_cache.cache_rootdir()             # <<<<<<<<<<<<<<
 *         basedir = os.path.join(rootdir, dirname)
 *         pathlib.Path(basedir).mkdir(parents=True, exist_ok=True)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_busgraph_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_cache_rootdir); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_rootdir = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":449
 *         """
 *         rootdir = busgraph_cache.cache_rootdir()
 *         basedir = os.path.join(rootdir, dirname)             # <<<<<<<<<<<<<<
 *         pathlib.Path(basedir).mkdir(parents=True, exist_ok=True)
 *         lockfile = os.path.join(basedir, ".lock")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_join); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_rootdir, __pyx_v_dirname};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_rootdir, __pyx_v_dirname};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_rootdir);
    __Pyx_GIVEREF(__pyx_v_rootdir);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_rootdir);
    __Pyx_INCREF(__pyx_v_dirname);
    __Pyx_GIVEREF(__pyx_v_dirname);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_dirname);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_basedir = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":450
 *         rootdir = busgraph_cache.cache_rootdir()
 *         basedir = os.path.join(rootdir, dirname)
 *         pathlib.Path(basedir).mkdir(parents=True, exist_ok=True)             # <<<<<<<<<<<<<<
 *         lockfile = os.path.join(basedir, ".lock")
 *         lrufile = os.path.join(basedir, ".lru")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pathlib); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Path); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_basedir) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_basedir);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_mkdir); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_parents, Py_True) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_exist_ok, Py_True) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":451
 *         basedir = os.path.join(rootdir, dirname)
 *         pathlib.Path(basedir).mkdir(parents=True, exist_ok=True)
 *         lockfile = os.path.join(basedir, ".lock")             # <<<<<<<<<<<<<<
 *         lrufile = os.path.join(basedir, ".lru")
 *         identifier = self._graph.identifier
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_path); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_join); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_basedir, __pyx_kp_u_lock};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_basedir, __pyx_kp_u_lock};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_basedir);
    __Pyx_GIVEREF(__pyx_v_basedir);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_4, __pyx_v_basedir);
    __Pyx_INCREF(__pyx_kp_u_lock);
    __Pyx_GIVEREF(__pyx_kp_u_lock);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_4, __pyx_kp_u_lock);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_lockfile = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":452
 *         pathlib.Path(basedir).mkdir(parents=True, exist_ok=True)
 *         lockfile = os.path.join(basedir, ".lock")
 *         lrufile = os.path.join(basedir, ".lru")             # <<<<<<<<<<<<<<
 *         identifier = self._graph.identifier
 *         shortcode = self._graph.short_identifier
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_join); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_basedir, __pyx_kp_u_lru};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_basedir, __pyx_kp_u_lru};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_basedir);
    __Pyx_GIVEREF(__pyx_v_basedir);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_basedir);
    __Pyx_INCREF(__pyx_kp_u_lru);
    __Pyx_GIVEREF(__pyx_kp_u_lru);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_kp_u_lru);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_lrufile = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":453
 *         lockfile = os.path.join(basedir, ".lock")
 *         lrufile = os.path.join(basedir, ".lru")
 *         identifier = self._graph.identifier             # <<<<<<<<<<<<<<
 *         shortcode = self._graph.short_identifier
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_identifier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_identifier = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":454
 *         lrufile = os.path.join(basedir, ".lru")
 *         identifier = self._graph.identifier
 *         shortcode = self._graph.short_identifier             # <<<<<<<<<<<<<<
 * 
 *         #this makes our writes to the filesystem atomic
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_short_identifier); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_shortcode = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":457
 * 
 *         #this makes our writes to the filesystem atomic
 *         file_context = copy_on_close_context()             # <<<<<<<<<<<<<<
 * 
 *         #this solves thread-safety?
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_copy_on_close_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_file_context = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":460
 * 
 *         #this solves thread-safety?
 *         with __global_locks[dirname]:             # <<<<<<<<<<<<<<
 *             #this solves inter-process safety?
 *             with fasteners.InterProcessLock(lockfile):
 */
  /*with:*/ {
    if (unlikely(__pyx_v_10minorminer_9busclique___global_locks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 460, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_10minorminer_9busclique___global_locks, __pyx_v_dirname); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_exit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "minorminer/busclique.pyx":462
 *         with __global_locks[dirname]:
 *             #this solves inter-process safety?
 *             with fasteners.InterProcessLock(lockfile):             # <<<<<<<<<<<<<<
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):
 */
          /*with:*/ {
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_fasteners); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_InterProcessLock); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 462, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_v_lockfile) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_lockfile);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_10 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_exit); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 462, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_enter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L13_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 462, __pyx_L13_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            /*try:*/ {
              {
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
                __Pyx_XGOTREF(__pyx_t_11);
                __Pyx_XGOTREF(__pyx_t_12);
                __Pyx_XGOTREF(__pyx_t_13);
                /*try:*/ {

                  /* "minorminer/busclique.pyx":463
 *             #this solves inter-process safety?
 *             with fasteners.InterProcessLock(lockfile):
 *                 cachefile = os.path.join(basedir, str(shortcode))             # <<<<<<<<<<<<<<
 *                 if os.path.exists(cachefile):
 *                     with open(cachefile, 'rb') as filecache:
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_os); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 463, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_join); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 463, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_shortcode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __pyx_t_2 = NULL;
                  __pyx_t_4 = 0;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
                    if (likely(__pyx_t_2)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                      __Pyx_INCREF(__pyx_t_2);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_5, function);
                      __pyx_t_4 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_5)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_basedir, __pyx_t_1};
                    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L17_error)
                    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_basedir, __pyx_t_1};
                    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L17_error)
                    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  } else
                  #endif
                  {
                    __pyx_t_14 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 463, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    if (__pyx_t_2) {
                      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_2); __pyx_t_2 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_basedir);
                    __Pyx_GIVEREF(__pyx_v_basedir);
                    PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_4, __pyx_v_basedir);
                    __Pyx_GIVEREF(__pyx_t_1);
                    PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_4, __pyx_t_1);
                    __pyx_t_1 = 0;
                    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __pyx_v_cachefile = __pyx_t_3;
                  __pyx_t_3 = 0;

                  /* "minorminer/busclique.pyx":464
 *             with fasteners.InterProcessLock(lockfile):
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):             # <<<<<<<<<<<<<<
 *                     with open(cachefile, 'rb') as filecache:
 *                         currcache = load(filecache)
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_os); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 464, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_path); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 464, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_14);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_exists); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 464, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  __pyx_t_14 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
                    if (likely(__pyx_t_14)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                      __Pyx_INCREF(__pyx_t_14);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_5, function);
                    }
                  }
                  __pyx_t_3 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_14, __pyx_v_cachefile) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_cachefile);
                  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 464, __pyx_L17_error)
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  if (__pyx_t_15) {

                    /* "minorminer/busclique.pyx":465
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):
 *                     with open(cachefile, 'rb') as filecache:             # <<<<<<<<<<<<<<
 *                         currcache = load(filecache)
 *                 else:
 */
                    /*with:*/ {
                      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_3);
                      __Pyx_INCREF(__pyx_v_cachefile);
                      __Pyx_GIVEREF(__pyx_v_cachefile);
                      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_cachefile);
                      __Pyx_INCREF(__pyx_n_u_rb);
                      __Pyx_GIVEREF(__pyx_n_u_rb);
                      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_rb);
                      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 465, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                      __pyx_t_16 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 465, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_16);
                      __pyx_t_14 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 465, __pyx_L24_error)
                      __Pyx_GOTREF(__pyx_t_14);
                      __pyx_t_1 = NULL;
                      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
                        if (likely(__pyx_t_1)) {
                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                          __Pyx_INCREF(__pyx_t_1);
                          __Pyx_INCREF(function);
                          __Pyx_DECREF_SET(__pyx_t_14, function);
                        }
                      }
                      __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_14);
                      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L24_error)
                      __Pyx_GOTREF(__pyx_t_3);
                      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                      __pyx_t_14 = __pyx_t_3;
                      __pyx_t_3 = 0;
                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                      /*try:*/ {
                        {
                          __Pyx_PyThreadState_declare
                          __Pyx_PyThreadState_assign
                          __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
                          __Pyx_XGOTREF(__pyx_t_17);
                          __Pyx_XGOTREF(__pyx_t_18);
                          __Pyx_XGOTREF(__pyx_t_19);
                          /*try:*/ {
                            __pyx_v_filecache = __pyx_t_14;
                            __pyx_t_14 = 0;

                            /* "minorminer/busclique.pyx":466
 *                 if os.path.exists(cachefile):
 *                     with open(cachefile, 'rb') as filecache:
 *                         currcache = load(filecache)             # <<<<<<<<<<<<<<
 *                 else:
 *                     currcache = {}
 */
                            __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_load); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 466, __pyx_L28_error)
                            __Pyx_GOTREF(__pyx_t_5);
                            __pyx_t_3 = NULL;
                            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
                              if (likely(__pyx_t_3)) {
                                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                                __Pyx_INCREF(__pyx_t_3);
                                __Pyx_INCREF(function);
                                __Pyx_DECREF_SET(__pyx_t_5, function);
                              }
                            }
                            __pyx_t_14 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_filecache) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_filecache);
                            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 466, __pyx_L28_error)
                            __Pyx_GOTREF(__pyx_t_14);
                            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                            __pyx_v_currcache = __pyx_t_14;
                            __pyx_t_14 = 0;

                            /* "minorminer/busclique.pyx":465
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):
 *                     with open(cachefile, 'rb') as filecache:             # <<<<<<<<<<<<<<
 *                         currcache = load(filecache)
 *                 else:
 */
                          }
                          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                          goto __pyx_L33_try_end;
                          __pyx_L28_error:;
                          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                          /*except:*/ {
                            __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                            if (__Pyx_GetException(&__pyx_t_14, &__pyx_t_5, &__pyx_t_3) < 0) __PYX_ERR(0, 465, __pyx_L30_except_error)
                            __Pyx_GOTREF(__pyx_t_14);
                            __Pyx_GOTREF(__pyx_t_5);
                            __Pyx_GOTREF(__pyx_t_3);
                            __pyx_t_1 = PyTuple_Pack(3, __pyx_t_14, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L30_except_error)
                            __Pyx_GOTREF(__pyx_t_1);
                            __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_1, NULL);
                            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                            if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 465, __pyx_L30_except_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_20);
                            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                            if (__pyx_t_15 < 0) __PYX_ERR(0, 465, __pyx_L30_except_error)
                            __pyx_t_21 = ((!(__pyx_t_15 != 0)) != 0);
                            if (__pyx_t_21) {
                              __Pyx_GIVEREF(__pyx_t_14);
                              __Pyx_GIVEREF(__pyx_t_5);
                              __Pyx_XGIVEREF(__pyx_t_3);
                              __Pyx_ErrRestoreWithState(__pyx_t_14, __pyx_t_5, __pyx_t_3);
                              __pyx_t_14 = 0; __pyx_t_5 = 0; __pyx_t_3 = 0; 
                              __PYX_ERR(0, 465, __pyx_L30_except_error)
                            }
                            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                            goto __pyx_L29_exception_handled;
                          }
                          __pyx_L30_except_error:;
                          __Pyx_XGIVEREF(__pyx_t_17);
                          __Pyx_XGIVEREF(__pyx_t_18);
                          __Pyx_XGIVEREF(__pyx_t_19);
                          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
                          goto __pyx_L17_error;
                          __pyx_L29_exception_handled:;
                          __Pyx_XGIVEREF(__pyx_t_17);
                          __Pyx_XGIVEREF(__pyx_t_18);
                          __Pyx_XGIVEREF(__pyx_t_19);
                          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
                          __pyx_L33_try_end:;
                        }
                      }
                      /*finally:*/ {
                        /*normal exit:*/{
                          if (__pyx_t_16) {
                            __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_tuple__4, NULL);
                            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                            if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 465, __pyx_L17_error)
                            __Pyx_GOTREF(__pyx_t_19);
                            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                          }
                          goto __pyx_L27;
                        }
                        __pyx_L27:;
                      }
                      goto __pyx_L37;
                      __pyx_L24_error:;
                      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                      goto __pyx_L17_error;
                      __pyx_L37:;
                    }

                    /* "minorminer/busclique.pyx":464
 *             with fasteners.InterProcessLock(lockfile):
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):             # <<<<<<<<<<<<<<
 *                     with open(cachefile, 'rb') as filecache:
 *                         currcache = load(filecache)
 */
                    goto __pyx_L23;
                  }

                  /* "minorminer/busclique.pyx":468
 *                         currcache = load(filecache)
 *                 else:
 *                     currcache = {}             # <<<<<<<<<<<<<<
 *                 cache = None if force_write else currcache.get(identifier)
 *                 if cache is None:
 */
                  /*else*/ {
                    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 468, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __pyx_v_currcache = __pyx_t_3;
                    __pyx_t_3 = 0;
                  }
                  __pyx_L23:;

                  /* "minorminer/busclique.pyx":469
 *                 else:
 *                     currcache = {}
 *                 cache = None if force_write else currcache.get(identifier)             # <<<<<<<<<<<<<<
 *                 if cache is None:
 *                     cache = compute()
 */
                  __pyx_t_21 = __Pyx_PyObject_IsTrue(__pyx_v_force_write); if (unlikely(__pyx_t_21 < 0)) __PYX_ERR(0, 469, __pyx_L17_error)
                  if (__pyx_t_21) {
                    __Pyx_INCREF(Py_None);
                    __pyx_t_3 = Py_None;
                  } else {
                    if (unlikely(!__pyx_v_currcache)) { __Pyx_RaiseUnboundLocalError("currcache"); __PYX_ERR(0, 469, __pyx_L17_error) }
                    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_currcache, __pyx_n_s_get); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 469, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __pyx_t_1 = NULL;
                    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
                      if (likely(__pyx_t_1)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                        __Pyx_INCREF(__pyx_t_1);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_14, function);
                      }
                    }
                    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_1, __pyx_v_identifier) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_v_identifier);
                    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 469, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    __pyx_t_3 = __pyx_t_5;
                    __pyx_t_5 = 0;
                  }
                  __pyx_v_cache = __pyx_t_3;
                  __pyx_t_3 = 0;

                  /* "minorminer/busclique.pyx":470
 *                     currcache = {}
 *                 cache = None if force_write else currcache.get(identifier)
 *                 if cache is None:             # <<<<<<<<<<<<<<
 *                     cache = compute()
 *                     currcache[identifier] = cache
 */
                  __pyx_t_21 = (__pyx_v_cache == Py_None);
                  __pyx_t_15 = (__pyx_t_21 != 0);
                  if (__pyx_t_15) {

                    /* "minorminer/busclique.pyx":471
 *                 cache = None if force_write else currcache.get(identifier)
 *                 if cache is None:
 *                     cache = compute()             # <<<<<<<<<<<<<<
 *                     currcache[identifier] = cache
 *                     with file_context.open(cachefile, 'wb') as filecache:
 */
                    __Pyx_INCREF(__pyx_v_compute);
                    __pyx_t_5 = __pyx_v_compute; __pyx_t_14 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
                      if (likely(__pyx_t_14)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                        __Pyx_INCREF(__pyx_t_14);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_5, function);
                      }
                    }
                    __pyx_t_3 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
                    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 471, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __Pyx_DECREF_SET(__pyx_v_cache, __pyx_t_3);
                    __pyx_t_3 = 0;

                    /* "minorminer/busclique.pyx":472
 *                 if cache is None:
 *                     cache = compute()
 *                     currcache[identifier] = cache             # <<<<<<<<<<<<<<
 *                     with file_context.open(cachefile, 'wb') as filecache:
 *                         dump(currcache, filecache)
 */
                    if (unlikely(!__pyx_v_currcache)) { __Pyx_RaiseUnboundLocalError("currcache"); __PYX_ERR(0, 472, __pyx_L17_error) }
                    if (unlikely(PyObject_SetItem(__pyx_v_currcache, __pyx_v_identifier, __pyx_v_cache) < 0)) __PYX_ERR(0, 472, __pyx_L17_error)

                    /* "minorminer/busclique.pyx":473
 *                     cache = compute()
 *                     currcache[identifier] = cache
 *                     with file_context.open(cachefile, 'wb') as filecache:             # <<<<<<<<<<<<<<
 *                         dump(currcache, filecache)
 * 
 */
                    /*with:*/ {
                      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_context, __pyx_n_s_open); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      __pyx_t_14 = NULL;
                      __pyx_t_4 = 0;
                      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
                        if (likely(__pyx_t_14)) {
                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                          __Pyx_INCREF(__pyx_t_14);
                          __Pyx_INCREF(function);
                          __Pyx_DECREF_SET(__pyx_t_5, function);
                          __pyx_t_4 = 1;
                        }
                      }
                      #if CYTHON_FAST_PYCALL
                      if (PyFunction_Check(__pyx_t_5)) {
                        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_cachefile, __pyx_n_u_wb};
                        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L17_error)
                        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                        __Pyx_GOTREF(__pyx_t_3);
                      } else
                      #endif
                      #if CYTHON_FAST_PYCCALL
                      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_cachefile, __pyx_n_u_wb};
                        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L17_error)
                        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                        __Pyx_GOTREF(__pyx_t_3);
                      } else
                      #endif
                      {
                        __pyx_t_1 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L17_error)
                        __Pyx_GOTREF(__pyx_t_1);
                        if (__pyx_t_14) {
                          __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_14); __pyx_t_14 = NULL;
                        }
                        __Pyx_INCREF(__pyx_v_cachefile);
                        __Pyx_GIVEREF(__pyx_v_cachefile);
                        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_4, __pyx_v_cachefile);
                        __Pyx_INCREF(__pyx_n_u_wb);
                        __Pyx_GIVEREF(__pyx_n_u_wb);
                        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_4, __pyx_n_u_wb);
                        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L17_error)
                        __Pyx_GOTREF(__pyx_t_3);
                        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                      }
                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                      __pyx_t_16 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_exit); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 473, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_16);
                      __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_enter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L39_error)
                      __Pyx_GOTREF(__pyx_t_1);
                      __pyx_t_14 = NULL;
                      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_1);
                        if (likely(__pyx_t_14)) {
                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                          __Pyx_INCREF(__pyx_t_14);
                          __Pyx_INCREF(function);
                          __Pyx_DECREF_SET(__pyx_t_1, function);
                        }
                      }
                      __pyx_t_5 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
                      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L39_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                      __pyx_t_1 = __pyx_t_5;
                      __pyx_t_5 = 0;
                      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                      /*try:*/ {
                        {
                          __Pyx_PyThreadState_declare
                          __Pyx_PyThreadState_assign
                          __Pyx_ExceptionSave(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
                          __Pyx_XGOTREF(__pyx_t_19);
                          __Pyx_XGOTREF(__pyx_t_18);
                          __Pyx_XGOTREF(__pyx_t_17);
                          /*try:*/ {
                            __Pyx_XDECREF_SET(__pyx_v_filecache, __pyx_t_1);
                            __pyx_t_1 = 0;

                            /* "minorminer/busclique.pyx":474
 *                     currcache[identifier] = cache
 *                     with file_context.open(cachefile, 'wb') as filecache:
 *                         dump(currcache, filecache)             # <<<<<<<<<<<<<<
 * 
 *                 #now, update the LRU cache -- if this was being done in-memory,
 */
                            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_dump); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L43_error)
                            __Pyx_GOTREF(__pyx_t_3);
                            if (unlikely(!__pyx_v_currcache)) { __Pyx_RaiseUnboundLocalError("currcache"); __PYX_ERR(0, 474, __pyx_L43_error) }
                            __pyx_t_5 = NULL;
                            __pyx_t_4 = 0;
                            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                              if (likely(__pyx_t_5)) {
                                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                                __Pyx_INCREF(__pyx_t_5);
                                __Pyx_INCREF(function);
                                __Pyx_DECREF_SET(__pyx_t_3, function);
                                __pyx_t_4 = 1;
                              }
                            }
                            #if CYTHON_FAST_PYCALL
                            if (PyFunction_Check(__pyx_t_3)) {
                              PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_currcache, __pyx_v_filecache};
                              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L43_error)
                              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                              __Pyx_GOTREF(__pyx_t_1);
                            } else
                            #endif
                            #if CYTHON_FAST_PYCCALL
                            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                              PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_currcache, __pyx_v_filecache};
                              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L43_error)
                              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                              __Pyx_GOTREF(__pyx_t_1);
                            } else
                            #endif
                            {
                              __pyx_t_14 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 474, __pyx_L43_error)
                              __Pyx_GOTREF(__pyx_t_14);
                              if (__pyx_t_5) {
                                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_5); __pyx_t_5 = NULL;
                              }
                              __Pyx_INCREF(__pyx_v_currcache);
                              __Pyx_GIVEREF(__pyx_v_currcache);
                              PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_4, __pyx_v_currcache);
                              __Pyx_INCREF(__pyx_v_filecache);
                              __Pyx_GIVEREF(__pyx_v_filecache);
                              PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_4, __pyx_v_filecache);
                              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L43_error)
                              __Pyx_GOTREF(__pyx_t_1);
                              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                            }
                            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

                            /* "minorminer/busclique.pyx":473
 *                     cache = compute()
 *                     currcache[identifier] = cache
 *                     with file_context.open(cachefile, 'wb') as filecache:             # <<<<<<<<<<<<<<
 *                         dump(currcache, filecache)
 * 
 */
                          }
                          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                          goto __pyx_L48_try_end;
                          __pyx_L43_error:;
                          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                          /*except:*/ {
                            __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_14) < 0) __PYX_ERR(0, 473, __pyx_L45_except_error)
                            __Pyx_GOTREF(__pyx_t_1);
                            __Pyx_GOTREF(__pyx_t_3);
                            __Pyx_GOTREF(__pyx_t_14);
                            __pyx_t_5 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_3, __pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L45_except_error)
                            __Pyx_GOTREF(__pyx_t_5);
                            __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_5, NULL);
                            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                            if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 473, __pyx_L45_except_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_20);
                            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                            if (__pyx_t_15 < 0) __PYX_ERR(0, 473, __pyx_L45_except_error)
                            __pyx_t_21 = ((!(__pyx_t_15 != 0)) != 0);
                            if (__pyx_t_21) {
                              __Pyx_GIVEREF(__pyx_t_1);
                              __Pyx_GIVEREF(__pyx_t_3);
                              __Pyx_XGIVEREF(__pyx_t_14);
                              __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_14);
                              __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_14 = 0; 
                              __PYX_ERR(0, 473, __pyx_L45_except_error)
                            }
                            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                            goto __pyx_L44_exception_handled;
                          }
                          __pyx_L45_except_error:;
                          __Pyx_XGIVEREF(__pyx_t_19);
                          __Pyx_XGIVEREF(__pyx_t_18);
                          __Pyx_XGIVEREF(__pyx_t_17);
                          __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
                          goto __pyx_L17_error;
                          __pyx_L44_exception_handled:;
                          __Pyx_XGIVEREF(__pyx_t_19);
                          __Pyx_XGIVEREF(__pyx_t_18);
                          __Pyx_XGIVEREF(__pyx_t_17);
                          __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
                          __pyx_L48_try_end:;
                        }
                      }
                      /*finally:*/ {
                        /*normal exit:*/{
                          if (__pyx_t_16) {
                            __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_tuple__4, NULL);
                            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                            if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 473, __pyx_L17_error)
                            __Pyx_GOTREF(__pyx_t_17);
                            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                          }
                          goto __pyx_L42;
                        }
                        __pyx_L42:;
                      }
                      goto __pyx_L52;
                      __pyx_L39_error:;
                      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                      goto __pyx_L17_error;
                      __pyx_L52:;
                    }

                    /* "minorminer/busclique.pyx":470
 *                     currcache = {}
 *                 cache = None if force_write else currcache.get(identifier)
 *                 if cache is None:             # <<<<<<<<<<<<<<
 *                     cache = compute()
 *                     currcache[identifier] = cache
 */
                  }

                  /* "minorminer/busclique.pyx":479
 *                 #there are more efficient algorithms... but since we're doing
 *                 #linear-time read&write, might as well do it the lazy way.
 *                 try:             # <<<<<<<<<<<<<<
 *                     with open(lrufile, 'rb') as lru:
 *                         LRU = load(lru)
 */
                  {
                    __Pyx_PyThreadState_declare
                    __Pyx_PyThreadState_assign
                    __Pyx_ExceptionSave(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
                    __Pyx_XGOTREF(__pyx_t_16);
                    __Pyx_XGOTREF(__pyx_t_17);
                    __Pyx_XGOTREF(__pyx_t_18);
                    /*try:*/ {

                      /* "minorminer/busclique.pyx":480
 *                 #linear-time read&write, might as well do it the lazy way.
 *                 try:
 *                     with open(lrufile, 'rb') as lru:             # <<<<<<<<<<<<<<
 *                         LRU = load(lru)
 *                 except FileNotFoundError:
 */
                      /*with:*/ {
                        __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 480, __pyx_L53_error)
                        __Pyx_GOTREF(__pyx_t_14);
                        __Pyx_INCREF(__pyx_v_lrufile);
                        __Pyx_GIVEREF(__pyx_v_lrufile);
                        PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_lrufile);
                        __Pyx_INCREF(__pyx_n_u_rb);
                        __Pyx_GIVEREF(__pyx_n_u_rb);
                        PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_u_rb);
                        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L53_error)
                        __Pyx_GOTREF(__pyx_t_3);
                        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                        __pyx_t_19 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_exit); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 480, __pyx_L53_error)
                        __Pyx_GOTREF(__pyx_t_19);
                        __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_enter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_1);
                        __pyx_t_5 = NULL;
                        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
                          if (likely(__pyx_t_5)) {
                            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                            __Pyx_INCREF(__pyx_t_5);
                            __Pyx_INCREF(function);
                            __Pyx_DECREF_SET(__pyx_t_1, function);
                          }
                        }
                        __pyx_t_14 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
                        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                        if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 480, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_14);
                        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                        __pyx_t_1 = __pyx_t_14;
                        __pyx_t_14 = 0;
                        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                        /*try:*/ {
                          {
                            __Pyx_PyThreadState_declare
                            __Pyx_PyThreadState_assign
                            __Pyx_ExceptionSave(&__pyx_t_20, &__pyx_t_22, &__pyx_t_23);
                            __Pyx_XGOTREF(__pyx_t_20);
                            __Pyx_XGOTREF(__pyx_t_22);
                            __Pyx_XGOTREF(__pyx_t_23);
                            /*try:*/ {
                              __pyx_v_lru = __pyx_t_1;
                              __pyx_t_1 = 0;

                              /* "minorminer/busclique.pyx":481
 *                 try:
 *                     with open(lrufile, 'rb') as lru:
 *                         LRU = load(lru)             # <<<<<<<<<<<<<<
 *                 except FileNotFoundError:
 *                     LRU = []
 */
                              __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_load); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L63_error)
                              __Pyx_GOTREF(__pyx_t_3);
                              __pyx_t_14 = NULL;
                              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_3);
                                if (likely(__pyx_t_14)) {
                                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                                  __Pyx_INCREF(__pyx_t_14);
                                  __Pyx_INCREF(function);
                                  __Pyx_DECREF_SET(__pyx_t_3, function);
                                }
                              }
                              __pyx_t_1 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_14, __pyx_v_lru) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_lru);
                              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                              if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L63_error)
                              __Pyx_GOTREF(__pyx_t_1);
                              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                              __pyx_v_LRU = __pyx_t_1;
                              __pyx_t_1 = 0;

                              /* "minorminer/busclique.pyx":480
 *                 #linear-time read&write, might as well do it the lazy way.
 *                 try:
 *                     with open(lrufile, 'rb') as lru:             # <<<<<<<<<<<<<<
 *                         LRU = load(lru)
 *                 except FileNotFoundError:
 */
                            }
                            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                            __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                            __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                            goto __pyx_L68_try_end;
                            __pyx_L63_error:;
                            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                            /*except:*/ {
                              __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                              if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_14) < 0) __PYX_ERR(0, 480, __pyx_L65_except_error)
                              __Pyx_GOTREF(__pyx_t_1);
                              __Pyx_GOTREF(__pyx_t_3);
                              __Pyx_GOTREF(__pyx_t_14);
                              __pyx_t_5 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_3, __pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L65_except_error)
                              __Pyx_GOTREF(__pyx_t_5);
                              __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_5, NULL);
                              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                              if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 480, __pyx_L65_except_error)
                              __Pyx_GOTREF(__pyx_t_24);
                              __pyx_t_21 = __Pyx_PyObject_IsTrue(__pyx_t_24);
                              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                              if (__pyx_t_21 < 0) __PYX_ERR(0, 480, __pyx_L65_except_error)
                              __pyx_t_15 = ((!(__pyx_t_21 != 0)) != 0);
                              if (__pyx_t_15) {
                                __Pyx_GIVEREF(__pyx_t_1);
                                __Pyx_GIVEREF(__pyx_t_3);
                                __Pyx_XGIVEREF(__pyx_t_14);
                                __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_14);
                                __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_14 = 0; 
                                __PYX_ERR(0, 480, __pyx_L65_except_error)
                              }
                              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                              goto __pyx_L64_exception_handled;
                            }
                            __pyx_L65_except_error:;
                            __Pyx_XGIVEREF(__pyx_t_20);
                            __Pyx_XGIVEREF(__pyx_t_22);
                            __Pyx_XGIVEREF(__pyx_t_23);
                            __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_22, __pyx_t_23);
                            goto __pyx_L53_error;
                            __pyx_L64_exception_handled:;
                            __Pyx_XGIVEREF(__pyx_t_20);
                            __Pyx_XGIVEREF(__pyx_t_22);
                            __Pyx_XGIVEREF(__pyx_t_23);
                            __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_22, __pyx_t_23);
                            __pyx_L68_try_end:;
                          }
                        }
                        /*finally:*/ {
                          /*normal exit:*/{
                            if (__pyx_t_19) {
                              __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_tuple__4, NULL);
                              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                              if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 480, __pyx_L53_error)
                              __Pyx_GOTREF(__pyx_t_23);
                              __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                            }
                            goto __pyx_L62;
                          }
                          __pyx_L62:;
                        }
                        goto __pyx_L72;
                        __pyx_L59_error:;
                        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                        goto __pyx_L53_error;
                        __pyx_L72:;
                      }

                      /* "minorminer/busclique.pyx":479
 *                 #there are more efficient algorithms... but since we're doing
 *                 #linear-time read&write, might as well do it the lazy way.
 *                 try:             # <<<<<<<<<<<<<<
 *                     with open(lrufile, 'rb') as lru:
 *                         LRU = load(lru)
 */
                    }
                    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                    goto __pyx_L58_try_end;
                    __pyx_L53_error:;
                    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

                    /* "minorminer/busclique.pyx":482
 *                     with open(lrufile, 'rb') as lru:
 *                         LRU = load(lru)
 *                 except FileNotFoundError:             # <<<<<<<<<<<<<<
 *                     LRU = []
 *                 newLRU = [shortcode]
 */
                    __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_3, &__pyx_t_1);
                    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_FileNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 482, __pyx_L55_except_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    __pyx_t_4 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_14, __pyx_t_5);
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __Pyx_ErrRestore(__pyx_t_14, __pyx_t_3, __pyx_t_1);
                    __pyx_t_14 = 0; __pyx_t_3 = 0; __pyx_t_1 = 0;
                    if (__pyx_t_4) {
                      __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_14) < 0) __PYX_ERR(0, 482, __pyx_L55_except_error)
                      __Pyx_GOTREF(__pyx_t_1);
                      __Pyx_GOTREF(__pyx_t_3);
                      __Pyx_GOTREF(__pyx_t_14);

                      /* "minorminer/busclique.pyx":483
 *                         LRU = load(lru)
 *                 except FileNotFoundError:
 *                     LRU = []             # <<<<<<<<<<<<<<
 *                 newLRU = [shortcode]
 *                 for x in LRU:
 */
                      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L55_except_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      __Pyx_XDECREF_SET(__pyx_v_LRU, __pyx_t_5);
                      __pyx_t_5 = 0;
                      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                      goto __pyx_L54_exception_handled;
                    }
                    goto __pyx_L55_except_error;
                    __pyx_L55_except_error:;

                    /* "minorminer/busclique.pyx":479
 *                 #there are more efficient algorithms... but since we're doing
 *                 #linear-time read&write, might as well do it the lazy way.
 *                 try:             # <<<<<<<<<<<<<<
 *                     with open(lrufile, 'rb') as lru:
 *                         LRU = load(lru)
 */
                    __Pyx_XGIVEREF(__pyx_t_16);
                    __Pyx_XGIVEREF(__pyx_t_17);
                    __Pyx_XGIVEREF(__pyx_t_18);
                    __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
                    goto __pyx_L17_error;
                    __pyx_L54_exception_handled:;
                    __Pyx_XGIVEREF(__pyx_t_16);
                    __Pyx_XGIVEREF(__pyx_t_17);
                    __Pyx_XGIVEREF(__pyx_t_18);
                    __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
                    __pyx_L58_try_end:;
                  }

                  /* "minorminer/busclique.pyx":484
 *                 except FileNotFoundError:
 *                     LRU = []
 *                 newLRU = [shortcode]             # <<<<<<<<<<<<<<
 *                 for x in LRU:
 *                     if x != shortcode:
 */
                  __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 484, __pyx_L17_error)
                  __Pyx_GOTREF(__pyx_t_14);
                  __Pyx_INCREF(__pyx_v_shortcode);
                  __Pyx_GIVEREF(__pyx_v_shortcode);
                  PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_shortcode);
                  __pyx_v_newLRU = ((PyObject*)__pyx_t_14);
                  __pyx_t_14 = 0;

                  /* "minorminer/busclique.pyx":485
 *                     LRU = []
 *                 newLRU = [shortcode]
 *                 for x in LRU:             # <<<<<<<<<<<<<<
 *                     if x != shortcode:
 *                         newLRU.append(x)
 */
                  if (unlikely(!__pyx_v_LRU)) { __Pyx_RaiseUnboundLocalError("LRU"); __PYX_ERR(0, 485, __pyx_L17_error) }
                  if (likely(PyList_CheckExact(__pyx_v_LRU)) || PyTuple_CheckExact(__pyx_v_LRU)) {
                    __pyx_t_14 = __pyx_v_LRU; __Pyx_INCREF(__pyx_t_14); __pyx_t_25 = 0;
                    __pyx_t_26 = NULL;
                  } else {
                    __pyx_t_25 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_v_LRU); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 485, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __pyx_t_26 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 485, __pyx_L17_error)
                  }
                  for (;;) {
                    if (likely(!__pyx_t_26)) {
                      if (likely(PyList_CheckExact(__pyx_t_14))) {
                        if (__pyx_t_25 >= PyList_GET_SIZE(__pyx_t_14)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_25); __Pyx_INCREF(__pyx_t_3); __pyx_t_25++; if (unlikely(0 < 0)) __PYX_ERR(0, 485, __pyx_L17_error)
                        #else
                        __pyx_t_3 = PySequence_ITEM(__pyx_t_14, __pyx_t_25); __pyx_t_25++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L17_error)
                        __Pyx_GOTREF(__pyx_t_3);
                        #endif
                      } else {
                        if (__pyx_t_25 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_25); __Pyx_INCREF(__pyx_t_3); __pyx_t_25++; if (unlikely(0 < 0)) __PYX_ERR(0, 485, __pyx_L17_error)
                        #else
                        __pyx_t_3 = PySequence_ITEM(__pyx_t_14, __pyx_t_25); __pyx_t_25++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L17_error)
                        __Pyx_GOTREF(__pyx_t_3);
                        #endif
                      }
                    } else {
                      __pyx_t_3 = __pyx_t_26(__pyx_t_14);
                      if (unlikely(!__pyx_t_3)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 485, __pyx_L17_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_3);
                    }
                    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_3);
                    __pyx_t_3 = 0;

                    /* "minorminer/busclique.pyx":486
 *                 newLRU = [shortcode]
 *                 for x in LRU:
 *                     if x != shortcode:             # <<<<<<<<<<<<<<
 *                         newLRU.append(x)
 * 
 */
                    __pyx_t_3 = PyObject_RichCompare(__pyx_v_x, __pyx_v_shortcode, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L17_error)
                    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 486, __pyx_L17_error)
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                    if (__pyx_t_15) {

                      /* "minorminer/busclique.pyx":487
 *                 for x in LRU:
 *                     if x != shortcode:
 *                         newLRU.append(x)             # <<<<<<<<<<<<<<
 * 
 *                 #this violates atomicity, but it's okay -- at worst, we'll need
 */
                      __pyx_t_27 = __Pyx_PyList_Append(__pyx_v_newLRU, __pyx_v_x); if (unlikely(__pyx_t_27 == ((int)-1))) __PYX_ERR(0, 487, __pyx_L17_error)

                      /* "minorminer/busclique.pyx":486
 *                 newLRU = [shortcode]
 *                 for x in LRU:
 *                     if x != shortcode:             # <<<<<<<<<<<<<<
 *                         newLRU.append(x)
 * 
 */
                    }

                    /* "minorminer/busclique.pyx":485
 *                     LRU = []
 *                 newLRU = [shortcode]
 *                 for x in LRU:             # <<<<<<<<<<<<<<
 *                     if x != shortcode:
 *                         newLRU.append(x)
 */
                  }
                  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

                  /* "minorminer/busclique.pyx":491
 *                 #this violates atomicity, but it's okay -- at worst, we'll need
 *                 #to recompute a deleted file that still has an entry in the LRU
 *                 while len(newLRU) > __lru_size:             # <<<<<<<<<<<<<<
 *                     oldkey = newLRU.pop()
 *                     os.remove(os.path.join(basedir, oldkey))
 */
                  while (1) {
                    __pyx_t_25 = PyList_GET_SIZE(__pyx_v_newLRU); if (unlikely(__pyx_t_25 == ((Py_ssize_t)-1))) __PYX_ERR(0, 491, __pyx_L17_error)
                    __pyx_t_15 = ((__pyx_t_25 > __pyx_v_10minorminer_9busclique___lru_size) != 0);
                    if (!__pyx_t_15) break;

                    /* "minorminer/busclique.pyx":492
 *                 #to recompute a deleted file that still has an entry in the LRU
 *                 while len(newLRU) > __lru_size:
 *                     oldkey = newLRU.pop()             # <<<<<<<<<<<<<<
 *                     os.remove(os.path.join(basedir, oldkey))
 *                 with file_context.open(lrufile, 'wb') as lru:
 */
                    __pyx_t_14 = __Pyx_PyList_Pop(__pyx_v_newLRU); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 492, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __Pyx_XDECREF_SET(__pyx_v_oldkey, __pyx_t_14);
                    __pyx_t_14 = 0;

                    /* "minorminer/busclique.pyx":493
 *                 while len(newLRU) > __lru_size:
 *                     oldkey = newLRU.pop()
 *                     os.remove(os.path.join(basedir, oldkey))             # <<<<<<<<<<<<<<
 *                 with file_context.open(lrufile, 'wb') as lru:
 *                     dump(newLRU, lru)
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_os); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 493, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 493, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_join); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 493, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __pyx_t_2 = NULL;
                    __pyx_t_4 = 0;
                    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
                      if (likely(__pyx_t_2)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                        __Pyx_INCREF(__pyx_t_2);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_5, function);
                        __pyx_t_4 = 1;
                      }
                    }
                    #if CYTHON_FAST_PYCALL
                    if (PyFunction_Check(__pyx_t_5)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_basedir, __pyx_v_oldkey};
                      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L17_error)
                      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                      __Pyx_GOTREF(__pyx_t_3);
                    } else
                    #endif
                    #if CYTHON_FAST_PYCCALL
                    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_basedir, __pyx_v_oldkey};
                      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L17_error)
                      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                      __Pyx_GOTREF(__pyx_t_3);
                    } else
                    #endif
                    {
                      __pyx_t_28 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 493, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_28);
                      if (__pyx_t_2) {
                        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2); __pyx_t_2 = NULL;
                      }
                      __Pyx_INCREF(__pyx_v_basedir);
                      __Pyx_GIVEREF(__pyx_v_basedir);
                      PyTuple_SET_ITEM(__pyx_t_28, 0+__pyx_t_4, __pyx_v_basedir);
                      __Pyx_INCREF(__pyx_v_oldkey);
                      __Pyx_GIVEREF(__pyx_v_oldkey);
                      PyTuple_SET_ITEM(__pyx_t_28, 1+__pyx_t_4, __pyx_v_oldkey);
                      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_3);
                      __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __pyx_t_5 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
                      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
                      if (likely(__pyx_t_5)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                        __Pyx_INCREF(__pyx_t_5);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_1, function);
                      }
                    }
                    __pyx_t_14 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
                    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 493, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  }

                  /* "minorminer/busclique.pyx":494
 *                     oldkey = newLRU.pop()
 *                     os.remove(os.path.join(basedir, oldkey))
 *                 with file_context.open(lrufile, 'wb') as lru:             # <<<<<<<<<<<<<<
 *                     dump(newLRU, lru)
 * 
 */
                  /*with:*/ {
                    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_context, __pyx_n_s_open); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __pyx_t_3 = NULL;
                    __pyx_t_4 = 0;
                    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
                      if (likely(__pyx_t_3)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                        __Pyx_INCREF(__pyx_t_3);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_1, function);
                        __pyx_t_4 = 1;
                      }
                    }
                    #if CYTHON_FAST_PYCALL
                    if (PyFunction_Check(__pyx_t_1)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_lrufile, __pyx_n_u_wb};
                      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 494, __pyx_L17_error)
                      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                      __Pyx_GOTREF(__pyx_t_14);
                    } else
                    #endif
                    #if CYTHON_FAST_PYCCALL
                    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_lrufile, __pyx_n_u_wb};
                      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 494, __pyx_L17_error)
                      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                      __Pyx_GOTREF(__pyx_t_14);
                    } else
                    #endif
                    {
                      __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 494, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      if (__pyx_t_3) {
                        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
                      }
                      __Pyx_INCREF(__pyx_v_lrufile);
                      __Pyx_GIVEREF(__pyx_v_lrufile);
                      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_lrufile);
                      __Pyx_INCREF(__pyx_n_u_wb);
                      __Pyx_GIVEREF(__pyx_n_u_wb);
                      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_u_wb);
                      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 494, __pyx_L17_error)
                      __Pyx_GOTREF(__pyx_t_14);
                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __pyx_t_18 = __Pyx_PyObject_LookupSpecial(__pyx_t_14, __pyx_n_s_exit); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 494, __pyx_L17_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_14, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 494, __pyx_L80_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    __pyx_t_3 = NULL;
                    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
                      if (likely(__pyx_t_3)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                        __Pyx_INCREF(__pyx_t_3);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_5, function);
                      }
                    }
                    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
                    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L80_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __pyx_t_5 = __pyx_t_1;
                    __pyx_t_1 = 0;
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    /*try:*/ {
                      {
                        __Pyx_PyThreadState_declare
                        __Pyx_PyThreadState_assign
                        __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_16, &__pyx_t_19);
                        __Pyx_XGOTREF(__pyx_t_17);
                        __Pyx_XGOTREF(__pyx_t_16);
                        __Pyx_XGOTREF(__pyx_t_19);
                        /*try:*/ {
                          __Pyx_XDECREF_SET(__pyx_v_lru, __pyx_t_5);
                          __pyx_t_5 = 0;

                          /* "minorminer/busclique.pyx":495
 *                     os.remove(os.path.join(basedir, oldkey))
 *                 with file_context.open(lrufile, 'wb') as lru:
 *                     dump(newLRU, lru)             # <<<<<<<<<<<<<<
 * 
 *         #perform the copy-on-write for the lrufile and cachefile
 */
                          __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_dump); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 495, __pyx_L84_error)
                          __Pyx_GOTREF(__pyx_t_14);
                          __pyx_t_1 = NULL;
                          __pyx_t_4 = 0;
                          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
                            if (likely(__pyx_t_1)) {
                              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                              __Pyx_INCREF(__pyx_t_1);
                              __Pyx_INCREF(function);
                              __Pyx_DECREF_SET(__pyx_t_14, function);
                              __pyx_t_4 = 1;
                            }
                          }
                          #if CYTHON_FAST_PYCALL
                          if (PyFunction_Check(__pyx_t_14)) {
                            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_newLRU, __pyx_v_lru};
                            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L84_error)
                            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                            __Pyx_GOTREF(__pyx_t_5);
                          } else
                          #endif
                          #if CYTHON_FAST_PYCCALL
                          if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
                            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_newLRU, __pyx_v_lru};
                            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L84_error)
                            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                            __Pyx_GOTREF(__pyx_t_5);
                          } else
                          #endif
                          {
                            __pyx_t_3 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 495, __pyx_L84_error)
                            __Pyx_GOTREF(__pyx_t_3);
                            if (__pyx_t_1) {
                              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
                            }
                            __Pyx_INCREF(__pyx_v_newLRU);
                            __Pyx_GIVEREF(__pyx_v_newLRU);
                            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_newLRU);
                            __Pyx_INCREF(__pyx_v_lru);
                            __Pyx_GIVEREF(__pyx_v_lru);
                            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_lru);
                            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L84_error)
                            __Pyx_GOTREF(__pyx_t_5);
                            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                          }
                          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                          /* "minorminer/busclique.pyx":494
 *                     oldkey = newLRU.pop()
 *                     os.remove(os.path.join(basedir, oldkey))
 *                 with file_context.open(lrufile, 'wb') as lru:             # <<<<<<<<<<<<<<
 *                     dump(newLRU, lru)
 * 
 */
                        }
                        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                        goto __pyx_L89_try_end;
                        __pyx_L84_error:;
                        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                        __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
                        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                        /*except:*/ {
                          __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_14, &__pyx_t_3) < 0) __PYX_ERR(0, 494, __pyx_L86_except_error)
                          __Pyx_GOTREF(__pyx_t_5);
                          __Pyx_GOTREF(__pyx_t_14);
                          __Pyx_GOTREF(__pyx_t_3);
                          __pyx_t_1 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_14, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L86_except_error)
                          __Pyx_GOTREF(__pyx_t_1);
                          __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_t_1, NULL);
                          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                          if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 494, __pyx_L86_except_error)
                          __Pyx_GOTREF(__pyx_t_23);
                          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_23);
                          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                          if (__pyx_t_15 < 0) __PYX_ERR(0, 494, __pyx_L86_except_error)
                          __pyx_t_21 = ((!(__pyx_t_15 != 0)) != 0);
                          if (__pyx_t_21) {
                            __Pyx_GIVEREF(__pyx_t_5);
                            __Pyx_GIVEREF(__pyx_t_14);
                            __Pyx_XGIVEREF(__pyx_t_3);
                            __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_14, __pyx_t_3);
                            __pyx_t_5 = 0; __pyx_t_14 = 0; __pyx_t_3 = 0; 
                            __PYX_ERR(0, 494, __pyx_L86_except_error)
                          }
                          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                          goto __pyx_L85_exception_handled;
                        }
                        __pyx_L86_except_error:;
                        __Pyx_XGIVEREF(__pyx_t_17);
                        __Pyx_XGIVEREF(__pyx_t_16);
                        __Pyx_XGIVEREF(__pyx_t_19);
                        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_19);
                        goto __pyx_L17_error;
                        __pyx_L85_exception_handled:;
                        __Pyx_XGIVEREF(__pyx_t_17);
                        __Pyx_XGIVEREF(__pyx_t_16);
                        __Pyx_XGIVEREF(__pyx_t_19);
                        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_19);
                        __pyx_L89_try_end:;
                      }
                    }
                    /*finally:*/ {
                      /*normal exit:*/{
                        if (__pyx_t_18) {
                          __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_tuple__4, NULL);
                          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                          if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 494, __pyx_L17_error)
                          __Pyx_GOTREF(__pyx_t_19);
                          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                        }
                        goto __pyx_L83;
                      }
                      __pyx_L83:;
                    }
                    goto __pyx_L93;
                    __pyx_L80_error:;
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    goto __pyx_L17_error;
                    __pyx_L93:;
                  }

                  /* "minorminer/busclique.pyx":462
 *         with __global_locks[dirname]:
 *             #this solves inter-process safety?
 *             with fasteners.InterProcessLock(lockfile):             # <<<<<<<<<<<<<<
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):
 */
                }
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                goto __pyx_L22_try_end;
                __pyx_L17_error:;
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                /*except:*/ {
                  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                  if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_14, &__pyx_t_5) < 0) __PYX_ERR(0, 462, __pyx_L19_except_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GOTREF(__pyx_t_14);
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_1 = PyTuple_Pack(3, __pyx_t_3, __pyx_t_14, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L19_except_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_1, NULL);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 462, __pyx_L19_except_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_21 = __Pyx_PyObject_IsTrue(__pyx_t_18);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  if (__pyx_t_21 < 0) __PYX_ERR(0, 462, __pyx_L19_except_error)
                  __pyx_t_15 = ((!(__pyx_t_21 != 0)) != 0);
                  if (__pyx_t_15) {
                    __Pyx_GIVEREF(__pyx_t_3);
                    __Pyx_GIVEREF(__pyx_t_14);
                    __Pyx_XGIVEREF(__pyx_t_5);
                    __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_14, __pyx_t_5);
                    __pyx_t_3 = 0; __pyx_t_14 = 0; __pyx_t_5 = 0; 
                    __PYX_ERR(0, 462, __pyx_L19_except_error)
                  }
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  goto __pyx_L18_exception_handled;
                }
                __pyx_L19_except_error:;
                __Pyx_XGIVEREF(__pyx_t_11);
                __Pyx_XGIVEREF(__pyx_t_12);
                __Pyx_XGIVEREF(__pyx_t_13);
                __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
                goto __pyx_L7_error;
                __pyx_L18_exception_handled:;
                __Pyx_XGIVEREF(__pyx_t_11);
                __Pyx_XGIVEREF(__pyx_t_12);
                __Pyx_XGIVEREF(__pyx_t_13);
                __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
                __pyx_L22_try_end:;
              }
            }
            /*finally:*/ {
              /*normal exit:*/{
                if (__pyx_t_10) {
                  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__4, NULL);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 462, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                }
                goto __pyx_L16;
              }
              __pyx_L16:;
            }
            goto __pyx_L97;
            __pyx_L13_error:;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            goto __pyx_L7_error;
            __pyx_L97:;
          }

          /* "minorminer/busclique.pyx":460
 * 
 *         #this solves thread-safety?
 *         with __global_locks[dirname]:             # <<<<<<<<<<<<<<
 *             #this solves inter-process safety?
 *             with fasteners.InterProcessLock(lockfile):
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_14, &__pyx_t_3) < 0) __PYX_ERR(0, 460, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_14, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 460, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_15 < 0) __PYX_ERR(0, 460, __pyx_L9_except_error)
          __pyx_t_21 = ((!(__pyx_t_15 != 0)) != 0);
          if (__pyx_t_21) {
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_3);
            __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_14, __pyx_t_3);
            __pyx_t_5 = 0; __pyx_t_14 = 0; __pyx_t_3 = 0; 
            __PYX_ERR(0, 460, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_6) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__4, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L101;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L1_error;
    __pyx_L101:;
  }

  /* "minorminer/busclique.pyx":498
 * 
 *         #perform the copy-on-write for the lrufile and cachefile
 *         file_context.close()             # <<<<<<<<<<<<<<
 *         return cache
 * 
 */
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_context, __pyx_n_s_close); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":499
 *         #perform the copy-on-write for the lrufile and cachefile
 *         file_context.close()
 *         return cache             # <<<<<<<<<<<<<<
 * 
 *     def largest_clique(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_cache)) { __Pyx_RaiseUnboundLocalError("cache"); __PYX_ERR(0, 499, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_v_cache);
  __pyx_r = __pyx_v_cache;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":421
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def _fetch_cache(self, dirname, compute, force_write=False):             # <<<<<<<<<<<<<<
 *         """This is an ad-hoc implementation of a file-cache using a LRU strategy.
 *         It's intended to be platform independent, thread- and multiprocess-safe,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache._fetch_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rootdir);
  __Pyx_XDECREF(__pyx_v_basedir);
  __Pyx_XDECREF(__pyx_v_lockfile);
  __Pyx_XDECREF(__pyx_v_lrufile);
  __Pyx_XDECREF(__pyx_v_identifier);
  __Pyx_XDECREF(__pyx_v_shortcode);
  __Pyx_XDECREF(__pyx_v_file_context);
  __Pyx_XDECREF(__pyx_v_cachefile);
  __Pyx_XDECREF(__pyx_v_filecache);
  __Pyx_XDECREF(__pyx_v_currcache);
  __Pyx_XDECREF(__pyx_v_cache);
  __Pyx_XDECREF(__pyx_v_lru);
  __Pyx_XDECREF(__pyx_v_LRU);
  __Pyx_XDECREF(__pyx_v_newLRU);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_oldkey);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":501
 *         return cache
 * 
 *     def largest_clique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_21largest_clique(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_20largest_clique[] = "Returns the largest-found clique in the clique cache.\n\n        This will compute the entire clique cache if it is missing from the\n        filesystem.\n\n        Returns:\n            dict: An embedding of node labels from ``range(len(embedding))``\n            mapped to chains of the largest-found clique.\n\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_21largest_clique = {"largest_clique", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_21largest_clique, METH_O, __pyx_doc_10minorminer_9busclique_14busgraph_cache_20largest_clique};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_21largest_clique(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("largest_clique (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_20largest_clique(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_20largest_clique(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_embs = NULL;
  PyObject *__pyx_v_keys = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("largest_clique", 0);

  /* "minorminer/busclique.pyx":512
 * 
 *         """
 *         self._ensure_clique_cache()             # <<<<<<<<<<<<<<
 *         embs = self._cliques['raw']
 *         keys = self._cliques['size']
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":513
 *         """
 *         self._ensure_clique_cache()
 *         embs = self._cliques['raw']             # <<<<<<<<<<<<<<
 *         keys = self._cliques['size']
 *         if keys:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_raw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_embs = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":514
 *         self._ensure_clique_cache()
 *         embs = self._cliques['raw']
 *         keys = self._cliques['size']             # <<<<<<<<<<<<<<
 *         if keys:
 *             return self._graph.relabel(dict(enumerate(embs[keys[max(keys)]])))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_keys = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":515
 *         embs = self._cliques['raw']
 *         keys = self._cliques['size']
 *         if keys:             # <<<<<<<<<<<<<<
 *             return self._graph.relabel(dict(enumerate(embs[keys[max(keys)]])))
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_keys); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 515, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":516
 *         keys = self._cliques['size']
 *         if keys:
 *             return self._graph.relabel(dict(enumerate(embs[keys[max(keys)]])))             # <<<<<<<<<<<<<<
 *         else:
 *             return {}
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_relabel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_keys, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_embs, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":515
 *         embs = self._cliques['raw']
 *         keys = self._cliques['size']
 *         if keys:             # <<<<<<<<<<<<<<
 *             return self._graph.relabel(dict(enumerate(embs[keys[max(keys)]])))
 *         else:
 */
  }

  /* "minorminer/busclique.pyx":518
 *             return self._graph.relabel(dict(enumerate(embs[keys[max(keys)]])))
 *         else:
 *             return {}             # <<<<<<<<<<<<<<
 * 
 *     def largest_clique_by_chainlength(self, chainlength):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":501
 *         return cache
 * 
 *     def largest_clique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.largest_clique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_embs);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":520
 *             return {}
 * 
 *     def largest_clique_by_chainlength(self, chainlength):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache, with a specified
 *         maximum ``chainlength``.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_23largest_clique_by_chainlength(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_22largest_clique_by_chainlength[] = "Returns the largest-found clique in the clique cache, with a specified\n        maximum ``chainlength``.\n\n        This will compute the entire clique cache if it is missing from the\n        filesystem.\n\n        Args:\n            chainlength (int):\n                Max chain length.\n\n        Returns:\n            dict: An embedding of node labels from ``range(len(embedding))``\n            mapped to chains of the largest-found clique with maximum ``chainlength``.\n\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_23largest_clique_by_chainlength = {"largest_clique_by_chainlength", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_23largest_clique_by_chainlength, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_22largest_clique_by_chainlength};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_23largest_clique_by_chainlength(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_chainlength = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("largest_clique_by_chainlength (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_chainlength,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chainlength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("largest_clique_by_chainlength", 1, 2, 2, 1); __PYX_ERR(0, 520, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "largest_clique_by_chainlength") < 0)) __PYX_ERR(0, 520, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_chainlength = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("largest_clique_by_chainlength", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 520, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.largest_clique_by_chainlength", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_22largest_clique_by_chainlength(__pyx_self, __pyx_v_self, __pyx_v_chainlength);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_22largest_clique_by_chainlength(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_chainlength) {
  PyObject *__pyx_v_embs = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("largest_clique_by_chainlength", 0);

  /* "minorminer/busclique.pyx":536
 * 
 *         """
 *         self._ensure_clique_cache()             # <<<<<<<<<<<<<<
 *         embs = self._cliques['raw']
 *         length = self._cliques['length'] or [0]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":537
 *         """
 *         self._ensure_clique_cache()
 *         embs = self._cliques['raw']             # <<<<<<<<<<<<<<
 *         length = self._cliques['length'] or [0]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_raw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_embs = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":538
 *         self._ensure_clique_cache()
 *         embs = self._cliques['raw']
 *         length = self._cliques['length'] or [0]             # <<<<<<<<<<<<<<
 * 
 *         if 0 < chainlength <= max(length):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 538, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_v_length = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":540
 *         length = self._cliques['length'] or [0]
 * 
 *         if 0 < chainlength <= max(length):             # <<<<<<<<<<<<<<
 *             return self._graph.relabel(dict(enumerate(embs[length[chainlength]])))
 *         else:
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_int_0, __pyx_v_chainlength, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  if (__Pyx_PyObject_IsTrue(__pyx_t_2)) {
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_chainlength, __pyx_t_3, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":541
 * 
 *         if 0 < chainlength <= max(length):
 *             return self._graph.relabel(dict(enumerate(embs[length[chainlength]])))             # <<<<<<<<<<<<<<
 *         else:
 *             return {}
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_relabel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_length, __pyx_v_chainlength); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_embs, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":540
 *         length = self._cliques['length'] or [0]
 * 
 *         if 0 < chainlength <= max(length):             # <<<<<<<<<<<<<<
 *             return self._graph.relabel(dict(enumerate(embs[length[chainlength]])))
 *         else:
 */
  }

  /* "minorminer/busclique.pyx":543
 *             return self._graph.relabel(dict(enumerate(embs[length[chainlength]])))
 *         else:
 *             return {}             # <<<<<<<<<<<<<<
 * 
 *     def find_clique_embedding(self, nn):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":520
 *             return {}
 * 
 *     def largest_clique_by_chainlength(self, chainlength):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache, with a specified
 *         maximum ``chainlength``.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.largest_clique_by_chainlength", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_embs);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":545
 *             return {}
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """Returns a clique embedding, minimizing the maximum chainlength given
 *         its size.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_25find_clique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_24find_clique_embedding[] = "Returns a clique embedding, minimizing the maximum chainlength given\n        its size.\n\n        This will compute the entire clique cache if it is missing from\n        the filesystem.\n\n        Args:\n            nn (int/iterable):\n                A number (indicating the size of the desired clique) or an\n                iterable (specifying the node labels of the desired clique).\n\n        Returns:\n            dict: An embedding of node labels (either ``nn``, or ``range(nn)``)\n            mapped to chains of a clique embedding.\n\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_25find_clique_embedding = {"find_clique_embedding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_25find_clique_embedding, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_24find_clique_embedding};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_25find_clique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_clique_embedding (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nn,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_clique_embedding", 1, 2, 2, 1); __PYX_ERR(0, 545, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_clique_embedding") < 0)) __PYX_ERR(0, 545, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_nn = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_clique_embedding", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 545, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_24find_clique_embedding(__pyx_self, __pyx_v_self, __pyx_v_nn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_24find_clique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_keys = NULL;
  PyObject *__pyx_v_raw = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_emb = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_clique_embedding", 0);

  /* "minorminer/busclique.pyx":562
 * 
 *         """
 *         num, nodes = _num_nodes(nn)             # <<<<<<<<<<<<<<
 *         self._ensure_clique_cache()
 *         keys = self._cliques['size']
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_num_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_nn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_nn);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 562, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 562, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 562, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_num = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":563
 *         """
 *         num, nodes = _num_nodes(nn)
 *         self._ensure_clique_cache()             # <<<<<<<<<<<<<<
 *         keys = self._cliques['size']
 *         raw = self._cliques['raw']
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_clique_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":564
 *         num, nodes = _num_nodes(nn)
 *         self._ensure_clique_cache()
 *         keys = self._cliques['size']             # <<<<<<<<<<<<<<
 *         raw = self._cliques['raw']
 *         key = keys.get(num)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_keys = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":565
 *         self._ensure_clique_cache()
 *         keys = self._cliques['size']
 *         raw = self._cliques['raw']             # <<<<<<<<<<<<<<
 *         key = keys.get(num)
 *         if key is None:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cliques); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_raw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_raw = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":566
 *         keys = self._cliques['size']
 *         raw = self._cliques['raw']
 *         key = keys.get(num)             # <<<<<<<<<<<<<<
 *         if key is None:
 *             return {}
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_keys, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_num) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_num);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_key = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":567
 *         raw = self._cliques['raw']
 *         key = keys.get(num)
 *         if key is None:             # <<<<<<<<<<<<<<
 *             return {}
 *         emb = dict(zip(nodes, raw[key]))
 */
  __pyx_t_6 = (__pyx_v_key == Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "minorminer/busclique.pyx":568
 *         key = keys.get(num)
 *         if key is None:
 *             return {}             # <<<<<<<<<<<<<<
 *         emb = dict(zip(nodes, raw[key]))
 *         return self._graph.relabel(emb)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":567
 *         raw = self._cliques['raw']
 *         key = keys.get(num)
 *         if key is None:             # <<<<<<<<<<<<<<
 *             return {}
 *         emb = dict(zip(nodes, raw[key]))
 */
  }

  /* "minorminer/busclique.pyx":569
 *         if key is None:
 *             return {}
 *         emb = dict(zip(nodes, raw[key]))             # <<<<<<<<<<<<<<
 *         return self._graph.relabel(emb)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_raw, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_emb = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":570
 *             return {}
 *         emb = dict(zip(nodes, raw[key]))
 *         return self._graph.relabel(emb)             # <<<<<<<<<<<<<<
 * 
 *     def largest_balanced_biclique(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_relabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_emb) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_emb);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":545
 *             return {}
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """Returns a clique embedding, minimizing the maximum chainlength given
 *         its size.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":572
 *         return self._graph.relabel(emb)
 * 
 *     def largest_balanced_biclique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-size biclique where both sides have equal size.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_27largest_balanced_biclique(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_26largest_balanced_biclique[] = "Returns the largest-size biclique where both sides have equal size.\n\n        Nodes of the embedding dict are from ``range(len(embedding))``, where the\n        nodes ``range(len(embedding)/2)`` are completely connected to the nodes\n        ``range(len(embedding)/2, len(embedding))``.\n\n        This will compute the entire biclique cache if it is missing from the\n        filesystem.\n\n        Returns:\n            dict: An embedding of node labels (described above) mapped to chains\n            of the largest balanced biclique.\n\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_27largest_balanced_biclique = {"largest_balanced_biclique", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_27largest_balanced_biclique, METH_O, __pyx_doc_10minorminer_9busclique_14busgraph_cache_26largest_balanced_biclique};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_27largest_balanced_biclique(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("largest_balanced_biclique (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_26largest_balanced_biclique(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":590
 *         biggest = self._bicliques['max_side']
 *         embs = self._bicliques['raw']
 *         s0, s1 = key = max(embs, key=lambda x: min(x))             # <<<<<<<<<<<<<<
 *         raw_emb = embs[key]
 *         raw_emb0 = sorted(raw_emb[:s0], key=len)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_25largest_balanced_biclique_lambda3(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_25largest_balanced_biclique_lambda3 = {"lambda3", (PyCFunction)__pyx_pw_10minorminer_9busclique_14busgraph_cache_25largest_balanced_biclique_lambda3, METH_O, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_25largest_balanced_biclique_lambda3(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda3 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda3(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda3", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.largest_balanced_biclique.lambda3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":572
 *         return self._graph.relabel(emb)
 * 
 *     def largest_balanced_biclique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-size biclique where both sides have equal size.
 * 
 */

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_26largest_balanced_biclique(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *__pyx_v_biggest = NULL;
  PyObject *__pyx_v_embs = NULL;
  PyObject *__pyx_v_s0 = NULL;
  PyObject *__pyx_v_s1 = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_raw_emb = NULL;
  PyObject *__pyx_v_raw_emb0 = NULL;
  PyObject *__pyx_v_raw_emb1 = NULL;
  PyObject *__pyx_v_s_min = NULL;
  PyObject *__pyx_v_emb0 = NULL;
  PyObject *__pyx_v_emb1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("largest_balanced_biclique", 0);

  /* "minorminer/busclique.pyx":587
 * 
 *         """
 *         self._ensure_biclique_cache()             # <<<<<<<<<<<<<<
 *         biggest = self._bicliques['max_side']
 *         embs = self._bicliques['raw']
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_biclique_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":588
 *         """
 *         self._ensure_biclique_cache()
 *         biggest = self._bicliques['max_side']             # <<<<<<<<<<<<<<
 *         embs = self._bicliques['raw']
 *         s0, s1 = key = max(embs, key=lambda x: min(x))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bicliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_max_side); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_biggest = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":589
 *         self._ensure_biclique_cache()
 *         biggest = self._bicliques['max_side']
 *         embs = self._bicliques['raw']             # <<<<<<<<<<<<<<
 *         s0, s1 = key = max(embs, key=lambda x: min(x))
 *         raw_emb = embs[key]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bicliques); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_raw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_embs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":590
 *         biggest = self._bicliques['max_side']
 *         embs = self._bicliques['raw']
 *         s0, s1 = key = max(embs, key=lambda x: min(x))             # <<<<<<<<<<<<<<
 *         raw_emb = embs[key]
 *         raw_emb0 = sorted(raw_emb[:s0], key=len)
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_embs);
  __Pyx_GIVEREF(__pyx_v_embs);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_embs);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_25largest_balanced_biclique_lambda3, 0, __pyx_n_s_busgraph_cache_largest_balanced, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 590, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 590, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 590, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_s0 = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_s1 = __pyx_t_1;
  __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_key = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":591
 *         embs = self._bicliques['raw']
 *         s0, s1 = key = max(embs, key=lambda x: min(x))
 *         raw_emb = embs[key]             # <<<<<<<<<<<<<<
 *         raw_emb0 = sorted(raw_emb[:s0], key=len)
 *         raw_emb1 = sorted(raw_emb[s0:s0+s1], key=len)
 */
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_embs, __pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_raw_emb = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":592
 *         s0, s1 = key = max(embs, key=lambda x: min(x))
 *         raw_emb = embs[key]
 *         raw_emb0 = sorted(raw_emb[:s0], key=len)             # <<<<<<<<<<<<<<
 *         raw_emb1 = sorted(raw_emb[s0:s0+s1], key=len)
 *         s_min = min(s0, s1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_raw_emb, 0, 0, NULL, &__pyx_v_s0, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_key, __pyx_t_2) < 0) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_raw_emb0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":593
 *         raw_emb = embs[key]
 *         raw_emb0 = sorted(raw_emb[:s0], key=len)
 *         raw_emb1 = sorted(raw_emb[s0:s0+s1], key=len)             # <<<<<<<<<<<<<<
 *         s_min = min(s0, s1)
 *         emb0 = raw_emb0[:s_min]
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_s0, __pyx_v_s1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_raw_emb, 0, 0, &__pyx_v_s0, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_key, __pyx_t_1) < 0) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_raw_emb1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":594
 *         raw_emb0 = sorted(raw_emb[:s0], key=len)
 *         raw_emb1 = sorted(raw_emb[s0:s0+s1], key=len)
 *         s_min = min(s0, s1)             # <<<<<<<<<<<<<<
 *         emb0 = raw_emb0[:s_min]
 *         emb1 = raw_emb1[:s_min]
 */
  __Pyx_INCREF(__pyx_v_s1);
  __pyx_t_1 = __pyx_v_s1;
  __Pyx_INCREF(__pyx_v_s0);
  __pyx_t_3 = __pyx_v_s0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 594, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_s_min = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":595
 *         raw_emb1 = sorted(raw_emb[s0:s0+s1], key=len)
 *         s_min = min(s0, s1)
 *         emb0 = raw_emb0[:s_min]             # <<<<<<<<<<<<<<
 *         emb1 = raw_emb1[:s_min]
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_raw_emb0, 0, 0, NULL, &__pyx_v_s_min, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_emb0 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":596
 *         s_min = min(s0, s1)
 *         emb0 = raw_emb0[:s_min]
 *         emb1 = raw_emb1[:s_min]             # <<<<<<<<<<<<<<
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_raw_emb1, 0, 0, NULL, &__pyx_v_s_min, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_emb1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":597
 *         emb0 = raw_emb0[:s_min]
 *         emb1 = raw_emb1[:s_min]
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))             # <<<<<<<<<<<<<<
 * 
 *     def find_biclique_embedding(self, nn, mm):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_relabel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_v_emb0, __pyx_v_emb1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":572
 *         return self._graph.relabel(emb)
 * 
 *     def largest_balanced_biclique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-size biclique where both sides have equal size.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.largest_balanced_biclique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_biggest);
  __Pyx_XDECREF(__pyx_v_embs);
  __Pyx_XDECREF(__pyx_v_s0);
  __Pyx_XDECREF(__pyx_v_s1);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_raw_emb);
  __Pyx_XDECREF(__pyx_v_raw_emb0);
  __Pyx_XDECREF(__pyx_v_raw_emb1);
  __Pyx_XDECREF(__pyx_v_s_min);
  __Pyx_XDECREF(__pyx_v_emb0);
  __Pyx_XDECREF(__pyx_v_emb1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":599
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))
 * 
 *     def find_biclique_embedding(self, nn, mm):             # <<<<<<<<<<<<<<
 *         """Returns a biclique embedding, minimizing the maximum chain length
 *         given its size.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_29find_biclique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_14busgraph_cache_28find_biclique_embedding[] = "Returns a biclique embedding, minimizing the maximum chain length\n        given its size.\n\n        This will compute the entire biclique cache if it is missing from the\n        filesystem.\n\n        Args:\n            nn (int/iterable):\n                A number (indicating the size of one side of the desired biclique)\n                or an iterable (specifying the node labels of one side the desired\n                biclique).\n\n            mm (int/iterable):\n                Same as ``nn``, for the other side of the desired biclique.\n\n        In the case that ``nn`` is a number, the first side will have nodes\n        labeled from ``range(nn)``. In the case that ``mm`` is a number, the\n        second side will have nodes labeled from ``range(n, n + mm)``; where\n        ``n`` is either ``nn`` or ``len(nn)``.\n\n        Returns:\n            dict: An embedding of node labels (described above) mapped to chains\n            of a biclique embedding.\n        ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_29find_biclique_embedding = {"find_biclique_embedding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_29find_biclique_embedding, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_14busgraph_cache_28find_biclique_embedding};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_29find_biclique_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nn = 0;
  PyObject *__pyx_v_mm = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_biclique_embedding (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nn,&__pyx_n_s_mm,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_biclique_embedding", 1, 3, 3, 1); __PYX_ERR(0, 599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mm)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_biclique_embedding", 1, 3, 3, 2); __PYX_ERR(0, 599, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_biclique_embedding") < 0)) __PYX_ERR(0, 599, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_nn = values[1];
    __pyx_v_mm = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_biclique_embedding", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 599, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.find_biclique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_28find_biclique_embedding(__pyx_self, __pyx_v_self, __pyx_v_nn, __pyx_v_mm);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_28find_biclique_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nn, PyObject *__pyx_v_mm) {
  PyObject *__pyx_v_biggest = NULL;
  PyObject *__pyx_v_by_size = NULL;
  PyObject *__pyx_v_raw = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_N = NULL;
  PyObject *__pyx_v_m = NULL;
  PyObject *__pyx_v_M = NULL;
  PyObject *__pyx_v_emb = NULL;
  PyObject *__pyx_v_nmax = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_mmax = NULL;
  PyObject *__pyx_v_row = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_s0 = NULL;
  PyObject *__pyx_v_s1 = NULL;
  PyObject *__pyx_v_emb0 = NULL;
  PyObject *__pyx_v_emb1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_biclique_embedding", 0);

  /* "minorminer/busclique.pyx":624
 *             of a biclique embedding.
 *         """
 *         self._ensure_biclique_cache()             # <<<<<<<<<<<<<<
 *         biggest = self._bicliques['max_side']
 *         by_size = self._bicliques['size']
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_biclique_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":625
 *         """
 *         self._ensure_biclique_cache()
 *         biggest = self._bicliques['max_side']             # <<<<<<<<<<<<<<
 *         by_size = self._bicliques['size']
 *         raw = self._bicliques['raw']
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bicliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_max_side); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_biggest = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":626
 *         self._ensure_biclique_cache()
 *         biggest = self._bicliques['max_side']
 *         by_size = self._bicliques['size']             # <<<<<<<<<<<<<<
 *         raw = self._bicliques['raw']
 *         n, N = _num_nodes(nn)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bicliques); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_by_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":627
 *         biggest = self._bicliques['max_side']
 *         by_size = self._bicliques['size']
 *         raw = self._bicliques['raw']             # <<<<<<<<<<<<<<
 *         n, N = _num_nodes(nn)
 *         m, M = _num_nodes(mm, offset = n)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bicliques); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_raw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_raw = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":628
 *         by_size = self._bicliques['size']
 *         raw = self._bicliques['raw']
 *         n, N = _num_nodes(nn)             # <<<<<<<<<<<<<<
 *         m, M = _num_nodes(mm, offset = n)
 *         if(m < n):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_num_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_nn) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_nn);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 628, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 628, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 628, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_n = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_N = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":629
 *         raw = self._bicliques['raw']
 *         n, N = _num_nodes(nn)
 *         m, M = _num_nodes(mm, offset = n)             # <<<<<<<<<<<<<<
 *         if(m < n):
 *             n, N, m, M = m, M, n, N
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_num_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_mm);
  __Pyx_GIVEREF(__pyx_v_mm);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_mm);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_offset, __pyx_v_n) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 629, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 2) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 629, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_v_m = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_M = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":630
 *         n, N = _num_nodes(nn)
 *         m, M = _num_nodes(mm, offset = n)
 *         if(m < n):             # <<<<<<<<<<<<<<
 *             n, N, m, M = m, M, n, N
 *         if m == 0:
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_m, __pyx_v_n, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 630, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":631
 *         m, M = _num_nodes(mm, offset = n)
 *         if(m < n):
 *             n, N, m, M = m, M, n, N             # <<<<<<<<<<<<<<
 *         if m == 0:
 *             emb = self._graph.independent_set(n)
 */
    __pyx_t_7 = __pyx_v_m;
    __pyx_t_8 = __pyx_v_M;
    __pyx_t_9 = __pyx_v_n;
    __pyx_t_10 = __pyx_v_N;
    __pyx_v_n = __pyx_t_7;
    __pyx_t_7 = 0;
    __pyx_v_N = __pyx_t_8;
    __pyx_t_8 = 0;
    __pyx_v_m = __pyx_t_9;
    __pyx_t_9 = 0;
    __pyx_v_M = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":630
 *         n, N = _num_nodes(nn)
 *         m, M = _num_nodes(mm, offset = n)
 *         if(m < n):             # <<<<<<<<<<<<<<
 *             n, N, m, M = m, M, n, N
 *         if m == 0:
 */
  }

  /* "minorminer/busclique.pyx":632
 *         if(m < n):
 *             n, N, m, M = m, M, n, N
 *         if m == 0:             # <<<<<<<<<<<<<<
 *             emb = self._graph.independent_set(n)
 *             if emb is None:
 */
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_v_m, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":633
 *             n, N, m, M = m, M, n, N
 *         if m == 0:
 *             emb = self._graph.independent_set(n)             # <<<<<<<<<<<<<<
 *             if emb is None:
 *                 return {}
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_independent_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_n) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_n);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_emb = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":634
 *         if m == 0:
 *             emb = self._graph.independent_set(n)
 *             if emb is None:             # <<<<<<<<<<<<<<
 *                 return {}
 *             return dict(zip(N, emb))
 */
    __pyx_t_6 = (__pyx_v_emb == Py_None);
    __pyx_t_11 = (__pyx_t_6 != 0);
    if (__pyx_t_11) {

      /* "minorminer/busclique.pyx":635
 *             emb = self._graph.independent_set(n)
 *             if emb is None:
 *                 return {}             # <<<<<<<<<<<<<<
 *             return dict(zip(N, emb))
 *         nmax = biggest[None]
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "minorminer/busclique.pyx":634
 *         if m == 0:
 *             emb = self._graph.independent_set(n)
 *             if emb is None:             # <<<<<<<<<<<<<<
 *                 return {}
 *             return dict(zip(N, emb))
 */
    }

    /* "minorminer/busclique.pyx":636
 *             if emb is None:
 *                 return {}
 *             return dict(zip(N, emb))             # <<<<<<<<<<<<<<
 *         nmax = biggest[None]
 *         for i in range(n, nmax + 1):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_N);
    __Pyx_GIVEREF(__pyx_v_N);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_N);
    __Pyx_INCREF(__pyx_v_emb);
    __Pyx_GIVEREF(__pyx_v_emb);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_emb);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":632
 *         if(m < n):
 *             n, N, m, M = m, M, n, N
 *         if m == 0:             # <<<<<<<<<<<<<<
 *             emb = self._graph.independent_set(n)
 *             if emb is None:
 */
  }

  /* "minorminer/busclique.pyx":637
 *                 return {}
 *             return dict(zip(N, emb))
 *         nmax = biggest[None]             # <<<<<<<<<<<<<<
 *         for i in range(n, nmax + 1):
 *             mmax = biggest.get(i, 0)
 */
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_biggest, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_nmax = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "minorminer/busclique.pyx":638
 *             return dict(zip(N, emb))
 *         nmax = biggest[None]
 *         for i in range(n, nmax + 1):             # <<<<<<<<<<<<<<
 *             mmax = biggest.get(i, 0)
 *             if mmax < m:
 */
  __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_nmax, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_n);
  __Pyx_GIVEREF(__pyx_v_n);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_n);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_12 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_12 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 638, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 638, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 638, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_13(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 638, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":639
 *         nmax = biggest[None]
 *         for i in range(n, nmax + 1):
 *             mmax = biggest.get(i, 0)             # <<<<<<<<<<<<<<
 *             if mmax < m:
 *                 continue
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_biggest, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_14 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_14 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_i, __pyx_int_0};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_i, __pyx_int_0};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_15 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_v_i);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, __pyx_int_0);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_15, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_mmax, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":640
 *         for i in range(n, nmax + 1):
 *             mmax = biggest.get(i, 0)
 *             if mmax < m:             # <<<<<<<<<<<<<<
 *                 continue
 *             row = by_size.get(i, {0:None})
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_mmax, __pyx_v_m, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 640, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_11) {

      /* "minorminer/busclique.pyx":641
 *             mmax = biggest.get(i, 0)
 *             if mmax < m:
 *                 continue             # <<<<<<<<<<<<<<
 *             row = by_size.get(i, {0:None})
 *             for j in range(m, mmax + 1):
 */
      goto __pyx_L10_continue;

      /* "minorminer/busclique.pyx":640
 *         for i in range(n, nmax + 1):
 *             mmax = biggest.get(i, 0)
 *             if mmax < m:             # <<<<<<<<<<<<<<
 *                 continue
 *             row = by_size.get(i, {0:None})
 */
    }

    /* "minorminer/busclique.pyx":642
 *             if mmax < m:
 *                 continue
 *             row = by_size.get(i, {0:None})             # <<<<<<<<<<<<<<
 *             for j in range(m, mmax + 1):
 *                 key = row.get(j)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_by_size, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (PyDict_SetItem(__pyx_t_15, __pyx_int_0, Py_None) < 0) __PYX_ERR(0, 642, __pyx_L1_error)
    __pyx_t_2 = NULL;
    __pyx_t_14 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_14 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_i, __pyx_t_15};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_i, __pyx_t_15};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else
    #endif
    {
      __pyx_t_16 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_14, __pyx_v_i);
      __Pyx_GIVEREF(__pyx_t_15);
      PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_14, __pyx_t_15);
      __pyx_t_15 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_row, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":643
 *                 continue
 *             row = by_size.get(i, {0:None})
 *             for j in range(m, mmax + 1):             # <<<<<<<<<<<<<<
 *                 key = row.get(j)
 *                 if key is None:
 */
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_mmax, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_m);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3); __pyx_t_17 = 0;
      __pyx_t_18 = NULL;
    } else {
      __pyx_t_17 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_18 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 643, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_18)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_17); __Pyx_INCREF(__pyx_t_4); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_17); __Pyx_INCREF(__pyx_t_4); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_18(__pyx_t_3);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 643, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":644
 *             row = by_size.get(i, {0:None})
 *             for j in range(m, mmax + 1):
 *                 key = row.get(j)             # <<<<<<<<<<<<<<
 *                 if key is None:
 *                     continue
 */
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_row, __pyx_n_s_get); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_15 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_16, function);
        }
      }
      __pyx_t_4 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_15, __pyx_v_j) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_v_j);
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":645
 *             for j in range(m, mmax + 1):
 *                 key = row.get(j)
 *                 if key is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 emb = raw[key]
 */
      __pyx_t_11 = (__pyx_v_key == Py_None);
      __pyx_t_6 = (__pyx_t_11 != 0);
      if (__pyx_t_6) {

        /* "minorminer/busclique.pyx":646
 *                 key = row.get(j)
 *                 if key is None:
 *                     continue             # <<<<<<<<<<<<<<
 *                 emb = raw[key]
 *                 s0, s1 = key
 */
        goto __pyx_L13_continue;

        /* "minorminer/busclique.pyx":645
 *             for j in range(m, mmax + 1):
 *                 key = row.get(j)
 *                 if key is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 emb = raw[key]
 */
      }

      /* "minorminer/busclique.pyx":647
 *                 if key is None:
 *                     continue
 *                 emb = raw[key]             # <<<<<<<<<<<<<<
 *                 s0, s1 = key
 *                 emb0 = emb[:s0]
 */
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_raw, __pyx_v_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 647, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_emb = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":648
 *                     continue
 *                 emb = raw[key]
 *                 s0, s1 = key             # <<<<<<<<<<<<<<
 *                 emb0 = emb[:s0]
 *                 emb1 = emb[s0:]
 */
      if ((likely(PyTuple_CheckExact(__pyx_v_key))) || (PyList_CheckExact(__pyx_v_key))) {
        PyObject* sequence = __pyx_v_key;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 648, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_16);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_15 = PyObject_GetIter(__pyx_v_key); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_5 = Py_TYPE(__pyx_t_15)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_5(__pyx_t_15); if (unlikely(!__pyx_t_4)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_16 = __pyx_t_5(__pyx_t_15); if (unlikely(!__pyx_t_16)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_16);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_15), 2) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
        __pyx_t_5 = NULL;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_5 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 648, __pyx_L1_error)
        __pyx_L17_unpacking_done:;
      }
      __pyx_v_s0 = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v_s1 = __pyx_t_16;
      __pyx_t_16 = 0;

      /* "minorminer/busclique.pyx":649
 *                 emb = raw[key]
 *                 s0, s1 = key
 *                 emb0 = emb[:s0]             # <<<<<<<<<<<<<<
 *                 emb1 = emb[s0:]
 *                 if s0 <= n and s1 <= m:
 */
      __pyx_t_16 = __Pyx_PyObject_GetSlice(__pyx_v_emb, 0, 0, NULL, &__pyx_v_s0, NULL, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_v_emb0 = __pyx_t_16;
      __pyx_t_16 = 0;

      /* "minorminer/busclique.pyx":650
 *                 s0, s1 = key
 *                 emb0 = emb[:s0]
 *                 emb1 = emb[s0:]             # <<<<<<<<<<<<<<
 *                 if s0 <= n and s1 <= m:
 *                     emb = dict(zip(N + M, emb0[:n] + emb1))
 */
      __pyx_t_16 = __Pyx_PyObject_GetSlice(__pyx_v_emb, 0, 0, &__pyx_v_s0, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_v_emb1 = __pyx_t_16;
      __pyx_t_16 = 0;

      /* "minorminer/busclique.pyx":651
 *                 emb0 = emb[:s0]
 *                 emb1 = emb[s0:]
 *                 if s0 <= n and s1 <= m:             # <<<<<<<<<<<<<<
 *                     emb = dict(zip(N + M, emb0[:n] + emb1))
 *                 else:
 */
      __pyx_t_16 = PyObject_RichCompare(__pyx_v_s0, __pyx_v_n, Py_LE); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 651, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      if (__pyx_t_11) {
      } else {
        __pyx_t_6 = __pyx_t_11;
        goto __pyx_L19_bool_binop_done;
      }
      __pyx_t_16 = PyObject_RichCompare(__pyx_v_s1, __pyx_v_m, Py_LE); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 651, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_6 = __pyx_t_11;
      __pyx_L19_bool_binop_done:;
      if (__pyx_t_6) {

        /* "minorminer/busclique.pyx":652
 *                 emb1 = emb[s0:]
 *                 if s0 <= n and s1 <= m:
 *                     emb = dict(zip(N + M, emb0[:n] + emb1))             # <<<<<<<<<<<<<<
 *                 else:
 *                     emb = dict(zip(N + M, emb1[:n] + emb0))
 */
        __pyx_t_16 = PyNumber_Add(__pyx_v_N, __pyx_v_M); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_emb0, 0, 0, NULL, &__pyx_v_n, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_15 = PyNumber_Add(__pyx_t_4, __pyx_v_emb1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_15);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_15);
        __pyx_t_16 = 0;
        __pyx_t_15 = 0;
        __pyx_t_15 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF_SET(__pyx_v_emb, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "minorminer/busclique.pyx":651
 *                 emb0 = emb[:s0]
 *                 emb1 = emb[s0:]
 *                 if s0 <= n and s1 <= m:             # <<<<<<<<<<<<<<
 *                     emb = dict(zip(N + M, emb0[:n] + emb1))
 *                 else:
 */
        goto __pyx_L18;
      }

      /* "minorminer/busclique.pyx":654
 *                     emb = dict(zip(N + M, emb0[:n] + emb1))
 *                 else:
 *                     emb = dict(zip(N + M, emb1[:n] + emb0))             # <<<<<<<<<<<<<<
 *                 return self._graph.relabel(emb)
 *         return {}
 */
      /*else*/ {
        __pyx_t_4 = PyNumber_Add(__pyx_v_N, __pyx_v_M); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_15 = __Pyx_PyObject_GetSlice(__pyx_v_emb1, 0, 0, NULL, &__pyx_v_n, NULL, 0, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_16 = PyNumber_Add(__pyx_t_15, __pyx_v_emb0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_16);
        __pyx_t_4 = 0;
        __pyx_t_16 = 0;
        __pyx_t_16 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_15, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF_SET(__pyx_v_emb, __pyx_t_15);
        __pyx_t_15 = 0;
      }
      __pyx_L18:;

      /* "minorminer/busclique.pyx":655
 *                 else:
 *                     emb = dict(zip(N + M, emb1[:n] + emb0))
 *                 return self._graph.relabel(emb)             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_relabel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_15 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_16, __pyx_v_emb) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_emb);
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_15;
      __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "minorminer/busclique.pyx":643
 *                 continue
 *             row = by_size.get(i, {0:None})
 *             for j in range(m, mmax + 1):             # <<<<<<<<<<<<<<
 *                 key = row.get(j)
 *                 if key is None:
 */
      __pyx_L13_continue:;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":638
 *             return dict(zip(N, emb))
 *         nmax = biggest[None]
 *         for i in range(n, nmax + 1):             # <<<<<<<<<<<<<<
 *             mmax = biggest.get(i, 0)
 *             if mmax < m:
 */
    __pyx_L10_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":656
 *                     emb = dict(zip(N + M, emb1[:n] + emb0))
 *                 return self._graph.relabel(emb)
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     def draw_fragment_embedding(self, emb, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":599
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))
 * 
 *     def find_biclique_embedding(self, nn, mm):             # <<<<<<<<<<<<<<
 *         """Returns a biclique embedding, minimizing the maximum chain length
 *         given its size.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.find_biclique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_biggest);
  __Pyx_XDECREF(__pyx_v_by_size);
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_N);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_M);
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XDECREF(__pyx_v_nmax);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_mmax);
  __Pyx_XDECREF(__pyx_v_row);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_s0);
  __Pyx_XDECREF(__pyx_v_s1);
  __Pyx_XDECREF(__pyx_v_emb0);
  __Pyx_XDECREF(__pyx_v_emb1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":658
 *         return {}
 * 
 *     def draw_fragment_embedding(self, emb, **kwargs):             # <<<<<<<<<<<<<<
 *         m, n, t, nodes, edges = self._graph.fragment_graph_spec()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_31draw_fragment_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_14busgraph_cache_31draw_fragment_embedding = {"draw_fragment_embedding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_14busgraph_cache_31draw_fragment_embedding, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_14busgraph_cache_31draw_fragment_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_emb = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_fragment_embedding (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_emb,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_emb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_fragment_embedding", 1, 2, 2, 1); __PYX_ERR(0, 658, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "draw_fragment_embedding") < 0)) __PYX_ERR(0, 658, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_emb = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_fragment_embedding", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 658, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.draw_fragment_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_14busgraph_cache_30draw_fragment_embedding(__pyx_self, __pyx_v_self, __pyx_v_emb, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_14busgraph_cache_30draw_fragment_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_emb, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_m = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_nodes = NULL;
  PyObject *__pyx_v_edges = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_f_emb = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_k = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("draw_fragment_embedding", 0);

  /* "minorminer/busclique.pyx":659
 * 
 *     def draw_fragment_embedding(self, emb, **kwargs):
 *         m, n, t, nodes, edges = self._graph.fragment_graph_spec()             # <<<<<<<<<<<<<<
 * 
 *         f = dnx.chimera_graph(m, n=n, t=t, node_list=nodes, edge_list=edges)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fragment_graph_spec); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 5)) {
      if (size > 5) __Pyx_RaiseTooManyValuesError(5);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 659, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 4); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 4); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6};
      for (i=0; i < 5; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 659, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6};
    __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
    for (index=0; index < 5; index++) {
      PyObject* item = __pyx_t_8(__pyx_t_7); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 5) < 0) __PYX_ERR(0, 659, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 659, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_m = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_n = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_t = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_nodes = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_edges = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "minorminer/busclique.pyx":661
 *         m, n, t, nodes, edges = self._graph.fragment_graph_spec()
 * 
 *         f = dnx.chimera_graph(m, n=n, t=t, node_list=nodes, edge_list=edges)             # <<<<<<<<<<<<<<
 *         f_emb = {
 *             k : self._graph.fragment_nodes(c)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_dnx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_chimera_graph); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_m);
  __Pyx_GIVEREF(__pyx_v_m);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_m);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_n, __pyx_v_n) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_t, __pyx_v_t) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_node_list, __pyx_v_nodes) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_edge_list, __pyx_v_edges) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_f = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "minorminer/busclique.pyx":662
 * 
 *         f = dnx.chimera_graph(m, n=n, t=t, node_list=nodes, edge_list=edges)
 *         f_emb = {             # <<<<<<<<<<<<<<
 *             k : self._graph.fragment_nodes(c)
 *             for k, c in self._graph.delabel(emb).items()
 */
  { /* enter inner scope */
    __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "minorminer/busclique.pyx":664
 *         f_emb = {
 *             k : self._graph.fragment_nodes(c)
 *             for k, c in self._graph.delabel(emb).items()             # <<<<<<<<<<<<<<
 *         }
 *         dnx.draw_chimera_embedding(f, f_emb, **kwargs)
 */
    __pyx_t_9 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 664, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_delabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_emb) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_emb);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 664, __pyx_L7_error)
    }
    __pyx_t_2 = __Pyx_dict_iterator(__pyx_t_1, 0, __pyx_n_s_items, (&__pyx_t_10), (&__pyx_t_11)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_2;
    __pyx_t_2 = 0;
    while (1) {
      __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_10, &__pyx_t_9, &__pyx_t_2, &__pyx_t_1, NULL, __pyx_t_11);
      if (unlikely(__pyx_t_12 == 0)) break;
      if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 664, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_k, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_c, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":663
 *         f = dnx.chimera_graph(m, n=n, t=t, node_list=nodes, edge_list=edges)
 *         f_emb = {
 *             k : self._graph.fragment_nodes(c)             # <<<<<<<<<<<<<<
 *             for k, c in self._graph.delabel(emb).items()
 *         }
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fragment_nodes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_8genexpr5__pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_8genexpr5__pyx_v_c);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_4, (PyObject*)__pyx_8genexpr5__pyx_v_k, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 663, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_c); __pyx_8genexpr5__pyx_v_c = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_k); __pyx_8genexpr5__pyx_v_k = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_c); __pyx_8genexpr5__pyx_v_c = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_k); __pyx_8genexpr5__pyx_v_k = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_f_emb = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "minorminer/busclique.pyx":666
 *             for k, c in self._graph.delabel(emb).items()
 *         }
 *         dnx.draw_chimera_embedding(f, f_emb, **kwargs)             # <<<<<<<<<<<<<<
 * 
 * cdef dict _make_clique_cache(vector[embedding_t] &embs):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_dnx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_draw_chimera_embedding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_f);
  __Pyx_GIVEREF(__pyx_v_f);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_f);
  __Pyx_INCREF(__pyx_v_f_emb);
  __Pyx_GIVEREF(__pyx_v_f_emb);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_f_emb);
  __pyx_t_1 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "minorminer/busclique.pyx":658
 *         return {}
 * 
 *     def draw_fragment_embedding(self, emb, **kwargs):             # <<<<<<<<<<<<<<
 *         m, n, t, nodes, edges = self._graph.fragment_graph_spec()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("minorminer.busclique.busgraph_cache.draw_fragment_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_edges);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_f_emb);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_k);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":668
 *         dnx.draw_chimera_embedding(f, f_emb, **kwargs)
 * 
 * cdef dict _make_clique_cache(vector[embedding_t] &embs):             # <<<<<<<<<<<<<<
 *     """
 *     Process the raw clique cache produced by the c++ code.  Store both the raw
 */

static PyObject *__pyx_f_10minorminer_9busclique__make_clique_cache(std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  &__pyx_v_embs) {
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_v_emb;
  size_t __pyx_v_maxsize;
  CYTHON_UNUSED size_t __pyx_v_maxlength;
  size_t __pyx_v_i;
  PyObject *__pyx_v_by_size = 0;
  PyObject *__pyx_v_by_length = 0;
  PyObject *__pyx_v_raw = 0;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  size_t __pyx_t_9;
  std::vector<std::vector<size_t> > ::size_type __pyx_t_10;
  std::vector<std::vector<size_t> > ::size_type __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_clique_cache", 0);

  /* "minorminer/busclique.pyx":677
 *     """
 *     cdef embedding_t emb
 *     cdef size_t maxsize = 0             # <<<<<<<<<<<<<<
 *     cdef size_t maxlength = 0
 *     cdef size_t i, j
 */
  __pyx_v_maxsize = 0;

  /* "minorminer/busclique.pyx":678
 *     cdef embedding_t emb
 *     cdef size_t maxsize = 0
 *     cdef size_t maxlength = 0             # <<<<<<<<<<<<<<
 *     cdef size_t i, j
 *     cdef dict by_size = {}
 */
  __pyx_v_maxlength = 0;

  /* "minorminer/busclique.pyx":680
 *     cdef size_t maxlength = 0
 *     cdef size_t i, j
 *     cdef dict by_size = {}             # <<<<<<<<<<<<<<
 *     cdef dict by_length = {}
 *     cdef list raw = []
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_by_size = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":681
 *     cdef size_t i, j
 *     cdef dict by_size = {}
 *     cdef dict by_length = {}             # <<<<<<<<<<<<<<
 *     cdef list raw = []
 *     for length, emb in enumerate(embs):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_by_length = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":682
 *     cdef dict by_size = {}
 *     cdef dict by_length = {}
 *     cdef list raw = []             # <<<<<<<<<<<<<<
 *     for length, emb in enumerate(embs):
 *         by_length[length] = length
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_raw = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":683
 *     cdef dict by_length = {}
 *     cdef list raw = []
 *     for length, emb in enumerate(embs):             # <<<<<<<<<<<<<<
 *         by_length[length] = length
 *         raw.append(tuple(map(tuple, emb)))
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  __pyx_t_2 = __pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t(__pyx_v_embs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 683, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_6 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_emb = __pyx_t_6;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":684
 *     cdef list raw = []
 *     for length, emb in enumerate(embs):
 *         by_length[length] = length             # <<<<<<<<<<<<<<
 *         raw.append(tuple(map(tuple, emb)))
 *         maxsize = max(emb.size(), maxsize)
 */
    if (unlikely(PyDict_SetItem(__pyx_v_by_length, __pyx_v_length, __pyx_v_length) < 0)) __PYX_ERR(0, 684, __pyx_L1_error)

    /* "minorminer/busclique.pyx":685
 *     for length, emb in enumerate(embs):
 *         by_length[length] = length
 *         raw.append(tuple(map(tuple, emb)))             # <<<<<<<<<<<<<<
 *         maxsize = max(emb.size(), maxsize)
 *         for i in range(maxsize, -1, -1):
 */
    __pyx_t_2 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_v_emb); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)(&PyTuple_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyTuple_Type)));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)(&PyTuple_Type)));
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_raw, __pyx_t_7); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "minorminer/busclique.pyx":686
 *         by_length[length] = length
 *         raw.append(tuple(map(tuple, emb)))
 *         maxsize = max(emb.size(), maxsize)             # <<<<<<<<<<<<<<
 *         for i in range(maxsize, -1, -1):
 *             if by_size.setdefault(i, length) != length:
 */
    __pyx_t_9 = __pyx_v_maxsize;
    __pyx_t_10 = __pyx_v_emb.size();
    if (((__pyx_t_9 > __pyx_t_10) != 0)) {
      __pyx_t_11 = __pyx_t_9;
    } else {
      __pyx_t_11 = __pyx_t_10;
    }
    __pyx_v_maxsize = __pyx_t_11;

    /* "minorminer/busclique.pyx":687
 *         raw.append(tuple(map(tuple, emb)))
 *         maxsize = max(emb.size(), maxsize)
 *         for i in range(maxsize, -1, -1):             # <<<<<<<<<<<<<<
 *             if by_size.setdefault(i, length) != length:
 *                 break
 */
    for (__pyx_t_9 = __pyx_v_maxsize + 1; __pyx_t_9 > -1L + 1; ) { __pyx_t_9-=1;
      __pyx_v_i = __pyx_t_9;

      /* "minorminer/busclique.pyx":688
 *         maxsize = max(emb.size(), maxsize)
 *         for i in range(maxsize, -1, -1):
 *             if by_size.setdefault(i, length) != length:             # <<<<<<<<<<<<<<
 *                 break
 *     return {'raw': raw, 'size': by_size, 'length': by_length}
 */
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyDict_SetDefault(__pyx_v_by_size, __pyx_t_7, __pyx_v_length, -1L); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_v_length, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_12) {

        /* "minorminer/busclique.pyx":689
 *         for i in range(maxsize, -1, -1):
 *             if by_size.setdefault(i, length) != length:
 *                 break             # <<<<<<<<<<<<<<
 *     return {'raw': raw, 'size': by_size, 'length': by_length}
 * 
 */
        goto __pyx_L6_break;

        /* "minorminer/busclique.pyx":688
 *         maxsize = max(emb.size(), maxsize)
 *         for i in range(maxsize, -1, -1):
 *             if by_size.setdefault(i, length) != length:             # <<<<<<<<<<<<<<
 *                 break
 *     return {'raw': raw, 'size': by_size, 'length': by_length}
 */
      }
    }
    __pyx_L6_break:;

    /* "minorminer/busclique.pyx":683
 *     cdef dict by_length = {}
 *     cdef list raw = []
 *     for length, emb in enumerate(embs):             # <<<<<<<<<<<<<<
 *         by_length[length] = length
 *         raw.append(tuple(map(tuple, emb)))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":690
 *             if by_size.setdefault(i, length) != length:
 *                 break
 *     return {'raw': raw, 'size': by_size, 'length': by_length}             # <<<<<<<<<<<<<<
 * 
 * cdef _keep_biclique_key(tuple key0, tuple key1, dict chainlength):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_raw, __pyx_v_raw) < 0) __PYX_ERR(0, 690, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_size, __pyx_v_by_size) < 0) __PYX_ERR(0, 690, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_length, __pyx_v_by_length) < 0) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":668
 *         dnx.draw_chimera_embedding(f, f_emb, **kwargs)
 * 
 * cdef dict _make_clique_cache(vector[embedding_t] &embs):             # <<<<<<<<<<<<<<
 *     """
 *     Process the raw clique cache produced by the c++ code.  Store both the raw
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("minorminer.busclique._make_clique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_by_size);
  __Pyx_XDECREF(__pyx_v_by_length);
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":692
 *     return {'raw': raw, 'size': by_size, 'length': by_length}
 * 
 * cdef _keep_biclique_key(tuple key0, tuple key1, dict chainlength):             # <<<<<<<<<<<<<<
 *     """
 *     Helper function for `_make_biclique_cache`.  Each "key" is a pair of ints
 */

static PyObject *__pyx_f_10minorminer_9busclique__keep_biclique_key(PyObject *__pyx_v_key0, PyObject *__pyx_v_key1, PyObject *__pyx_v_chainlength) {
  PyObject *__pyx_v_long0 = NULL;
  PyObject *__pyx_v_short0 = NULL;
  PyObject *__pyx_v_long1 = NULL;
  PyObject *__pyx_v_short1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_keep_biclique_key", 0);

  /* "minorminer/busclique.pyx":705
 *     are preferred).
 *     """
 *     if key0 == key1:             # <<<<<<<<<<<<<<
 *         return True
 *     long0, short0 = chainlength[key0]
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_key0, __pyx_v_key1, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":706
 *     """
 *     if key0 == key1:
 *         return True             # <<<<<<<<<<<<<<
 *     long0, short0 = chainlength[key0]
 *     long1, short1 = chainlength[key1]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":705
 *     are preferred).
 *     """
 *     if key0 == key1:             # <<<<<<<<<<<<<<
 *         return True
 *     long0, short0 = chainlength[key0]
 */
  }

  /* "minorminer/busclique.pyx":707
 *     if key0 == key1:
 *         return True
 *     long0, short0 = chainlength[key0]             # <<<<<<<<<<<<<<
 *     long1, short1 = chainlength[key1]
 *     if long0 < long1 or (long0 == long1 and short0 <= short1):
 */
  if (unlikely(__pyx_v_chainlength == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 707, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_chainlength, __pyx_v_key0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 707, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 707, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 707, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_v_long0 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_short0 = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "minorminer/busclique.pyx":708
 *         return True
 *     long0, short0 = chainlength[key0]
 *     long1, short1 = chainlength[key1]             # <<<<<<<<<<<<<<
 *     if long0 < long1 or (long0 == long1 and short0 <= short1):
 *         return True
 */
  if (unlikely(__pyx_v_chainlength == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 708, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_chainlength, __pyx_v_key1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 708, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 708, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 708, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v_long1 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_short1 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":709
 *     long0, short0 = chainlength[key0]
 *     long1, short1 = chainlength[key1]
 *     if long0 < long1 or (long0 == long1 and short0 <= short1):             # <<<<<<<<<<<<<<
 *         return True
 *     else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_long0, __pyx_v_long1, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_long0, __pyx_v_long1, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_short0, __pyx_v_short1, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_7;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":710
 *     long1, short1 = chainlength[key1]
 *     if long0 < long1 or (long0 == long1 and short0 <= short1):
 *         return True             # <<<<<<<<<<<<<<
 *     else:
 *         return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":709
 *     long0, short0 = chainlength[key0]
 *     long1, short1 = chainlength[key1]
 *     if long0 < long1 or (long0 == long1 and short0 <= short1):             # <<<<<<<<<<<<<<
 *         return True
 *     else:
 */
  }

  /* "minorminer/busclique.pyx":712
 *         return True
 *     else:
 *         return False             # <<<<<<<<<<<<<<
 * 
 * cdef dict _make_biclique_cache(vector[pair[pair[size_t, size_t],
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":692
 *     return {'raw': raw, 'size': by_size, 'length': by_length}
 * 
 * cdef _keep_biclique_key(tuple key0, tuple key1, dict chainlength):             # <<<<<<<<<<<<<<
 *     """
 *     Helper function for `_make_biclique_cache`.  Each "key" is a pair of ints
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("minorminer.busclique._keep_biclique_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_long0);
  __Pyx_XDECREF(__pyx_v_short0);
  __Pyx_XDECREF(__pyx_v_long1);
  __Pyx_XDECREF(__pyx_v_short1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":714
 *         return False
 * 
 * cdef dict _make_biclique_cache(vector[pair[pair[size_t, size_t],             # <<<<<<<<<<<<<<
 *                                embedding_t]] &embs):
 *     """
 */

static PyObject *__pyx_f_10minorminer_9busclique__make_biclique_cache(std::vector<std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t> >  &__pyx_v_embs) {
  PyObject *__pyx_v_raw = 0;
  PyObject *__pyx_v_chainlength = 0;
  PyObject *__pyx_v_by_size = 0;
  CYTHON_UNUSED PyObject *__pyx_v_by_length = 0;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s0 = NULL;
  PyObject *__pyx_v_s1 = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_emb0 = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_emb1 = NULL;
  PyObject *__pyx_v_realkey = NULL;
  PyObject *__pyx_v_max_side = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_s0 = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_s1 = NULL;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_8genexpr6__pyx_v_emb;
  PyObject *__pyx_8genexpr7__pyx_v_key = NULL;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_8genexpr7__pyx_v_emb;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t> > ::iterator __pyx_t_2;
  std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_biclique_cache", 0);

  /* "minorminer/busclique.pyx":727
 *     """
 *     cdef embedding_t emb
 *     cdef dict raw = {(s0, s1): emb for (s0, s1), emb in embs if emb.size()}             # <<<<<<<<<<<<<<
 *     cdef dict chainlength = {key: (max(map(len, emb)), min(map(len, emb)))
 *                              for key, emb in embs if emb.size()}
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_embs.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_embs.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_t_4 = __pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 727, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 727, __pyx_L5_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 727, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __pyx_t_9 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 727, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_10);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_8(__pyx_t_11); if (unlikely(!__pyx_t_7)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_10 = __pyx_t_8(__pyx_t_11); if (unlikely(!__pyx_t_10)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_10);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_11), 2) < 0) __PYX_ERR(0, 727, __pyx_L5_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L11_unpacking_done;
        __pyx_L10_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 727, __pyx_L5_error)
        __pyx_L11_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_s0, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_s1, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_8genexpr6__pyx_v_emb = __pyx_t_9;
      __pyx_t_12 = (__pyx_8genexpr6__pyx_v_emb.size() != 0);
      if (__pyx_t_12) {
        __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_8genexpr6__pyx_v_s0);
        __Pyx_GIVEREF(__pyx_8genexpr6__pyx_v_s0);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_8genexpr6__pyx_v_s0);
        __Pyx_INCREF(__pyx_8genexpr6__pyx_v_s1);
        __Pyx_GIVEREF(__pyx_8genexpr6__pyx_v_s1);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_8genexpr6__pyx_v_s1);
        __pyx_t_6 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_8genexpr6__pyx_v_emb); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_4, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 727, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_s0); __pyx_8genexpr6__pyx_v_s0 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_s1); __pyx_8genexpr6__pyx_v_s1 = 0;
    goto __pyx_L13_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_s0); __pyx_8genexpr6__pyx_v_s0 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_s1); __pyx_8genexpr6__pyx_v_s1 = 0;
    goto __pyx_L1_error;
    __pyx_L13_exit_scope:;
  } /* exit inner scope */
  __pyx_v_raw = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":728
 *     cdef embedding_t emb
 *     cdef dict raw = {(s0, s1): emb for (s0, s1), emb in embs if emb.size()}
 *     cdef dict chainlength = {key: (max(map(len, emb)), min(map(len, emb)))             # <<<<<<<<<<<<<<
 *                              for key, emb in embs if emb.size()}
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "minorminer/busclique.pyx":729
 *     cdef dict raw = {(s0, s1): emb for (s0, s1), emb in embs if emb.size()}
 *     cdef dict chainlength = {key: (max(map(len, emb)), min(map(len, emb)))
 *                              for key, emb in embs if emb.size()}             # <<<<<<<<<<<<<<
 * 
 *     cdef dict by_size = {}
 */
    __pyx_t_2 = __pyx_v_embs.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_embs.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_t_6 = __pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 729, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_6);
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 729, __pyx_L16_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 729, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_10 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 729, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_10)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_10); if (unlikely(!__pyx_t_4)) goto __pyx_L19_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_5 = __pyx_t_8(__pyx_t_10); if (unlikely(!__pyx_t_5)) goto __pyx_L19_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_10), 2) < 0) __PYX_ERR(0, 729, __pyx_L16_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L20_unpacking_done;
        __pyx_L19_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 729, __pyx_L16_error)
        __pyx_L20_unpacking_done:;
      }
      __pyx_t_9 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_key, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_8genexpr7__pyx_v_emb = __pyx_t_9;
      __pyx_t_12 = (__pyx_8genexpr7__pyx_v_emb.size() != 0);
      if (__pyx_t_12) {

        /* "minorminer/busclique.pyx":728
 *     cdef embedding_t emb
 *     cdef dict raw = {(s0, s1): emb for (s0, s1), emb in embs if emb.size()}
 *     cdef dict chainlength = {key: (max(map(len, emb)), min(map(len, emb)))             # <<<<<<<<<<<<<<
 *                              for key, emb in embs if emb.size()}
 * 
 */
        __pyx_t_6 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_8genexpr7__pyx_v_emb); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
        __pyx_t_6 = 0;
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_8genexpr7__pyx_v_emb); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
        __pyx_t_5 = 0;
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_10);
        __pyx_t_4 = 0;
        __pyx_t_10 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr7__pyx_v_key, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 728, __pyx_L16_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "minorminer/busclique.pyx":729
 *     cdef dict raw = {(s0, s1): emb for (s0, s1), emb in embs if emb.size()}
 *     cdef dict chainlength = {key: (max(map(len, emb)), min(map(len, emb)))
 *                              for key, emb in embs if emb.size()}             # <<<<<<<<<<<<<<
 * 
 *     cdef dict by_size = {}
 */
      }
    }
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_key); __pyx_8genexpr7__pyx_v_key = 0;
    goto __pyx_L22_exit_scope;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_key); __pyx_8genexpr7__pyx_v_key = 0;
    goto __pyx_L1_error;
    __pyx_L22_exit_scope:;
  } /* exit inner scope */
  __pyx_v_chainlength = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":731
 *                              for key, emb in embs if emb.size()}
 * 
 *     cdef dict by_size = {}             # <<<<<<<<<<<<<<
 *     cdef dict by_length = {}
 *     for key, val in raw.items():
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_by_size = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":732
 * 
 *     cdef dict by_size = {}
 *     cdef dict by_length = {}             # <<<<<<<<<<<<<<
 *     for key, val in raw.items():
 *         s0, s1 = key
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_by_length = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":733
 *     cdef dict by_size = {}
 *     cdef dict by_length = {}
 *     for key, val in raw.items():             # <<<<<<<<<<<<<<
 *         s0, s1 = key
 *         emb0 = tuple(val[:s0])
 */
  __pyx_t_13 = 0;
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_raw, 1, __pyx_n_s_items, (&__pyx_t_14), (&__pyx_t_15)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_16 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_14, &__pyx_t_13, &__pyx_t_6, &__pyx_t_10, NULL, __pyx_t_15);
    if (unlikely(__pyx_t_16 == 0)) break;
    if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":734
 *     cdef dict by_length = {}
 *     for key, val in raw.items():
 *         s0, s1 = key             # <<<<<<<<<<<<<<
 *         emb0 = tuple(val[:s0])
 *         emb1 = tuple(val[s0:])
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_key))) || (PyList_CheckExact(__pyx_v_key))) {
      PyObject* sequence = __pyx_v_key;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 734, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 734, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 734, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_v_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 734, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_8(__pyx_t_4); if (unlikely(!__pyx_t_10)) goto __pyx_L25_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_4); if (unlikely(!__pyx_t_6)) goto __pyx_L25_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_4), 2) < 0) __PYX_ERR(0, 734, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L26_unpacking_done;
      __pyx_L25_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 734, __pyx_L1_error)
      __pyx_L26_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s0, __pyx_t_10);
    __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s1, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "minorminer/busclique.pyx":735
 *     for key, val in raw.items():
 *         s0, s1 = key
 *         emb0 = tuple(val[:s0])             # <<<<<<<<<<<<<<
 *         emb1 = tuple(val[s0:])
 *         if(s0 < s1):
 */
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_val, 0, 0, NULL, &__pyx_v_s0, NULL, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PySequence_Tuple(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_emb0, ((PyObject*)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":736
 *         s0, s1 = key
 *         emb0 = tuple(val[:s0])
 *         emb1 = tuple(val[s0:])             # <<<<<<<<<<<<<<
 *         if(s0 < s1):
 *             s1, s0 = s0, s1
 */
    __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_v_val, 0, 0, &__pyx_v_s0, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_emb1, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "minorminer/busclique.pyx":737
 *         emb0 = tuple(val[:s0])
 *         emb1 = tuple(val[s0:])
 *         if(s0 < s1):             # <<<<<<<<<<<<<<
 *             s1, s0 = s0, s1
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)
 */
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_s0, __pyx_v_s1, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 737, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 737, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_12) {

      /* "minorminer/busclique.pyx":738
 *         emb1 = tuple(val[s0:])
 *         if(s0 < s1):
 *             s1, s0 = s0, s1             # <<<<<<<<<<<<<<
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)
 *         if not _keep_biclique_key(realkey, key, chainlength):
 */
      __pyx_t_17 = __pyx_v_s0;
      __pyx_t_18 = __pyx_v_s1;
      __pyx_v_s1 = __pyx_t_17;
      __pyx_t_17 = 0;
      __pyx_v_s0 = __pyx_t_18;
      __pyx_t_18 = 0;

      /* "minorminer/busclique.pyx":737
 *         emb0 = tuple(val[:s0])
 *         emb1 = tuple(val[s0:])
 *         if(s0 < s1):             # <<<<<<<<<<<<<<
 *             s1, s0 = s0, s1
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)
 */
    }

    /* "minorminer/busclique.pyx":739
 *         if(s0 < s1):
 *             s1, s0 = s0, s1
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)             # <<<<<<<<<<<<<<
 *         if not _keep_biclique_key(realkey, key, chainlength):
 *             by_size[s0][s1] = key
 */
    __pyx_t_10 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = __Pyx_PyDict_SetDefault(__pyx_v_by_size, __pyx_v_s0, __pyx_t_10, -1L); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_setdefault); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_16 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_16 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_s1, __pyx_v_key};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 739, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_s1, __pyx_v_key};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 739, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 739, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_s1);
      __Pyx_GIVEREF(__pyx_v_s1);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_16, __pyx_v_s1);
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_16, __pyx_v_key);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 739, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_realkey, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "minorminer/busclique.pyx":740
 *             s1, s0 = s0, s1
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)
 *         if not _keep_biclique_key(realkey, key, chainlength):             # <<<<<<<<<<<<<<
 *             by_size[s0][s1] = key
 * 
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v_realkey))||((__pyx_v_realkey) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_realkey)->tp_name), 0))) __PYX_ERR(0, 740, __pyx_L1_error)
    if (!(likely(PyTuple_CheckExact(__pyx_v_key))||((__pyx_v_key) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_key)->tp_name), 0))) __PYX_ERR(0, 740, __pyx_L1_error)
    __pyx_t_6 = __pyx_f_10minorminer_9busclique__keep_biclique_key(((PyObject*)__pyx_v_realkey), ((PyObject*)__pyx_v_key), __pyx_v_chainlength); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_19 = ((!__pyx_t_12) != 0);
    if (__pyx_t_19) {

      /* "minorminer/busclique.pyx":741
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)
 *         if not _keep_biclique_key(realkey, key, chainlength):
 *             by_size[s0][s1] = key             # <<<<<<<<<<<<<<
 * 
 *     max_side = {None: max(by_size)}
 */
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_by_size, __pyx_v_s0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 741, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_v_s1, __pyx_v_key) < 0)) __PYX_ERR(0, 741, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "minorminer/busclique.pyx":740
 *             s1, s0 = s0, s1
 *         realkey = by_size.setdefault(s0, {}).setdefault(s1, key)
 *         if not _keep_biclique_key(realkey, key, chainlength):             # <<<<<<<<<<<<<<
 *             by_size[s0][s1] = key
 * 
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":743
 *             by_size[s0][s1] = key
 * 
 *     max_side = {None: max(by_size)}             # <<<<<<<<<<<<<<
 *     for key, val in by_size.items():
 *         max_side[key] = max(val)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_by_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_1, Py_None, __pyx_t_6) < 0) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_max_side = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":744
 * 
 *     max_side = {None: max(by_size)}
 *     for key, val in by_size.items():             # <<<<<<<<<<<<<<
 *         max_side[key] = max(val)
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}
 */
  __pyx_t_14 = 0;
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_by_size, 1, __pyx_n_s_items, (&__pyx_t_13), (&__pyx_t_15)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_16 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_13, &__pyx_t_14, &__pyx_t_6, &__pyx_t_10, NULL, __pyx_t_15);
    if (unlikely(__pyx_t_16 == 0)) break;
    if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "minorminer/busclique.pyx":745
 *     max_side = {None: max(by_size)}
 *     for key, val in by_size.items():
 *         max_side[key] = max(val)             # <<<<<<<<<<<<<<
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_val); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (unlikely(PyDict_SetItem(__pyx_v_max_side, __pyx_v_key, __pyx_t_10) < 0)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":746
 *     for key, val in by_size.items():
 *         max_side[key] = max(val)
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}             # <<<<<<<<<<<<<<
 * 
 * def _trivial_relabeler(emb):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_raw, __pyx_v_raw) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_size, __pyx_v_by_size) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_max_side, __pyx_v_max_side) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":714
 *         return False
 * 
 * cdef dict _make_biclique_cache(vector[pair[pair[size_t, size_t],             # <<<<<<<<<<<<<<
 *                                embedding_t]] &embs):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("minorminer.busclique._make_biclique_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_chainlength);
  __Pyx_XDECREF(__pyx_v_by_size);
  __Pyx_XDECREF(__pyx_v_by_length);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s0);
  __Pyx_XDECREF(__pyx_v_s1);
  __Pyx_XDECREF(__pyx_v_emb0);
  __Pyx_XDECREF(__pyx_v_emb1);
  __Pyx_XDECREF(__pyx_v_realkey);
  __Pyx_XDECREF(__pyx_v_max_side);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_s0);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_s1);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":748
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}
 * 
 * def _trivial_relabeler(emb):             # <<<<<<<<<<<<<<
 *     """This doesn't relabel anything"""
 *     return {v: tuple(chain) for v, chain in emb.items()}
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_5_trivial_relabeler(PyObject *__pyx_self, PyObject *__pyx_v_emb); /*proto*/
static char __pyx_doc_10minorminer_9busclique_4_trivial_relabeler[] = "This doesn't relabel anything";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_5_trivial_relabeler = {"_trivial_relabeler", (PyCFunction)__pyx_pw_10minorminer_9busclique_5_trivial_relabeler, METH_O, __pyx_doc_10minorminer_9busclique_4_trivial_relabeler};
static PyObject *__pyx_pw_10minorminer_9busclique_5_trivial_relabeler(PyObject *__pyx_self, PyObject *__pyx_v_emb) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_trivial_relabeler (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_4_trivial_relabeler(__pyx_self, ((PyObject *)__pyx_v_emb));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_4_trivial_relabeler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_emb) {
  PyObject *__pyx_8genexpr8__pyx_v_v = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_chain = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_trivial_relabeler", 0);

  /* "minorminer/busclique.pyx":750
 * def _trivial_relabeler(emb):
 *     """This doesn't relabel anything"""
 *     return {v: tuple(chain) for v, chain in emb.items()}             # <<<<<<<<<<<<<<
 * 
 * def _make_relabeler(f):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_emb == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 750, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_emb, 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 750, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 750, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_v, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_chain, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_8genexpr8__pyx_v_chain); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 750, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr8__pyx_v_v, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 750, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_chain); __pyx_8genexpr8__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_v); __pyx_8genexpr8__pyx_v_v = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_chain); __pyx_8genexpr8__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_v); __pyx_8genexpr8__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":748
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}
 * 
 * def _trivial_relabeler(emb):             # <<<<<<<<<<<<<<
 *     """This doesn't relabel anything"""
 *     return {v: tuple(chain) for v, chain in emb.items()}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("minorminer.busclique._trivial_relabeler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_v);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_chain);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":752
 *     return {v: tuple(chain) for v, chain in emb.items()}
 * 
 * def _make_relabeler(f):             # <<<<<<<<<<<<<<
 *     """
 *     This returns an embedding-relabeling function, which applies `f` to each
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_7_make_relabeler(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static char __pyx_doc_10minorminer_9busclique_6_make_relabeler[] = "\n    This returns an embedding-relabeling function, which applies `f` to each\n    chain in the embedding.\n    ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_7_make_relabeler = {"_make_relabeler", (PyCFunction)__pyx_pw_10minorminer_9busclique_7_make_relabeler, METH_O, __pyx_doc_10minorminer_9busclique_6_make_relabeler};
static PyObject *__pyx_pw_10minorminer_9busclique_7_make_relabeler(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_relabeler (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_6_make_relabeler(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":757
 *     chain in the embedding.
 *     """
 *     def _relabeler(emb):             # <<<<<<<<<<<<<<
 *         return {v: tuple(f(chain)) for v, chain in emb.items()}
 *     return _relabeler
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_15_make_relabeler_1_relabeler(PyObject *__pyx_self, PyObject *__pyx_v_emb); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_15_make_relabeler_1_relabeler = {"_relabeler", (PyCFunction)__pyx_pw_10minorminer_9busclique_15_make_relabeler_1_relabeler, METH_O, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_15_make_relabeler_1_relabeler(PyObject *__pyx_self, PyObject *__pyx_v_emb) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_relabeler (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_15_make_relabeler__relabeler(__pyx_self, ((PyObject *)__pyx_v_emb));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_15_make_relabeler__relabeler(PyObject *__pyx_self, PyObject *__pyx_v_emb) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *__pyx_cur_scope;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *__pyx_outer_scope;
  PyObject *__pyx_8genexpr9__pyx_v_v = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_chain = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_relabeler", 0);
  __pyx_outer_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "minorminer/busclique.pyx":758
 *     """
 *     def _relabeler(emb):
 *         return {v: tuple(f(chain)) for v, chain in emb.items()}             # <<<<<<<<<<<<<<
 *     return _relabeler
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_emb == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 758, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_emb, 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 758, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 758, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_v, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_chain, __pyx_t_7);
      __pyx_t_7 = 0;
      if (unlikely(!__pyx_cur_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 758, __pyx_L5_error) }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
      __pyx_t_6 = __pyx_cur_scope->__pyx_v_f; __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_9, __pyx_8genexpr9__pyx_v_chain) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_8genexpr9__pyx_v_chain);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 758, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 758, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr9__pyx_v_v, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 758, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_chain); __pyx_8genexpr9__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_v); __pyx_8genexpr9__pyx_v_v = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_chain); __pyx_8genexpr9__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_v); __pyx_8genexpr9__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":757
 *     chain in the embedding.
 *     """
 *     def _relabeler(emb):             # <<<<<<<<<<<<<<
 *         return {v: tuple(f(chain)) for v, chain in emb.items()}
 *     return _relabeler
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("minorminer.busclique._make_relabeler._relabeler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_v);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_chain);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":752
 *     return {v: tuple(chain) for v, chain in emb.items()}
 * 
 * def _make_relabeler(f):             # <<<<<<<<<<<<<<
 *     """
 *     This returns an embedding-relabeling function, which applies `f` to each
 */

static PyObject *__pyx_pf_10minorminer_9busclique_6_make_relabeler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *__pyx_cur_scope;
  PyObject *__pyx_v__relabeler = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_relabeler", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 752, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_f);

  /* "minorminer/busclique.pyx":757
 *     chain in the embedding.
 *     """
 *     def _relabeler(emb):             # <<<<<<<<<<<<<<
 *         return {v: tuple(f(chain)) for v, chain in emb.items()}
 *     return _relabeler
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_15_make_relabeler_1_relabeler, 0, __pyx_n_s_make_relabeler_locals__relabele, ((PyObject*)__pyx_cur_scope), __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__relabeler = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":759
 *     def _relabeler(emb):
 *         return {v: tuple(f(chain)) for v, chain in emb.items()}
 *     return _relabeler             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused topo_cache_t:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v__relabeler);
  __pyx_r = __pyx_v__relabeler;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":752
 *     return {v: tuple(chain) for v, chain in emb.items()}
 * 
 * def _make_relabeler(f):             # <<<<<<<<<<<<<<
 *     """
 *     This returns an embedding-relabeling function, which applies `f` to each
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._make_relabeler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__relabeler);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":766
 *     topo_cache[chimera_spec]
 * 
 * cdef _serialize(topo_cache_t *topo, str family):             # <<<<<<<<<<<<<<
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 */

static PyObject *__pyx_fuse_0__pyx_f_10minorminer_9busclique__serialize(busclique::topo_cache<busclique::zephyr_spec>  *__pyx_v_topo, PyObject *__pyx_v_family) {
  size_t __pyx_v_nchars;
  char *__pyx_v_tmp0;
  uint8_t *__pyx_v_tmp;
  uint64_t *__pyx_v_tmp64;
  uint64_t __pyx_v_shortcode_state;
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_v_shortcode = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_serialize", 0);

  /* "minorminer/busclique.pyx":767
 * 
 * cdef _serialize(topo_cache_t *topo, str family):
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)             # <<<<<<<<<<<<<<
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 */
  __pyx_v_nchars = __pyx_v_topo->serialize(busclique::serialize_size_tag(), NULL);

  /* "minorminer/busclique.pyx":768
 * cdef _serialize(topo_cache_t *topo, str family):
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)             # <<<<<<<<<<<<<<
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 */
  __pyx_v_tmp0 = ((char *)PyMem_Malloc(__pyx_v_nchars));

  /* "minorminer/busclique.pyx":769
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0             # <<<<<<<<<<<<<<
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 *     cdef uint64_t shortcode_state = 0
 */
  __pyx_v_tmp = ((uint8_t *)__pyx_v_tmp0);

  /* "minorminer/busclique.pyx":770
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));             # <<<<<<<<<<<<<<
 *     cdef uint64_t shortcode_state = 0
 *     cdef uint64_t c
 */
  __pyx_v_tmp64 = ((uint64_t *)((void *)__pyx_v_tmp));

  /* "minorminer/busclique.pyx":771
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 *     cdef uint64_t shortcode_state = 0             # <<<<<<<<<<<<<<
 *     cdef uint64_t c
 *     cdef object ret, shortcode
 */
  __pyx_v_shortcode_state = 0;

  /* "minorminer/busclique.pyx":776
 *     cdef size_t i, j, d, _
 * 
 *     if tmp == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "minorminer/busclique.pyx":777
 * 
 *     if tmp == NULL:
 *         raise MemoryError("could not allocate memory for topology serialization")             # <<<<<<<<<<<<<<
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 777, __pyx_L1_error)

    /* "minorminer/busclique.pyx":776
 *     cdef size_t i, j, d, _
 * 
 *     if tmp == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 */
  }

  /* "minorminer/busclique.pyx":779
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 *     topo.serialize(serialize_write_tag(), tmp);             # <<<<<<<<<<<<<<
 *     ret = PyBytes_FromStringAndSize(tmp0, nchars);
 * 
 */
  (void)(__pyx_v_topo->serialize(busclique::serialize_write_tag(), __pyx_v_tmp));

  /* "minorminer/busclique.pyx":780
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 *     ret = PyBytes_FromStringAndSize(tmp0, nchars);             # <<<<<<<<<<<<<<
 * 
 *     # this is a hash-like operation.  It is *not* a good hash.  It's optimized
 */
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_tmp0, __pyx_v_nchars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ret = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":784
 *     # this is a hash-like operation.  It is *not* a good hash.  It's optimized
 *     # for speed and produces a 16-byte key which is reasonably sensitive to input.
 *     for i in range(nchars//8):             # <<<<<<<<<<<<<<
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):
 */
  __pyx_t_3 = (__pyx_v_nchars / 8);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "minorminer/busclique.pyx":785
 *     # for speed and produces a 16-byte key which is reasonably sensitive to input.
 *     for i in range(nchars//8):
 *         shortcode_state += tmp64[i]             # <<<<<<<<<<<<<<
 *     for j in range(nchars-(nchars&7), nchars):
 *         shortcode_state += tmp[j] << (j&3)
 */
    __pyx_v_shortcode_state = (__pyx_v_shortcode_state + (__pyx_v_tmp64[__pyx_v_i]));
  }

  /* "minorminer/busclique.pyx":786
 *     for i in range(nchars//8):
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):             # <<<<<<<<<<<<<<
 *         shortcode_state += tmp[j] << (j&3)
 * 
 */
  __pyx_t_3 = __pyx_v_nchars;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = (__pyx_v_nchars - (__pyx_v_nchars & 7)); __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_j = __pyx_t_5;

    /* "minorminer/busclique.pyx":787
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):
 *         shortcode_state += tmp[j] << (j&3)             # <<<<<<<<<<<<<<
 * 
 *     PyMem_Free(tmp0)
 */
    __pyx_v_shortcode_state = (__pyx_v_shortcode_state + ((__pyx_v_tmp[__pyx_v_j]) << (__pyx_v_j & 3)));
  }

  /* "minorminer/busclique.pyx":789
 *         shortcode_state += tmp[j] << (j&3)
 * 
 *     PyMem_Free(tmp0)             # <<<<<<<<<<<<<<
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)
 *     return ret, shortcode
 */
  PyMem_Free(__pyx_v_tmp0);

  /* "minorminer/busclique.pyx":790
 * 
 *     PyMem_Free(tmp0)
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)             # <<<<<<<<<<<<<<
 *     return ret, shortcode
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_16x, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_shortcode_state); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_family, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_family, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_family);
    __Pyx_GIVEREF(__pyx_v_family);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_family);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_shortcode = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":791
 *     PyMem_Free(tmp0)
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)
 *     return ret, shortcode             # <<<<<<<<<<<<<<
 * 
 * cdef class _zephyr_busgraph:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ret);
  __Pyx_INCREF(__pyx_v_shortcode);
  __Pyx_GIVEREF(__pyx_v_shortcode);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_shortcode);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":766
 *     topo_cache[chimera_spec]
 * 
 * cdef _serialize(topo_cache_t *topo, str family):             # <<<<<<<<<<<<<<
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("minorminer.busclique._serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_shortcode);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_1__pyx_f_10minorminer_9busclique__serialize(busclique::topo_cache<busclique::pegasus_spec>  *__pyx_v_topo, PyObject *__pyx_v_family) {
  size_t __pyx_v_nchars;
  char *__pyx_v_tmp0;
  uint8_t *__pyx_v_tmp;
  uint64_t *__pyx_v_tmp64;
  uint64_t __pyx_v_shortcode_state;
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_v_shortcode = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_serialize", 0);

  /* "minorminer/busclique.pyx":767
 * 
 * cdef _serialize(topo_cache_t *topo, str family):
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)             # <<<<<<<<<<<<<<
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 */
  __pyx_v_nchars = __pyx_v_topo->serialize(busclique::serialize_size_tag(), NULL);

  /* "minorminer/busclique.pyx":768
 * cdef _serialize(topo_cache_t *topo, str family):
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)             # <<<<<<<<<<<<<<
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 */
  __pyx_v_tmp0 = ((char *)PyMem_Malloc(__pyx_v_nchars));

  /* "minorminer/busclique.pyx":769
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0             # <<<<<<<<<<<<<<
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 *     cdef uint64_t shortcode_state = 0
 */
  __pyx_v_tmp = ((uint8_t *)__pyx_v_tmp0);

  /* "minorminer/busclique.pyx":770
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));             # <<<<<<<<<<<<<<
 *     cdef uint64_t shortcode_state = 0
 *     cdef uint64_t c
 */
  __pyx_v_tmp64 = ((uint64_t *)((void *)__pyx_v_tmp));

  /* "minorminer/busclique.pyx":771
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 *     cdef uint64_t shortcode_state = 0             # <<<<<<<<<<<<<<
 *     cdef uint64_t c
 *     cdef object ret, shortcode
 */
  __pyx_v_shortcode_state = 0;

  /* "minorminer/busclique.pyx":776
 *     cdef size_t i, j, d, _
 * 
 *     if tmp == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "minorminer/busclique.pyx":777
 * 
 *     if tmp == NULL:
 *         raise MemoryError("could not allocate memory for topology serialization")             # <<<<<<<<<<<<<<
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 777, __pyx_L1_error)

    /* "minorminer/busclique.pyx":776
 *     cdef size_t i, j, d, _
 * 
 *     if tmp == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 */
  }

  /* "minorminer/busclique.pyx":779
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 *     topo.serialize(serialize_write_tag(), tmp);             # <<<<<<<<<<<<<<
 *     ret = PyBytes_FromStringAndSize(tmp0, nchars);
 * 
 */
  (void)(__pyx_v_topo->serialize(busclique::serialize_write_tag(), __pyx_v_tmp));

  /* "minorminer/busclique.pyx":780
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 *     ret = PyBytes_FromStringAndSize(tmp0, nchars);             # <<<<<<<<<<<<<<
 * 
 *     # this is a hash-like operation.  It is *not* a good hash.  It's optimized
 */
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_tmp0, __pyx_v_nchars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ret = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":784
 *     # this is a hash-like operation.  It is *not* a good hash.  It's optimized
 *     # for speed and produces a 16-byte key which is reasonably sensitive to input.
 *     for i in range(nchars//8):             # <<<<<<<<<<<<<<
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):
 */
  __pyx_t_3 = (__pyx_v_nchars / 8);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "minorminer/busclique.pyx":785
 *     # for speed and produces a 16-byte key which is reasonably sensitive to input.
 *     for i in range(nchars//8):
 *         shortcode_state += tmp64[i]             # <<<<<<<<<<<<<<
 *     for j in range(nchars-(nchars&7), nchars):
 *         shortcode_state += tmp[j] << (j&3)
 */
    __pyx_v_shortcode_state = (__pyx_v_shortcode_state + (__pyx_v_tmp64[__pyx_v_i]));
  }

  /* "minorminer/busclique.pyx":786
 *     for i in range(nchars//8):
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):             # <<<<<<<<<<<<<<
 *         shortcode_state += tmp[j] << (j&3)
 * 
 */
  __pyx_t_3 = __pyx_v_nchars;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = (__pyx_v_nchars - (__pyx_v_nchars & 7)); __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_j = __pyx_t_5;

    /* "minorminer/busclique.pyx":787
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):
 *         shortcode_state += tmp[j] << (j&3)             # <<<<<<<<<<<<<<
 * 
 *     PyMem_Free(tmp0)
 */
    __pyx_v_shortcode_state = (__pyx_v_shortcode_state + ((__pyx_v_tmp[__pyx_v_j]) << (__pyx_v_j & 3)));
  }

  /* "minorminer/busclique.pyx":789
 *         shortcode_state += tmp[j] << (j&3)
 * 
 *     PyMem_Free(tmp0)             # <<<<<<<<<<<<<<
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)
 *     return ret, shortcode
 */
  PyMem_Free(__pyx_v_tmp0);

  /* "minorminer/busclique.pyx":790
 * 
 *     PyMem_Free(tmp0)
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)             # <<<<<<<<<<<<<<
 *     return ret, shortcode
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_16x, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_shortcode_state); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_family, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_family, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_family);
    __Pyx_GIVEREF(__pyx_v_family);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_family);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_shortcode = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":791
 *     PyMem_Free(tmp0)
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)
 *     return ret, shortcode             # <<<<<<<<<<<<<<
 * 
 * cdef class _zephyr_busgraph:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ret);
  __Pyx_INCREF(__pyx_v_shortcode);
  __Pyx_GIVEREF(__pyx_v_shortcode);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_shortcode);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":766
 *     topo_cache[chimera_spec]
 * 
 * cdef _serialize(topo_cache_t *topo, str family):             # <<<<<<<<<<<<<<
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("minorminer.busclique._serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_shortcode);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_2__pyx_f_10minorminer_9busclique__serialize(busclique::topo_cache<busclique::chimera_spec>  *__pyx_v_topo, PyObject *__pyx_v_family) {
  size_t __pyx_v_nchars;
  char *__pyx_v_tmp0;
  uint8_t *__pyx_v_tmp;
  uint64_t *__pyx_v_tmp64;
  uint64_t __pyx_v_shortcode_state;
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_v_shortcode = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_2_serialize", 0);

  /* "minorminer/busclique.pyx":767
 * 
 * cdef _serialize(topo_cache_t *topo, str family):
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)             # <<<<<<<<<<<<<<
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 */
  __pyx_v_nchars = __pyx_v_topo->serialize(busclique::serialize_size_tag(), NULL);

  /* "minorminer/busclique.pyx":768
 * cdef _serialize(topo_cache_t *topo, str family):
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)             # <<<<<<<<<<<<<<
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 */
  __pyx_v_tmp0 = ((char *)PyMem_Malloc(__pyx_v_nchars));

  /* "minorminer/busclique.pyx":769
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0             # <<<<<<<<<<<<<<
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 *     cdef uint64_t shortcode_state = 0
 */
  __pyx_v_tmp = ((uint8_t *)__pyx_v_tmp0);

  /* "minorminer/busclique.pyx":770
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));             # <<<<<<<<<<<<<<
 *     cdef uint64_t shortcode_state = 0
 *     cdef uint64_t c
 */
  __pyx_v_tmp64 = ((uint64_t *)((void *)__pyx_v_tmp));

  /* "minorminer/busclique.pyx":771
 *     cdef uint8_t *tmp = <uint8_t *>tmp0
 *     cdef uint64_t *tmp64 = <uint64_t*>(<void *>(tmp));
 *     cdef uint64_t shortcode_state = 0             # <<<<<<<<<<<<<<
 *     cdef uint64_t c
 *     cdef object ret, shortcode
 */
  __pyx_v_shortcode_state = 0;

  /* "minorminer/busclique.pyx":776
 *     cdef size_t i, j, d, _
 * 
 *     if tmp == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "minorminer/busclique.pyx":777
 * 
 *     if tmp == NULL:
 *         raise MemoryError("could not allocate memory for topology serialization")             # <<<<<<<<<<<<<<
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 777, __pyx_L1_error)

    /* "minorminer/busclique.pyx":776
 *     cdef size_t i, j, d, _
 * 
 *     if tmp == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 */
  }

  /* "minorminer/busclique.pyx":779
 *         raise MemoryError("could not allocate memory for topology serialization")
 * 
 *     topo.serialize(serialize_write_tag(), tmp);             # <<<<<<<<<<<<<<
 *     ret = PyBytes_FromStringAndSize(tmp0, nchars);
 * 
 */
  (void)(__pyx_v_topo->serialize(busclique::serialize_write_tag(), __pyx_v_tmp));

  /* "minorminer/busclique.pyx":780
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 *     ret = PyBytes_FromStringAndSize(tmp0, nchars);             # <<<<<<<<<<<<<<
 * 
 *     # this is a hash-like operation.  It is *not* a good hash.  It's optimized
 */
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_tmp0, __pyx_v_nchars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ret = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":784
 *     # this is a hash-like operation.  It is *not* a good hash.  It's optimized
 *     # for speed and produces a 16-byte key which is reasonably sensitive to input.
 *     for i in range(nchars//8):             # <<<<<<<<<<<<<<
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):
 */
  __pyx_t_3 = (__pyx_v_nchars / 8);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "minorminer/busclique.pyx":785
 *     # for speed and produces a 16-byte key which is reasonably sensitive to input.
 *     for i in range(nchars//8):
 *         shortcode_state += tmp64[i]             # <<<<<<<<<<<<<<
 *     for j in range(nchars-(nchars&7), nchars):
 *         shortcode_state += tmp[j] << (j&3)
 */
    __pyx_v_shortcode_state = (__pyx_v_shortcode_state + (__pyx_v_tmp64[__pyx_v_i]));
  }

  /* "minorminer/busclique.pyx":786
 *     for i in range(nchars//8):
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):             # <<<<<<<<<<<<<<
 *         shortcode_state += tmp[j] << (j&3)
 * 
 */
  __pyx_t_3 = __pyx_v_nchars;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = (__pyx_v_nchars - (__pyx_v_nchars & 7)); __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_j = __pyx_t_5;

    /* "minorminer/busclique.pyx":787
 *         shortcode_state += tmp64[i]
 *     for j in range(nchars-(nchars&7), nchars):
 *         shortcode_state += tmp[j] << (j&3)             # <<<<<<<<<<<<<<
 * 
 *     PyMem_Free(tmp0)
 */
    __pyx_v_shortcode_state = (__pyx_v_shortcode_state + ((__pyx_v_tmp[__pyx_v_j]) << (__pyx_v_j & 3)));
  }

  /* "minorminer/busclique.pyx":789
 *         shortcode_state += tmp[j] << (j&3)
 * 
 *     PyMem_Free(tmp0)             # <<<<<<<<<<<<<<
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)
 *     return ret, shortcode
 */
  PyMem_Free(__pyx_v_tmp0);

  /* "minorminer/busclique.pyx":790
 * 
 *     PyMem_Free(tmp0)
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)             # <<<<<<<<<<<<<<
 *     return ret, shortcode
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_16x, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_shortcode_state); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_family, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_family, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_family);
    __Pyx_GIVEREF(__pyx_v_family);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_family);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_shortcode = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":791
 *     PyMem_Free(tmp0)
 *     shortcode = "{}_{:16x}".format(family, shortcode_state)
 *     return ret, shortcode             # <<<<<<<<<<<<<<
 * 
 * cdef class _zephyr_busgraph:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ret);
  __Pyx_INCREF(__pyx_v_shortcode);
  __Pyx_GIVEREF(__pyx_v_shortcode);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_shortcode);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":766
 *     topo_cache[chimera_spec]
 * 
 * cdef _serialize(topo_cache_t *topo, str family):             # <<<<<<<<<<<<<<
 *     cdef size_t nchars = topo.serialize(serialize_size_tag(), NULL)
 *     cdef char *tmp0 = <char *>PyMem_Malloc(nchars)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("minorminer.busclique._serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_shortcode);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":801
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):             # <<<<<<<<<<<<<<
 *         """
 *         This is a class which manages a single zephyr graph, and dispatches
 */

/* Python wrapper */
static int __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_compute_identifier = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_g,&__pyx_n_s_seed,&__pyx_n_s_compute_identifier,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_identifier);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 801, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_g = values[0];
    __pyx_v_seed = values[1];
    __pyx_v_compute_identifier = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 801, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph___cinit__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self), __pyx_v_g, __pyx_v_seed, __pyx_v_compute_identifier);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph___cinit__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_compute_identifier) {
  size_t __pyx_v_rows;
  CYTHON_UNUSED size_t __pyx_v_cols;
  size_t __pyx_v_tile;
  uint32_t __pyx_v_internal_seed;
  PyObject *__pyx_v_seed_obj = NULL;
  busclique::zephyr_spec *__pyx_v_zep;
  PyObject *__pyx_v_coordinates = NULL;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_v_edges;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  uint8_t *__pyx_t_8;
  uint32_t __pyx_t_9;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_t_10;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "minorminer/busclique.pyx":806
 *         various structure-aware c++ embedding functions on it.
 *         """
 *         cdef size_t rows = g.graph['rows']             # <<<<<<<<<<<<<<
 *         cdef size_t cols = g.graph['columns']
 *         cdef size_t tile = g.graph['tile']
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_rows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_rows = __pyx_t_3;

  /* "minorminer/busclique.pyx":807
 *         """
 *         cdef size_t rows = g.graph['rows']
 *         cdef size_t cols = g.graph['columns']             # <<<<<<<<<<<<<<
 *         cdef size_t tile = g.graph['tile']
 *         if tile > 4:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_columns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cols = __pyx_t_3;

  /* "minorminer/busclique.pyx":808
 *         cdef size_t rows = g.graph['rows']
 *         cdef size_t cols = g.graph['columns']
 *         cdef size_t tile = g.graph['tile']             # <<<<<<<<<<<<<<
 *         if tile > 4:
 *             raise NotImplementedError(("this clique embedder supports zephyr "
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_tile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tile = __pyx_t_3;

  /* "minorminer/busclique.pyx":809
 *         cdef size_t cols = g.graph['columns']
 *         cdef size_t tile = g.graph['tile']
 *         if tile > 4:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError(("this clique embedder supports zephyr "
 *                                        "graphs with a tile size of 4 or less"))
 */
  __pyx_t_4 = ((__pyx_v_tile > 4) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "minorminer/busclique.pyx":810
 *         cdef size_t tile = g.graph['tile']
 *         if tile > 4:
 *             raise NotImplementedError(("this clique embedder supports zephyr "             # <<<<<<<<<<<<<<
 *                                        "graphs with a tile size of 4 or less"))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 810, __pyx_L1_error)

    /* "minorminer/busclique.pyx":809
 *         cdef size_t cols = g.graph['columns']
 *         cdef size_t tile = g.graph['tile']
 *         if tile > 4:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError(("this clique embedder supports zephyr "
 *                                        "graphs with a tile size of 4 or less"))
 */
  }

  /* "minorminer/busclique.pyx":814
 * 
 *         cdef uint32_t internal_seed
 *         if seed is None:             # <<<<<<<<<<<<<<
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 */
  __pyx_t_4 = (__pyx_v_seed == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "minorminer/busclique.pyx":815
 *         cdef uint32_t internal_seed
 *         if seed is None:
 *             seed_obj = os.urandom(sizeof(uint32_t))             # <<<<<<<<<<<<<<
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_urandom); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(uint32_t))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_seed_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":816
 *         if seed is None:
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]             # <<<<<<<<<<<<<<
 *         else:
 *             internal_seed = seed
 */
    __pyx_t_8 = __Pyx_PyObject_AsWritableUString(__pyx_v_seed_obj); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L1_error)
    __pyx_v_internal_seed = (((uint32_t *)((void *)((uint8_t *)__pyx_t_8)))[0]);

    /* "minorminer/busclique.pyx":814
 * 
 *         cdef uint32_t internal_seed
 *         if seed is None:             # <<<<<<<<<<<<<<
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 */
    goto __pyx_L4;
  }

  /* "minorminer/busclique.pyx":818
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 *         else:
 *             internal_seed = seed             # <<<<<<<<<<<<<<
 * 
 *         cdef zephyr_spec *zep = new zephyr_spec(rows, tile, internal_seed)
 */
  /*else*/ {
    __pyx_t_9 = __Pyx_PyInt_As_uint32_t(__pyx_v_seed); if (unlikely((__pyx_t_9 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 818, __pyx_L1_error)
    __pyx_v_internal_seed = __pyx_t_9;
  }
  __pyx_L4:;

  /* "minorminer/busclique.pyx":820
 *             internal_seed = seed
 * 
 *         cdef zephyr_spec *zep = new zephyr_spec(rows, tile, internal_seed)             # <<<<<<<<<<<<<<
 *         coordinates = dnx.zephyr_coordinates(rows)
 *         cdef edges_t edges
 */
  __pyx_v_zep = new busclique::zephyr_spec(__pyx_v_rows, __pyx_v_tile, __pyx_v_internal_seed);

  /* "minorminer/busclique.pyx":821
 * 
 *         cdef zephyr_spec *zep = new zephyr_spec(rows, tile, internal_seed)
 *         coordinates = dnx.zephyr_coordinates(rows)             # <<<<<<<<<<<<<<
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_dnx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zephyr_coordinates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_rows); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_coordinates = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":823
 *         coordinates = dnx.zephyr_coordinates(rows)
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':             # <<<<<<<<<<<<<<
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_labels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_int, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "minorminer/busclique.pyx":824
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()             # <<<<<<<<<<<<<<
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __pyx_convert_vector_from_py_size_t(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->nodes = __pyx_t_10;

    /* "minorminer/busclique.pyx":825
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()
 *             edges = g.edges()             # <<<<<<<<<<<<<<
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_edges = __pyx_t_11;

    /* "minorminer/busclique.pyx":826
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler             # <<<<<<<<<<<<<<
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_trivial_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":827
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler             # <<<<<<<<<<<<<<
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_zephyr_to_linear(g.nodes())
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_trivial_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":823
 *         coordinates = dnx.zephyr_coordinates(rows)
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':             # <<<<<<<<<<<<<<
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":828
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_zephyr_to_linear(g.nodes())
 *             edges = coordinates.iter_zephyr_to_linear_pairs(g.edges())
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_labels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_coordinate, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_5)) {

    /* "minorminer/busclique.pyx":829
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_zephyr_to_linear(g.nodes())             # <<<<<<<<<<<<<<
 *             edges = coordinates.iter_zephyr_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_zephyr)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_zephyr_to_linear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = __pyx_convert_vector_from_py_size_t(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_self->nodes = __pyx_t_10;

    /* "minorminer/busclique.pyx":830
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_zephyr_to_linear(g.nodes())
 *             edges = coordinates.iter_zephyr_to_linear_pairs(g.edges())             # <<<<<<<<<<<<<<
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_zephyr)
 *             self.delabel = _make_relabeler(coordinates.iter_zephyr_to_linear)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_zephyr_to_linear_pairs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_edges = __pyx_t_11;

    /* "minorminer/busclique.pyx":831
 *             self.nodes = coordinates.iter_zephyr_to_linear(g.nodes())
 *             edges = coordinates.iter_zephyr_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_zephyr)             # <<<<<<<<<<<<<<
 *             self.delabel = _make_relabeler(coordinates.iter_zephyr_to_linear)
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_linear_to_zephyr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":832
 *             edges = coordinates.iter_zephyr_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_zephyr)
 *             self.delabel = _make_relabeler(coordinates.iter_zephyr_to_linear)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("unrecognized graph labeling")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_zephyr_to_linear); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":828
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_zephyr_to_linear(g.nodes())
 *             edges = coordinates.iter_zephyr_to_linear_pairs(g.edges())
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":834
 *             self.delabel = _make_relabeler(coordinates.iter_zephyr_to_linear)
 *         else:
 *             raise ValueError("unrecognized graph labeling")             # <<<<<<<<<<<<<<
 * 
 *         self.topo = new topo_cache[zephyr_spec](zep[0], self.nodes, edges)
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 834, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "minorminer/busclique.pyx":836
 *             raise ValueError("unrecognized graph labeling")
 * 
 *         self.topo = new topo_cache[zephyr_spec](zep[0], self.nodes, edges)             # <<<<<<<<<<<<<<
 *         short_clique(zep[0], self.nodes, edges, self.emb_1)
 * 
 */
  __pyx_v_self->topo = new busclique::topo_cache<busclique::zephyr_spec> ((__pyx_v_zep[0]), __pyx_v_self->nodes, __pyx_v_edges);

  /* "minorminer/busclique.pyx":837
 * 
 *         self.topo = new topo_cache[zephyr_spec](zep[0], self.nodes, edges)
 *         short_clique(zep[0], self.nodes, edges, self.emb_1)             # <<<<<<<<<<<<<<
 * 
 *         if compute_identifier:
 */
  (void)(busclique::short_clique<busclique::zephyr_spec>((__pyx_v_zep[0]), __pyx_v_self->nodes, __pyx_v_edges, __pyx_v_self->emb_1));

  /* "minorminer/busclique.pyx":839
 *         short_clique(zep[0], self.nodes, edges, self.emb_1)
 * 
 *         if compute_identifier:             # <<<<<<<<<<<<<<
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'zephyr')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_compute_identifier); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 839, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "minorminer/busclique.pyx":840
 * 
 *         if compute_identifier:
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'zephyr')             # <<<<<<<<<<<<<<
 * 
 *         del zep
 */
    __pyx_t_2 = __pyx_fuse_0__pyx_f_10minorminer_9busclique__serialize(__pyx_v_self->topo, __pyx_n_u_zephyr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 840, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 840, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 840, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_13 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_6 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_7), 2) < 0) __PYX_ERR(0, 840, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 840, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->identifier);
    __Pyx_DECREF(__pyx_v_self->identifier);
    __pyx_v_self->identifier = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_v_self->short_identifier);
    __Pyx_DECREF(__pyx_v_self->short_identifier);
    __pyx_v_self->short_identifier = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "minorminer/busclique.pyx":839
 *         short_clique(zep[0], self.nodes, edges, self.emb_1)
 * 
 *         if compute_identifier:             # <<<<<<<<<<<<<<
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'zephyr')
 * 
 */
  }

  /* "minorminer/busclique.pyx":842
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'zephyr')
 * 
 *         del zep             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  delete __pyx_v_zep;

  /* "minorminer/busclique.pyx":801
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):             # <<<<<<<<<<<<<<
 *         """
 *         This is a class which manages a single zephyr graph, and dispatches
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_seed_obj);
  __Pyx_XDECREF(__pyx_v_coordinates);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":844
 *         del zep
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.topo
 * 
 */

/* Python wrapper */
static void __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_2__dealloc__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_2__dealloc__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "minorminer/busclique.pyx":845
 * 
 *     def __dealloc__(self):
 *         del self.topo             # <<<<<<<<<<<<<<
 * 
 *     def set_mask_bound(self, num_masks):
 */
  delete __pyx_v_self->topo;

  /* "minorminer/busclique.pyx":844
 *         del zep
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.topo
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "minorminer/busclique.pyx":847
 *         del self.topo
 * 
 *     def set_mask_bound(self, num_masks):             # <<<<<<<<<<<<<<
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_5set_mask_bound(PyObject *__pyx_v_self, PyObject *__pyx_v_num_masks); /*proto*/
static char __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_4set_mask_bound[] = "Sets the maximum number of masks to consider.";
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_5set_mask_bound(PyObject *__pyx_v_self, PyObject *__pyx_v_num_masks) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_mask_bound (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_4set_mask_bound(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_num_masks));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_4set_mask_bound(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_num_masks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_mask_bound", 0);

  /* "minorminer/busclique.pyx":849
 *     def set_mask_bound(self, num_masks):
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)             # <<<<<<<<<<<<<<
 * 
 *     def bicliques(self):
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_num_masks); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 849, __pyx_L1_error)
  (__pyx_v_self->topo[0]).set_mask_bound(__pyx_t_1);

  /* "minorminer/busclique.pyx":847
 *         del self.topo
 * 
 *     def set_mask_bound(self, num_masks):             # <<<<<<<<<<<<<<
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.set_mask_bound", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":851
 *         self.topo[0].set_mask_bound(num_masks)
 * 
 *     def bicliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a biclique cache -- see _make_biclique_cache for more info.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7bicliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_6bicliques[] = "\n        Returns a biclique cache -- see _make_biclique_cache for more info.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7bicliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bicliques (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_6bicliques(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_6bicliques(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  std::vector<std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t> >  __pyx_v_embs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bicliques", 0);

  /* "minorminer/busclique.pyx":856
 *         """
 *         cdef vector[pair[pair[size_t, size_t], embedding_t]] embs
 *         best_bicliques[zephyr_spec](self.topo[0], embs)             # <<<<<<<<<<<<<<
 *         return _make_biclique_cache(embs)
 * 
 */
  busclique::best_bicliques<busclique::zephyr_spec>((__pyx_v_self->topo[0]), __pyx_v_embs);

  /* "minorminer/busclique.pyx":857
 *         cdef vector[pair[pair[size_t, size_t], embedding_t]] embs
 *         best_bicliques[zephyr_spec](self.topo[0], embs)
 *         return _make_biclique_cache(embs)             # <<<<<<<<<<<<<<
 * 
 *     def cliques(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10minorminer_9busclique__make_biclique_cache(__pyx_v_embs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":851
 *         self.topo[0].set_mask_bound(num_masks)
 * 
 *     def bicliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a biclique cache -- see _make_biclique_cache for more info.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.bicliques", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":859
 *         return _make_biclique_cache(embs)
 * 
 *     def cliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a clique cache -- see _make_clique_cache for more info.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_9cliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_8cliques[] = "\n        Returns a clique cache -- see _make_clique_cache for more info.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_9cliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cliques (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_8cliques(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_8cliques(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_v_embs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cliques", 0);

  /* "minorminer/busclique.pyx":864
 *         """
 *         cdef vector[embedding_t] embs
 *         best_cliques[zephyr_spec](self.topo[0], embs, self.emb_1)             # <<<<<<<<<<<<<<
 *         return _make_clique_cache(embs)
 * 
 */
  busclique::best_cliques<busclique::zephyr_spec>((__pyx_v_self->topo[0]), __pyx_v_embs, __pyx_v_self->emb_1);

  /* "minorminer/busclique.pyx":865
 *         cdef vector[embedding_t] embs
 *         best_cliques[zephyr_spec](self.topo[0], embs, self.emb_1)
 *         return _make_clique_cache(embs)             # <<<<<<<<<<<<<<
 * 
 *     def independent_set(self, size):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10minorminer_9busclique__make_clique_cache(__pyx_v_embs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":859
 *         return _make_biclique_cache(embs)
 * 
 *     def cliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a clique cache -- see _make_clique_cache for more info.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.cliques", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":867
 *         return _make_clique_cache(embs)
 * 
 *     def independent_set(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         This is extremely silly, but what else should we do when a user requests
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_11independent_set(PyObject *__pyx_v_self, PyObject *__pyx_v_size); /*proto*/
static char __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_10independent_set[] = "\n        This is extremely silly, but what else should we do when a user requests\n        a K_{0, n}?\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_11independent_set(PyObject *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("independent_set (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_10independent_set(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_10independent_set(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_v_emb = NULL;
  int __pyx_9genexpr10__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("independent_set", 0);

  /* "minorminer/busclique.pyx":873
 *         """
 *         cdef int i
 *         if size < self.nodes.size():             # <<<<<<<<<<<<<<
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->nodes.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_size, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "minorminer/busclique.pyx":874
 *         cdef int i
 *         if size < self.nodes.size():
 *             emb = [self.nodes[i] for i in range(size)]             # <<<<<<<<<<<<<<
 *             return emb
 * 
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 874, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_v_size); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 874, __pyx_L1_error)
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_9genexpr10__pyx_v_i = __pyx_t_6;
        __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_self->nodes[__pyx_9genexpr10__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 874, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    } /* exit inner scope */
    __pyx_v_emb = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":875
 *         if size < self.nodes.size():
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb             # <<<<<<<<<<<<<<
 * 
 *     def find_clique_embedding(self, nn):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_emb);
    __pyx_r = __pyx_v_emb;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":873
 *         """
 *         cdef int i
 *         if size < self.nodes.size():             # <<<<<<<<<<<<<<
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb
 */
  }

  /* "minorminer/busclique.pyx":867
 *         return _make_clique_cache(embs)
 * 
 *     def independent_set(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         This is extremely silly, but what else should we do when a user requests
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.independent_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":877
 *             return emb
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """
 *         Perfoms a "one-shot" embedding procedure to find K_n where n is either
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_13find_clique_embedding(PyObject *__pyx_v_self, PyObject *__pyx_v_nn); /*proto*/
static char __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_12find_clique_embedding[] = "\n        Perfoms a \"one-shot\" embedding procedure to find K_n where n is either\n        int(nn) or len(tuple(nn)), without generating a cache.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_13find_clique_embedding(PyObject *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_clique_embedding (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_12find_clique_embedding(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_nn));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_12find_clique_embedding(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_v_nodes = NULL;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_v_emb;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_clique_embedding", 0);

  /* "minorminer/busclique.pyx":882
 *         int(nn) or len(tuple(nn)), without generating a cache.
 *         """
 *         num, nodes = _num_nodes(nn)             # <<<<<<<<<<<<<<
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_num_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_nn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_nn);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 882, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 882, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 882, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_num = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":884
 *         num, nodes = _num_nodes(nn)
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():             # <<<<<<<<<<<<<<
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->emb_1.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_num, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":885
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1             # <<<<<<<<<<<<<<
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}
 */
    __pyx_t_7 = __pyx_v_self->emb_1;
    __pyx_v_emb = __pyx_t_7;

    /* "minorminer/busclique.pyx":884
 *         num, nodes = _num_nodes(nn)
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():             # <<<<<<<<<<<<<<
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":886
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):             # <<<<<<<<<<<<<<
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_num); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 886, __pyx_L1_error)
  __pyx_t_6 = ((!(busclique::find_clique<busclique::zephyr_spec>((__pyx_v_self->topo[0]), __pyx_t_8, __pyx_v_emb) != 0)) != 0);
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":887
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}             # <<<<<<<<<<<<<<
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":886
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):             # <<<<<<<<<<<<<<
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))
 */
  }
  __pyx_L5:;

  /* "minorminer/busclique.pyx":888
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))             # <<<<<<<<<<<<<<
 * 
 *     def fragment_graph_spec(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_v_emb); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_v_self->relabel);
  __pyx_t_1 = __pyx_v_self->relabel; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":877
 *             return emb
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """
 *         Perfoms a "one-shot" embedding procedure to find K_n where n is either
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":890
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 *     def fragment_graph_spec(self):             # <<<<<<<<<<<<<<
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_15fragment_graph_spec(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_15fragment_graph_spec(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fragment_graph_spec (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_14fragment_graph_spec(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_14fragment_graph_spec(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  size_t __pyx_v_m;
  size_t __pyx_v_n;
  size_t __pyx_v_t;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_v_nodes;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_v_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fragment_graph_spec", 0);

  /* "minorminer/busclique.pyx":891
 * 
 *     def fragment_graph_spec(self):
 *         m = coordinate_index(self.topo.topo.dim_y)             # <<<<<<<<<<<<<<
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore
 */
  __pyx_v_m = busclique::coordinate_index(__pyx_v_self->topo->topo.dim_y);

  /* "minorminer/busclique.pyx":892
 *     def fragment_graph_spec(self):
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)             # <<<<<<<<<<<<<<
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()
 */
  __pyx_v_n = busclique::coordinate_index(__pyx_v_self->topo->topo.dim_x);

  /* "minorminer/busclique.pyx":893
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore             # <<<<<<<<<<<<<<
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()
 */
  __pyx_t_1 = __pyx_v_self->topo->topo.shore;
  __pyx_v_t = __pyx_t_1;

  /* "minorminer/busclique.pyx":894
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()             # <<<<<<<<<<<<<<
 *         edges = self.topo.fragment_edges()
 *         return m, n, t, nodes, edges
 */
  __pyx_v_nodes = __pyx_v_self->topo->fragment_nodes();

  /* "minorminer/busclique.pyx":895
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()             # <<<<<<<<<<<<<<
 *         return m, n, t, nodes, edges
 * 
 */
  __pyx_v_edges = __pyx_v_self->topo->fragment_edges();

  /* "minorminer/busclique.pyx":896
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()
 *         return m, n, t, nodes, edges             # <<<<<<<<<<<<<<
 * 
 *     def fragment_nodes(self, nodes = None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_t); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_convert_vector_to_py_size_t(__pyx_v_nodes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_v_edges); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":890
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 *     def fragment_graph_spec(self):             # <<<<<<<<<<<<<<
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.fragment_graph_spec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":898
 *         return m, n, t, nodes, edges
 * 
 *     def fragment_nodes(self, nodes = None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             return self.topo.fragment_nodes()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_17fragment_nodes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_17fragment_nodes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fragment_nodes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fragment_nodes") < 0)) __PYX_ERR(0, 898, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nodes = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fragment_nodes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 898, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.fragment_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_16fragment_nodes(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self), __pyx_v_nodes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_16fragment_nodes(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, PyObject *__pyx_v_nodes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fragment_nodes", 0);

  /* "minorminer/busclique.pyx":899
 * 
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             return self.topo.fragment_nodes()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_nodes == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":900
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:
 *             return self.topo.fragment_nodes()             # <<<<<<<<<<<<<<
 *         else:
 *             return self.topo.topo.fragment_nodes(nodes)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_convert_vector_to_py_size_t(__pyx_v_self->topo->fragment_nodes()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":899
 * 
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             return self.topo.fragment_nodes()
 *         else:
 */
  }

  /* "minorminer/busclique.pyx":902
 *             return self.topo.fragment_nodes()
 *         else:
 *             return self.topo.topo.fragment_nodes(nodes)             # <<<<<<<<<<<<<<
 * 
 * cdef class _pegasus_busgraph:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_convert_vector_from_py_size_t(__pyx_v_nodes); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)
    __pyx_t_3 = __pyx_convert_vector_to_py_size_t(__pyx_v_self->topo->topo.fragment_nodes(__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":898
 *         return m, n, t, nodes, edges
 * 
 *     def fragment_nodes(self, nodes = None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             return self.topo.fragment_nodes()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.fragment_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":797
 *     cdef nodes_t nodes
 *     cdef embedding_t emb_1
 *     cdef readonly object relabel             # <<<<<<<<<<<<<<
 *     cdef readonly object delabel
 *     cdef readonly object identifier
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7relabel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7relabel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_7relabel___get__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_7relabel___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->relabel);
  __pyx_r = __pyx_v_self->relabel;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":798
 *     cdef embedding_t emb_1
 *     cdef readonly object relabel
 *     cdef readonly object delabel             # <<<<<<<<<<<<<<
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7delabel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7delabel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_7delabel___get__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_7delabel___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->delabel);
  __pyx_r = __pyx_v_self->delabel;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":799
 *     cdef readonly object relabel
 *     cdef readonly object delabel
 *     cdef readonly object identifier             # <<<<<<<<<<<<<<
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_10identifier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_10identifier_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_10identifier___get__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_10identifier___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->identifier);
  __pyx_r = __pyx_v_self->identifier;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":800
 *     cdef readonly object delabel
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier             # <<<<<<<<<<<<<<
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_16short_identifier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_16short_identifier_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_16short_identifier___get__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_16short_identifier___get__(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->short_identifier);
  __pyx_r = __pyx_v_self->short_identifier;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_18__reduce_cython__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_20__setstate_cython__(((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_16_zephyr_busgraph_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._zephyr_busgraph.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":912
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):             # <<<<<<<<<<<<<<
 *         """
 *         This is a class which manages a single pegasus graph, and dispatches
 */

/* Python wrapper */
static int __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_compute_identifier = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_g,&__pyx_n_s_seed,&__pyx_n_s_compute_identifier,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_identifier);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 912, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_g = values[0];
    __pyx_v_seed = values[1];
    __pyx_v_compute_identifier = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 912, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph___cinit__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self), __pyx_v_g, __pyx_v_seed, __pyx_v_compute_identifier);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph___cinit__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_compute_identifier) {
  PyObject *__pyx_v_rows = NULL;
  PyObject *__pyx_v_voff = NULL;
  PyObject *__pyx_v_hoff = NULL;
  uint32_t __pyx_v_internal_seed;
  PyObject *__pyx_v_seed_obj = NULL;
  busclique::pegasus_spec *__pyx_v_peg;
  PyObject *__pyx_v_coordinates = NULL;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_v_edges;
  PyObject *__pyx_9genexpr11__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  uint8_t *__pyx_t_9;
  uint32_t __pyx_t_10;
  size_t __pyx_t_11;
  std::vector<uint8_t>  __pyx_t_12;
  std::vector<uint8_t>  __pyx_t_13;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_t_14;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *(*__pyx_t_17)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "minorminer/busclique.pyx":917
 *         various structure-aware c++ embedding functions on it.
 *         """
 *         rows = g.graph['rows']             # <<<<<<<<<<<<<<
 *         voff = [o//2 for o in g.graph['vertical_offsets'][::2]]
 *         hoff = [o//2 for o in g.graph['horizontal_offsets'][::2]]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_rows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rows = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":918
 *         """
 *         rows = g.graph['rows']
 *         voff = [o//2 for o in g.graph['vertical_offsets'][::2]]             # <<<<<<<<<<<<<<
 *         hoff = [o//2 for o in g.graph['horizontal_offsets'][::2]]
 * 
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_vertical_offsets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_slice__12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 918, __pyx_L5_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 918, __pyx_L5_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 918, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_o, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_FloorDivideObjC(__pyx_9genexpr11__pyx_v_o, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 918, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_o); __pyx_9genexpr11__pyx_v_o = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_o); __pyx_9genexpr11__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_voff = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":919
 *         rows = g.graph['rows']
 *         voff = [o//2 for o in g.graph['vertical_offsets'][::2]]
 *         hoff = [o//2 for o in g.graph['horizontal_offsets'][::2]]             # <<<<<<<<<<<<<<
 * 
 *         cdef uint32_t internal_seed
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 919, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_horizontal_offsets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_slice__12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 919, __pyx_L11_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 919, __pyx_L11_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 919, __pyx_L11_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 919, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_o, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyInt_FloorDivideObjC(__pyx_9genexpr12__pyx_v_o, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 919, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_o); __pyx_9genexpr12__pyx_v_o = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_o); __pyx_9genexpr12__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  __pyx_v_hoff = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":922
 * 
 *         cdef uint32_t internal_seed
 *         if seed is None:             # <<<<<<<<<<<<<<
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 */
  __pyx_t_6 = (__pyx_v_seed == Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "minorminer/busclique.pyx":923
 *         cdef uint32_t internal_seed
 *         if seed is None:
 *             seed_obj = os.urandom(sizeof(uint32_t))             # <<<<<<<<<<<<<<
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_urandom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(uint32_t))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_seed_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":924
 *         if seed is None:
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]             # <<<<<<<<<<<<<<
 *         else:
 *             internal_seed = seed
 */
    __pyx_t_9 = __Pyx_PyObject_AsWritableUString(__pyx_v_seed_obj); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 924, __pyx_L1_error)
    __pyx_v_internal_seed = (((uint32_t *)((void *)((uint8_t *)__pyx_t_9)))[0]);

    /* "minorminer/busclique.pyx":922
 * 
 *         cdef uint32_t internal_seed
 *         if seed is None:             # <<<<<<<<<<<<<<
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 */
    goto __pyx_L15;
  }

  /* "minorminer/busclique.pyx":926
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 *         else:
 *             internal_seed = seed             # <<<<<<<<<<<<<<
 * 
 *         cdef pegasus_spec *peg = new pegasus_spec(rows, voff, hoff, internal_seed)
 */
  /*else*/ {
    __pyx_t_10 = __Pyx_PyInt_As_uint32_t(__pyx_v_seed); if (unlikely((__pyx_t_10 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 926, __pyx_L1_error)
    __pyx_v_internal_seed = __pyx_t_10;
  }
  __pyx_L15:;

  /* "minorminer/busclique.pyx":928
 *             internal_seed = seed
 * 
 *         cdef pegasus_spec *peg = new pegasus_spec(rows, voff, hoff, internal_seed)             # <<<<<<<<<<<<<<
 *         coordinates = dnx.pegasus_coordinates(rows)
 *         cdef edges_t edges
 */
  __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_v_rows); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 928, __pyx_L1_error)
  __pyx_t_12 = __pyx_convert_vector_from_py_uint8_t(__pyx_v_voff); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 928, __pyx_L1_error)
  __pyx_t_13 = __pyx_convert_vector_from_py_uint8_t(__pyx_v_hoff); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 928, __pyx_L1_error)
  __pyx_v_peg = new busclique::pegasus_spec(__pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_v_internal_seed);

  /* "minorminer/busclique.pyx":929
 * 
 *         cdef pegasus_spec *peg = new pegasus_spec(rows, voff, hoff, internal_seed)
 *         coordinates = dnx.pegasus_coordinates(rows)             # <<<<<<<<<<<<<<
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_dnx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_pegasus_coordinates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_rows) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_rows);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_coordinates = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":931
 *         coordinates = dnx.pegasus_coordinates(rows)
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':             # <<<<<<<<<<<<<<
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_labels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_int, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "minorminer/busclique.pyx":932
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()             # <<<<<<<<<<<<<<
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_14 = __pyx_convert_vector_from_py_size_t(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->nodes = __pyx_t_14;

    /* "minorminer/busclique.pyx":933
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()
 *             edges = g.edges()             # <<<<<<<<<<<<<<
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_15 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_edges = __pyx_t_15;

    /* "minorminer/busclique.pyx":934
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler             # <<<<<<<<<<<<<<
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_trivial_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":935
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler             # <<<<<<<<<<<<<<
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_pegasus_to_linear(g.nodes())
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_trivial_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":931
 *         coordinates = dnx.pegasus_coordinates(rows)
 *         cdef edges_t edges
 *         if g.graph['labels'] == 'int':             # <<<<<<<<<<<<<<
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 */
    goto __pyx_L16;
  }

  /* "minorminer/busclique.pyx":936
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_pegasus_to_linear(g.nodes())
 *             edges = coordinates.iter_pegasus_to_linear_pairs(g.edges())
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_labels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_coordinate, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {

    /* "minorminer/busclique.pyx":937
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_pegasus_to_linear(g.nodes())             # <<<<<<<<<<<<<<
 *             edges = coordinates.iter_pegasus_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_pegasus)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_pegasus_to_linear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_16 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_3 = (__pyx_t_16) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_16) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_14 = __pyx_convert_vector_from_py_size_t(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_self->nodes = __pyx_t_14;

    /* "minorminer/busclique.pyx":938
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_pegasus_to_linear(g.nodes())
 *             edges = coordinates.iter_pegasus_to_linear_pairs(g.edges())             # <<<<<<<<<<<<<<
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_pegasus)
 *             self.delabel = _make_relabeler(coordinates.iter_pegasus_to_linear)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_pegasus_to_linear_pairs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_16 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_3 = (__pyx_t_16) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_16) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_15 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_edges = __pyx_t_15;

    /* "minorminer/busclique.pyx":939
 *             self.nodes = coordinates.iter_pegasus_to_linear(g.nodes())
 *             edges = coordinates.iter_pegasus_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_pegasus)             # <<<<<<<<<<<<<<
 *             self.delabel = _make_relabeler(coordinates.iter_pegasus_to_linear)
 *         elif g.graph['labels'] == 'nice':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_linear_to_pegasus); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":940
 *             edges = coordinates.iter_pegasus_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_pegasus)
 *             self.delabel = _make_relabeler(coordinates.iter_pegasus_to_linear)             # <<<<<<<<<<<<<<
 *         elif g.graph['labels'] == 'nice':
 *             self.nodes = coordinates.iter_nice_to_linear(g.nodes())
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_pegasus_to_linear); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":936
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_pegasus_to_linear(g.nodes())
 *             edges = coordinates.iter_pegasus_to_linear_pairs(g.edges())
 */
    goto __pyx_L16;
  }

  /* "minorminer/busclique.pyx":941
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_pegasus)
 *             self.delabel = _make_relabeler(coordinates.iter_pegasus_to_linear)
 *         elif g.graph['labels'] == 'nice':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_nice_to_linear(g.nodes())
 *             edges = coordinates.iter_nice_to_linear_pairs(g.edges())
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_labels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_nice, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_7)) {

    /* "minorminer/busclique.pyx":942
 *             self.delabel = _make_relabeler(coordinates.iter_pegasus_to_linear)
 *         elif g.graph['labels'] == 'nice':
 *             self.nodes = coordinates.iter_nice_to_linear(g.nodes())             # <<<<<<<<<<<<<<
 *             edges = coordinates.iter_nice_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_nice)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_nice_to_linear); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_16 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_3 = (__pyx_t_16) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_16) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_14 = __pyx_convert_vector_from_py_size_t(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->nodes = __pyx_t_14;

    /* "minorminer/busclique.pyx":943
 *         elif g.graph['labels'] == 'nice':
 *             self.nodes = coordinates.iter_nice_to_linear(g.nodes())
 *             edges = coordinates.iter_nice_to_linear_pairs(g.edges())             # <<<<<<<<<<<<<<
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_nice)
 *             self.delabel = _make_relabeler(coordinates.iter_nice_to_linear)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_nice_to_linear_pairs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_16 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_3 = (__pyx_t_16) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_16) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_15 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_edges = __pyx_t_15;

    /* "minorminer/busclique.pyx":944
 *             self.nodes = coordinates.iter_nice_to_linear(g.nodes())
 *             edges = coordinates.iter_nice_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_nice)             # <<<<<<<<<<<<<<
 *             self.delabel = _make_relabeler(coordinates.iter_nice_to_linear)
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_linear_to_nice); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":945
 *             edges = coordinates.iter_nice_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_nice)
 *             self.delabel = _make_relabeler(coordinates.iter_nice_to_linear)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("unrecognized graph labeling")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_nice_to_linear); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":941
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_pegasus)
 *             self.delabel = _make_relabeler(coordinates.iter_pegasus_to_linear)
 *         elif g.graph['labels'] == 'nice':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_nice_to_linear(g.nodes())
 *             edges = coordinates.iter_nice_to_linear_pairs(g.edges())
 */
    goto __pyx_L16;
  }

  /* "minorminer/busclique.pyx":947
 *             self.delabel = _make_relabeler(coordinates.iter_nice_to_linear)
 *         else:
 *             raise ValueError("unrecognized graph labeling")             # <<<<<<<<<<<<<<
 * 
 *         self.topo = new topo_cache[pegasus_spec](peg[0], self.nodes, edges)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 947, __pyx_L1_error)
  }
  __pyx_L16:;

  /* "minorminer/busclique.pyx":949
 *             raise ValueError("unrecognized graph labeling")
 * 
 *         self.topo = new topo_cache[pegasus_spec](peg[0], self.nodes, edges)             # <<<<<<<<<<<<<<
 *         short_clique(peg[0], self.nodes, edges, self.emb_1)
 * 
 */
  __pyx_v_self->topo = new busclique::topo_cache<busclique::pegasus_spec> ((__pyx_v_peg[0]), __pyx_v_self->nodes, __pyx_v_edges);

  /* "minorminer/busclique.pyx":950
 * 
 *         self.topo = new topo_cache[pegasus_spec](peg[0], self.nodes, edges)
 *         short_clique(peg[0], self.nodes, edges, self.emb_1)             # <<<<<<<<<<<<<<
 * 
 *         if compute_identifier:
 */
  (void)(busclique::short_clique<busclique::pegasus_spec>((__pyx_v_peg[0]), __pyx_v_self->nodes, __pyx_v_edges, __pyx_v_self->emb_1));

  /* "minorminer/busclique.pyx":952
 *         short_clique(peg[0], self.nodes, edges, self.emb_1)
 * 
 *         if compute_identifier:             # <<<<<<<<<<<<<<
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'pegasus')
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_compute_identifier); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 952, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "minorminer/busclique.pyx":953
 * 
 *         if compute_identifier:
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'pegasus')             # <<<<<<<<<<<<<<
 * 
 *         del peg
 */
    __pyx_t_1 = __pyx_fuse_1__pyx_f_10minorminer_9busclique__serialize(__pyx_v_self->topo, __pyx_n_u_pegasus); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 953, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_17 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_17(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L18_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_3 = __pyx_t_17(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L18_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_8), 2) < 0) __PYX_ERR(0, 953, __pyx_L1_error)
      __pyx_t_17 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L19_unpacking_done;
      __pyx_L18_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_17 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 953, __pyx_L1_error)
      __pyx_L19_unpacking_done:;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->identifier);
    __Pyx_DECREF(__pyx_v_self->identifier);
    __pyx_v_self->identifier = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->short_identifier);
    __Pyx_DECREF(__pyx_v_self->short_identifier);
    __pyx_v_self->short_identifier = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":952
 *         short_clique(peg[0], self.nodes, edges, self.emb_1)
 * 
 *         if compute_identifier:             # <<<<<<<<<<<<<<
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'pegasus')
 * 
 */
  }

  /* "minorminer/busclique.pyx":955
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'pegasus')
 * 
 *         del peg             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  delete __pyx_v_peg;

  /* "minorminer/busclique.pyx":912
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):             # <<<<<<<<<<<<<<
 *         """
 *         This is a class which manages a single pegasus graph, and dispatches
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rows);
  __Pyx_XDECREF(__pyx_v_voff);
  __Pyx_XDECREF(__pyx_v_hoff);
  __Pyx_XDECREF(__pyx_v_seed_obj);
  __Pyx_XDECREF(__pyx_v_coordinates);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":957
 *         del peg
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.topo
 * 
 */

/* Python wrapper */
static void __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_2__dealloc__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_2__dealloc__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "minorminer/busclique.pyx":958
 * 
 *     def __dealloc__(self):
 *         del self.topo             # <<<<<<<<<<<<<<
 * 
 *     def set_mask_bound(self, num_masks):
 */
  delete __pyx_v_self->topo;

  /* "minorminer/busclique.pyx":957
 *         del peg
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.topo
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "minorminer/busclique.pyx":960
 *         del self.topo
 * 
 *     def set_mask_bound(self, num_masks):             # <<<<<<<<<<<<<<
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_5set_mask_bound(PyObject *__pyx_v_self, PyObject *__pyx_v_num_masks); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_4set_mask_bound[] = "Sets the maximum number of masks to consider.";
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_5set_mask_bound(PyObject *__pyx_v_self, PyObject *__pyx_v_num_masks) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_mask_bound (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_4set_mask_bound(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_num_masks));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_4set_mask_bound(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_num_masks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_mask_bound", 0);

  /* "minorminer/busclique.pyx":962
 *     def set_mask_bound(self, num_masks):
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)             # <<<<<<<<<<<<<<
 * 
 *     def bicliques(self):
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_num_masks); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L1_error)
  (__pyx_v_self->topo[0]).set_mask_bound(__pyx_t_1);

  /* "minorminer/busclique.pyx":960
 *         del self.topo
 * 
 *     def set_mask_bound(self, num_masks):             # <<<<<<<<<<<<<<
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.set_mask_bound", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":964
 *         self.topo[0].set_mask_bound(num_masks)
 * 
 *     def bicliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a biclique cache -- see _make_biclique_cache for more info.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7bicliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_6bicliques[] = "\n        Returns a biclique cache -- see _make_biclique_cache for more info.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7bicliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bicliques (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_6bicliques(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_6bicliques(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  std::vector<std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t> >  __pyx_v_embs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bicliques", 0);

  /* "minorminer/busclique.pyx":969
 *         """
 *         cdef vector[pair[pair[size_t, size_t], embedding_t]] embs
 *         best_bicliques[pegasus_spec](self.topo[0], embs)             # <<<<<<<<<<<<<<
 *         return _make_biclique_cache(embs)
 * 
 */
  busclique::best_bicliques<busclique::pegasus_spec>((__pyx_v_self->topo[0]), __pyx_v_embs);

  /* "minorminer/busclique.pyx":970
 *         cdef vector[pair[pair[size_t, size_t], embedding_t]] embs
 *         best_bicliques[pegasus_spec](self.topo[0], embs)
 *         return _make_biclique_cache(embs)             # <<<<<<<<<<<<<<
 * 
 *     def cliques(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10minorminer_9busclique__make_biclique_cache(__pyx_v_embs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":964
 *         self.topo[0].set_mask_bound(num_masks)
 * 
 *     def bicliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a biclique cache -- see _make_biclique_cache for more info.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.bicliques", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":972
 *         return _make_biclique_cache(embs)
 * 
 *     def cliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a clique cache -- see _make_clique_cache for more info.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_9cliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_8cliques[] = "\n        Returns a clique cache -- see _make_clique_cache for more info.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_9cliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cliques (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_8cliques(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_8cliques(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_v_embs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cliques", 0);

  /* "minorminer/busclique.pyx":977
 *         """
 *         cdef vector[embedding_t] embs
 *         best_cliques[pegasus_spec](self.topo[0], embs, self.emb_1)             # <<<<<<<<<<<<<<
 *         return _make_clique_cache(embs)
 * 
 */
  busclique::best_cliques<busclique::pegasus_spec>((__pyx_v_self->topo[0]), __pyx_v_embs, __pyx_v_self->emb_1);

  /* "minorminer/busclique.pyx":978
 *         cdef vector[embedding_t] embs
 *         best_cliques[pegasus_spec](self.topo[0], embs, self.emb_1)
 *         return _make_clique_cache(embs)             # <<<<<<<<<<<<<<
 * 
 *     def independent_set(self, size):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10minorminer_9busclique__make_clique_cache(__pyx_v_embs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":972
 *         return _make_biclique_cache(embs)
 * 
 *     def cliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a clique cache -- see _make_clique_cache for more info.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.cliques", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":980
 *         return _make_clique_cache(embs)
 * 
 *     def independent_set(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         This is extremely silly, but what else should we do when a user requests
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_11independent_set(PyObject *__pyx_v_self, PyObject *__pyx_v_size); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_10independent_set[] = "\n        This is extremely silly, but what else should we do when a user requests\n        a K_{0, n}?\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_11independent_set(PyObject *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("independent_set (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_10independent_set(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_10independent_set(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_v_emb = NULL;
  int __pyx_9genexpr13__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("independent_set", 0);

  /* "minorminer/busclique.pyx":986
 *         """
 *         cdef int i
 *         if size < self.nodes.size():             # <<<<<<<<<<<<<<
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->nodes.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_size, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "minorminer/busclique.pyx":987
 *         cdef int i
 *         if size < self.nodes.size():
 *             emb = [self.nodes[i] for i in range(size)]             # <<<<<<<<<<<<<<
 *             return emb
 * 
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_v_size); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 987, __pyx_L1_error)
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_9genexpr13__pyx_v_i = __pyx_t_6;
        __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_self->nodes[__pyx_9genexpr13__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 987, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    } /* exit inner scope */
    __pyx_v_emb = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":988
 *         if size < self.nodes.size():
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb             # <<<<<<<<<<<<<<
 * 
 *     def find_clique_embedding(self, nn):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_emb);
    __pyx_r = __pyx_v_emb;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":986
 *         """
 *         cdef int i
 *         if size < self.nodes.size():             # <<<<<<<<<<<<<<
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb
 */
  }

  /* "minorminer/busclique.pyx":980
 *         return _make_clique_cache(embs)
 * 
 *     def independent_set(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         This is extremely silly, but what else should we do when a user requests
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.independent_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":990
 *             return emb
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """
 *         Perfoms a "one-shot" embedding procedure to find K_n where n is either
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_13find_clique_embedding(PyObject *__pyx_v_self, PyObject *__pyx_v_nn); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_12find_clique_embedding[] = "\n        Perfoms a \"one-shot\" embedding procedure to find K_n where n is either\n        int(nn) or len(tuple(nn)), without generating a cache.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_13find_clique_embedding(PyObject *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_clique_embedding (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_12find_clique_embedding(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_nn));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_12find_clique_embedding(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_v_nodes = NULL;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_v_emb;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_clique_embedding", 0);

  /* "minorminer/busclique.pyx":995
 *         int(nn) or len(tuple(nn)), without generating a cache.
 *         """
 *         num, nodes = _num_nodes(nn)             # <<<<<<<<<<<<<<
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_num_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_nn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_nn);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 995, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 995, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_num = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":997
 *         num, nodes = _num_nodes(nn)
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():             # <<<<<<<<<<<<<<
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->emb_1.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_num, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":998
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1             # <<<<<<<<<<<<<<
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}
 */
    __pyx_t_7 = __pyx_v_self->emb_1;
    __pyx_v_emb = __pyx_t_7;

    /* "minorminer/busclique.pyx":997
 *         num, nodes = _num_nodes(nn)
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():             # <<<<<<<<<<<<<<
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":999
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):             # <<<<<<<<<<<<<<
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_num); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 999, __pyx_L1_error)
  __pyx_t_6 = ((!(busclique::find_clique<busclique::pegasus_spec>((__pyx_v_self->topo[0]), __pyx_t_8, __pyx_v_emb) != 0)) != 0);
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":1000
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}             # <<<<<<<<<<<<<<
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":999
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):             # <<<<<<<<<<<<<<
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))
 */
  }
  __pyx_L5:;

  /* "minorminer/busclique.pyx":1001
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))             # <<<<<<<<<<<<<<
 * 
 *     def fragment_graph_spec(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_v_emb); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_v_self->relabel);
  __pyx_t_1 = __pyx_v_self->relabel; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":990
 *             return emb
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """
 *         Perfoms a "one-shot" embedding procedure to find K_n where n is either
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1003
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 *     def fragment_graph_spec(self):             # <<<<<<<<<<<<<<
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_15fragment_graph_spec(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_15fragment_graph_spec(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fragment_graph_spec (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_14fragment_graph_spec(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_14fragment_graph_spec(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  size_t __pyx_v_m;
  size_t __pyx_v_n;
  size_t __pyx_v_t;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_v_nodes;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_v_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fragment_graph_spec", 0);

  /* "minorminer/busclique.pyx":1004
 * 
 *     def fragment_graph_spec(self):
 *         m = coordinate_index(self.topo.topo.dim_y)             # <<<<<<<<<<<<<<
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore
 */
  __pyx_v_m = busclique::coordinate_index(__pyx_v_self->topo->topo.dim_y);

  /* "minorminer/busclique.pyx":1005
 *     def fragment_graph_spec(self):
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)             # <<<<<<<<<<<<<<
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()
 */
  __pyx_v_n = busclique::coordinate_index(__pyx_v_self->topo->topo.dim_x);

  /* "minorminer/busclique.pyx":1006
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore             # <<<<<<<<<<<<<<
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()
 */
  __pyx_t_1 = __pyx_v_self->topo->topo.shore;
  __pyx_v_t = __pyx_t_1;

  /* "minorminer/busclique.pyx":1007
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()             # <<<<<<<<<<<<<<
 *         edges = self.topo.fragment_edges()
 *         return m, n, t, nodes, edges
 */
  __pyx_v_nodes = __pyx_v_self->topo->fragment_nodes();

  /* "minorminer/busclique.pyx":1008
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()             # <<<<<<<<<<<<<<
 *         return m, n, t, nodes, edges
 * 
 */
  __pyx_v_edges = __pyx_v_self->topo->fragment_edges();

  /* "minorminer/busclique.pyx":1009
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()
 *         return m, n, t, nodes, edges             # <<<<<<<<<<<<<<
 * 
 *     def fragment_nodes(self, nodes = None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_t); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_convert_vector_to_py_size_t(__pyx_v_nodes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_v_edges); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1003
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 *     def fragment_graph_spec(self):             # <<<<<<<<<<<<<<
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.fragment_graph_spec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1011
 *         return m, n, t, nodes, edges
 * 
 *     def fragment_nodes(self, nodes = None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             return self.topo.fragment_nodes()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_17fragment_nodes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_17fragment_nodes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fragment_nodes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fragment_nodes") < 0)) __PYX_ERR(0, 1011, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nodes = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fragment_nodes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1011, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.fragment_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_16fragment_nodes(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self), __pyx_v_nodes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_16fragment_nodes(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, PyObject *__pyx_v_nodes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fragment_nodes", 0);

  /* "minorminer/busclique.pyx":1012
 * 
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             return self.topo.fragment_nodes()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_nodes == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":1013
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:
 *             return self.topo.fragment_nodes()             # <<<<<<<<<<<<<<
 *         else:
 *             return self.topo.topo.fragment_nodes(nodes)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_convert_vector_to_py_size_t(__pyx_v_self->topo->fragment_nodes()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":1012
 * 
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             return self.topo.fragment_nodes()
 *         else:
 */
  }

  /* "minorminer/busclique.pyx":1015
 *             return self.topo.fragment_nodes()
 *         else:
 *             return self.topo.topo.fragment_nodes(nodes)             # <<<<<<<<<<<<<<
 * 
 * cdef class _chimera_busgraph:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_convert_vector_from_py_size_t(__pyx_v_nodes); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1015, __pyx_L1_error)
    __pyx_t_3 = __pyx_convert_vector_to_py_size_t(__pyx_v_self->topo->topo.fragment_nodes(__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":1011
 *         return m, n, t, nodes, edges
 * 
 *     def fragment_nodes(self, nodes = None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             return self.topo.fragment_nodes()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.fragment_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":908
 *     cdef nodes_t nodes
 *     cdef embedding_t emb_1
 *     cdef readonly object relabel             # <<<<<<<<<<<<<<
 *     cdef readonly object delabel
 *     cdef readonly object identifier
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7relabel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7relabel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_7relabel___get__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_7relabel___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->relabel);
  __pyx_r = __pyx_v_self->relabel;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":909
 *     cdef embedding_t emb_1
 *     cdef readonly object relabel
 *     cdef readonly object delabel             # <<<<<<<<<<<<<<
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7delabel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7delabel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_7delabel___get__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_7delabel___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->delabel);
  __pyx_r = __pyx_v_self->delabel;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":910
 *     cdef readonly object relabel
 *     cdef readonly object delabel
 *     cdef readonly object identifier             # <<<<<<<<<<<<<<
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_10identifier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_10identifier_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_10identifier___get__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_10identifier___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->identifier);
  __pyx_r = __pyx_v_self->identifier;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":911
 *     cdef readonly object delabel
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier             # <<<<<<<<<<<<<<
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_16short_identifier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_16short_identifier_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_16short_identifier___get__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_16short_identifier___get__(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->short_identifier);
  __pyx_r = __pyx_v_self->short_identifier;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_18__reduce_cython__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_20__setstate_cython__(((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_pegasus_busgraph_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._pegasus_busgraph.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1035
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):             # <<<<<<<<<<<<<<
 *         rows = g.graph['rows']
 *         cols = g.graph['columns']
 */

/* Python wrapper */
static int __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_compute_identifier = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_g,&__pyx_n_s_seed,&__pyx_n_s_compute_identifier,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compute_identifier);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1035, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_g = values[0];
    __pyx_v_seed = values[1];
    __pyx_v_compute_identifier = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1035, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph___cinit__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self), __pyx_v_g, __pyx_v_seed, __pyx_v_compute_identifier);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10minorminer_9busclique_17_chimera_busgraph___cinit__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_g, PyObject *__pyx_v_seed, PyObject *__pyx_v_compute_identifier) {
  PyObject *__pyx_v_rows = NULL;
  PyObject *__pyx_v_cols = NULL;
  PyObject *__pyx_v_tile = NULL;
  uint32_t __pyx_v_internal_seed;
  PyObject *__pyx_v_seed_obj = NULL;
  busclique::chimera_spec *__pyx_v_chim;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_v_edges;
  PyObject *__pyx_v_coordinates = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  uint8_t *__pyx_t_7;
  uint32_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  uint8_t __pyx_t_11;
  int __pyx_t_12;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_t_13;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "minorminer/busclique.pyx":1036
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 *         rows = g.graph['rows']             # <<<<<<<<<<<<<<
 *         cols = g.graph['columns']
 *         tile = g.graph['tile']
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_rows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rows = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1037
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 *         rows = g.graph['rows']
 *         cols = g.graph['columns']             # <<<<<<<<<<<<<<
 *         tile = g.graph['tile']
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_columns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cols = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1038
 *         rows = g.graph['rows']
 *         cols = g.graph['columns']
 *         tile = g.graph['tile']             # <<<<<<<<<<<<<<
 * 
 *         if tile > 8:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_tile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_tile = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1040
 *         tile = g.graph['tile']
 * 
 *         if tile > 8:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError(("this clique embedder supports chimera "
 *                                        "graphs with a tile size of 8 or less"))
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_tile, __pyx_int_8, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "minorminer/busclique.pyx":1041
 * 
 *         if tile > 8:
 *             raise NotImplementedError(("this clique embedder supports chimera "             # <<<<<<<<<<<<<<
 *                                        "graphs with a tile size of 8 or less"))
 *         cdef uint32_t internal_seed
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1041, __pyx_L1_error)

    /* "minorminer/busclique.pyx":1040
 *         tile = g.graph['tile']
 * 
 *         if tile > 8:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError(("this clique embedder supports chimera "
 *                                        "graphs with a tile size of 8 or less"))
 */
  }

  /* "minorminer/busclique.pyx":1044
 *                                        "graphs with a tile size of 8 or less"))
 *         cdef uint32_t internal_seed
 *         if seed is None:             # <<<<<<<<<<<<<<
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 */
  __pyx_t_3 = (__pyx_v_seed == Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":1045
 *         cdef uint32_t internal_seed
 *         if seed is None:
 *             seed_obj = os.urandom(sizeof(uint32_t))             # <<<<<<<<<<<<<<
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_urandom); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(uint32_t))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_seed_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1046
 *         if seed is None:
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]             # <<<<<<<<<<<<<<
 *         else:
 *             internal_seed = seed
 */
    __pyx_t_7 = __Pyx_PyObject_AsWritableUString(__pyx_v_seed_obj); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 1046, __pyx_L1_error)
    __pyx_v_internal_seed = (((uint32_t *)((void *)((uint8_t *)__pyx_t_7)))[0]);

    /* "minorminer/busclique.pyx":1044
 *                                        "graphs with a tile size of 8 or less"))
 *         cdef uint32_t internal_seed
 *         if seed is None:             # <<<<<<<<<<<<<<
 *             seed_obj = os.urandom(sizeof(uint32_t))
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 */
    goto __pyx_L4;
  }

  /* "minorminer/busclique.pyx":1048
 *             internal_seed = (<uint32_t *>(<void *>(<uint8_t *>(seed_obj))))[0]
 *         else:
 *             internal_seed = seed             # <<<<<<<<<<<<<<
 * 
 *         cdef chimera_spec *chim = new chimera_spec(rows, cols, tile, internal_seed)
 */
  /*else*/ {
    __pyx_t_8 = __Pyx_PyInt_As_uint32_t(__pyx_v_seed); if (unlikely((__pyx_t_8 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L1_error)
    __pyx_v_internal_seed = __pyx_t_8;
  }
  __pyx_L4:;

  /* "minorminer/busclique.pyx":1050
 *             internal_seed = seed
 * 
 *         cdef chimera_spec *chim = new chimera_spec(rows, cols, tile, internal_seed)             # <<<<<<<<<<<<<<
 *         cdef edges_t edges
 *         coordinates = dnx.chimera_coordinates(rows, cols, tile)
 */
  __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_v_rows); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_cols); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyInt_As_uint8_t(__pyx_v_tile); if (unlikely((__pyx_t_11 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L1_error)
  __pyx_v_chim = new busclique::chimera_spec(__pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_v_internal_seed);

  /* "minorminer/busclique.pyx":1052
 *         cdef chimera_spec *chim = new chimera_spec(rows, cols, tile, internal_seed)
 *         cdef edges_t edges
 *         coordinates = dnx.chimera_coordinates(rows, cols, tile)             # <<<<<<<<<<<<<<
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_dnx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_chimera_coordinates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_rows, __pyx_v_cols, __pyx_v_tile};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_rows, __pyx_v_cols, __pyx_v_tile};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_rows);
    __Pyx_GIVEREF(__pyx_v_rows);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_12, __pyx_v_rows);
    __Pyx_INCREF(__pyx_v_cols);
    __Pyx_GIVEREF(__pyx_v_cols);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_12, __pyx_v_cols);
    __Pyx_INCREF(__pyx_v_tile);
    __Pyx_GIVEREF(__pyx_v_tile);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_12, __pyx_v_tile);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_coordinates = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1053
 *         cdef edges_t edges
 *         coordinates = dnx.chimera_coordinates(rows, cols, tile)
 *         if g.graph['labels'] == 'int':             # <<<<<<<<<<<<<<
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_labels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_int, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":1054
 *         coordinates = dnx.chimera_coordinates(rows, cols, tile)
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()             # <<<<<<<<<<<<<<
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = __pyx_convert_vector_from_py_size_t(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->nodes = __pyx_t_13;

    /* "minorminer/busclique.pyx":1055
 *         if g.graph['labels'] == 'int':
 *             self.nodes = g.nodes()
 *             edges = g.edges()             # <<<<<<<<<<<<<<
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_14 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_edges = __pyx_t_14;

    /* "minorminer/busclique.pyx":1056
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler             # <<<<<<<<<<<<<<
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_trivial_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":1057
 *             edges = g.edges()
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler             # <<<<<<<<<<<<<<
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_chimera_to_linear(g.nodes())
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_trivial_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":1053
 *         cdef edges_t edges
 *         coordinates = dnx.chimera_coordinates(rows, cols, tile)
 *         if g.graph['labels'] == 'int':             # <<<<<<<<<<<<<<
 *             self.nodes = g.nodes()
 *             edges = g.edges()
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":1058
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_chimera_to_linear(g.nodes())
 *             edges = coordinates.iter_chimera_to_linear_pairs(g.edges())
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_labels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_coordinate, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_4)) {

    /* "minorminer/busclique.pyx":1059
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_chimera_to_linear(g.nodes())             # <<<<<<<<<<<<<<
 *             edges = coordinates.iter_chimera_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_chimera)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_chimera_to_linear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_nodes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_15 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_15) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_13 = __pyx_convert_vector_from_py_size_t(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_self->nodes = __pyx_t_13;

    /* "minorminer/busclique.pyx":1060
 *         elif g.graph['labels'] == 'coordinate':
 *             self.nodes = coordinates.iter_chimera_to_linear(g.nodes())
 *             edges = coordinates.iter_chimera_to_linear_pairs(g.edges())             # <<<<<<<<<<<<<<
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_chimera)
 *             self.delabel = _make_relabeler(coordinates.iter_chimera_to_linear)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_chimera_to_linear_pairs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_edges); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_15 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_15) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_14 = __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_edges = __pyx_t_14;

    /* "minorminer/busclique.pyx":1061
 *             self.nodes = coordinates.iter_chimera_to_linear(g.nodes())
 *             edges = coordinates.iter_chimera_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_chimera)             # <<<<<<<<<<<<<<
 *             self.delabel = _make_relabeler(coordinates.iter_chimera_to_linear)
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_linear_to_chimera); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->relabel);
    __Pyx_DECREF(__pyx_v_self->relabel);
    __pyx_v_self->relabel = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1062
 *             edges = coordinates.iter_chimera_to_linear_pairs(g.edges())
 *             self.relabel = _make_relabeler(coordinates.iter_linear_to_chimera)
 *             self.delabel = _make_relabeler(coordinates.iter_chimera_to_linear)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("unrecognized graph labeling")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_make_relabeler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_coordinates, __pyx_n_s_iter_chimera_to_linear); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->delabel);
    __Pyx_DECREF(__pyx_v_self->delabel);
    __pyx_v_self->delabel = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1058
 *             self.relabel = _trivial_relabeler
 *             self.delabel = _trivial_relabeler
 *         elif g.graph['labels'] == 'coordinate':             # <<<<<<<<<<<<<<
 *             self.nodes = coordinates.iter_chimera_to_linear(g.nodes())
 *             edges = coordinates.iter_chimera_to_linear_pairs(g.edges())
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":1064
 *             self.delabel = _make_relabeler(coordinates.iter_chimera_to_linear)
 *         else:
 *             raise ValueError("unrecognized graph labeling")             # <<<<<<<<<<<<<<
 * 
 *         self.topo = new topo_cache[chimera_spec](chim[0], self.nodes, edges)
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1064, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "minorminer/busclique.pyx":1066
 *             raise ValueError("unrecognized graph labeling")
 * 
 *         self.topo = new topo_cache[chimera_spec](chim[0], self.nodes, edges)             # <<<<<<<<<<<<<<
 *         short_clique(chim[0], self.nodes, edges, self.emb_1)
 * 
 */
  __pyx_v_self->topo = new busclique::topo_cache<busclique::chimera_spec> ((__pyx_v_chim[0]), __pyx_v_self->nodes, __pyx_v_edges);

  /* "minorminer/busclique.pyx":1067
 * 
 *         self.topo = new topo_cache[chimera_spec](chim[0], self.nodes, edges)
 *         short_clique(chim[0], self.nodes, edges, self.emb_1)             # <<<<<<<<<<<<<<
 * 
 *         if compute_identifier:
 */
  (void)(busclique::short_clique<busclique::chimera_spec>((__pyx_v_chim[0]), __pyx_v_self->nodes, __pyx_v_edges, __pyx_v_self->emb_1));

  /* "minorminer/busclique.pyx":1069
 *         short_clique(chim[0], self.nodes, edges, self.emb_1)
 * 
 *         if compute_identifier:             # <<<<<<<<<<<<<<
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'chimera')
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_compute_identifier); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1069, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "minorminer/busclique.pyx":1070
 * 
 *         if compute_identifier:
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'chimera')             # <<<<<<<<<<<<<<
 * 
 *         del chim
 */
    __pyx_t_2 = __pyx_fuse_2__pyx_f_10minorminer_9busclique__serialize(__pyx_v_self->topo, __pyx_n_u_chimera); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1070, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_16 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_16(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_6 = __pyx_t_16(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1070, __pyx_L1_error)
      __pyx_t_16 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_16 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1070, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->identifier);
    __Pyx_DECREF(__pyx_v_self->identifier);
    __pyx_v_self->identifier = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_v_self->short_identifier);
    __Pyx_DECREF(__pyx_v_self->short_identifier);
    __pyx_v_self->short_identifier = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "minorminer/busclique.pyx":1069
 *         short_clique(chim[0], self.nodes, edges, self.emb_1)
 * 
 *         if compute_identifier:             # <<<<<<<<<<<<<<
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'chimera')
 * 
 */
  }

  /* "minorminer/busclique.pyx":1072
 *             self.identifier, self.short_identifier = _serialize(self.topo, 'chimera')
 * 
 *         del chim             # <<<<<<<<<<<<<<
 * 
 * 
 */
  delete __pyx_v_chim;

  /* "minorminer/busclique.pyx":1035
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):             # <<<<<<<<<<<<<<
 *         rows = g.graph['rows']
 *         cols = g.graph['columns']
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rows);
  __Pyx_XDECREF(__pyx_v_cols);
  __Pyx_XDECREF(__pyx_v_tile);
  __Pyx_XDECREF(__pyx_v_seed_obj);
  __Pyx_XDECREF(__pyx_v_coordinates);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1075
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.topo
 * 
 */

/* Python wrapper */
static void __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_2__dealloc__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_2__dealloc__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "minorminer/busclique.pyx":1076
 * 
 *     def __dealloc__(self):
 *         del self.topo             # <<<<<<<<<<<<<<
 * 
 *     def set_mask_bound(self, num_masks):
 */
  delete __pyx_v_self->topo;

  /* "minorminer/busclique.pyx":1075
 * 
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self.topo
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "minorminer/busclique.pyx":1078
 *         del self.topo
 * 
 *     def set_mask_bound(self, num_masks):             # <<<<<<<<<<<<<<
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_5set_mask_bound(PyObject *__pyx_v_self, PyObject *__pyx_v_num_masks); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_4set_mask_bound[] = "Sets the maximum number of masks to consider.";
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_5set_mask_bound(PyObject *__pyx_v_self, PyObject *__pyx_v_num_masks) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_mask_bound (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_4set_mask_bound(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_num_masks));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_4set_mask_bound(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_num_masks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_mask_bound", 0);

  /* "minorminer/busclique.pyx":1080
 *     def set_mask_bound(self, num_masks):
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)             # <<<<<<<<<<<<<<
 * 
 *     def bicliques(self):
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_num_masks); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L1_error)
  (__pyx_v_self->topo[0]).set_mask_bound(__pyx_t_1);

  /* "minorminer/busclique.pyx":1078
 *         del self.topo
 * 
 *     def set_mask_bound(self, num_masks):             # <<<<<<<<<<<<<<
 *         """Sets the maximum number of masks to consider."""
 *         self.topo[0].set_mask_bound(num_masks)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.set_mask_bound", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1082
 *         self.topo[0].set_mask_bound(num_masks)
 * 
 *     def bicliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a biclique cache -- see _make_biclique_cache for more info.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7bicliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_6bicliques[] = "\n        Returns a biclique cache -- see _make_biclique_cache for more info.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7bicliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bicliques (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_6bicliques(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_6bicliques(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  std::vector<std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t> >  __pyx_v_embs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bicliques", 0);

  /* "minorminer/busclique.pyx":1087
 *         """
 *         cdef vector[pair[pair[size_t, size_t], embedding_t]] embs
 *         best_bicliques(self.topo[0], embs)             # <<<<<<<<<<<<<<
 *         return _make_biclique_cache(embs)
 * 
 */
  busclique::best_bicliques<busclique::chimera_spec>((__pyx_v_self->topo[0]), __pyx_v_embs);

  /* "minorminer/busclique.pyx":1088
 *         cdef vector[pair[pair[size_t, size_t], embedding_t]] embs
 *         best_bicliques(self.topo[0], embs)
 *         return _make_biclique_cache(embs)             # <<<<<<<<<<<<<<
 * 
 *     def cliques(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10minorminer_9busclique__make_biclique_cache(__pyx_v_embs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1082
 *         self.topo[0].set_mask_bound(num_masks)
 * 
 *     def bicliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a biclique cache -- see _make_biclique_cache for more info.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.bicliques", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1090
 *         return _make_biclique_cache(embs)
 * 
 *     def cliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a clique cache -- see _make_clique_cache for more info.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_9cliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_8cliques[] = "\n        Returns a clique cache -- see _make_clique_cache for more info.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_9cliques(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cliques (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_8cliques(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_8cliques(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_v_embs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cliques", 0);

  /* "minorminer/busclique.pyx":1095
 *         """
 *         cdef vector[embedding_t] embs
 *         best_cliques(self.topo[0], embs, self.emb_1)             # <<<<<<<<<<<<<<
 *         return _make_clique_cache(embs)
 * 
 */
  busclique::best_cliques<busclique::chimera_spec>((__pyx_v_self->topo[0]), __pyx_v_embs, __pyx_v_self->emb_1);

  /* "minorminer/busclique.pyx":1096
 *         cdef vector[embedding_t] embs
 *         best_cliques(self.topo[0], embs, self.emb_1)
 *         return _make_clique_cache(embs)             # <<<<<<<<<<<<<<
 * 
 *     def independent_set(self, size):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10minorminer_9busclique__make_clique_cache(__pyx_v_embs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1090
 *         return _make_biclique_cache(embs)
 * 
 *     def cliques(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a clique cache -- see _make_clique_cache for more info.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.cliques", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1098
 *         return _make_clique_cache(embs)
 * 
 *     def independent_set(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         This is extremely silly, but what else should we do when a user requests
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_11independent_set(PyObject *__pyx_v_self, PyObject *__pyx_v_size); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_10independent_set[] = "\n        This is extremely silly, but what else should we do when a user requests\n        a K_{0, n}?\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_11independent_set(PyObject *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("independent_set (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_10independent_set(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_10independent_set(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_v_emb = NULL;
  int __pyx_9genexpr14__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("independent_set", 0);

  /* "minorminer/busclique.pyx":1104
 *         """
 *         cdef int i
 *         if size < self.nodes.size():             # <<<<<<<<<<<<<<
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->nodes.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_size, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "minorminer/busclique.pyx":1105
 *         cdef int i
 *         if size < self.nodes.size():
 *             emb = [self.nodes[i] for i in range(size)]             # <<<<<<<<<<<<<<
 *             return emb
 * 
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_v_size); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L1_error)
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_9genexpr14__pyx_v_i = __pyx_t_6;
        __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_self->nodes[__pyx_9genexpr14__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 1105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    } /* exit inner scope */
    __pyx_v_emb = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1106
 *         if size < self.nodes.size():
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb             # <<<<<<<<<<<<<<
 * 
 *     def find_clique_embedding(self, nn):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_emb);
    __pyx_r = __pyx_v_emb;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":1104
 *         """
 *         cdef int i
 *         if size < self.nodes.size():             # <<<<<<<<<<<<<<
 *             emb = [self.nodes[i] for i in range(size)]
 *             return emb
 */
  }

  /* "minorminer/busclique.pyx":1098
 *         return _make_clique_cache(embs)
 * 
 *     def independent_set(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         This is extremely silly, but what else should we do when a user requests
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.independent_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1108
 *             return emb
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """
 *         Perfoms a "one-shot" embedding procedure to find K_n where n is either
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_13find_clique_embedding(PyObject *__pyx_v_self, PyObject *__pyx_v_nn); /*proto*/
static char __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_12find_clique_embedding[] = "\n        Perfoms a \"one-shot\" embedding procedure to find K_n where n is either\n        int(nn) or len(tuple(nn)), without generating a cache.\n        ";
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_13find_clique_embedding(PyObject *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_clique_embedding (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_12find_clique_embedding(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v_nn));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_12find_clique_embedding(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_nn) {
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_v_nodes = NULL;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_v_emb;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_clique_embedding", 0);

  /* "minorminer/busclique.pyx":1113
 *         int(nn) or len(tuple(nn)), without generating a cache.
 *         """
 *         num, nodes = _num_nodes(nn)             # <<<<<<<<<<<<<<
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_num_nodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_nn) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_nn);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1113, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1113, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1113, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_num = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_nodes = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":1115
 *         num, nodes = _num_nodes(nn)
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():             # <<<<<<<<<<<<<<
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->emb_1.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_num, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":1116
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1             # <<<<<<<<<<<<<<
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}
 */
    __pyx_t_7 = __pyx_v_self->emb_1;
    __pyx_v_emb = __pyx_t_7;

    /* "minorminer/busclique.pyx":1115
 *         num, nodes = _num_nodes(nn)
 *         cdef embedding_t emb
 *         if num <= self.emb_1.size():             # <<<<<<<<<<<<<<
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 */
    goto __pyx_L5;
  }

  /* "minorminer/busclique.pyx":1117
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):             # <<<<<<<<<<<<<<
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))
 */
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_v_num); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L1_error)
  __pyx_t_6 = ((!(busclique::find_clique<busclique::chimera_spec>((__pyx_v_self->topo[0]), __pyx_t_8, __pyx_v_emb) != 0)) != 0);
  if (__pyx_t_6) {

    /* "minorminer/busclique.pyx":1118
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}             # <<<<<<<<<<<<<<
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":1117
 *         if num <= self.emb_1.size():
 *             emb = self.emb_1
 *         elif not find_clique(self.topo[0], num, emb):             # <<<<<<<<<<<<<<
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))
 */
  }
  __pyx_L5:;

  /* "minorminer/busclique.pyx":1119
 *         elif not find_clique(self.topo[0], num, emb):
 *             return {}
 *         return self.relabel(dict(zip(nodes, emb)))             # <<<<<<<<<<<<<<
 * 
 *     def fragment_graph_spec(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_v_emb); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_v_nodes);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nodes);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_v_self->relabel);
  __pyx_t_1 = __pyx_v_self->relabel; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1108
 *             return emb
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """
 *         Perfoms a "one-shot" embedding procedure to find K_n where n is either
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.find_clique_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1121
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 *     def fragment_graph_spec(self):             # <<<<<<<<<<<<<<
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_15fragment_graph_spec(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_15fragment_graph_spec(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fragment_graph_spec (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_14fragment_graph_spec(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_14fragment_graph_spec(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  size_t __pyx_v_m;
  size_t __pyx_v_n;
  size_t __pyx_v_t;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_v_nodes;
  __pyx_t_10minorminer_9busclique_edges_t __pyx_v_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fragment_graph_spec", 0);

  /* "minorminer/busclique.pyx":1122
 * 
 *     def fragment_graph_spec(self):
 *         m = coordinate_index(self.topo.topo.dim_y)             # <<<<<<<<<<<<<<
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore
 */
  __pyx_v_m = busclique::coordinate_index(__pyx_v_self->topo->topo.dim_y);

  /* "minorminer/busclique.pyx":1123
 *     def fragment_graph_spec(self):
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)             # <<<<<<<<<<<<<<
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()
 */
  __pyx_v_n = busclique::coordinate_index(__pyx_v_self->topo->topo.dim_x);

  /* "minorminer/busclique.pyx":1124
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore             # <<<<<<<<<<<<<<
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()
 */
  __pyx_t_1 = __pyx_v_self->topo->topo.shore;
  __pyx_v_t = __pyx_t_1;

  /* "minorminer/busclique.pyx":1125
 *         n = coordinate_index(self.topo.topo.dim_x)
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()             # <<<<<<<<<<<<<<
 *         edges = self.topo.fragment_edges()
 *         return m, n, t, nodes, edges
 */
  __pyx_v_nodes = __pyx_v_self->topo->fragment_nodes();

  /* "minorminer/busclique.pyx":1126
 *         t = self.topo.topo.shore
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()             # <<<<<<<<<<<<<<
 *         return m, n, t, nodes, edges
 * 
 */
  __pyx_v_edges = __pyx_v_self->topo->fragment_edges();

  /* "minorminer/busclique.pyx":1127
 *         nodes = self.topo.fragment_nodes()
 *         edges = self.topo.fragment_edges()
 *         return m, n, t, nodes, edges             # <<<<<<<<<<<<<<
 * 
 *     def fragment_nodes(self, nodes = None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_t); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_convert_vector_to_py_size_t(__pyx_v_nodes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(__pyx_v_edges); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1121
 *         return self.relabel(dict(zip(nodes, emb)))
 * 
 *     def fragment_graph_spec(self):             # <<<<<<<<<<<<<<
 *         m = coordinate_index(self.topo.topo.dim_y)
 *         n = coordinate_index(self.topo.topo.dim_x)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.fragment_graph_spec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1129
 *         return m, n, t, nodes, edges
 * 
 *     def fragment_nodes(self, nodes = None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             return self.topo.fragment_nodes()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_17fragment_nodes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_17fragment_nodes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fragment_nodes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fragment_nodes") < 0)) __PYX_ERR(0, 1129, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nodes = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fragment_nodes", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1129, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.fragment_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_16fragment_nodes(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self), __pyx_v_nodes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_16fragment_nodes(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, PyObject *__pyx_v_nodes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_10minorminer_9busclique_nodes_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fragment_nodes", 0);

  /* "minorminer/busclique.pyx":1130
 * 
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             return self.topo.fragment_nodes()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_nodes == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "minorminer/busclique.pyx":1131
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:
 *             return self.topo.fragment_nodes()             # <<<<<<<<<<<<<<
 *         else:
 *             return self.topo.topo.fragment_nodes(nodes)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_convert_vector_to_py_size_t(__pyx_v_self->topo->fragment_nodes()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "minorminer/busclique.pyx":1130
 * 
 *     def fragment_nodes(self, nodes = None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             return self.topo.fragment_nodes()
 *         else:
 */
  }

  /* "minorminer/busclique.pyx":1133
 *             return self.topo.fragment_nodes()
 *         else:
 *             return self.topo.topo.fragment_nodes(nodes)             # <<<<<<<<<<<<<<
 * 
 * def _default_quality_function(emb):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_convert_vector_from_py_size_t(__pyx_v_nodes); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1133, __pyx_L1_error)
    __pyx_t_3 = __pyx_convert_vector_to_py_size_t(__pyx_v_self->topo->topo.fragment_nodes(__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":1129
 *         return m, n, t, nodes, edges
 * 
 *     def fragment_nodes(self, nodes = None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             return self.topo.fragment_nodes()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.fragment_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1031
 *     cdef embedding_t emb_1
 *     cdef nodes_t nodes
 *     cdef readonly object relabel             # <<<<<<<<<<<<<<
 *     cdef readonly object delabel
 *     cdef readonly object identifier
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7relabel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7relabel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_7relabel___get__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_7relabel___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->relabel);
  __pyx_r = __pyx_v_self->relabel;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1032
 *     cdef nodes_t nodes
 *     cdef readonly object relabel
 *     cdef readonly object delabel             # <<<<<<<<<<<<<<
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7delabel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7delabel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_7delabel___get__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_7delabel___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->delabel);
  __pyx_r = __pyx_v_self->delabel;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1033
 *     cdef readonly object relabel
 *     cdef readonly object delabel
 *     cdef readonly object identifier             # <<<<<<<<<<<<<<
 *     cdef readonly object short_identifier
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_10identifier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_10identifier_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_10identifier___get__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_10identifier___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->identifier);
  __pyx_r = __pyx_v_self->identifier;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1034
 *     cdef readonly object delabel
 *     cdef readonly object identifier
 *     cdef readonly object short_identifier             # <<<<<<<<<<<<<<
 *     def __cinit__(self, g, seed = 0, compute_identifier = False):
 *         rows = g.graph['rows']
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_16short_identifier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_16short_identifier_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_16short_identifier___get__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_16short_identifier___get__(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->short_identifier);
  __pyx_r = __pyx_v_self->short_identifier;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_18__reduce_cython__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_17_chimera_busgraph_20__setstate_cython__(((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_17_chimera_busgraph_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique._chimera_busgraph.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1135
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * def _default_quality_function(emb):             # <<<<<<<<<<<<<<
 *     """A function that returns a tuple corresponds to the "quality" of the embedding.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_9_default_quality_function(PyObject *__pyx_self, PyObject *__pyx_v_emb); /*proto*/
static char __pyx_doc_10minorminer_9busclique_8_default_quality_function[] = "A function that returns a tuple corresponds to the \"quality\" of the embedding.\n\n    The \"quality\" of the embedding is measured by the maximum chainlength (to be\n    minimized), and the number of chains with that length (to be maximized).\n\n    Args:\n        emb: (list/tuple of lists/tuples)\n            The embedding.\n\n    Returns:\n        quality: (tuple)\n\n    Usage:\n        better_emb = max(emb0, emb1, key = _default_quality_function)\n\n    ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_9_default_quality_function = {"_default_quality_function", (PyCFunction)__pyx_pw_10minorminer_9busclique_9_default_quality_function, METH_O, __pyx_doc_10minorminer_9busclique_8_default_quality_function};
static PyObject *__pyx_pw_10minorminer_9busclique_9_default_quality_function(PyObject *__pyx_self, PyObject *__pyx_v_emb) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_default_quality_function (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_8_default_quality_function(__pyx_self, ((PyObject *)__pyx_v_emb));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10minorminer_9busclique_25_default_quality_function_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "minorminer/busclique.pyx":1153
 *     """
 *     maxlen = max(map(len, emb))
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)             # <<<<<<<<<<<<<<
 * 
 * def _regularize_embedding(g, emb):
 */

static PyObject *__pyx_pf_10minorminer_9busclique_25_default_quality_function_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_5_genexpr(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1153, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10minorminer_9busclique_25_default_quality_function_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_default_quality_function_locals, __pyx_n_s_minorminer_busclique); if (unlikely(!gen)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("minorminer.busclique._default_quality_function.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10minorminer_9busclique_25_default_quality_function_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1153, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_emb)) { __Pyx_RaiseClosureNameError("emb"); __PYX_ERR(0, 1153, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_emb)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_emb)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_emb; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_emb); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1153, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1153, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1153, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1153, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1153, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1153, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_chain);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_chain, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_chain); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1153, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_maxlen)) { __Pyx_RaiseClosureNameError("maxlen"); __PYX_ERR(0, 1153, __pyx_L1_error) }
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_maxlen, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1153, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1135
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * def _default_quality_function(emb):             # <<<<<<<<<<<<<<
 *     """A function that returns a tuple corresponds to the "quality" of the embedding.
 * 
 */

static PyObject *__pyx_pf_10minorminer_9busclique_8_default_quality_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_emb) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *__pyx_cur_scope;
  PyObject *__pyx_gb_10minorminer_9busclique_25_default_quality_function_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_default_quality_function", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1135, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_emb = __pyx_v_emb;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_emb);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_emb);

  /* "minorminer/busclique.pyx":1152
 * 
 *     """
 *     maxlen = max(map(len, emb))             # <<<<<<<<<<<<<<
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 */
  __pyx_t_1 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_emb);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_emb);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_emb);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1153
 *     """
 *     maxlen = max(map(len, emb))
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)             # <<<<<<<<<<<<<<
 * 
 * def _regularize_embedding(g, emb):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyNumber_Negative(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_pf_10minorminer_9busclique_25_default_quality_function_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1135
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * def _default_quality_function(emb):             # <<<<<<<<<<<<<<
 *     """A function that returns a tuple corresponds to the "quality" of the embedding.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("minorminer.busclique._default_quality_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_10minorminer_9busclique_25_default_quality_function_2generator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1155
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 * def _regularize_embedding(g, emb):             # <<<<<<<<<<<<<<
 *     """Best-effort augmented embedding, with hopefully uniform chainlegnths.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_11_regularize_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_10_regularize_embedding[] = "Best-effort augmented embedding, with hopefully uniform chainlegnths.\n\n    We augment ``emb`` by adding nodes to chains that are shorter than the chain\n    with maximum length.\n\n    Args:\n        emb: (dict with list/tuple values)\n            The embedding\n\n    Returns:\n        emb: (dict with tuple values)\n            The augmented embedding.\n\n    ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_11_regularize_embedding = {"_regularize_embedding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_11_regularize_embedding, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_10_regularize_embedding};
static PyObject *__pyx_pw_10minorminer_9busclique_11_regularize_embedding(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_emb = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_regularize_embedding (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_g,&__pyx_n_s_emb,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_emb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_regularize_embedding", 1, 2, 2, 1); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_regularize_embedding") < 0)) __PYX_ERR(0, 1155, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_g = values[0];
    __pyx_v_emb = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_regularize_embedding", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1155, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique._regularize_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_10_regularize_embedding(__pyx_self, __pyx_v_g, __pyx_v_emb);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10minorminer_9busclique_21_regularize_embedding_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "minorminer/busclique.pyx":1187
 *                 chain_node = (i, 'chain', j)
 *                 top_nodes.append(chain_node)
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)             # <<<<<<<<<<<<<<
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 *         for chain_node in top_nodes:
 */

static PyObject *__pyx_pf_10minorminer_9busclique_21_regularize_embedding_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_7_genexpr(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_7_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1187, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10minorminer_9busclique_21_regularize_embedding_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_regularize_embedding_locals_gen, __pyx_n_s_minorminer_busclique); if (unlikely(!gen)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("minorminer.busclique._regularize_embedding.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10minorminer_9busclique_21_regularize_embedding_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1187, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_boundary)) { __Pyx_RaiseClosureNameError("boundary"); __PYX_ERR(0, 1187, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_boundary)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_boundary)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_boundary; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_boundary); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1187, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1187, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1187, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1187, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_q);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_q, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_chain_node)) { __Pyx_RaiseClosureNameError("chain_node"); __PYX_ERR(0, 1187, __pyx_L1_error) }
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_chain_node);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_chain_node);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_chain_node);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_q);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_q);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_cur_scope->__pyx_v_q);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1187, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1155
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 * def _regularize_embedding(g, emb):             # <<<<<<<<<<<<<<
 *     """Best-effort augmented embedding, with hopefully uniform chainlegnths.
 * 
 */

static PyObject *__pyx_pf_10minorminer_9busclique_10_regularize_embedding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_g, PyObject *__pyx_v_emb) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *__pyx_cur_scope;
  PyObject *__pyx_v_progress = NULL;
  PyObject *__pyx_v_max_length = NULL;
  PyObject *__pyx_v_remainder = NULL;
  PyObject *__pyx_v_matching_graph = NULL;
  PyObject *__pyx_v_top_nodes = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_chain = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_matching = NULL;
  PyObject *__pyx_v_matched = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_i = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_chain = NULL;
  PyObject *__pyx_gb_10minorminer_9busclique_21_regularize_embedding_2generator1 = 0;
  PyObject *__pyx_9genexpr18__pyx_v_i = NULL;
  PyObject *__pyx_9genexpr18__pyx_v_chain = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_regularize_embedding", 0);
  __pyx_cur_scope = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *)__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding(__pyx_ptype_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1155, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_INCREF(__pyx_v_emb);

  /* "minorminer/busclique.pyx":1170
 * 
 *     """
 *     progress = len(set(map(len, emb.values()))) != 1             # <<<<<<<<<<<<<<
 *     max_length = max(map(len, emb.values()))
 *     emb = {i: list(chain) for i, chain in emb.items()}
 */
  __pyx_t_1 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_emb, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PySet_New(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = PySet_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_t_5 != 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_progress = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":1171
 *     """
 *     progress = len(set(map(len, emb.values()))) != 1
 *     max_length = max(map(len, emb.values()))             # <<<<<<<<<<<<<<
 *     emb = {i: list(chain) for i, chain in emb.items()}
 *     remainder = set(g) - set().union(*emb.values())
 */
  __pyx_t_3 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_emb, __pyx_n_s_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_max_length = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1172
 *     progress = len(set(map(len, emb.values()))) != 1
 *     max_length = max(map(len, emb.values()))
 *     emb = {i: list(chain) for i, chain in emb.items()}             # <<<<<<<<<<<<<<
 *     remainder = set(g) - set().union(*emb.values())
 *     while progress:
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = 0;
    if (unlikely(__pyx_v_emb == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 1172, __pyx_L5_error)
    }
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_emb, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1172, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_6, &__pyx_t_5, &__pyx_t_3, &__pyx_t_4, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 1172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_chain, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_9genexpr16__pyx_v_chain); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1172, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr16__pyx_v_i, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1172, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_chain); __pyx_9genexpr16__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_i); __pyx_9genexpr16__pyx_v_i = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_chain); __pyx_9genexpr16__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_i); __pyx_9genexpr16__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_emb, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1173
 *     max_length = max(map(len, emb.values()))
 *     emb = {i: list(chain) for i, chain in emb.items()}
 *     remainder = set(g) - set().union(*emb.values())             # <<<<<<<<<<<<<<
 *     while progress:
 *         progress = False
 */
  __pyx_t_1 = PySet_New(__pyx_v_g); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_union); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_emb, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_remainder = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":1174
 *     emb = {i: list(chain) for i, chain in emb.items()}
 *     remainder = set(g) - set().union(*emb.values())
 *     while progress:             # <<<<<<<<<<<<<<
 *         progress = False
 *         matching_graph = nx.Graph()
 */
  while (1) {
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_progress); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1174, __pyx_L1_error)
    if (!__pyx_t_10) break;

    /* "minorminer/busclique.pyx":1175
 *     remainder = set(g) - set().union(*emb.values())
 *     while progress:
 *         progress = False             # <<<<<<<<<<<<<<
 *         matching_graph = nx.Graph()
 *         top_nodes = []
 */
    __Pyx_INCREF(Py_False);
    __Pyx_DECREF_SET(__pyx_v_progress, Py_False);

    /* "minorminer/busclique.pyx":1176
 *     while progress:
 *         progress = False
 *         matching_graph = nx.Graph()             # <<<<<<<<<<<<<<
 *         top_nodes = []
 *         for i, chain in emb.items():
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_matching_graph, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":1177
 *         progress = False
 *         matching_graph = nx.Graph()
 *         top_nodes = []             # <<<<<<<<<<<<<<
 *         for i, chain in emb.items():
 *             if len(chain) == max_length:
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_top_nodes, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":1178
 *         matching_graph = nx.Graph()
 *         top_nodes = []
 *         for i, chain in emb.items():             # <<<<<<<<<<<<<<
 *             if len(chain) == max_length:
 *                 continue
 */
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_dict_iterator(__pyx_v_emb, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_1;
    __pyx_t_1 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_6, &__pyx_t_1, &__pyx_t_2, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 1178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_chain, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "minorminer/busclique.pyx":1179
 *         top_nodes = []
 *         for i, chain in emb.items():
 *             if len(chain) == max_length:             # <<<<<<<<<<<<<<
 *                 continue
 *             boundary = nx.node_boundary(g, chain, remainder)
 */
      __pyx_t_11 = PyObject_Length(__pyx_v_chain); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1179, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_v_max_length, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1179, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1179, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_10) {

        /* "minorminer/busclique.pyx":1180
 *         for i, chain in emb.items():
 *             if len(chain) == max_length:
 *                 continue             # <<<<<<<<<<<<<<
 *             boundary = nx.node_boundary(g, chain, remainder)
 *             if not boundary:
 */
        goto __pyx_L11_continue;

        /* "minorminer/busclique.pyx":1179
 *         top_nodes = []
 *         for i, chain in emb.items():
 *             if len(chain) == max_length:             # <<<<<<<<<<<<<<
 *                 continue
 *             boundary = nx.node_boundary(g, chain, remainder)
 */
      }

      /* "minorminer/busclique.pyx":1181
 *             if len(chain) == max_length:
 *                 continue
 *             boundary = nx.node_boundary(g, chain, remainder)             # <<<<<<<<<<<<<<
 *             if not boundary:
 *                 continue
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_node_boundary); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_g, __pyx_v_chain, __pyx_v_remainder};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1181, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_g, __pyx_v_chain, __pyx_v_remainder};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1181, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_g);
        __Pyx_GIVEREF(__pyx_v_g);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_g);
        __Pyx_INCREF(__pyx_v_chain);
        __Pyx_GIVEREF(__pyx_v_chain);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_chain);
        __Pyx_INCREF(__pyx_v_remainder);
        __Pyx_GIVEREF(__pyx_v_remainder);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_remainder);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_boundary);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_boundary, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":1182
 *                 continue
 *             boundary = nx.node_boundary(g, chain, remainder)
 *             if not boundary:             # <<<<<<<<<<<<<<
 *                 continue
 *             for j in range(len(chain), max_length):
 */
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_boundary); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1182, __pyx_L1_error)
      __pyx_t_12 = ((!__pyx_t_10) != 0);
      if (__pyx_t_12) {

        /* "minorminer/busclique.pyx":1183
 *             boundary = nx.node_boundary(g, chain, remainder)
 *             if not boundary:
 *                 continue             # <<<<<<<<<<<<<<
 *             for j in range(len(chain), max_length):
 *                 chain_node = (i, 'chain', j)
 */
        goto __pyx_L11_continue;

        /* "minorminer/busclique.pyx":1182
 *                 continue
 *             boundary = nx.node_boundary(g, chain, remainder)
 *             if not boundary:             # <<<<<<<<<<<<<<
 *                 continue
 *             for j in range(len(chain), max_length):
 */
      }

      /* "minorminer/busclique.pyx":1184
 *             if not boundary:
 *                 continue
 *             for j in range(len(chain), max_length):             # <<<<<<<<<<<<<<
 *                 chain_node = (i, 'chain', j)
 *                 top_nodes.append(chain_node)
 */
      __pyx_t_11 = PyObject_Length(__pyx_v_chain); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1184, __pyx_L1_error)
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_max_length);
      __Pyx_GIVEREF(__pyx_v_max_length);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_max_length);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_11 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_13 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1184, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_11); __Pyx_INCREF(__pyx_t_1); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 1184, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_11); __Pyx_INCREF(__pyx_t_1); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 1184, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_13(__pyx_t_4);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1184, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "minorminer/busclique.pyx":1185
 *                 continue
 *             for j in range(len(chain), max_length):
 *                 chain_node = (i, 'chain', j)             # <<<<<<<<<<<<<<
 *                 top_nodes.append(chain_node)
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)
 */
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_n_u_chain);
        __Pyx_GIVEREF(__pyx_n_u_chain);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_chain);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_j);
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_chain_node);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_chain_node, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_1);
        __pyx_t_1 = 0;

        /* "minorminer/busclique.pyx":1186
 *             for j in range(len(chain), max_length):
 *                 chain_node = (i, 'chain', j)
 *                 top_nodes.append(chain_node)             # <<<<<<<<<<<<<<
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 */
        __pyx_t_1 = __pyx_cur_scope->__pyx_v_chain_node;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_top_nodes, __pyx_t_1); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 1186, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "minorminer/busclique.pyx":1187
 *                 chain_node = (i, 'chain', j)
 *                 top_nodes.append(chain_node)
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)             # <<<<<<<<<<<<<<
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 *         for chain_node in top_nodes:
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_matching_graph, __pyx_n_s_add_edges_from); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_2 = __pyx_pf_10minorminer_9busclique_21_regularize_embedding_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_1 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_15, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "minorminer/busclique.pyx":1184
 *             if not boundary:
 *                 continue
 *             for j in range(len(chain), max_length):             # <<<<<<<<<<<<<<
 *                 chain_node = (i, 'chain', j)
 *                 top_nodes.append(chain_node)
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_L11_continue:;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":1188
 *                 top_nodes.append(chain_node)
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)             # <<<<<<<<<<<<<<
 *         for chain_node in top_nodes:
 *             matched = matching.get(chain_node)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_nx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_bipartite); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_maximum_matching); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_matching_graph, __pyx_v_top_nodes};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1188, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_matching_graph, __pyx_v_top_nodes};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1188, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_matching_graph);
      __Pyx_GIVEREF(__pyx_v_matching_graph);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_7, __pyx_v_matching_graph);
      __Pyx_INCREF(__pyx_v_top_nodes);
      __Pyx_GIVEREF(__pyx_v_top_nodes);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_7, __pyx_v_top_nodes);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_matching, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":1189
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 *         for chain_node in top_nodes:             # <<<<<<<<<<<<<<
 *             matched = matching.get(chain_node)
 *             if matched is not None:
 */
    __pyx_t_3 = __pyx_v_top_nodes; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1189, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_chain_node);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_chain_node, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":1190
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 *         for chain_node in top_nodes:
 *             matched = matching.get(chain_node)             # <<<<<<<<<<<<<<
 *             if matched is not None:
 *                 remainder.discard(matched)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_matching, __pyx_n_s_get); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_cur_scope->__pyx_v_chain_node) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_cur_scope->__pyx_v_chain_node);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_matched, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":1191
 *         for chain_node in top_nodes:
 *             matched = matching.get(chain_node)
 *             if matched is not None:             # <<<<<<<<<<<<<<
 *                 remainder.discard(matched)
 *                 progress = True
 */
      __pyx_t_12 = (__pyx_v_matched != Py_None);
      __pyx_t_10 = (__pyx_t_12 != 0);
      if (__pyx_t_10) {

        /* "minorminer/busclique.pyx":1192
 *             matched = matching.get(chain_node)
 *             if matched is not None:
 *                 remainder.discard(matched)             # <<<<<<<<<<<<<<
 *                 progress = True
 *                 emb[chain_node[0]].append(matched)
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_remainder, __pyx_n_s_discard); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_matched) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_matched);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "minorminer/busclique.pyx":1193
 *             if matched is not None:
 *                 remainder.discard(matched)
 *                 progress = True             # <<<<<<<<<<<<<<
 *                 emb[chain_node[0]].append(matched)
 *         if progress:
 */
        __Pyx_INCREF(Py_True);
        __Pyx_DECREF_SET(__pyx_v_progress, Py_True);

        /* "minorminer/busclique.pyx":1194
 *                 remainder.discard(matched)
 *                 progress = True
 *                 emb[chain_node[0]].append(matched)             # <<<<<<<<<<<<<<
 *         if progress:
 *             progress = len(set(map(len, emb.values()))) != 1
 */
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_chain_node, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1194, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_emb, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1194, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_14 = __Pyx_PyObject_Append(__pyx_t_9, __pyx_v_matched); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 1194, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "minorminer/busclique.pyx":1191
 *         for chain_node in top_nodes:
 *             matched = matching.get(chain_node)
 *             if matched is not None:             # <<<<<<<<<<<<<<
 *                 remainder.discard(matched)
 *                 progress = True
 */
      }

      /* "minorminer/busclique.pyx":1189
 *                 matching_graph.add_edges_from((chain_node, q) for q in boundary)
 *         matching = nx.bipartite.maximum_matching(matching_graph, top_nodes)
 *         for chain_node in top_nodes:             # <<<<<<<<<<<<<<
 *             matched = matching.get(chain_node)
 *             if matched is not None:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "minorminer/busclique.pyx":1195
 *                 progress = True
 *                 emb[chain_node[0]].append(matched)
 *         if progress:             # <<<<<<<<<<<<<<
 *             progress = len(set(map(len, emb.values()))) != 1
 *     return {i: tuple(chain) for i, chain in emb.items()}
 */
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_progress); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1195, __pyx_L1_error)
    if (__pyx_t_10) {

      /* "minorminer/busclique.pyx":1196
 *                 emb[chain_node[0]].append(matched)
 *         if progress:
 *             progress = len(set(map(len, emb.values()))) != 1             # <<<<<<<<<<<<<<
 *     return {i: tuple(chain) for i, chain in emb.items()}
 * 
 */
      __pyx_t_3 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_emb, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
      __pyx_t_3 = 0;
      __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_4, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySet_New(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_5 = PySet_GET_SIZE(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyBool_FromLong((__pyx_t_5 != 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_progress, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":1195
 *                 progress = True
 *                 emb[chain_node[0]].append(matched)
 *         if progress:             # <<<<<<<<<<<<<<
 *             progress = len(set(map(len, emb.values()))) != 1
 *     return {i: tuple(chain) for i, chain in emb.items()}
 */
    }
  }

  /* "minorminer/busclique.pyx":1197
 *         if progress:
 *             progress = len(set(map(len, emb.values()))) != 1
 *     return {i: tuple(chain) for i, chain in emb.items()}             # <<<<<<<<<<<<<<
 * 
 * def mine_clique_embeddings(
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_emb, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1197, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_9);
    __pyx_t_9 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_9, __pyx_t_6, &__pyx_t_5, &__pyx_t_3, &__pyx_t_1, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 1197, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_chain, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_9genexpr18__pyx_v_chain); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1197, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(PyDict_SetItem(__pyx_t_4, (PyObject*)__pyx_9genexpr18__pyx_v_i, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 1197, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_chain); __pyx_9genexpr18__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_i); __pyx_9genexpr18__pyx_v_i = 0;
    goto __pyx_L26_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_chain); __pyx_9genexpr18__pyx_v_chain = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_i); __pyx_9genexpr18__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L26_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1155
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 * def _regularize_embedding(g, emb):             # <<<<<<<<<<<<<<
 *     """Best-effort augmented embedding, with hopefully uniform chainlegnths.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("minorminer.busclique._regularize_embedding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_progress);
  __Pyx_XDECREF(__pyx_v_max_length);
  __Pyx_XDECREF(__pyx_v_remainder);
  __Pyx_XDECREF(__pyx_v_matching_graph);
  __Pyx_XDECREF(__pyx_v_top_nodes);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_chain);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_matching);
  __Pyx_XDECREF(__pyx_v_matched);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_chain);
  __Pyx_XDECREF(__pyx_gb_10minorminer_9busclique_21_regularize_embedding_2generator1);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_chain);
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1199
 *     return {i: tuple(chain) for i, chain in emb.items()}
 * 
 * def mine_clique_embeddings(             # <<<<<<<<<<<<<<
 *         g,
 *         quality_function=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_13mine_clique_embeddings(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10minorminer_9busclique_12mine_clique_embeddings[] = "Use heuristic and polynomial methods to construct a cache of clique embeddings.\n\n    This function uses both the heuristic :func:`minorminer.find_embedding` and\n    the polynomial algorithm in :func:`busclique.find_clique_embedding` to find\n    clique embeddings, and collects the best results of the two methods into a\n    single :class:`busgraph_cache` object.  One purpose of this function is to\n    identify optimized embeddings for hardware graphs, to be provided through\n    Ocean structured solvers and samplers.\n\n    If a disk-cache for ``g`` already exists, it will be augmented with the\n    embeddings discovered during execution.  The resulting cache will be writte\n    to disk afterwards, so repeated calls to this function can only improve the\n    quality of the cache.\n\n    The method operates by\n        (1) Load or construct a cache of embeddings for ``g`` with\n            :class:`busgraph_cache`.\n        (2) For 0 <= n < num_seeds, compute a :class:`busgraph_cache` with a\n            random seed, and insert the corresponding embeddings into the cache.\n        (3) For 3 <= n <= heuristic_bound, find clique embeddings of size n\n            into ``g`` with :func:`minorminer.find_embedding`, and insert them\n            into the cache if they improve the quality.\n\n    Args:\n        g: (networkx.Graph)\n            The graph to compute a clique cache for.\n\n        quality_function: (callable, optional, default=None)\n            A function that assesses the \"quality\" of an embedding and returns\n            an object that can be compared to other values returned by\n            ``quality_function``.  The embeddings are represented as lists of\n            lists of ints corresponding to the integer-labeled topology of ``g``.\n            If None, we use :func:`_default_quality_function`.\n\n        regularize: (bool, optional, default=True)\n            If True, we attempt to augment the found embeddings by adding extra\n            nodes i""n order to produce embeddings with all-equal chainlengths.\n\n        num_seeds: (int, optional, default=16)\n            The number of random seeds to compute :class:`busgraph_cache`.\n\n        mask_bound: (int, optional, default=64)\n            The number of topologies considered internal to :class:`busgraph_cache`.\n\n        heuristic_bound: (int, optional, default=None)\n            The maximum clique size to attempt heuristic embedding.  If None,\n            a default value is computed based on the topology parameters of ``g``.\n\n        heuristic_args: (dict, optional, default=None)\n            Keyword arguments to be used when calling the heuristic embedder.\n            See :func:`minorminer.find_embedding` for more details.\n\n        heurisitc_runs: (int, optional, default=10)\n            The number of times to attempt heuristic embedding per size.\n\n    Returns:\n        cache: (busgraph_cache)\n\n    ";
static PyMethodDef __pyx_mdef_10minorminer_9busclique_13mine_clique_embeddings = {"mine_clique_embeddings", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_13mine_clique_embeddings, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10minorminer_9busclique_12mine_clique_embeddings};
static PyObject *__pyx_pw_10minorminer_9busclique_13mine_clique_embeddings(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_g = 0;
  PyObject *__pyx_v_quality_function = 0;
  PyObject *__pyx_v_regularize = 0;
  PyObject *__pyx_v_num_seeds = 0;
  PyObject *__pyx_v_mask_bound = 0;
  PyObject *__pyx_v_heuristic_bound = 0;
  PyObject *__pyx_v_heuristic_args = 0;
  PyObject *__pyx_v_heuristic_runs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mine_clique_embeddings (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_g,&__pyx_n_s_quality_function,&__pyx_n_s_regularize,&__pyx_n_s_num_seeds,&__pyx_n_s_mask_bound,&__pyx_n_s_heuristic_bound,&__pyx_n_s_heuristic_args,&__pyx_n_s_heuristic_runs,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};

    /* "minorminer/busclique.pyx":1201
 * def mine_clique_embeddings(
 *         g,
 *         quality_function=None,             # <<<<<<<<<<<<<<
 *         regularize=True,
 *         num_seeds=16,
 */
    values[1] = ((PyObject *)Py_None);

    /* "minorminer/busclique.pyx":1202
 *         g,
 *         quality_function=None,
 *         regularize=True,             # <<<<<<<<<<<<<<
 *         num_seeds=16,
 *         mask_bound=64,
 */
    values[2] = ((PyObject *)Py_True);
    values[3] = ((PyObject *)__pyx_int_16);
    values[4] = ((PyObject *)__pyx_int_64);

    /* "minorminer/busclique.pyx":1205
 *         num_seeds=16,
 *         mask_bound=64,
 *         heuristic_bound=None,             # <<<<<<<<<<<<<<
 *         heuristic_args=None,
 *         heuristic_runs=10,
 */
    values[5] = ((PyObject *)Py_None);

    /* "minorminer/busclique.pyx":1206
 *         mask_bound=64,
 *         heuristic_bound=None,
 *         heuristic_args=None,             # <<<<<<<<<<<<<<
 *         heuristic_runs=10,
 *     ):
 */
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)__pyx_int_10);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quality_function);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_regularize);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_seeds);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask_bound);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_heuristic_bound);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_heuristic_args);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_heuristic_runs);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mine_clique_embeddings") < 0)) __PYX_ERR(0, 1199, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_g = values[0];
    __pyx_v_quality_function = values[1];
    __pyx_v_regularize = values[2];
    __pyx_v_num_seeds = values[3];
    __pyx_v_mask_bound = values[4];
    __pyx_v_heuristic_bound = values[5];
    __pyx_v_heuristic_args = values[6];
    __pyx_v_heuristic_runs = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mine_clique_embeddings", 0, 1, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1199, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.mine_clique_embeddings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_12mine_clique_embeddings(__pyx_self, __pyx_v_g, __pyx_v_quality_function, __pyx_v_regularize, __pyx_v_num_seeds, __pyx_v_mask_bound, __pyx_v_heuristic_bound, __pyx_v_heuristic_args, __pyx_v_heuristic_runs);

  /* "minorminer/busclique.pyx":1199
 *     return {i: tuple(chain) for i, chain in emb.items()}
 * 
 * def mine_clique_embeddings(             # <<<<<<<<<<<<<<
 *         g,
 *         quality_function=None,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_12mine_clique_embeddings(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_g, PyObject *__pyx_v_quality_function, PyObject *__pyx_v_regularize, PyObject *__pyx_v_num_seeds, PyObject *__pyx_v_mask_bound, PyObject *__pyx_v_heuristic_bound, PyObject *__pyx_v_heuristic_args, PyObject *__pyx_v_heuristic_runs) {
  PyObject *__pyx_v_miner = NULL;
  PyObject *__pyx_v_logging = NULL;
  PyObject *__pyx_v_logger = NULL;
  PyObject *__pyx_v_bgc = NULL;
  PyObject *__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_seed = NULL;
  PyObject *__pyx_v_bgc_i = NULL;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_emb = NULL;
  PyObject *__pyx_v_tile = NULL;
  PyObject *__pyx_v_rows = NULL;
  PyObject *__pyx_v_cols = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_hm = NULL;
  PyObject *__pyx_v_emb1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  long __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mine_clique_embeddings", 0);
  __Pyx_INCREF(__pyx_v_quality_function);
  __Pyx_INCREF(__pyx_v_heuristic_bound);
  __Pyx_INCREF(__pyx_v_heuristic_args);

  /* "minorminer/busclique.pyx":1269
 *     """
 * 
 *     from minorminer import miner             # <<<<<<<<<<<<<<
 *     import logging
 *     logger = logging.getLogger(__name__)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_miner);
  __Pyx_GIVEREF(__pyx_n_s_miner);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_miner);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_minorminer, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_miner); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_miner = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1270
 * 
 *     from minorminer import miner
 *     import logging             # <<<<<<<<<<<<<<
 *     logger = logging.getLogger(__name__)
 *     bgc = busgraph_cache(g)
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_logging, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_logging = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1271
 *     from minorminer import miner
 *     import logging
 *     logger = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 *     bgc = busgraph_cache(g)
 *     for i in range(num_seeds):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_logging, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_logger = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1272
 *     import logging
 *     logger = logging.getLogger(__name__)
 *     bgc = busgraph_cache(g)             # <<<<<<<<<<<<<<
 *     for i in range(num_seeds):
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_busgraph_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_g) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_g);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_bgc = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1273
 *     logger = logging.getLogger(__name__)
 *     bgc = busgraph_cache(g)
 *     for i in range(num_seeds):             # <<<<<<<<<<<<<<
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)
 *         seed = random.randint(0, 2**32-1)
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_num_seeds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1273, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1273, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1273, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1273, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1274
 *     bgc = busgraph_cache(g)
 *     for i in range(num_seeds):
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)             # <<<<<<<<<<<<<<
 *         seed = random.randint(0, 2**32-1)
 *         bgc_i = busgraph_cache(g, seed=i)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_logger, __pyx_n_s_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_kp_u_polynomial_embedder_run_d_of_d, __pyx_t_4, __pyx_v_num_seeds};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_kp_u_polynomial_embedder_run_d_of_d, __pyx_t_4, __pyx_v_num_seeds};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_polynomial_embedder_run_d_of_d);
      __Pyx_GIVEREF(__pyx_kp_u_polynomial_embedder_run_d_of_d);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_kp_u_polynomial_embedder_run_d_of_d);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_num_seeds);
      __Pyx_GIVEREF(__pyx_v_num_seeds);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_num_seeds);
      __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1275
 *     for i in range(num_seeds):
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)
 *         seed = random.randint(0, 2**32-1)             # <<<<<<<<<<<<<<
 *         bgc_i = busgraph_cache(g, seed=i)
 *         bgc_i._graph.set_mask_bound(mask_bound)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_random); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_randint); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_seed, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1276
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)
 *         seed = random.randint(0, 2**32-1)
 *         bgc_i = busgraph_cache(g, seed=i)             # <<<<<<<<<<<<<<
 *         bgc_i._graph.set_mask_bound(mask_bound)
 *         bgc.merge_clique_cache(bgc_i, write_to_disk=False, quality_function=quality_function)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_busgraph_cache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_g);
    __Pyx_GIVEREF(__pyx_v_g);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_g);
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_seed, __pyx_v_i) < 0) __PYX_ERR(0, 1276, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_bgc_i, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":1277
 *         seed = random.randint(0, 2**32-1)
 *         bgc_i = busgraph_cache(g, seed=i)
 *         bgc_i._graph.set_mask_bound(mask_bound)             # <<<<<<<<<<<<<<
 *         bgc.merge_clique_cache(bgc_i, write_to_disk=False, quality_function=quality_function)
 *         if regularize:
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc_i, __pyx_n_s_graph_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_set_mask_bound); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_v_mask_bound) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_mask_bound);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "minorminer/busclique.pyx":1278
 *         bgc_i = busgraph_cache(g, seed=i)
 *         bgc_i._graph.set_mask_bound(mask_bound)
 *         bgc.merge_clique_cache(bgc_i, write_to_disk=False, quality_function=quality_function)             # <<<<<<<<<<<<<<
 *         if regularize:
 *             for n in range(3, len(bgc_i.largest_clique())+1):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_merge_clique_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_bgc_i);
    __Pyx_GIVEREF(__pyx_v_bgc_i);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_bgc_i);
    __pyx_t_9 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_write_to_disk, Py_False) < 0) __PYX_ERR(0, 1278, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_quality_function, __pyx_v_quality_function) < 0) __PYX_ERR(0, 1278, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1279
 *         bgc_i._graph.set_mask_bound(mask_bound)
 *         bgc.merge_clique_cache(bgc_i, write_to_disk=False, quality_function=quality_function)
 *         if regularize:             # <<<<<<<<<<<<<<
 *             for n in range(3, len(bgc_i.largest_clique())+1):
 *                 emb = _regularize_embedding(g, bgc_i.find_clique_embedding(n))
 */
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_regularize); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1279, __pyx_L1_error)
    if (__pyx_t_10) {

      /* "minorminer/busclique.pyx":1280
 *         bgc.merge_clique_cache(bgc_i, write_to_disk=False, quality_function=quality_function)
 *         if regularize:
 *             for n in range(3, len(bgc_i.largest_clique())+1):             # <<<<<<<<<<<<<<
 *                 emb = _regularize_embedding(g, bgc_i.find_clique_embedding(n))
 *                 bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc_i, __pyx_n_s_largest_clique); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1280, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = (__pyx_t_11 + 1);
      __pyx_t_11 = __pyx_t_12;
      for (__pyx_t_13 = 3; __pyx_t_13 < __pyx_t_11; __pyx_t_13+=1) {
        __pyx_v_n = __pyx_t_13;

        /* "minorminer/busclique.pyx":1281
 *         if regularize:
 *             for n in range(3, len(bgc_i.largest_clique())+1):
 *                 emb = _regularize_embedding(g, bgc_i.find_clique_embedding(n))             # <<<<<<<<<<<<<<
 *                 bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_regularize_embedding); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc_i, __pyx_n_s_find_clique_embedding); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_14, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_g, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_g, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1281, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_g);
          __Pyx_GIVEREF(__pyx_v_g);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_8, __pyx_v_g);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_8, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF_SET(__pyx_v_emb, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "minorminer/busclique.pyx":1282
 *             for n in range(3, len(bgc_i.largest_clique())+1):
 *                 emb = _regularize_embedding(g, bgc_i.find_clique_embedding(n))
 *                 bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)             # <<<<<<<<<<<<<<
 * 
 *     if heuristic_bound is None:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_insert_clique_embedding); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_v_emb);
        __Pyx_GIVEREF(__pyx_v_emb);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_emb);
        __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_write_to_disk, Py_False) < 0) __PYX_ERR(0, 1282, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_quality_function, __pyx_v_quality_function) < 0) __PYX_ERR(0, 1282, __pyx_L1_error)
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "minorminer/busclique.pyx":1279
 *         bgc_i._graph.set_mask_bound(mask_bound)
 *         bgc.merge_clique_cache(bgc_i, write_to_disk=False, quality_function=quality_function)
 *         if regularize:             # <<<<<<<<<<<<<<
 *             for n in range(3, len(bgc_i.largest_clique())+1):
 *                 emb = _regularize_embedding(g, bgc_i.find_clique_embedding(n))
 */
    }

    /* "minorminer/busclique.pyx":1273
 *     logger = logging.getLogger(__name__)
 *     bgc = busgraph_cache(g)
 *     for i in range(num_seeds):             # <<<<<<<<<<<<<<
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)
 *         seed = random.randint(0, 2**32-1)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1284
 *                 bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 * 
 *     if heuristic_bound is None:             # <<<<<<<<<<<<<<
 *         if bgc._family == "zephyr":
 *             tile = g.graph['tile']
 */
  __pyx_t_10 = (__pyx_v_heuristic_bound == Py_None);
  __pyx_t_15 = (__pyx_t_10 != 0);
  if (__pyx_t_15) {

    /* "minorminer/busclique.pyx":1285
 * 
 *     if heuristic_bound is None:
 *         if bgc._family == "zephyr":             # <<<<<<<<<<<<<<
 *             tile = g.graph['tile']
 *             rows = g.graph['rows']
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_family_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_15 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_zephyr, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1285, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_15) {

      /* "minorminer/busclique.pyx":1286
 *     if heuristic_bound is None:
 *         if bgc._family == "zephyr":
 *             tile = g.graph['tile']             # <<<<<<<<<<<<<<
 *             rows = g.graph['rows']
 *             heuristic_bound = min(8*tile, 2*tile*(rows - 1))
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_tile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_tile = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "minorminer/busclique.pyx":1287
 *         if bgc._family == "zephyr":
 *             tile = g.graph['tile']
 *             rows = g.graph['rows']             # <<<<<<<<<<<<<<
 *             heuristic_bound = min(8*tile, 2*tile*(rows - 1))
 *         elif bgc._family == "pegasus":
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_rows); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_rows = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":1288
 *             tile = g.graph['tile']
 *             rows = g.graph['rows']
 *             heuristic_bound = min(8*tile, 2*tile*(rows - 1))             # <<<<<<<<<<<<<<
 *         elif bgc._family == "pegasus":
 *             rows = g.graph['rows']
 */
      __pyx_t_1 = PyNumber_Multiply(__pyx_int_2, __pyx_v_tile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_rows, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyNumber_Multiply(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Multiply(__pyx_int_8, __pyx_v_tile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_15) {
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
      } else {
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_1 = __pyx_t_3;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_heuristic_bound, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "minorminer/busclique.pyx":1285
 * 
 *     if heuristic_bound is None:
 *         if bgc._family == "zephyr":             # <<<<<<<<<<<<<<
 *             tile = g.graph['tile']
 *             rows = g.graph['rows']
 */
      goto __pyx_L9;
    }

    /* "minorminer/busclique.pyx":1289
 *             rows = g.graph['rows']
 *             heuristic_bound = min(8*tile, 2*tile*(rows - 1))
 *         elif bgc._family == "pegasus":             # <<<<<<<<<<<<<<
 *             rows = g.graph['rows']
 *             heuristic_bound = min(36, 12*(rows-1))
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_family_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_15 = (__Pyx_PyUnicode_Equals(__pyx_t_7, __pyx_n_u_pegasus, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_15) {

      /* "minorminer/busclique.pyx":1290
 *             heuristic_bound = min(8*tile, 2*tile*(rows - 1))
 *         elif bgc._family == "pegasus":
 *             rows = g.graph['rows']             # <<<<<<<<<<<<<<
 *             heuristic_bound = min(36, 12*(rows-1))
 *         elif bgc._family == "chimera":
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_u_rows); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_rows = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":1291
 *         elif bgc._family == "pegasus":
 *             rows = g.graph['rows']
 *             heuristic_bound = min(36, 12*(rows-1))             # <<<<<<<<<<<<<<
 *         elif bgc._family == "chimera":
 *             rows = g.graph['rows']
 */
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_rows, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyNumber_Multiply(__pyx_int_12, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_16 = 36;
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1291, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1291, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_15) {
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
      } else {
        __pyx_t_9 = __Pyx_PyInt_From_long(__pyx_t_16); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_1 = __pyx_t_9;
        __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_heuristic_bound, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "minorminer/busclique.pyx":1289
 *             rows = g.graph['rows']
 *             heuristic_bound = min(8*tile, 2*tile*(rows - 1))
 *         elif bgc._family == "pegasus":             # <<<<<<<<<<<<<<
 *             rows = g.graph['rows']
 *             heuristic_bound = min(36, 12*(rows-1))
 */
      goto __pyx_L9;
    }

    /* "minorminer/busclique.pyx":1292
 *             rows = g.graph['rows']
 *             heuristic_bound = min(36, 12*(rows-1))
 *         elif bgc._family == "chimera":             # <<<<<<<<<<<<<<
 *             rows = g.graph['rows']
 *             cols = g.graph['cols']
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_family_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_15 = (__Pyx_PyUnicode_Equals(__pyx_t_7, __pyx_n_u_chimera, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (likely(__pyx_t_15)) {

      /* "minorminer/busclique.pyx":1293
 *             heuristic_bound = min(36, 12*(rows-1))
 *         elif bgc._family == "chimera":
 *             rows = g.graph['rows']             # <<<<<<<<<<<<<<
 *             cols = g.graph['cols']
 *             tile = g.graph['tile']
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_u_rows); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_rows = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":1294
 *         elif bgc._family == "chimera":
 *             rows = g.graph['rows']
 *             cols = g.graph['cols']             # <<<<<<<<<<<<<<
 *             tile = g.graph['tile']
 *             heuristic_bound = (min(tile*rows, tile*cols)+3)/4
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_cols); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_cols = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "minorminer/busclique.pyx":1295
 *             rows = g.graph['rows']
 *             cols = g.graph['cols']
 *             tile = g.graph['tile']             # <<<<<<<<<<<<<<
 *             heuristic_bound = (min(tile*rows, tile*cols)+3)/4
 *         else:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_g, __pyx_n_s_graph); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_u_tile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_tile = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "minorminer/busclique.pyx":1296
 *             cols = g.graph['cols']
 *             tile = g.graph['tile']
 *             heuristic_bound = (min(tile*rows, tile*cols)+3)/4             # <<<<<<<<<<<<<<
 *         else:
 *             raise NotImplementedError("graph family not supported")
 */
      __pyx_t_1 = PyNumber_Multiply(__pyx_v_tile, __pyx_v_cols); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyNumber_Multiply(__pyx_v_tile, __pyx_v_rows); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_15) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_9 = __pyx_t_1;
      } else {
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_9 = __pyx_t_7;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_9, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_1, __pyx_int_4, 4, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_heuristic_bound, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "minorminer/busclique.pyx":1292
 *             rows = g.graph['rows']
 *             heuristic_bound = min(36, 12*(rows-1))
 *         elif bgc._family == "chimera":             # <<<<<<<<<<<<<<
 *             rows = g.graph['rows']
 *             cols = g.graph['cols']
 */
      goto __pyx_L9;
    }

    /* "minorminer/busclique.pyx":1298
 *             heuristic_bound = (min(tile*rows, tile*cols)+3)/4
 *         else:
 *             raise NotImplementedError("graph family not supported")             # <<<<<<<<<<<<<<
 * 
 *     if quality_function is None:
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 1298, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "minorminer/busclique.pyx":1284
 *                 bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 * 
 *     if heuristic_bound is None:             # <<<<<<<<<<<<<<
 *         if bgc._family == "zephyr":
 *             tile = g.graph['tile']
 */
  }

  /* "minorminer/busclique.pyx":1300
 *             raise NotImplementedError("graph family not supported")
 * 
 *     if quality_function is None:             # <<<<<<<<<<<<<<
 *         quality_function = _default_quality_function
 * 
 */
  __pyx_t_15 = (__pyx_v_quality_function == Py_None);
  __pyx_t_10 = (__pyx_t_15 != 0);
  if (__pyx_t_10) {

    /* "minorminer/busclique.pyx":1301
 * 
 *     if quality_function is None:
 *         quality_function = _default_quality_function             # <<<<<<<<<<<<<<
 * 
 *     if heuristic_args is None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_default_quality_function); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF_SET(__pyx_v_quality_function, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "minorminer/busclique.pyx":1300
 *             raise NotImplementedError("graph family not supported")
 * 
 *     if quality_function is None:             # <<<<<<<<<<<<<<
 *         quality_function = _default_quality_function
 * 
 */
  }

  /* "minorminer/busclique.pyx":1303
 *         quality_function = _default_quality_function
 * 
 *     if heuristic_args is None:             # <<<<<<<<<<<<<<
 *         heuristic_args = dict(chainlength_patience=100, max_beta=3, tries=1)
 * 
 */
  __pyx_t_10 = (__pyx_v_heuristic_args == Py_None);
  __pyx_t_15 = (__pyx_t_10 != 0);
  if (__pyx_t_15) {

    /* "minorminer/busclique.pyx":1304
 * 
 *     if heuristic_args is None:
 *         heuristic_args = dict(chainlength_patience=100, max_beta=3, tries=1)             # <<<<<<<<<<<<<<
 * 
 *     for size in range(3, heuristic_bound+1):
 */
    __pyx_t_9 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_chainlength_patience, __pyx_int_100) < 0) __PYX_ERR(0, 1304, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_max_beta, __pyx_int_3) < 0) __PYX_ERR(0, 1304, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_tries, __pyx_int_1) < 0) __PYX_ERR(0, 1304, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_heuristic_args, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "minorminer/busclique.pyx":1303
 *         quality_function = _default_quality_function
 * 
 *     if heuristic_args is None:             # <<<<<<<<<<<<<<
 *         heuristic_args = dict(chainlength_patience=100, max_beta=3, tries=1)
 * 
 */
  }

  /* "minorminer/busclique.pyx":1306
 *         heuristic_args = dict(chainlength_patience=100, max_beta=3, tries=1)
 * 
 *     for size in range(3, heuristic_bound+1):             # <<<<<<<<<<<<<<
 *         logger.info("running heuristic embeddings for size %d", size)
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 */
  __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_v_heuristic_bound, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_int_3);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_1, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_9)) || PyTuple_CheckExact(__pyx_t_9)) {
    __pyx_t_1 = __pyx_t_9; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1306, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1306, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1306, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1306, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "minorminer/busclique.pyx":1307
 * 
 *     for size in range(3, heuristic_bound+1):
 *         logger.info("running heuristic embeddings for size %d", size)             # <<<<<<<<<<<<<<
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_logger, __pyx_n_s_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_kp_u_running_heuristic_embeddings_for, __pyx_v_size};
      __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_9);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_kp_u_running_heuristic_embeddings_for, __pyx_v_size};
      __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_9);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_running_heuristic_embeddings_for);
      __Pyx_GIVEREF(__pyx_kp_u_running_heuristic_embeddings_for);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_kp_u_running_heuristic_embeddings_for);
      __Pyx_INCREF(__pyx_v_size);
      __Pyx_GIVEREF(__pyx_v_size);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_v_size);
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "minorminer/busclique.pyx":1308
 *     for size in range(3, heuristic_bound+1):
 *         logger.info("running heuristic embeddings for size %d", size)
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)             # <<<<<<<<<<<<<<
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):
 *             if not emb:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_nx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_complete_graph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_9 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_v_size) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_size);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_v_g);
    __Pyx_GIVEREF(__pyx_v_g);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_g);
    __pyx_t_9 = 0;
    if (unlikely(__pyx_v_heuristic_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
      __PYX_ERR(0, 1308, __pyx_L1_error)
    }
    if (likely(PyDict_CheckExact(__pyx_v_heuristic_args))) {
      __pyx_t_9 = PyDict_Copy(__pyx_v_heuristic_args); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    } else {
      __pyx_t_9 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_heuristic_args, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_v_miner, __pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_hm, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "minorminer/busclique.pyx":1309
 *         logger.info("running heuristic embeddings for size %d", size)
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):             # <<<<<<<<<<<<<<
 *             if not emb:
 *                 continue
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_hm, __pyx_n_s_find_embeddings); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_heuristic_runs);
    __Pyx_GIVEREF(__pyx_v_heuristic_runs);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_heuristic_runs);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_force, Py_False) < 0) __PYX_ERR(0, 1309, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
      __pyx_t_17 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_17 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1309, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_17)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_3); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1309, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1309, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_3); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1309, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1309, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_17(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1309, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_emb, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "minorminer/busclique.pyx":1310
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):
 *             if not emb:             # <<<<<<<<<<<<<<
 *                 continue
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 */
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_emb); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1310, __pyx_L1_error)
      __pyx_t_10 = ((!__pyx_t_15) != 0);
      if (__pyx_t_10) {

        /* "minorminer/busclique.pyx":1311
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):
 *             if not emb:
 *                 continue             # <<<<<<<<<<<<<<
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 *             if regularize:
 */
        goto __pyx_L14_continue;

        /* "minorminer/busclique.pyx":1310
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):
 *             if not emb:             # <<<<<<<<<<<<<<
 *                 continue
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 */
      }

      /* "minorminer/busclique.pyx":1312
 *             if not emb:
 *                 continue
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)             # <<<<<<<<<<<<<<
 *             if regularize:
 *                 emb1 = _regularize_embedding(g, emb)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_insert_clique_embedding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_emb);
      __Pyx_GIVEREF(__pyx_v_emb);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_emb);
      __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_write_to_disk, Py_False) < 0) __PYX_ERR(0, 1312, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_quality_function, __pyx_v_quality_function) < 0) __PYX_ERR(0, 1312, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "minorminer/busclique.pyx":1313
 *                 continue
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 *             if regularize:             # <<<<<<<<<<<<<<
 *                 emb1 = _regularize_embedding(g, emb)
 *                 bgc.insert_clique_embedding(emb1, write_to_disk=False, quality_function=quality_function)
 */
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_regularize); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1313, __pyx_L1_error)
      if (__pyx_t_10) {

        /* "minorminer/busclique.pyx":1314
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 *             if regularize:
 *                 emb1 = _regularize_embedding(g, emb)             # <<<<<<<<<<<<<<
 *                 bgc.insert_clique_embedding(emb1, write_to_disk=False, quality_function=quality_function)
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_regularize_embedding); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1314, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_9 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_g, __pyx_v_emb};
          __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1314, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_g, __pyx_v_emb};
          __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1314, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        {
          __pyx_t_3 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1314, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(__pyx_v_g);
          __Pyx_GIVEREF(__pyx_v_g);
          PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_8, __pyx_v_g);
          __Pyx_INCREF(__pyx_v_emb);
          __Pyx_GIVEREF(__pyx_v_emb);
          PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_8, __pyx_v_emb);
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1314, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_emb1, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "minorminer/busclique.pyx":1315
 *             if regularize:
 *                 emb1 = _regularize_embedding(g, emb)
 *                 bgc.insert_clique_embedding(emb1, write_to_disk=False, quality_function=quality_function)             # <<<<<<<<<<<<<<
 * 
 *     logger.info("writing embedding cache to disk")
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_insert_clique_embedding); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_emb1);
        __Pyx_GIVEREF(__pyx_v_emb1);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_emb1);
        __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_write_to_disk, Py_False) < 0) __PYX_ERR(0, 1315, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_quality_function, __pyx_v_quality_function) < 0) __PYX_ERR(0, 1315, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "minorminer/busclique.pyx":1313
 *                 continue
 *             bgc.insert_clique_embedding(emb, write_to_disk=False, quality_function=quality_function)
 *             if regularize:             # <<<<<<<<<<<<<<
 *                 emb1 = _regularize_embedding(g, emb)
 *                 bgc.insert_clique_embedding(emb1, write_to_disk=False, quality_function=quality_function)
 */
      }

      /* "minorminer/busclique.pyx":1309
 *         logger.info("running heuristic embeddings for size %d", size)
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 *         for emb in hm.find_embeddings(heuristic_runs, force=False):             # <<<<<<<<<<<<<<
 *             if not emb:
 *                 continue
 */
      __pyx_L14_continue:;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "minorminer/busclique.pyx":1306
 *         heuristic_args = dict(chainlength_patience=100, max_beta=3, tries=1)
 * 
 *     for size in range(3, heuristic_bound+1):             # <<<<<<<<<<<<<<
 *         logger.info("running heuristic embeddings for size %d", size)
 *         hm = miner(nx.complete_graph(size), g, **heuristic_args)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1317
 *                 bgc.insert_clique_embedding(emb1, write_to_disk=False, quality_function=quality_function)
 * 
 *     logger.info("writing embedding cache to disk")             # <<<<<<<<<<<<<<
 *     bgc.insert_clique_embedding({}, write_to_disk=True)
 *     return bgc
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_logger, __pyx_n_s_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_kp_u_writing_embedding_cache_to_disk) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_writing_embedding_cache_to_disk);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1318
 * 
 *     logger.info("writing embedding cache to disk")
 *     bgc.insert_clique_embedding({}, write_to_disk=True)             # <<<<<<<<<<<<<<
 *     return bgc
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgc, __pyx_n_s_insert_clique_embedding); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_write_to_disk, Py_True) < 0) __PYX_ERR(0, 1318, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":1319
 *     logger.info("writing embedding cache to disk")
 *     bgc.insert_clique_embedding({}, write_to_disk=True)
 *     return bgc             # <<<<<<<<<<<<<<
 * 
 * class copy_on_close_context:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_bgc);
  __pyx_r = __pyx_v_bgc;
  goto __pyx_L0;

  /* "minorminer/busclique.pyx":1199
 *     return {i: tuple(chain) for i, chain in emb.items()}
 * 
 * def mine_clique_embeddings(             # <<<<<<<<<<<<<<
 *         g,
 *         quality_function=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("minorminer.busclique.mine_clique_embeddings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_miner);
  __Pyx_XDECREF(__pyx_v_logging);
  __Pyx_XDECREF(__pyx_v_logger);
  __Pyx_XDECREF(__pyx_v_bgc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_seed);
  __Pyx_XDECREF(__pyx_v_bgc_i);
  __Pyx_XDECREF(__pyx_v_emb);
  __Pyx_XDECREF(__pyx_v_tile);
  __Pyx_XDECREF(__pyx_v_rows);
  __Pyx_XDECREF(__pyx_v_cols);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_hm);
  __Pyx_XDECREF(__pyx_v_emb1);
  __Pyx_XDECREF(__pyx_v_quality_function);
  __Pyx_XDECREF(__pyx_v_heuristic_bound);
  __Pyx_XDECREF(__pyx_v_heuristic_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1322
 * 
 * class copy_on_close_context:
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.files = {}
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_21copy_on_close_context_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_21copy_on_close_context_1__init__ = {"__init__", (PyCFunction)__pyx_pw_10minorminer_9busclique_21copy_on_close_context_1__init__, METH_O, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_21copy_on_close_context_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_21copy_on_close_context___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_21copy_on_close_context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "minorminer/busclique.pyx":1323
 * class copy_on_close_context:
 *     def __init__(self):
 *         self.files = {}             # <<<<<<<<<<<<<<
 * 
 *     def open(self, file_name, mode):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_files, __pyx_t_1) < 0) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1322
 * 
 * class copy_on_close_context:
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.files = {}
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("minorminer.busclique.copy_on_close_context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1325
 *         self.files = {}
 * 
 *     def open(self, file_name, mode):             # <<<<<<<<<<<<<<
 *         if file_name in self.files:
 *             self.files.clear()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_21copy_on_close_context_3open(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_21copy_on_close_context_3open = {"open", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_21copy_on_close_context_3open, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_21copy_on_close_context_3open(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_file_name = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_file_name,&__pyx_n_s_mode,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("open", 1, 3, 3, 1); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("open", 1, 3, 3, 2); __PYX_ERR(0, 1325, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "open") < 0)) __PYX_ERR(0, 1325, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_file_name = values[1];
    __pyx_v_mode = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("open", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1325, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("minorminer.busclique.copy_on_close_context.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10minorminer_9busclique_21copy_on_close_context_2open(__pyx_self, __pyx_v_self, __pyx_v_file_name, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_21copy_on_close_context_2open(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_file_name, PyObject *__pyx_v_mode) {
  PyObject *__pyx_v_temp_name = NULL;
  PyObject *__pyx_v_file_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open", 0);

  /* "minorminer/busclique.pyx":1326
 * 
 *     def open(self, file_name, mode):
 *         if file_name in self.files:             # <<<<<<<<<<<<<<
 *             self.files.clear()
 *             raise RuntimeError("copy_on_close_context can only open files once.  Discarding all files.")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_files); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_file_name, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "minorminer/busclique.pyx":1327
 *     def open(self, file_name, mode):
 *         if file_name in self.files:
 *             self.files.clear()             # <<<<<<<<<<<<<<
 *             raise RuntimeError("copy_on_close_context can only open files once.  Discarding all files.")
 *         else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_files); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_clear); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":1328
 *         if file_name in self.files:
 *             self.files.clear()
 *             raise RuntimeError("copy_on_close_context can only open files once.  Discarding all files.")             # <<<<<<<<<<<<<<
 *         else:
 *             temp_name = file_name + '~'
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1328, __pyx_L1_error)

    /* "minorminer/busclique.pyx":1326
 * 
 *     def open(self, file_name, mode):
 *         if file_name in self.files:             # <<<<<<<<<<<<<<
 *             self.files.clear()
 *             raise RuntimeError("copy_on_close_context can only open files once.  Discarding all files.")
 */
  }

  /* "minorminer/busclique.pyx":1330
 *             raise RuntimeError("copy_on_close_context can only open files once.  Discarding all files.")
 *         else:
 *             temp_name = file_name + '~'             # <<<<<<<<<<<<<<
 *             file_obj = open(temp_name, mode)
 *             self.files[file_name] = temp_name, file_obj
 */
  /*else*/ {
    __pyx_t_1 = PyNumber_Add(__pyx_v_file_name, __pyx_kp_u__21); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_temp_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "minorminer/busclique.pyx":1331
 *         else:
 *             temp_name = file_name + '~'
 *             file_obj = open(temp_name, mode)             # <<<<<<<<<<<<<<
 *             self.files[file_name] = temp_name, file_obj
 *             return file_obj
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_temp_name);
    __Pyx_GIVEREF(__pyx_v_temp_name);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_temp_name);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_mode);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_file_obj = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":1332
 *             temp_name = file_name + '~'
 *             file_obj = open(temp_name, mode)
 *             self.files[file_name] = temp_name, file_obj             # <<<<<<<<<<<<<<
 *             return file_obj
 * 
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_temp_name);
    __Pyx_GIVEREF(__pyx_v_temp_name);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_temp_name);
    __Pyx_INCREF(__pyx_v_file_obj);
    __Pyx_GIVEREF(__pyx_v_file_obj);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_file_obj);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_files); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_file_name, __pyx_t_5) < 0)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":1333
 *             file_obj = open(temp_name, mode)
 *             self.files[file_name] = temp_name, file_obj
 *             return file_obj             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_file_obj);
    __pyx_r = __pyx_v_file_obj;
    goto __pyx_L0;
  }

  /* "minorminer/busclique.pyx":1325
 *         self.files = {}
 * 
 *     def open(self, file_name, mode):             # <<<<<<<<<<<<<<
 *         if file_name in self.files:
 *             self.files.clear()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("minorminer.busclique.copy_on_close_context.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_temp_name);
  __Pyx_XDECREF(__pyx_v_file_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "minorminer/busclique.pyx":1335
 *             return file_obj
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         for file_name, (temp_name, file_obj) in self.files.items():
 *             file_obj.close()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10minorminer_9busclique_21copy_on_close_context_5close(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_10minorminer_9busclique_21copy_on_close_context_5close = {"close", (PyCFunction)__pyx_pw_10minorminer_9busclique_21copy_on_close_context_5close, METH_O, 0};
static PyObject *__pyx_pw_10minorminer_9busclique_21copy_on_close_context_5close(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_10minorminer_9busclique_21copy_on_close_context_4close(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10minorminer_9busclique_21copy_on_close_context_4close(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_file_name = NULL;
  PyObject *__pyx_v_temp_name = NULL;
  PyObject *__pyx_v_file_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "minorminer/busclique.pyx":1336
 * 
 *     def close(self):
 *         for file_name, (temp_name, file_obj) in self.files.items():             # <<<<<<<<<<<<<<
 *             file_obj.close()
 *             os.replace(temp_name, file_name)
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_files); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1336, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_6, &__pyx_t_5, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_file_name, __pyx_t_6);
    __pyx_t_6 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1336, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 1336, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1336, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_temp_name, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_file_obj, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "minorminer/busclique.pyx":1337
 *     def close(self):
 *         for file_name, (temp_name, file_obj) in self.files.items():
 *             file_obj.close()             # <<<<<<<<<<<<<<
 *             os.replace(temp_name, file_name)
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_obj, __pyx_n_s_close); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "minorminer/busclique.pyx":1338
 *         for file_name, (temp_name, file_obj) in self.files.items():
 *             file_obj.close()
 *             os.replace(temp_name, file_name)             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_os); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_replace); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_temp_name, __pyx_v_file_name};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_temp_name, __pyx_v_file_name};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_temp_name);
      __Pyx_GIVEREF(__pyx_v_temp_name);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_7, __pyx_v_temp_name);
      __Pyx_INCREF(__pyx_v_file_name);
      __Pyx_GIVEREF(__pyx_v_file_name);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_7, __pyx_v_file_name);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1335
 *             return file_obj
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         for file_name, (temp_name, file_obj) in self.files.items():
 *             file_obj.close()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("minorminer.busclique.copy_on_close_context.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_file_name);
  __Pyx_XDECREF(__pyx_v_temp_name);
  __Pyx_XDECREF(__pyx_v_file_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_size_t")
 * cdef vector[X] __pyx_convert_vector_from_py_size_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<size_t>  __pyx_convert_vector_from_py_size_t(PyObject *__pyx_v_o) {
  std::vector<size_t>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<size_t>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_size_t", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_size_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_v_item); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((size_t)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_size_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_size_t")
 * cdef vector[X] __pyx_convert_vector_from_py_size_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_size_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<std::vector<size_t> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(PyObject *__pyx_v_o) {
  std::vector<std::vector<size_t> >  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::vector<size_t> >  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::vector<size_t>  __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_vector_from_py_size_t(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::vector<size_t> )__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(PyObject *__pyx_v_o) {
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_10minorminer_9busclique_embedding_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((__pyx_t_10minorminer_9busclique_embedding_t)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t")
 * cdef vector[X] __pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py___pyx_t_10minorminer_9busclique_embedding_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_size_t")
 * cdef object __pyx_convert_vector_to_py_size_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_size_t(const std::vector<size_t>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_size_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_size_t")
 * cdef object __pyx_convert_vector_to_py_size_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_FromSize_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_size_t")
 * cdef object __pyx_convert_vector_to_py_size_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_size_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(const std::vector<std::vector<size_t> >  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __pyx_convert_vector_to_py_size_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t(const std::vector<__pyx_t_10minorminer_9busclique_embedding_t>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t")
 * cdef object __pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t")
 * cdef object __pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py___pyx_t_10minorminer_9busclique_embedding_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pair.to_py":158
 * 
 * @cname("__pyx_convert_pair_to_py_size_t____size_t")
 * cdef object __pyx_convert_pair_to_py_size_t____size_t(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

static PyObject *__pyx_convert_pair_to_py_size_t____size_t(std::pair<size_t,size_t>  const &__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_pair_to_py_size_t____size_t", 0);

  /* "pair.to_py":159
 * @cname("__pyx_convert_pair_to_py_size_t____size_t")
 * cdef object __pyx_convert_pair_to_py_size_t____size_t(const pair[X,Y]& p):
 *     return p.first, p.second             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_p.first); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_p.second); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pair.to_py":158
 * 
 * @cname("__pyx_convert_pair_to_py_size_t____size_t")
 * cdef object __pyx_convert_pair_to_py_size_t____size_t(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pair.to_py.__pyx_convert_pair_to_py_size_t____size_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t(std::pair<std::pair<size_t,size_t> ,__pyx_t_10minorminer_9busclique_embedding_t>  const &__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t", 0);

  /* "pair.to_py":159
 * @cname("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t")
 * cdef object __pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t(const pair[X,Y]& p):
 *     return p.first, p.second             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_pair_to_py_size_t____size_t(__pyx_v_p.first); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_size_t_3e___(__pyx_v_p.second); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pair.to_py":158
 * 
 * @cname("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t")
 * cdef object __pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pair.to_py.__pyx_convert_pair_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e_________pyx_t_10minorminer_9busclique_embedding_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pair.from_py":145
 * 
 * @cname("__pyx_convert_pair_from_py_size_t__and_size_t")
 * cdef pair[X,Y] __pyx_convert_pair_from_py_size_t__and_size_t(object o) except *:             # <<<<<<<<<<<<<<
 *     x, y = o
 *     return pair[X,Y](<X>x, <Y>y)
 */

static std::pair<size_t,size_t>  __pyx_convert_pair_from_py_size_t__and_size_t(PyObject *__pyx_v_o) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  std::pair<size_t,size_t>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_pair_from_py_size_t__and_size_t", 0);

  /* "pair.from_py":146
 * @cname("__pyx_convert_pair_from_py_size_t__and_size_t")
 * cdef pair[X,Y] __pyx_convert_pair_from_py_size_t__and_size_t(object o) except *:
 *     x, y = o             # <<<<<<<<<<<<<<
 *     return pair[X,Y](<X>x, <Y>y)
 * 
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_o))) || (PyList_CheckExact(__pyx_v_o))) {
    PyObject* sequence = __pyx_v_o;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 146, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(1, 146, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 146, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_y = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pair.from_py":147
 * cdef pair[X,Y] __pyx_convert_pair_from_py_size_t__and_size_t(object o) except *:
 *     x, y = o
 *     return pair[X,Y](<X>x, <Y>y)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_v_x); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 147, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_v_y); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 147, __pyx_L1_error)
  __pyx_r = std::pair<size_t,size_t> (((size_t)__pyx_t_5), ((size_t)__pyx_t_6));
  goto __pyx_L0;

  /* "pair.from_py":145
 * 
 * @cname("__pyx_convert_pair_from_py_size_t__and_size_t")
 * cdef pair[X,Y] __pyx_convert_pair_from_py_size_t__and_size_t(object o) except *:             # <<<<<<<<<<<<<<
 *     x, y = o
 *     return pair[X,Y](<X>x, <Y>y)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pair.from_py.__pyx_convert_pair_from_py_size_t__and_size_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<std::pair<size_t,size_t> >  __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(PyObject *__pyx_v_o) {
  std::vector<std::pair<size_t,size_t> >  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::pair<size_t,size_t> >  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::pair<size_t,size_t>  __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_pair_from_py_size_t__and_size_t(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::pair<size_t,size_t> )__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(const std::vector<std::pair<size_t,size_t> >  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __pyx_convert_pair_to_py_size_t____size_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_std_3a__3a_pair_3c_size_t_2c_size_t_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_uint8_t")
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<uint8_t>  __pyx_convert_vector_from_py_uint8_t(PyObject *__pyx_v_o) {
  std::vector<uint8_t>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<uint8_t>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_uint8_t", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_v_item); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((uint8_t)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_uint8_t")
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_uint8_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_10minorminer_9busclique__zephyr_busgraph(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)o);
  new((void*)&(p->nodes)) __pyx_t_10minorminer_9busclique_nodes_t();
  new((void*)&(p->emb_1)) __pyx_t_10minorminer_9busclique_embedding_t();
  p->relabel = Py_None; Py_INCREF(Py_None);
  p->delabel = Py_None; Py_INCREF(Py_None);
  p->identifier = Py_None; Py_INCREF(Py_None);
  p->short_identifier = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10minorminer_9busclique__zephyr_busgraph(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->nodes);
  __Pyx_call_destructor(p->emb_1);
  Py_CLEAR(p->relabel);
  Py_CLEAR(p->delabel);
  Py_CLEAR(p->identifier);
  Py_CLEAR(p->short_identifier);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10minorminer_9busclique__zephyr_busgraph(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)o;
  if (p->relabel) {
    e = (*v)(p->relabel, a); if (e) return e;
  }
  if (p->delabel) {
    e = (*v)(p->delabel, a); if (e) return e;
  }
  if (p->identifier) {
    e = (*v)(p->identifier, a); if (e) return e;
  }
  if (p->short_identifier) {
    e = (*v)(p->short_identifier, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique__zephyr_busgraph(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph *)o;
  tmp = ((PyObject*)p->relabel);
  p->relabel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->delabel);
  p->delabel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->identifier);
  p->identifier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->short_identifier);
  p->short_identifier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_relabel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7relabel_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_delabel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7delabel_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_identifier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_10identifier_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_short_identifier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_16short_identifier_1__get__(o);
}

static PyMethodDef __pyx_methods_10minorminer_9busclique__zephyr_busgraph[] = {
  {"set_mask_bound", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_5set_mask_bound, METH_O, __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_4set_mask_bound},
  {"bicliques", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_7bicliques, METH_NOARGS, __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_6bicliques},
  {"cliques", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_9cliques, METH_NOARGS, __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_8cliques},
  {"independent_set", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_11independent_set, METH_O, __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_10independent_set},
  {"find_clique_embedding", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_13find_clique_embedding, METH_O, __pyx_doc_10minorminer_9busclique_16_zephyr_busgraph_12find_clique_embedding},
  {"fragment_graph_spec", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_15fragment_graph_spec, METH_NOARGS, 0},
  {"fragment_nodes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_17fragment_nodes, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10minorminer_9busclique_16_zephyr_busgraph_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10minorminer_9busclique__zephyr_busgraph[] = {
  {(char *)"relabel", __pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_relabel, 0, (char *)0, 0},
  {(char *)"delabel", __pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_delabel, 0, (char *)0, 0},
  {(char *)"identifier", __pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_identifier, 0, (char *)0, 0},
  {(char *)"short_identifier", __pyx_getprop_10minorminer_9busclique_16_zephyr_busgraph_short_identifier, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10minorminer_9busclique__zephyr_busgraph = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique._zephyr_busgraph", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique__zephyr_busgraph), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique__zephyr_busgraph, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique__zephyr_busgraph, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique__zephyr_busgraph, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10minorminer_9busclique__zephyr_busgraph, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10minorminer_9busclique__zephyr_busgraph, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique__zephyr_busgraph, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_10minorminer_9busclique__pegasus_busgraph(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)o);
  new((void*)&(p->nodes)) __pyx_t_10minorminer_9busclique_nodes_t();
  new((void*)&(p->emb_1)) __pyx_t_10minorminer_9busclique_embedding_t();
  p->relabel = Py_None; Py_INCREF(Py_None);
  p->delabel = Py_None; Py_INCREF(Py_None);
  p->identifier = Py_None; Py_INCREF(Py_None);
  p->short_identifier = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10minorminer_9busclique__pegasus_busgraph(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->nodes);
  __Pyx_call_destructor(p->emb_1);
  Py_CLEAR(p->relabel);
  Py_CLEAR(p->delabel);
  Py_CLEAR(p->identifier);
  Py_CLEAR(p->short_identifier);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10minorminer_9busclique__pegasus_busgraph(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)o;
  if (p->relabel) {
    e = (*v)(p->relabel, a); if (e) return e;
  }
  if (p->delabel) {
    e = (*v)(p->delabel, a); if (e) return e;
  }
  if (p->identifier) {
    e = (*v)(p->identifier, a); if (e) return e;
  }
  if (p->short_identifier) {
    e = (*v)(p->short_identifier, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique__pegasus_busgraph(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph *)o;
  tmp = ((PyObject*)p->relabel);
  p->relabel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->delabel);
  p->delabel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->identifier);
  p->identifier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->short_identifier);
  p->short_identifier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_relabel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7relabel_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_delabel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7delabel_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_identifier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_10identifier_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_short_identifier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_16short_identifier_1__get__(o);
}

static PyMethodDef __pyx_methods_10minorminer_9busclique__pegasus_busgraph[] = {
  {"set_mask_bound", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_5set_mask_bound, METH_O, __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_4set_mask_bound},
  {"bicliques", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_7bicliques, METH_NOARGS, __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_6bicliques},
  {"cliques", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_9cliques, METH_NOARGS, __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_8cliques},
  {"independent_set", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_11independent_set, METH_O, __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_10independent_set},
  {"find_clique_embedding", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_13find_clique_embedding, METH_O, __pyx_doc_10minorminer_9busclique_17_pegasus_busgraph_12find_clique_embedding},
  {"fragment_graph_spec", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_15fragment_graph_spec, METH_NOARGS, 0},
  {"fragment_nodes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_17fragment_nodes, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_pegasus_busgraph_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10minorminer_9busclique__pegasus_busgraph[] = {
  {(char *)"relabel", __pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_relabel, 0, (char *)0, 0},
  {(char *)"delabel", __pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_delabel, 0, (char *)0, 0},
  {(char *)"identifier", __pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_identifier, 0, (char *)0, 0},
  {(char *)"short_identifier", __pyx_getprop_10minorminer_9busclique_17_pegasus_busgraph_short_identifier, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10minorminer_9busclique__pegasus_busgraph = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique._pegasus_busgraph", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique__pegasus_busgraph), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique__pegasus_busgraph, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique__pegasus_busgraph, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique__pegasus_busgraph, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10minorminer_9busclique__pegasus_busgraph, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10minorminer_9busclique__pegasus_busgraph, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique__pegasus_busgraph, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_10minorminer_9busclique__chimera_busgraph(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)o);
  new((void*)&(p->emb_1)) __pyx_t_10minorminer_9busclique_embedding_t();
  new((void*)&(p->nodes)) __pyx_t_10minorminer_9busclique_nodes_t();
  p->relabel = Py_None; Py_INCREF(Py_None);
  p->delabel = Py_None; Py_INCREF(Py_None);
  p->identifier = Py_None; Py_INCREF(Py_None);
  p->short_identifier = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10minorminer_9busclique__chimera_busgraph(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->emb_1);
  __Pyx_call_destructor(p->nodes);
  Py_CLEAR(p->relabel);
  Py_CLEAR(p->delabel);
  Py_CLEAR(p->identifier);
  Py_CLEAR(p->short_identifier);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10minorminer_9busclique__chimera_busgraph(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)o;
  if (p->relabel) {
    e = (*v)(p->relabel, a); if (e) return e;
  }
  if (p->delabel) {
    e = (*v)(p->delabel, a); if (e) return e;
  }
  if (p->identifier) {
    e = (*v)(p->identifier, a); if (e) return e;
  }
  if (p->short_identifier) {
    e = (*v)(p->short_identifier, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique__chimera_busgraph(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *p = (struct __pyx_obj_10minorminer_9busclique__chimera_busgraph *)o;
  tmp = ((PyObject*)p->relabel);
  p->relabel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->delabel);
  p->delabel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->identifier);
  p->identifier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->short_identifier);
  p->short_identifier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_relabel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7relabel_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_delabel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7delabel_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_identifier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_10identifier_1__get__(o);
}

static PyObject *__pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_short_identifier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10minorminer_9busclique_17_chimera_busgraph_16short_identifier_1__get__(o);
}

static PyMethodDef __pyx_methods_10minorminer_9busclique__chimera_busgraph[] = {
  {"set_mask_bound", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_5set_mask_bound, METH_O, __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_4set_mask_bound},
  {"bicliques", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_7bicliques, METH_NOARGS, __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_6bicliques},
  {"cliques", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_9cliques, METH_NOARGS, __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_8cliques},
  {"independent_set", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_11independent_set, METH_O, __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_10independent_set},
  {"find_clique_embedding", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_13find_clique_embedding, METH_O, __pyx_doc_10minorminer_9busclique_17_chimera_busgraph_12find_clique_embedding},
  {"fragment_graph_spec", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_15fragment_graph_spec, METH_NOARGS, 0},
  {"fragment_nodes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_17fragment_nodes, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10minorminer_9busclique_17_chimera_busgraph_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10minorminer_9busclique__chimera_busgraph[] = {
  {(char *)"relabel", __pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_relabel, 0, (char *)0, 0},
  {(char *)"delabel", __pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_delabel, 0, (char *)0, 0},
  {(char *)"identifier", __pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_identifier, 0, (char *)0, 0},
  {(char *)"short_identifier", __pyx_getprop_10minorminer_9busclique_17_chimera_busgraph_short_identifier, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10minorminer_9busclique__chimera_busgraph = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique._chimera_busgraph", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique__chimera_busgraph), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique__chimera_busgraph, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Class for managing a single Chimera graph, and dispatches various\n    structure-aware C++ embedding functions on it.\n\n    Note:\n        Due to internal optimizations, not all Chimera graphs are supported by\n        this code. Specifically, the graphs :func:`dwave_networkx.chimera_graph(m, n, t)`\n        are only supported for :math:`t<=8`. The code currently supports D-Wave\n        products, which have :math:`t=4`, but not all graphs. For graphs with\n        :math:`t>8`, use the legacy chimera-embedding package.\n    ", /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique__chimera_busgraph, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique__chimera_busgraph, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10minorminer_9busclique__chimera_busgraph, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10minorminer_9busclique__chimera_busgraph, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique__chimera_busgraph, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct__update_clique_cache[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct__update_clique_cache = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct__update_clique_cache(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct__update_clique_cache > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct__update_clique_cache[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct__update_clique_cache];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct__update_clique_cache(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct__update_clique_cache < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct__update_clique_cache[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct__update_clique_cache++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct__update_clique_cache(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct__update_clique_cache(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct__update_clique_cache", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct__update_clique_cache), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct__update_clique_cache, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct__update_clique_cache, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct__update_clique_cache, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct__update_clique_cache, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_1_insert_clique_embedding", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_2_merge_clique_cache", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_f);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *)o;
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler *)o;
  tmp = ((PyObject*)p->__pyx_v_f);
  p->__pyx_v_f = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_3__make_relabeler", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_emb);
  Py_CLEAR(p->__pyx_v_maxlen);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *)o;
  if (p->__pyx_v_emb) {
    e = (*v)(p->__pyx_v_emb, a); if (e) return e;
  }
  if (p->__pyx_v_maxlen) {
    e = (*v)(p->__pyx_v_maxlen, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function *)o;
  tmp = ((PyObject*)p->__pyx_v_emb);
  p->__pyx_v_emb = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_maxlen);
  p->__pyx_v_maxlen = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_4__default_quality_function", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_5_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_5_genexpr[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_chain);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_5_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_5_genexpr[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_chain) {
    e = (*v)(p->__pyx_v_chain, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_boundary);
  Py_CLEAR(p->__pyx_v_chain_node);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *)o;
  if (p->__pyx_v_boundary) {
    e = (*v)(p->__pyx_v_boundary, a); if (e) return e;
  }
  if (p->__pyx_v_chain_node) {
    e = (*v)(p->__pyx_v_chain_node, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding *)o;
  tmp = ((PyObject*)p->__pyx_v_boundary);
  p->__pyx_v_boundary = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_chain_node);
  p->__pyx_v_chain_node = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_6__regularize_embedding", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding, /*tp_traverse*/
  __pyx_tp_clear_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_7_genexpr[8];
static int __pyx_freecount_10minorminer_9busclique___pyx_scope_struct_7_genexpr = 0;

static PyObject *__pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_7_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_7_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr)))) {
    o = (PyObject*)__pyx_freelist_10minorminer_9busclique___pyx_scope_struct_7_genexpr[--__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_7_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_7_genexpr(PyObject *o) {
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_q);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_7_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr)))) {
    __pyx_freelist_10minorminer_9busclique___pyx_scope_struct_7_genexpr[__pyx_freecount_10minorminer_9busclique___pyx_scope_struct_7_genexpr++] = ((struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_7_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *p = (struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_q) {
    e = (*v)(p->__pyx_v_q, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "minorminer.busclique.__pyx_scope_struct_7_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10minorminer_9busclique___pyx_scope_struct_7_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10minorminer_9busclique___pyx_scope_struct_7_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10minorminer_9busclique___pyx_scope_struct_7_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10minorminer_9busclique___pyx_scope_struct_7_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_busclique(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_busclique},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "busclique",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_16x, __pyx_k_16x, sizeof(__pyx_k_16x), 0, 1, 0, 0},
  {&__pyx_kp_s_A_cache_class_for_Chimera_Pegasu, __pyx_k_A_cache_class_for_Chimera_Pegasu, sizeof(__pyx_k_A_cache_class_for_Chimera_Pegasu), 0, 0, 1, 0},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_FileNotFoundError, __pyx_k_FileNotFoundError, sizeof(__pyx_k_FileNotFoundError), 0, 0, 1, 1},
  {&__pyx_n_s_Graph, __pyx_k_Graph, sizeof(__pyx_k_Graph), 0, 0, 1, 1},
  {&__pyx_n_s_InterProcessLock, __pyx_k_InterProcessLock, sizeof(__pyx_k_InterProcessLock), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_LRU, __pyx_k_LRU, sizeof(__pyx_k_LRU), 0, 0, 1, 1},
  {&__pyx_n_s_Lock, __pyx_k_Lock, sizeof(__pyx_k_Lock), 0, 0, 1, 1},
  {&__pyx_n_s_M, __pyx_k_M, sizeof(__pyx_k_M), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_Path, __pyx_k_Path, sizeof(__pyx_k_Path), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 1, 0, 0},
  {&__pyx_n_s_add_edges_from, __pyx_k_add_edges_from, sizeof(__pyx_k_add_edges_from), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_basedir, __pyx_k_basedir, sizeof(__pyx_k_basedir), 0, 0, 1, 1},
  {&__pyx_n_s_bgc, __pyx_k_bgc, sizeof(__pyx_k_bgc), 0, 0, 1, 1},
  {&__pyx_n_s_bgc_i, __pyx_k_bgc_i, sizeof(__pyx_k_bgc_i), 0, 0, 1, 1},
  {&__pyx_n_u_biclique, __pyx_k_biclique, sizeof(__pyx_k_biclique), 0, 1, 0, 1},
  {&__pyx_n_s_bicliques, __pyx_k_bicliques, sizeof(__pyx_k_bicliques), 0, 0, 1, 1},
  {&__pyx_n_s_bicliques_2, __pyx_k_bicliques_2, sizeof(__pyx_k_bicliques_2), 0, 0, 1, 1},
  {&__pyx_n_s_biggest, __pyx_k_biggest, sizeof(__pyx_k_biggest), 0, 0, 1, 1},
  {&__pyx_n_s_bipartite, __pyx_k_bipartite, sizeof(__pyx_k_bipartite), 0, 0, 1, 1},
  {&__pyx_n_s_blank, __pyx_k_blank, sizeof(__pyx_k_blank), 0, 0, 1, 1},
  {&__pyx_n_s_boundary, __pyx_k_boundary, sizeof(__pyx_k_boundary), 0, 0, 1, 1},
  {&__pyx_n_u_busclique, __pyx_k_busclique, sizeof(__pyx_k_busclique), 0, 1, 0, 1},
  {&__pyx_n_s_busgraph, __pyx_k_busgraph, sizeof(__pyx_k_busgraph), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache, __pyx_k_busgraph_cache, sizeof(__pyx_k_busgraph_cache), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache___init, __pyx_k_busgraph_cache___init, sizeof(__pyx_k_busgraph_cache___init), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache__ensure_biclique, __pyx_k_busgraph_cache__ensure_biclique, sizeof(__pyx_k_busgraph_cache__ensure_biclique), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache__ensure_clique_ca, __pyx_k_busgraph_cache__ensure_clique_ca, sizeof(__pyx_k_busgraph_cache__ensure_clique_ca), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache__fetch_cache, __pyx_k_busgraph_cache__fetch_cache, sizeof(__pyx_k_busgraph_cache__fetch_cache), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_cache_rootdir, __pyx_k_busgraph_cache_cache_rootdir, sizeof(__pyx_k_busgraph_cache_cache_rootdir), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_clear_all_caches, __pyx_k_busgraph_cache_clear_all_caches, sizeof(__pyx_k_busgraph_cache_clear_all_caches), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_draw_fragment_emb, __pyx_k_busgraph_cache_draw_fragment_emb, sizeof(__pyx_k_busgraph_cache_draw_fragment_emb), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_find_biclique_emb, __pyx_k_busgraph_cache_find_biclique_emb, sizeof(__pyx_k_busgraph_cache_find_biclique_emb), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_find_clique_embed, __pyx_k_busgraph_cache_find_clique_embed, sizeof(__pyx_k_busgraph_cache_find_clique_embed), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_get_clique_cache, __pyx_k_busgraph_cache_get_clique_cache, sizeof(__pyx_k_busgraph_cache_get_clique_cache), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_insert_clique_emb, __pyx_k_busgraph_cache_insert_clique_emb, sizeof(__pyx_k_busgraph_cache_insert_clique_emb), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_insert_clique_emb_2, __pyx_k_busgraph_cache_insert_clique_emb_2, sizeof(__pyx_k_busgraph_cache_insert_clique_emb_2), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_largest_balanced, __pyx_k_busgraph_cache_largest_balanced, sizeof(__pyx_k_busgraph_cache_largest_balanced), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_largest_balanced_2, __pyx_k_busgraph_cache_largest_balanced_2, sizeof(__pyx_k_busgraph_cache_largest_balanced_2), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_largest_clique, __pyx_k_busgraph_cache_largest_clique, sizeof(__pyx_k_busgraph_cache_largest_clique), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_largest_clique_by, __pyx_k_busgraph_cache_largest_clique_by, sizeof(__pyx_k_busgraph_cache_largest_clique_by), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_merge_clique_cach, __pyx_k_busgraph_cache_merge_clique_cach, sizeof(__pyx_k_busgraph_cache_merge_clique_cach), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_merge_clique_cach_2, __pyx_k_busgraph_cache_merge_clique_cach_2, sizeof(__pyx_k_busgraph_cache_merge_clique_cach_2), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_update_clique_cac, __pyx_k_busgraph_cache_update_clique_cac, sizeof(__pyx_k_busgraph_cache_update_clique_cac), 0, 0, 1, 1},
  {&__pyx_n_s_busgraph_cache_update_clique_cac_2, __pyx_k_busgraph_cache_update_clique_cac_2, sizeof(__pyx_k_busgraph_cache_update_clique_cac_2), 0, 0, 1, 1},
  {&__pyx_n_s_by_length, __pyx_k_by_length, sizeof(__pyx_k_by_length), 0, 0, 1, 1},
  {&__pyx_n_s_by_size, __pyx_k_by_size, sizeof(__pyx_k_by_size), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
  {&__pyx_n_s_cache_rootdir, __pyx_k_cache_rootdir, sizeof(__pyx_k_cache_rootdir), 0, 0, 1, 1},
  {&__pyx_n_s_cachefile, __pyx_k_cachefile, sizeof(__pyx_k_cachefile), 0, 0, 1, 1},
  {&__pyx_n_s_chain, __pyx_k_chain, sizeof(__pyx_k_chain), 0, 0, 1, 1},
  {&__pyx_n_u_chain, __pyx_k_chain, sizeof(__pyx_k_chain), 0, 1, 0, 1},
  {&__pyx_n_s_chain_node, __pyx_k_chain_node, sizeof(__pyx_k_chain_node), 0, 0, 1, 1},
  {&__pyx_n_s_chainlength, __pyx_k_chainlength, sizeof(__pyx_k_chainlength), 0, 0, 1, 1},
  {&__pyx_n_s_chainlength_patience, __pyx_k_chainlength_patience, sizeof(__pyx_k_chainlength_patience), 0, 0, 1, 1},
  {&__pyx_n_u_chimera, __pyx_k_chimera, sizeof(__pyx_k_chimera), 0, 1, 0, 1},
  {&__pyx_n_s_chimera_busgraph, __pyx_k_chimera_busgraph, sizeof(__pyx_k_chimera_busgraph), 0, 0, 1, 1},
  {&__pyx_n_s_chimera_coordinates, __pyx_k_chimera_coordinates, sizeof(__pyx_k_chimera_coordinates), 0, 0, 1, 1},
  {&__pyx_n_s_chimera_graph, __pyx_k_chimera_graph, sizeof(__pyx_k_chimera_graph), 0, 0, 1, 1},
  {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
  {&__pyx_n_s_clear_all_caches, __pyx_k_clear_all_caches, sizeof(__pyx_k_clear_all_caches), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_u_clique, __pyx_k_clique, sizeof(__pyx_k_clique), 0, 1, 0, 1},
  {&__pyx_n_s_cliques, __pyx_k_cliques, sizeof(__pyx_k_cliques), 0, 0, 1, 1},
  {&__pyx_n_s_cliques_2, __pyx_k_cliques_2, sizeof(__pyx_k_cliques_2), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cols, __pyx_k_cols, sizeof(__pyx_k_cols), 0, 0, 1, 1},
  {&__pyx_n_u_cols, __pyx_k_cols, sizeof(__pyx_k_cols), 0, 1, 0, 1},
  {&__pyx_n_u_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 1, 0, 1},
  {&__pyx_n_s_complete_graph, __pyx_k_complete_graph, sizeof(__pyx_k_complete_graph), 0, 0, 1, 1},
  {&__pyx_n_s_compress, __pyx_k_compress, sizeof(__pyx_k_compress), 0, 0, 1, 1},
  {&__pyx_n_s_compute, __pyx_k_compute, sizeof(__pyx_k_compute), 0, 0, 1, 1},
  {&__pyx_n_s_compute_identifier, __pyx_k_compute_identifier, sizeof(__pyx_k_compute_identifier), 0, 0, 1, 1},
  {&__pyx_n_u_coordinate, __pyx_k_coordinate, sizeof(__pyx_k_coordinate), 0, 1, 0, 1},
  {&__pyx_n_s_copy_on_close_context, __pyx_k_copy_on_close_context, sizeof(__pyx_k_copy_on_close_context), 0, 0, 1, 1},
  {&__pyx_n_s_copy_on_close_context___init, __pyx_k_copy_on_close_context___init, sizeof(__pyx_k_copy_on_close_context___init), 0, 0, 1, 1},
  {&__pyx_kp_u_copy_on_close_context_can_only_o, __pyx_k_copy_on_close_context_can_only_o, sizeof(__pyx_k_copy_on_close_context_can_only_o), 0, 1, 0, 0},
  {&__pyx_n_s_copy_on_close_context_close, __pyx_k_copy_on_close_context_close, sizeof(__pyx_k_copy_on_close_context_close), 0, 0, 1, 1},
  {&__pyx_n_s_copy_on_close_context_open, __pyx_k_copy_on_close_context_open, sizeof(__pyx_k_copy_on_close_context_open), 0, 0, 1, 1},
  {&__pyx_kp_u_could_not_allocate_memory_for_to, __pyx_k_could_not_allocate_memory_for_to, sizeof(__pyx_k_could_not_allocate_memory_for_to), 0, 1, 0, 0},
  {&__pyx_n_s_currcache, __pyx_k_currcache, sizeof(__pyx_k_currcache), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_decompress, __pyx_k_decompress, sizeof(__pyx_k_decompress), 0, 0, 1, 1},
  {&__pyx_n_s_default_quality_function, __pyx_k_default_quality_function, sizeof(__pyx_k_default_quality_function), 0, 0, 1, 1},
  {&__pyx_n_s_default_quality_function_locals, __pyx_k_default_quality_function_locals, sizeof(__pyx_k_default_quality_function_locals), 0, 0, 1, 1},
  {&__pyx_n_s_delabel, __pyx_k_delabel, sizeof(__pyx_k_delabel), 0, 0, 1, 1},
  {&__pyx_n_s_direct, __pyx_k_direct, sizeof(__pyx_k_direct), 0, 0, 1, 1},
  {&__pyx_n_s_dirname, __pyx_k_dirname, sizeof(__pyx_k_dirname), 0, 0, 1, 1},
  {&__pyx_n_s_dirstack, __pyx_k_dirstack, sizeof(__pyx_k_dirstack), 0, 0, 1, 1},
  {&__pyx_n_s_discard, __pyx_k_discard, sizeof(__pyx_k_discard), 0, 0, 1, 1},
  {&__pyx_n_s_dnx, __pyx_k_dnx, sizeof(__pyx_k_dnx), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_draw_chimera_embedding, __pyx_k_draw_chimera_embedding, sizeof(__pyx_k_draw_chimera_embedding), 0, 0, 1, 1},
  {&__pyx_n_s_draw_fragment_embedding, __pyx_k_draw_fragment_embedding, sizeof(__pyx_k_draw_fragment_embedding), 0, 0, 1, 1},
  {&__pyx_n_s_dump, __pyx_k_dump, sizeof(__pyx_k_dump), 0, 0, 1, 1},
  {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
  {&__pyx_n_u_dwave, __pyx_k_dwave, sizeof(__pyx_k_dwave), 0, 1, 0, 1},
  {&__pyx_n_s_dwave_networkx, __pyx_k_dwave_networkx, sizeof(__pyx_k_dwave_networkx), 0, 0, 1, 1},
  {&__pyx_n_s_edge_list, __pyx_k_edge_list, sizeof(__pyx_k_edge_list), 0, 0, 1, 1},
  {&__pyx_n_s_edges, __pyx_k_edges, sizeof(__pyx_k_edges), 0, 0, 1, 1},
  {&__pyx_n_s_emb, __pyx_k_emb, sizeof(__pyx_k_emb), 0, 0, 1, 1},
  {&__pyx_n_s_emb0, __pyx_k_emb0, sizeof(__pyx_k_emb0), 0, 0, 1, 1},
  {&__pyx_n_s_emb1, __pyx_k_emb1, sizeof(__pyx_k_emb1), 0, 0, 1, 1},
  {&__pyx_n_s_emb_list, __pyx_k_emb_list, sizeof(__pyx_k_emb_list), 0, 0, 1, 1},
  {&__pyx_n_s_embs, __pyx_k_embs, sizeof(__pyx_k_embs), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_ensure_biclique_cache, __pyx_k_ensure_biclique_cache, sizeof(__pyx_k_ensure_biclique_cache), 0, 0, 1, 1},
  {&__pyx_n_s_ensure_clique_cache, __pyx_k_ensure_clique_cache, sizeof(__pyx_k_ensure_clique_cache), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_exist_ok, __pyx_k_exist_ok, sizeof(__pyx_k_exist_ok), 0, 0, 1, 1},
  {&__pyx_n_s_exists, __pyx_k_exists, sizeof(__pyx_k_exists), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_f_emb, __pyx_k_f_emb, sizeof(__pyx_k_f_emb), 0, 0, 1, 1},
  {&__pyx_n_s_family, __pyx_k_family, sizeof(__pyx_k_family), 0, 0, 1, 1},
  {&__pyx_n_u_family, __pyx_k_family, sizeof(__pyx_k_family), 0, 1, 0, 1},
  {&__pyx_n_s_family_2, __pyx_k_family_2, sizeof(__pyx_k_family_2), 0, 0, 1, 1},
  {&__pyx_n_s_fasteners, __pyx_k_fasteners, sizeof(__pyx_k_fasteners), 0, 0, 1, 1},
  {&__pyx_n_s_fetch_cache, __pyx_k_fetch_cache, sizeof(__pyx_k_fetch_cache), 0, 0, 1, 1},
  {&__pyx_n_s_file_context, __pyx_k_file_context, sizeof(__pyx_k_file_context), 0, 0, 1, 1},
  {&__pyx_n_s_file_name, __pyx_k_file_name, sizeof(__pyx_k_file_name), 0, 0, 1, 1},
  {&__pyx_n_s_file_obj, __pyx_k_file_obj, sizeof(__pyx_k_file_obj), 0, 0, 1, 1},
  {&__pyx_n_s_filecache, __pyx_k_filecache, sizeof(__pyx_k_filecache), 0, 0, 1, 1},
  {&__pyx_n_s_files, __pyx_k_files, sizeof(__pyx_k_files), 0, 0, 1, 1},
  {&__pyx_n_s_find_biclique_embedding, __pyx_k_find_biclique_embedding, sizeof(__pyx_k_find_biclique_embedding), 0, 0, 1, 1},
  {&__pyx_n_s_find_clique_embedding, __pyx_k_find_clique_embedding, sizeof(__pyx_k_find_clique_embedding), 0, 0, 1, 1},
  {&__pyx_n_s_find_embeddings, __pyx_k_find_embeddings, sizeof(__pyx_k_find_embeddings), 0, 0, 1, 1},
  {&__pyx_n_s_force, __pyx_k_force, sizeof(__pyx_k_force), 0, 0, 1, 1},
  {&__pyx_n_s_force_write, __pyx_k_force_write, sizeof(__pyx_k_force_write), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fragment_graph_spec, __pyx_k_fragment_graph_spec, sizeof(__pyx_k_fragment_graph_spec), 0, 0, 1, 1},
  {&__pyx_n_s_fragment_nodes, __pyx_k_fragment_nodes, sizeof(__pyx_k_fragment_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
  {&__pyx_n_s_get_clique_cache, __pyx_k_get_clique_cache, sizeof(__pyx_k_get_clique_cache), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_graph, __pyx_k_graph, sizeof(__pyx_k_graph), 0, 0, 1, 1},
  {&__pyx_n_s_graph_2, __pyx_k_graph_2, sizeof(__pyx_k_graph_2), 0, 0, 1, 1},
  {&__pyx_kp_u_graph_family_not_supported, __pyx_k_graph_family_not_supported, sizeof(__pyx_k_graph_family_not_supported), 0, 1, 0, 0},
  {&__pyx_n_s_graphclass, __pyx_k_graphclass, sizeof(__pyx_k_graphclass), 0, 0, 1, 1},
  {&__pyx_n_s_graphdata, __pyx_k_graphdata, sizeof(__pyx_k_graphdata), 0, 0, 1, 1},
  {&__pyx_n_s_gzip, __pyx_k_gzip, sizeof(__pyx_k_gzip), 0, 0, 1, 1},
  {&__pyx_n_s_heuristic_args, __pyx_k_heuristic_args, sizeof(__pyx_k_heuristic_args), 0, 0, 1, 1},
  {&__pyx_n_s_heuristic_bound, __pyx_k_heuristic_bound, sizeof(__pyx_k_heuristic_bound), 0, 0, 1, 1},
  {&__pyx_n_s_heuristic_runs, __pyx_k_heuristic_runs, sizeof(__pyx_k_heuristic_runs), 0, 0, 1, 1},
  {&__pyx_n_s_hm, __pyx_k_hm, sizeof(__pyx_k_hm), 0, 0, 1, 1},
  {&__pyx_n_s_homebase, __pyx_k_homebase, sizeof(__pyx_k_homebase), 0, 0, 1, 1},
  {&__pyx_n_u_horizontal_offsets, __pyx_k_horizontal_offsets, sizeof(__pyx_k_horizontal_offsets), 0, 1, 0, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_identifier, __pyx_k_identifier, sizeof(__pyx_k_identifier), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_independent_set, __pyx_k_independent_set, sizeof(__pyx_k_independent_set), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_kp_u_input_graph_must_either_be_a_dwa, __pyx_k_input_graph_must_either_be_a_dwa, sizeof(__pyx_k_input_graph_must_either_be_a_dwa), 0, 1, 0, 0},
  {&__pyx_n_s_insert_clique_embedding, __pyx_k_insert_clique_embedding, sizeof(__pyx_k_insert_clique_embedding), 0, 0, 1, 1},
  {&__pyx_n_u_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 1, 0, 1},
  {&__pyx_n_s_is_dir, __pyx_k_is_dir, sizeof(__pyx_k_is_dir), 0, 0, 1, 1},
  {&__pyx_n_s_item, __pyx_k_item, sizeof(__pyx_k_item), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_iter_chimera_to_linear, __pyx_k_iter_chimera_to_linear, sizeof(__pyx_k_iter_chimera_to_linear), 0, 0, 1, 1},
  {&__pyx_n_s_iter_chimera_to_linear_pairs, __pyx_k_iter_chimera_to_linear_pairs, sizeof(__pyx_k_iter_chimera_to_linear_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_iter_linear_to_chimera, __pyx_k_iter_linear_to_chimera, sizeof(__pyx_k_iter_linear_to_chimera), 0, 0, 1, 1},
  {&__pyx_n_s_iter_linear_to_nice, __pyx_k_iter_linear_to_nice, sizeof(__pyx_k_iter_linear_to_nice), 0, 0, 1, 1},
  {&__pyx_n_s_iter_linear_to_pegasus, __pyx_k_iter_linear_to_pegasus, sizeof(__pyx_k_iter_linear_to_pegasus), 0, 0, 1, 1},
  {&__pyx_n_s_iter_linear_to_zephyr, __pyx_k_iter_linear_to_zephyr, sizeof(__pyx_k_iter_linear_to_zephyr), 0, 0, 1, 1},
  {&__pyx_n_s_iter_nice_to_linear, __pyx_k_iter_nice_to_linear, sizeof(__pyx_k_iter_nice_to_linear), 0, 0, 1, 1},
  {&__pyx_n_s_iter_nice_to_linear_pairs, __pyx_k_iter_nice_to_linear_pairs, sizeof(__pyx_k_iter_nice_to_linear_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_iter_pegasus_to_linear, __pyx_k_iter_pegasus_to_linear, sizeof(__pyx_k_iter_pegasus_to_linear), 0, 0, 1, 1},
  {&__pyx_n_s_iter_pegasus_to_linear_pairs, __pyx_k_iter_pegasus_to_linear_pairs, sizeof(__pyx_k_iter_pegasus_to_linear_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_iter_zephyr_to_linear, __pyx_k_iter_zephyr_to_linear, sizeof(__pyx_k_iter_zephyr_to_linear), 0, 0, 1, 1},
  {&__pyx_n_s_iter_zephyr_to_linear_pairs, __pyx_k_iter_zephyr_to_linear_pairs, sizeof(__pyx_k_iter_zephyr_to_linear_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_iterdir, __pyx_k_iterdir, sizeof(__pyx_k_iterdir), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_json, __pyx_k_json, sizeof(__pyx_k_json), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key0, __pyx_k_key0, sizeof(__pyx_k_key0), 0, 0, 1, 1},
  {&__pyx_n_s_key1, __pyx_k_key1, sizeof(__pyx_k_key1), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_u_labels, __pyx_k_labels, sizeof(__pyx_k_labels), 0, 1, 0, 1},
  {&__pyx_n_s_largest_balanced_biclique, __pyx_k_largest_balanced_biclique, sizeof(__pyx_k_largest_balanced_biclique), 0, 0, 1, 1},
  {&__pyx_n_s_largest_clique, __pyx_k_largest_clique, sizeof(__pyx_k_largest_clique), 0, 0, 1, 1},
  {&__pyx_n_s_largest_clique_by_chainlength, __pyx_k_largest_clique_by_chainlength, sizeof(__pyx_k_largest_clique_by_chainlength), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_u_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 1, 0, 1},
  {&__pyx_n_s_length0, __pyx_k_length0, sizeof(__pyx_k_length0), 0, 0, 1, 1},
  {&__pyx_n_s_length1, __pyx_k_length1, sizeof(__pyx_k_length1), 0, 0, 1, 1},
  {&__pyx_n_s_load, __pyx_k_load, sizeof(__pyx_k_load), 0, 0, 1, 1},
  {&__pyx_n_s_loads, __pyx_k_loads, sizeof(__pyx_k_loads), 0, 0, 1, 1},
  {&__pyx_kp_u_lock, __pyx_k_lock, sizeof(__pyx_k_lock), 0, 1, 0, 0},
  {&__pyx_n_s_lockfile, __pyx_k_lockfile, sizeof(__pyx_k_lockfile), 0, 0, 1, 1},
  {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_kp_u_lru, __pyx_k_lru, sizeof(__pyx_k_lru), 0, 1, 0, 0},
  {&__pyx_n_s_lru_2, __pyx_k_lru_2, sizeof(__pyx_k_lru_2), 0, 0, 1, 1},
  {&__pyx_n_s_lrufile, __pyx_k_lrufile, sizeof(__pyx_k_lrufile), 0, 0, 1, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_relabeler, __pyx_k_make_relabeler, sizeof(__pyx_k_make_relabeler), 0, 0, 1, 1},
  {&__pyx_n_s_make_relabeler_locals__relabele, __pyx_k_make_relabeler_locals__relabele, sizeof(__pyx_k_make_relabeler_locals__relabele), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_mask_bound, __pyx_k_mask_bound, sizeof(__pyx_k_mask_bound), 0, 0, 1, 1},
  {&__pyx_n_s_matched, __pyx_k_matched, sizeof(__pyx_k_matched), 0, 0, 1, 1},
  {&__pyx_n_s_matching, __pyx_k_matching, sizeof(__pyx_k_matching), 0, 0, 1, 1},
  {&__pyx_n_s_matching_graph, __pyx_k_matching_graph, sizeof(__pyx_k_matching_graph), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max_beta, __pyx_k_max_beta, sizeof(__pyx_k_max_beta), 0, 0, 1, 1},
  {&__pyx_n_s_max_length, __pyx_k_max_length, sizeof(__pyx_k_max_length), 0, 0, 1, 1},
  {&__pyx_n_u_max_side, __pyx_k_max_side, sizeof(__pyx_k_max_side), 0, 1, 0, 1},
  {&__pyx_n_s_maximum_matching, __pyx_k_maximum_matching, sizeof(__pyx_k_maximum_matching), 0, 0, 1, 1},
  {&__pyx_n_s_maxlen, __pyx_k_maxlen, sizeof(__pyx_k_maxlen), 0, 0, 1, 1},
  {&__pyx_n_s_merge_clique_cache, __pyx_k_merge_clique_cache, sizeof(__pyx_k_merge_clique_cache), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_mine_clique_embeddings, __pyx_k_mine_clique_embeddings, sizeof(__pyx_k_mine_clique_embeddings), 0, 0, 1, 1},
  {&__pyx_n_s_miner, __pyx_k_miner, sizeof(__pyx_k_miner), 0, 0, 1, 1},
  {&__pyx_n_s_minorminer, __pyx_k_minorminer, sizeof(__pyx_k_minorminer), 0, 0, 1, 1},
  {&__pyx_n_s_minorminer_busclique, __pyx_k_minorminer_busclique, sizeof(__pyx_k_minorminer_busclique), 0, 0, 1, 1},
  {&__pyx_kp_s_minorminer_busclique_pyx, __pyx_k_minorminer_busclique_pyx, sizeof(__pyx_k_minorminer_busclique_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_mkdir, __pyx_k_mkdir, sizeof(__pyx_k_mkdir), 0, 0, 1, 1},
  {&__pyx_n_s_mm, __pyx_k_mm, sizeof(__pyx_k_mm), 0, 0, 1, 1},
  {&__pyx_n_s_mmax, __pyx_k_mmax, sizeof(__pyx_k_mmax), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_networkx, __pyx_k_networkx, sizeof(__pyx_k_networkx), 0, 0, 1, 1},
  {&__pyx_n_s_newLRU, __pyx_k_newLRU, sizeof(__pyx_k_newLRU), 0, 0, 1, 1},
  {&__pyx_n_u_nice, __pyx_k_nice, sizeof(__pyx_k_nice), 0, 1, 0, 1},
  {&__pyx_n_s_nmax, __pyx_k_nmax, sizeof(__pyx_k_nmax), 0, 0, 1, 1},
  {&__pyx_n_s_nn, __pyx_k_nn, sizeof(__pyx_k_nn), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_no_seed, __pyx_k_no_seed, sizeof(__pyx_k_no_seed), 0, 0, 1, 1},
  {&__pyx_n_s_node_boundary, __pyx_k_node_boundary, sizeof(__pyx_k_node_boundary), 0, 0, 1, 1},
  {&__pyx_n_s_node_list, __pyx_k_node_list, sizeof(__pyx_k_node_list), 0, 0, 1, 1},
  {&__pyx_n_s_nodes, __pyx_k_nodes, sizeof(__pyx_k_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_num, __pyx_k_num, sizeof(__pyx_k_num), 0, 0, 1, 1},
  {&__pyx_n_s_num_nodes, __pyx_k_num_nodes, sizeof(__pyx_k_num_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_num_seeds, __pyx_k_num_seeds, sizeof(__pyx_k_num_seeds), 0, 0, 1, 1},
  {&__pyx_n_s_nx, __pyx_k_nx, sizeof(__pyx_k_nx), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_oldkey, __pyx_k_oldkey, sizeof(__pyx_k_oldkey), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_parents, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_pathlib, __pyx_k_pathlib, sizeof(__pyx_k_pathlib), 0, 0, 1, 1},
  {&__pyx_n_u_pegasus, __pyx_k_pegasus, sizeof(__pyx_k_pegasus), 0, 1, 0, 1},
  {&__pyx_n_s_pegasus_busgraph, __pyx_k_pegasus_busgraph, sizeof(__pyx_k_pegasus_busgraph), 0, 0, 1, 1},
  {&__pyx_n_s_pegasus_coordinates, __pyx_k_pegasus_coordinates, sizeof(__pyx_k_pegasus_coordinates), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_kp_u_polynomial_embedder_run_d_of_d, __pyx_k_polynomial_embedder_run_d_of_d, sizeof(__pyx_k_polynomial_embedder_run_d_of_d), 0, 1, 0, 0},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_progress, __pyx_k_progress, sizeof(__pyx_k_progress), 0, 0, 1, 1},
  {&__pyx_n_s_qual0, __pyx_k_qual0, sizeof(__pyx_k_qual0), 0, 0, 1, 1},
  {&__pyx_n_s_qual1, __pyx_k_qual1, sizeof(__pyx_k_qual1), 0, 0, 1, 1},
  {&__pyx_n_s_quality_function, __pyx_k_quality_function, sizeof(__pyx_k_quality_function), 0, 0, 1, 1},
  {&__pyx_kp_u_quality_function_is_ignored_for, __pyx_k_quality_function_is_ignored_for, sizeof(__pyx_k_quality_function_is_ignored_for), 0, 1, 0, 0},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_randint, __pyx_k_randint, sizeof(__pyx_k_randint), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_raw, __pyx_k_raw, sizeof(__pyx_k_raw), 0, 0, 1, 1},
  {&__pyx_n_u_raw, __pyx_k_raw, sizeof(__pyx_k_raw), 0, 1, 0, 1},
  {&__pyx_n_s_raw0, __pyx_k_raw0, sizeof(__pyx_k_raw0), 0, 0, 1, 1},
  {&__pyx_n_s_raw1, __pyx_k_raw1, sizeof(__pyx_k_raw1), 0, 0, 1, 1},
  {&__pyx_n_s_raw_emb, __pyx_k_raw_emb, sizeof(__pyx_k_raw_emb), 0, 0, 1, 1},
  {&__pyx_n_s_raw_emb0, __pyx_k_raw_emb0, sizeof(__pyx_k_raw_emb0), 0, 0, 1, 1},
  {&__pyx_n_s_raw_emb1, __pyx_k_raw_emb1, sizeof(__pyx_k_raw_emb1), 0, 0, 1, 1},
  {&__pyx_n_u_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 1, 0, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_regularize, __pyx_k_regularize, sizeof(__pyx_k_regularize), 0, 0, 1, 1},
  {&__pyx_n_s_regularize_embedding, __pyx_k_regularize_embedding, sizeof(__pyx_k_regularize_embedding), 0, 0, 1, 1},
  {&__pyx_n_s_regularize_embedding_locals_gen, __pyx_k_regularize_embedding_locals_gen, sizeof(__pyx_k_regularize_embedding_locals_gen), 0, 0, 1, 1},
  {&__pyx_n_s_relabel, __pyx_k_relabel, sizeof(__pyx_k_relabel), 0, 0, 1, 1},
  {&__pyx_n_s_relabeler, __pyx_k_relabeler, sizeof(__pyx_k_relabeler), 0, 0, 1, 1},
  {&__pyx_n_s_remainder, __pyx_k_remainder, sizeof(__pyx_k_remainder), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_s_rmdir, __pyx_k_rmdir, sizeof(__pyx_k_rmdir), 0, 0, 1, 1},
  {&__pyx_n_s_rootdir, __pyx_k_rootdir, sizeof(__pyx_k_rootdir), 0, 0, 1, 1},
  {&__pyx_n_s_row, __pyx_k_row, sizeof(__pyx_k_row), 0, 0, 1, 1},
  {&__pyx_n_s_rows, __pyx_k_rows, sizeof(__pyx_k_rows), 0, 0, 1, 1},
  {&__pyx_n_u_rows, __pyx_k_rows, sizeof(__pyx_k_rows), 0, 1, 0, 1},
  {&__pyx_kp_u_running_heuristic_embeddings_for, __pyx_k_running_heuristic_embeddings_for, sizeof(__pyx_k_running_heuristic_embeddings_for), 0, 1, 0, 0},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_s0, __pyx_k_s0, sizeof(__pyx_k_s0), 0, 0, 1, 1},
  {&__pyx_n_s_s1, __pyx_k_s1, sizeof(__pyx_k_s1), 0, 0, 1, 1},
  {&__pyx_n_s_s_min, __pyx_k_s_min, sizeof(__pyx_k_s_min), 0, 0, 1, 1},
  {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set_mask_bound, __pyx_k_set_mask_bound, sizeof(__pyx_k_set_mask_bound), 0, 0, 1, 1},
  {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_short_identifier, __pyx_k_short_identifier, sizeof(__pyx_k_short_identifier), 0, 0, 1, 1},
  {&__pyx_n_s_shortcode, __pyx_k_shortcode, sizeof(__pyx_k_shortcode), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_u_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 1, 0, 1},
  {&__pyx_n_s_size0, __pyx_k_size0, sizeof(__pyx_k_size0), 0, 0, 1, 1},
  {&__pyx_n_s_size1, __pyx_k_size1, sizeof(__pyx_k_size1), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_sparse, __pyx_k_sparse, sizeof(__pyx_k_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_substack, __pyx_k_substack, sizeof(__pyx_k_substack), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_temp_name, __pyx_k_temp_name, sizeof(__pyx_k_temp_name), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_u_this_clique_embedder_supports_ch, __pyx_k_this_clique_embedder_supports_ch, sizeof(__pyx_k_this_clique_embedder_supports_ch), 0, 1, 0, 0},
  {&__pyx_kp_u_this_clique_embedder_supports_ze, __pyx_k_this_clique_embedder_supports_ze, sizeof(__pyx_k_this_clique_embedder_supports_ze), 0, 1, 0, 0},
  {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tile, __pyx_k_tile, sizeof(__pyx_k_tile), 0, 0, 1, 1},
  {&__pyx_n_u_tile, __pyx_k_tile, sizeof(__pyx_k_tile), 0, 1, 0, 1},
  {&__pyx_n_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 1},
  {&__pyx_n_s_top_nodes, __pyx_k_top_nodes, sizeof(__pyx_k_top_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_tries, __pyx_k_tries, sizeof(__pyx_k_tries), 0, 0, 1, 1},
  {&__pyx_n_s_trivial_relabeler, __pyx_k_trivial_relabeler, sizeof(__pyx_k_trivial_relabeler), 0, 0, 1, 1},
  {&__pyx_n_s_union, __pyx_k_union, sizeof(__pyx_k_union), 0, 0, 1, 1},
  {&__pyx_n_s_unlink, __pyx_k_unlink, sizeof(__pyx_k_unlink), 0, 0, 1, 1},
  {&__pyx_kp_u_unrecognized_graph_labeling, __pyx_k_unrecognized_graph_labeling, sizeof(__pyx_k_unrecognized_graph_labeling), 0, 1, 0, 0},
  {&__pyx_n_s_update_clique_cache, __pyx_k_update_clique_cache, sizeof(__pyx_k_update_clique_cache), 0, 0, 1, 1},
  {&__pyx_n_s_urandom, __pyx_k_urandom, sizeof(__pyx_k_urandom), 0, 0, 1, 1},
  {&__pyx_n_s_use_cache, __pyx_k_use_cache, sizeof(__pyx_k_use_cache), 0, 0, 1, 1},
  {&__pyx_n_s_used, __pyx_k_used, sizeof(__pyx_k_used), 0, 0, 1, 1},
  {&__pyx_n_s_user_data_dir, __pyx_k_user_data_dir, sizeof(__pyx_k_user_data_dir), 0, 0, 1, 1},
  {&__pyx_n_u_utf8, __pyx_k_utf8, sizeof(__pyx_k_utf8), 0, 1, 0, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {&__pyx_n_u_vertical_offsets, __pyx_k_vertical_offsets, sizeof(__pyx_k_vertical_offsets), 0, 1, 0, 1},
  {&__pyx_n_u_wb, __pyx_k_wb, sizeof(__pyx_k_wb), 0, 1, 0, 1},
  {&__pyx_n_s_write_to_disk, __pyx_k_write_to_disk, sizeof(__pyx_k_write_to_disk), 0, 0, 1, 1},
  {&__pyx_kp_u_writing_embedding_cache_to_disk, __pyx_k_writing_embedding_cache_to_disk, sizeof(__pyx_k_writing_embedding_cache_to_disk), 0, 1, 0, 0},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_u_zephyr, __pyx_k_zephyr, sizeof(__pyx_k_zephyr), 0, 1, 0, 1},
  {&__pyx_n_s_zephyr_busgraph, __pyx_k_zephyr_busgraph, sizeof(__pyx_k_zephyr_busgraph), 0, 0, 1, 1},
  {&__pyx_n_s_zephyr_coordinates, __pyx_k_zephyr_coordinates, sizeof(__pyx_k_zephyr_coordinates), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {&__pyx_n_s_zip_longest, __pyx_k_zip_longest, sizeof(__pyx_k_zip_longest), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 185, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 274, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 320, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 336, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 465, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 569, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 590, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 777, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 810, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 1153, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 1328, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "minorminer/busclique.pyx":120
 *                     'chimera': _chimera_busgraph}[family]
 *     except (AttributeError, KeyError):
 *         raise ValueError(("input graph must either be a "             # <<<<<<<<<<<<<<
 *                           "dwave_networkx.pegasus_graph, "
 *                           "dwave_networkx.chimera_graph or "
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_input_graph_must_either_be_a_dwa); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "minorminer/busclique.pyx":323
 *         if direct:
 *             if quality_function is not None:
 *                 raise ValueError("quality_function is ignored for direct insertion")             # <<<<<<<<<<<<<<
 *             self._ensure_clique_cache(blank=True)
 *             raw = self._cliques['raw']
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_quality_function_is_ignored_for); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "minorminer/busclique.pyx":465
 *                 cachefile = os.path.join(basedir, str(shortcode))
 *                 if os.path.exists(cachefile):
 *                     with open(cachefile, 'rb') as filecache:             # <<<<<<<<<<<<<<
 *                         currcache = load(filecache)
 *                 else:
 */
  __pyx_tuple__4 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "minorminer/busclique.pyx":757
 *     chain in the embedding.
 *     """
 *     def _relabeler(emb):             # <<<<<<<<<<<<<<
 *         return {v: tuple(f(chain)) for v, chain in emb.items()}
 *     return _relabeler
 */
  __pyx_tuple__5 = PyTuple_Pack(3, __pyx_n_s_emb, __pyx_n_s_v, __pyx_n_s_chain); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_codeobj__6 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__5, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_relabeler, 757, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__6)) __PYX_ERR(0, 757, __pyx_L1_error)

  /* "minorminer/busclique.pyx":777
 * 
 *     if tmp == NULL:
 *         raise MemoryError("could not allocate memory for topology serialization")             # <<<<<<<<<<<<<<
 * 
 *     topo.serialize(serialize_write_tag(), tmp);
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_could_not_allocate_memory_for_to); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "minorminer/busclique.pyx":810
 *         cdef size_t tile = g.graph['tile']
 *         if tile > 4:
 *             raise NotImplementedError(("this clique embedder supports zephyr "             # <<<<<<<<<<<<<<
 *                                        "graphs with a tile size of 4 or less"))
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_this_clique_embedder_supports_ze); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "minorminer/busclique.pyx":834
 *             self.delabel = _make_relabeler(coordinates.iter_zephyr_to_linear)
 *         else:
 *             raise ValueError("unrecognized graph labeling")             # <<<<<<<<<<<<<<
 * 
 *         self.topo = new topo_cache[zephyr_spec](zep[0], self.nodes, edges)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_unrecognized_graph_labeling); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "minorminer/busclique.pyx":918
 *         """
 *         rows = g.graph['rows']
 *         voff = [o//2 for o in g.graph['vertical_offsets'][::2]]             # <<<<<<<<<<<<<<
 *         hoff = [o//2 for o in g.graph['horizontal_offsets'][::2]]
 * 
 */
  __pyx_slice__12 = PySlice_New(Py_None, Py_None, __pyx_int_2); if (unlikely(!__pyx_slice__12)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__12);
  __Pyx_GIVEREF(__pyx_slice__12);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "minorminer/busclique.pyx":1041
 * 
 *         if tile > 8:
 *             raise NotImplementedError(("this clique embedder supports chimera "             # <<<<<<<<<<<<<<
 *                                        "graphs with a tile size of 8 or less"))
 *         cdef uint32_t internal_seed
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_u_this_clique_embedder_supports_ch); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "minorminer/busclique.pyx":1275
 *     for i in range(num_seeds):
 *         logger.info("polynomial embedder run %d of %d", i+1, num_seeds)
 *         seed = random.randint(0, 2**32-1)             # <<<<<<<<<<<<<<
 *         bgc_i = busgraph_cache(g, seed=i)
 *         bgc_i._graph.set_mask_bound(mask_bound)
 */
  __pyx_tuple__18 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_4294967295); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "minorminer/busclique.pyx":1298
 *             heuristic_bound = (min(tile*rows, tile*cols)+3)/4
 *         else:
 *             raise NotImplementedError("graph family not supported")             # <<<<<<<<<<<<<<
 * 
 *     if quality_function is None:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_graph_family_not_supported); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 1298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "minorminer/busclique.pyx":1328
 *         if file_name in self.files:
 *             self.files.clear()
 *             raise RuntimeError("copy_on_close_context can only open files once.  Discarding all files.")             # <<<<<<<<<<<<<<
 *         else:
 *             temp_name = file_name + '~'
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_copy_on_close_context_can_only_o); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "minorminer/busclique.pyx":46
 *                             'biclique': threading.Lock()}
 * 
 * def _num_nodes(nn, offset = 0):             # <<<<<<<<<<<<<<
 *     """Internal-use function to normalize inputs.
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(4, __pyx_n_s_nn, __pyx_n_s_offset, __pyx_n_s_nodes, __pyx_n_s_num); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_num_nodes, 46, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 46, __pyx_L1_error)

  /* "minorminer/busclique.pyx":67
 * 
 * _no_seed = object()
 * def find_clique_embedding(nodes, g, seed = _no_seed, use_cache = True):             # <<<<<<<<<<<<<<
 *     """Finds a clique embedding in the graph ``g`` using a polynomial-time
 *     algorithm.
 */
  __pyx_tuple__24 = PyTuple_Pack(7, __pyx_n_s_nodes, __pyx_n_s_g, __pyx_n_s_seed, __pyx_n_s_use_cache, __pyx_n_s_graphdata, __pyx_n_s_family, __pyx_n_s_busgraph); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_find_clique_embedding, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 67, __pyx_L1_error)

  /* "minorminer/busclique.pyx":155
 * 
 *     """
 *     def __init__(self, g, seed = 0):             # <<<<<<<<<<<<<<
 *         self._family = g.graph['family']
 *         graphclass = {'pegasus': _pegasus_busgraph,
 */
  __pyx_tuple__26 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_g, __pyx_n_s_seed, __pyx_n_s_graphclass); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_init, 155, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 155, __pyx_L1_error)
  __pyx_tuple__28 = PyTuple_Pack(1, ((PyObject *)__pyx_int_0)); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "minorminer/busclique.pyx":169
 *         self._bicliques = None
 * 
 *     def _ensure_clique_cache(self, blank=False):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:
 */
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_blank); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_ensure_clique_cache, 169, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 169, __pyx_L1_error)
  __pyx_tuple__31 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "minorminer/busclique.pyx":180
 *                 )
 * 
 *     def _ensure_biclique_cache(self):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_ensure_biclique_cache, 180, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 180, __pyx_L1_error)

  /* "minorminer/busclique.pyx":186
 * 
 *     @staticmethod
 *     def cache_rootdir(version=__cache_version):             # <<<<<<<<<<<<<<
 *         """Returns the directory corresponding to the provided cache version.
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_n_s_version); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_cache_rootdir, 186, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 186, __pyx_L1_error)

  /* "minorminer/busclique.pyx":199
 * 
 *     @staticmethod
 *     def clear_all_caches():             # <<<<<<<<<<<<<<
 *         """Removes all caches created by this class, up to and including the
 *         current version.
 */
  __pyx_tuple__36 = PyTuple_Pack(6, __pyx_n_s_dirstack, __pyx_n_s_i, __pyx_n_s_rootdir, __pyx_n_s_top, __pyx_n_s_substack, __pyx_n_s_item); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_clear_all_caches, 199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 199, __pyx_L1_error)

  /* "minorminer/busclique.pyx":225
 *                 dirstack.pop()
 * 
 *     def get_clique_cache(self, sparse=False, compress=True):             # <<<<<<<<<<<<<<
 *         """Returns an object representing the cliques stored in this cache.
 * 
 */
  __pyx_tuple__38 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_sparse, __pyx_n_s_compress, __pyx_n_s_raw, __pyx_n_s_size, __pyx_n_s_cache, __pyx_n_s_i, __pyx_n_s_s, __pyx_n_s_i); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(3, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_get_clique_cache, 225, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 225, __pyx_L1_error)
  __pyx_tuple__40 = PyTuple_Pack(2, ((PyObject *)Py_False), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "minorminer/busclique.pyx":254
 *         return cache
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):             # <<<<<<<<<<<<<<
 *         """Replaces or sets the cliques stored by this cache with the provided
 *         cache object.
 */
  __pyx_tuple__41 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_cache, __pyx_n_s_write_to_disk, __pyx_n_s_empty, __pyx_n_s_size, __pyx_n_s_emb); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_update_clique_cache, 254, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 254, __pyx_L1_error)
  __pyx_tuple__43 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "minorminer/busclique.pyx":287
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):             # <<<<<<<<<<<<<<
 *         """Inserts a clique embedding into the cache.
 * 
 */
  __pyx_tuple__44 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_emb, __pyx_n_s_write_to_disk, __pyx_n_s_quality_function, __pyx_n_s_direct, __pyx_n_s_raw, __pyx_n_s_key, __pyx_n_s_by_length, __pyx_n_s_length, __pyx_n_s_max_length, __pyx_n_s_emb_list, __pyx_n_s_cache, __pyx_n_s_i, __pyx_n_s_c); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(5, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_insert_clique_embedding, 287, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 287, __pyx_L1_error)
  __pyx_tuple__46 = PyTuple_Pack(3, ((PyObject *)Py_True), ((PyObject *)Py_None), ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "minorminer/busclique.pyx":341
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):             # <<<<<<<<<<<<<<
 *         """Merges the clique cache of another ``busgraph_cache`` object
 * 
 */
  __pyx_tuple__47 = PyTuple_Pack(23, __pyx_n_s_self, __pyx_n_s_cache, __pyx_n_s_quality_function, __pyx_n_s_write_to_disk, __pyx_n_s_embs, __pyx_n_s_raw0, __pyx_n_s_raw1, __pyx_n_s_used, __pyx_n_s_size, __pyx_n_s_size0, __pyx_n_s_size1, __pyx_n_s_i, __pyx_n_s_key0, __pyx_n_s_key1, __pyx_n_s_key, __pyx_n_s_qual0, __pyx_n_s_qual1, __pyx_n_s_length, __pyx_n_s_length0, __pyx_n_s_length1, __pyx_n_s_raw, __pyx_n_s_key, __pyx_n_s_i); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(4, 0, 23, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_merge_clique_cache, 341, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 341, __pyx_L1_error)
  __pyx_tuple__49 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "minorminer/busclique.pyx":421
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def _fetch_cache(self, dirname, compute, force_write=False):             # <<<<<<<<<<<<<<
 *         """This is an ad-hoc implementation of a file-cache using a LRU strategy.
 *         It's intended to be platform independent, thread- and multiprocess-safe,
 */
  __pyx_tuple__50 = PyTuple_Pack(20, __pyx_n_s_self, __pyx_n_s_dirname, __pyx_n_s_compute, __pyx_n_s_force_write, __pyx_n_s_rootdir, __pyx_n_s_basedir, __pyx_n_s_lockfile, __pyx_n_s_lrufile, __pyx_n_s_identifier, __pyx_n_s_shortcode, __pyx_n_s_file_context, __pyx_n_s_cachefile, __pyx_n_s_filecache, __pyx_n_s_currcache, __pyx_n_s_cache, __pyx_n_s_lru_2, __pyx_n_s_LRU, __pyx_n_s_newLRU, __pyx_n_s_x, __pyx_n_s_oldkey); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(4, 0, 20, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_fetch_cache, 421, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 421, __pyx_L1_error)
  __pyx_tuple__52 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "minorminer/busclique.pyx":501
 *         return cache
 * 
 *     def largest_clique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache.
 * 
 */
  __pyx_tuple__53 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_embs, __pyx_n_s_keys); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_largest_clique, 501, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 501, __pyx_L1_error)

  /* "minorminer/busclique.pyx":520
 *             return {}
 * 
 *     def largest_clique_by_chainlength(self, chainlength):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache, with a specified
 *         maximum ``chainlength``.
 */
  __pyx_tuple__55 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_chainlength, __pyx_n_s_embs, __pyx_n_s_length); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_largest_clique_by_chainlength, 520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 520, __pyx_L1_error)

  /* "minorminer/busclique.pyx":545
 *             return {}
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """Returns a clique embedding, minimizing the maximum chainlength given
 *         its size.
 */
  __pyx_tuple__57 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_nn, __pyx_n_s_num, __pyx_n_s_nodes, __pyx_n_s_keys, __pyx_n_s_raw, __pyx_n_s_key, __pyx_n_s_emb); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_find_clique_embedding, 545, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 545, __pyx_L1_error)

  /* "minorminer/busclique.pyx":572
 *         return self._graph.relabel(emb)
 * 
 *     def largest_balanced_biclique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-size biclique where both sides have equal size.
 * 
 */
  __pyx_tuple__59 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_biggest, __pyx_n_s_embs, __pyx_n_s_s0, __pyx_n_s_s1, __pyx_n_s_key, __pyx_n_s_raw_emb, __pyx_n_s_raw_emb0, __pyx_n_s_raw_emb1, __pyx_n_s_s_min, __pyx_n_s_emb0, __pyx_n_s_emb1); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(1, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_largest_balanced_biclique, 572, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 572, __pyx_L1_error)

  /* "minorminer/busclique.pyx":599
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))
 * 
 *     def find_biclique_embedding(self, nn, mm):             # <<<<<<<<<<<<<<
 *         """Returns a biclique embedding, minimizing the maximum chain length
 *         given its size.
 */
  __pyx_tuple__61 = PyTuple_Pack(21, __pyx_n_s_self, __pyx_n_s_nn, __pyx_n_s_mm, __pyx_n_s_biggest, __pyx_n_s_by_size, __pyx_n_s_raw, __pyx_n_s_n, __pyx_n_s_N, __pyx_n_s_m, __pyx_n_s_M, __pyx_n_s_emb, __pyx_n_s_nmax, __pyx_n_s_i, __pyx_n_s_mmax, __pyx_n_s_row, __pyx_n_s_j, __pyx_n_s_key, __pyx_n_s_s0, __pyx_n_s_s1, __pyx_n_s_emb0, __pyx_n_s_emb1); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(3, 0, 21, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_find_biclique_embedding, 599, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 599, __pyx_L1_error)

  /* "minorminer/busclique.pyx":658
 *         return {}
 * 
 *     def draw_fragment_embedding(self, emb, **kwargs):             # <<<<<<<<<<<<<<
 *         m, n, t, nodes, edges = self._graph.fragment_graph_spec()
 * 
 */
  __pyx_tuple__63 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_emb, __pyx_n_s_kwargs, __pyx_n_s_m, __pyx_n_s_n, __pyx_n_s_t, __pyx_n_s_nodes, __pyx_n_s_edges, __pyx_n_s_f, __pyx_n_s_f_emb, __pyx_n_s_k, __pyx_n_s_c); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_draw_fragment_embedding, 658, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 658, __pyx_L1_error)

  /* "minorminer/busclique.pyx":748
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}
 * 
 * def _trivial_relabeler(emb):             # <<<<<<<<<<<<<<
 *     """This doesn't relabel anything"""
 *     return {v: tuple(chain) for v, chain in emb.items()}
 */
  __pyx_tuple__65 = PyTuple_Pack(3, __pyx_n_s_emb, __pyx_n_s_v, __pyx_n_s_chain); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_trivial_relabeler, 748, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 748, __pyx_L1_error)

  /* "minorminer/busclique.pyx":752
 *     return {v: tuple(chain) for v, chain in emb.items()}
 * 
 * def _make_relabeler(f):             # <<<<<<<<<<<<<<
 *     """
 *     This returns an embedding-relabeling function, which applies `f` to each
 */
  __pyx_tuple__67 = PyTuple_Pack(3, __pyx_n_s_f, __pyx_n_s_relabeler, __pyx_n_s_relabeler); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_make_relabeler, 752, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 752, __pyx_L1_error)

  /* "minorminer/busclique.pyx":1135
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * def _default_quality_function(emb):             # <<<<<<<<<<<<<<
 *     """A function that returns a tuple corresponds to the "quality" of the embedding.
 * 
 */
  __pyx_tuple__69 = PyTuple_Pack(4, __pyx_n_s_emb, __pyx_n_s_maxlen, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_default_quality_function, 1135, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 1135, __pyx_L1_error)

  /* "minorminer/busclique.pyx":1155
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 * def _regularize_embedding(g, emb):             # <<<<<<<<<<<<<<
 *     """Best-effort augmented embedding, with hopefully uniform chainlegnths.
 * 
 */
  __pyx_tuple__71 = PyTuple_Pack(20, __pyx_n_s_g, __pyx_n_s_emb, __pyx_n_s_progress, __pyx_n_s_max_length, __pyx_n_s_remainder, __pyx_n_s_matching_graph, __pyx_n_s_top_nodes, __pyx_n_s_i, __pyx_n_s_chain, __pyx_n_s_boundary, __pyx_n_s_j, __pyx_n_s_chain_node, __pyx_n_s_matching, __pyx_n_s_matched, __pyx_n_s_i, __pyx_n_s_chain, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_i, __pyx_n_s_chain); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(2, 0, 20, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_regularize_embedding, 1155, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 1155, __pyx_L1_error)

  /* "minorminer/busclique.pyx":1199
 *     return {i: tuple(chain) for i, chain in emb.items()}
 * 
 * def mine_clique_embeddings(             # <<<<<<<<<<<<<<
 *         g,
 *         quality_function=None,
 */
  __pyx_tuple__73 = PyTuple_Pack(23, __pyx_n_s_g, __pyx_n_s_quality_function, __pyx_n_s_regularize, __pyx_n_s_num_seeds, __pyx_n_s_mask_bound, __pyx_n_s_heuristic_bound, __pyx_n_s_heuristic_args, __pyx_n_s_heuristic_runs, __pyx_n_s_miner, __pyx_n_s_logging, __pyx_n_s_logger, __pyx_n_s_bgc, __pyx_n_s_i, __pyx_n_s_seed, __pyx_n_s_bgc_i, __pyx_n_s_n, __pyx_n_s_emb, __pyx_n_s_tile, __pyx_n_s_rows, __pyx_n_s_cols, __pyx_n_s_size, __pyx_n_s_hm, __pyx_n_s_emb1); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(8, 0, 23, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_mine_clique_embeddings, 1199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 1199, __pyx_L1_error)

  /* "minorminer/busclique.pyx":1322
 * 
 * class copy_on_close_context:
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.files = {}
 * 
 */
  __pyx_tuple__75 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_init, 1322, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 1322, __pyx_L1_error)

  /* "minorminer/busclique.pyx":1325
 *         self.files = {}
 * 
 *     def open(self, file_name, mode):             # <<<<<<<<<<<<<<
 *         if file_name in self.files:
 *             self.files.clear()
 */
  __pyx_tuple__77 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_file_name, __pyx_n_s_mode, __pyx_n_s_temp_name, __pyx_n_s_file_obj); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__77, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_open, 1325, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 1325, __pyx_L1_error)

  /* "minorminer/busclique.pyx":1335
 *             return file_obj
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         for file_name, (temp_name, file_obj) in self.files.items():
 *             file_obj.close()
 */
  __pyx_tuple__79 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_file_name, __pyx_n_s_temp_name, __pyx_n_s_file_obj); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__79, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_minorminer_busclique_pyx, __pyx_n_s_close, 1335, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_setdefault.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_64 = PyInt_FromLong(64); if (unlikely(!__pyx_int_64)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967295 = PyInt_FromString((char *)"4294967295", 0, 0); if (unlikely(!__pyx_int_4294967295)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_10minorminer_9busclique___global_locks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique__zephyr_busgraph) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique__zephyr_busgraph.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique__zephyr_busgraph.tp_dictoffset && __pyx_type_10minorminer_9busclique__zephyr_busgraph.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique__zephyr_busgraph.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_zephyr_busgraph, (PyObject *)&__pyx_type_10minorminer_9busclique__zephyr_busgraph) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10minorminer_9busclique__zephyr_busgraph) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  __pyx_ptype_10minorminer_9busclique__zephyr_busgraph = &__pyx_type_10minorminer_9busclique__zephyr_busgraph;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique__pegasus_busgraph) < 0) __PYX_ERR(0, 904, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique__pegasus_busgraph.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique__pegasus_busgraph.tp_dictoffset && __pyx_type_10minorminer_9busclique__pegasus_busgraph.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique__pegasus_busgraph.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_pegasus_busgraph, (PyObject *)&__pyx_type_10minorminer_9busclique__pegasus_busgraph) < 0) __PYX_ERR(0, 904, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10minorminer_9busclique__pegasus_busgraph) < 0) __PYX_ERR(0, 904, __pyx_L1_error)
  __pyx_ptype_10minorminer_9busclique__pegasus_busgraph = &__pyx_type_10minorminer_9busclique__pegasus_busgraph;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique__chimera_busgraph) < 0) __PYX_ERR(0, 1017, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique__chimera_busgraph.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique__chimera_busgraph.tp_dictoffset && __pyx_type_10minorminer_9busclique__chimera_busgraph.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique__chimera_busgraph.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_chimera_busgraph, (PyObject *)&__pyx_type_10minorminer_9busclique__chimera_busgraph) < 0) __PYX_ERR(0, 1017, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10minorminer_9busclique__chimera_busgraph) < 0) __PYX_ERR(0, 1017, __pyx_L1_error)
  __pyx_ptype_10minorminer_9busclique__chimera_busgraph = &__pyx_type_10minorminer_9busclique__chimera_busgraph;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache) < 0) __PYX_ERR(0, 254, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct__update_clique_cache = &__pyx_type_10minorminer_9busclique___pyx_scope_struct__update_clique_cache;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding) < 0) __PYX_ERR(0, 287, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_1_insert_clique_embedding;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache) < 0) __PYX_ERR(0, 341, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_2_merge_clique_cache;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler) < 0) __PYX_ERR(0, 752, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_3__make_relabeler;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function) < 0) __PYX_ERR(0, 1135, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_4__default_quality_function;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(0, 1153, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_5_genexpr = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_5_genexpr;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding) < 0) __PYX_ERR(0, 1155, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_6__regularize_embedding;
  if (PyType_Ready(&__pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr) < 0) __PYX_ERR(0, 1187, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr.tp_dictoffset && __pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_10minorminer_9busclique___pyx_scope_struct_7_genexpr = &__pyx_type_10minorminer_9busclique___pyx_scope_struct_7_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initbusclique(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initbusclique(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_busclique(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_busclique(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_busclique(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'busclique' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_busclique(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("busclique", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_minorminer__busclique) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "minorminer.busclique")) {
      if (unlikely(PyDict_SetItemString(modules, "minorminer.busclique", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "minorminer/busclique.pyx":19
 * include "busclique_h.pxi"
 * 
 * import homebase             # <<<<<<<<<<<<<<
 * import os
 * import pathlib
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_homebase, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_homebase, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":20
 * 
 * import homebase
 * import os             # <<<<<<<<<<<<<<
 * import pathlib
 * import fasteners
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":21
 * import homebase
 * import os
 * import pathlib             # <<<<<<<<<<<<<<
 * import fasteners
 * import threading
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pathlib, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pathlib, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":22
 * import os
 * import pathlib
 * import fasteners             # <<<<<<<<<<<<<<
 * import threading
 * import random
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fasteners, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fasteners, __pyx_t_1) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":23
 * import pathlib
 * import fasteners
 * import threading             # <<<<<<<<<<<<<<
 * import random
 * import gzip
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_threading, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_threading, __pyx_t_1) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":24
 * import fasteners
 * import threading
 * import random             # <<<<<<<<<<<<<<
 * import gzip
 * from pickle import dump, load
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":25
 * import threading
 * import random
 * import gzip             # <<<<<<<<<<<<<<
 * from pickle import dump, load
 * from json import dumps, loads
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_gzip, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gzip, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":26
 * import random
 * import gzip
 * from pickle import dump, load             # <<<<<<<<<<<<<<
 * from json import dumps, loads
 * import networkx as nx
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_dump);
  __Pyx_GIVEREF(__pyx_n_s_dump);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_dump);
  __Pyx_INCREF(__pyx_n_s_load);
  __Pyx_GIVEREF(__pyx_n_s_load);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_load);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_dump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dump, __pyx_t_1) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_load); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_load, __pyx_t_1) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":27
 * import gzip
 * from pickle import dump, load
 * from json import dumps, loads             # <<<<<<<<<<<<<<
 * import networkx as nx
 * import dwave_networkx as dnx
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_dumps);
  __Pyx_GIVEREF(__pyx_n_s_dumps);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_dumps);
  __Pyx_INCREF(__pyx_n_s_loads);
  __Pyx_GIVEREF(__pyx_n_s_loads);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_loads);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_json, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_dumps); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dumps, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_loads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loads, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":28
 * from pickle import dump, load
 * from json import dumps, loads
 * import networkx as nx             # <<<<<<<<<<<<<<
 * import dwave_networkx as dnx
 * from itertools import zip_longest
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_networkx, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nx, __pyx_t_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":29
 * from json import dumps, loads
 * import networkx as nx
 * import dwave_networkx as dnx             # <<<<<<<<<<<<<<
 * from itertools import zip_longest
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_dwave_networkx, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dnx, __pyx_t_1) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":30
 * import networkx as nx
 * import dwave_networkx as dnx
 * from itertools import zip_longest             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_zip_longest);
  __Pyx_GIVEREF(__pyx_n_s_zip_longest);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_zip_longest);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_itertools, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_zip_longest); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_zip_longest, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":40
 * #when yield-improving changes are made to clique algorithms, or when bugs are
 * #fixed in the same.
 * cdef int __cache_version = 7             # <<<<<<<<<<<<<<
 * 
 * cdef int __lru_size = 100
 */
  __pyx_v_10minorminer_9busclique___cache_version = 7;

  /* "minorminer/busclique.pyx":42
 * cdef int __cache_version = 7
 * 
 * cdef int __lru_size = 100             # <<<<<<<<<<<<<<
 * cdef dict __global_locks = {'clique': threading.Lock(),
 *                             'biclique': threading.Lock()}
 */
  __pyx_v_10minorminer_9busclique___lru_size = 0x64;

  /* "minorminer/busclique.pyx":43
 * 
 * cdef int __lru_size = 100
 * cdef dict __global_locks = {'clique': threading.Lock(),             # <<<<<<<<<<<<<<
 *                             'biclique': threading.Lock()}
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_threading); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Lock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_clique, __pyx_t_1) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":44
 * cdef int __lru_size = 100
 * cdef dict __global_locks = {'clique': threading.Lock(),
 *                             'biclique': threading.Lock()}             # <<<<<<<<<<<<<<
 * 
 * def _num_nodes(nn, offset = 0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_threading); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Lock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_biclique, __pyx_t_1) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_10minorminer_9busclique___global_locks);
  __Pyx_DECREF_SET(__pyx_v_10minorminer_9busclique___global_locks, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":46
 *                             'biclique': threading.Lock()}
 * 
 * def _num_nodes(nn, offset = 0):             # <<<<<<<<<<<<<<
 *     """Internal-use function to normalize inputs.
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_1_num_nodes, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_num_nodes, __pyx_t_2) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":66
 *     return num, nodes
 * 
 * _no_seed = object()             # <<<<<<<<<<<<<<
 * def find_clique_embedding(nodes, g, seed = _no_seed, use_cache = True):
 *     """Finds a clique embedding in the graph ``g`` using a polynomial-time
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_no_seed, __pyx_t_2) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":67
 * 
 * _no_seed = object()
 * def find_clique_embedding(nodes, g, seed = _no_seed, use_cache = True):             # <<<<<<<<<<<<<<
 *     """Finds a clique embedding in the graph ``g`` using a polynomial-time
 *     algorithm.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_no_seed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_ = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_3find_clique_embedding, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_clique_embedding, __pyx_t_2) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":134
 *         return busgraph(g, seed=seed).find_clique_embedding(nodes)
 * 
 * class busgraph_cache:             # <<<<<<<<<<<<<<
 *     """A cache class for Chimera, Pegasus and Zephyr graphs, and their
 *     associated cliques and bicliques.
 */
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_busgraph_cache, __pyx_n_s_busgraph_cache, (PyObject *) NULL, __pyx_n_s_minorminer_busclique, __pyx_kp_s_A_cache_class_for_Chimera_Pegasu); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "minorminer/busclique.pyx":155
 * 
 *     """
 *     def __init__(self, g, seed = 0):             # <<<<<<<<<<<<<<
 *         self._family = g.graph['family']
 *         graphclass = {'pegasus': _pegasus_busgraph,
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_1__init__, 0, __pyx_n_s_busgraph_cache___init, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__28);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_1) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":169
 *         self._bicliques = None
 * 
 *     def _ensure_clique_cache(self, blank=False):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._cliques is None:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_3_ensure_clique_cache, 0, __pyx_n_s_busgraph_cache__ensure_clique_ca, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__31);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_ensure_clique_cache, __pyx_t_1) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":180
 *                 )
 * 
 *     def _ensure_biclique_cache(self):             # <<<<<<<<<<<<<<
 *         """Fetch/compute the clique cache, if it's not already in memory."""
 *         if self._bicliques is None:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_5_ensure_biclique_cache, 0, __pyx_n_s_busgraph_cache__ensure_biclique, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_ensure_biclique_cache, __pyx_t_1) < 0) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":186
 * 
 *     @staticmethod
 *     def cache_rootdir(version=__cache_version):             # <<<<<<<<<<<<<<
 *         """Returns the directory corresponding to the provided cache version.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_7cache_rootdir, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_busgraph_cache_cache_rootdir, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults), 1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_10minorminer_9busclique___cache_version); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_1)->__pyx_arg_version = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_10minorminer_9busclique_14__defaults__);

  /* "minorminer/busclique.pyx":185
 *             self._bicliques = self._fetch_cache('biclique', self._graph.bicliques)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def cache_rootdir(version=__cache_version):
 *         """Returns the directory corresponding to the provided cache version.
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_cache_rootdir, __pyx_t_3) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "minorminer/busclique.pyx":199
 * 
 *     @staticmethod
 *     def clear_all_caches():             # <<<<<<<<<<<<<<
 *         """Removes all caches created by this class, up to and including the
 *         current version.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_9clear_all_caches, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_busgraph_cache_clear_all_caches, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "minorminer/busclique.pyx":198
 *         return homebase.user_data_dir('busclique', 'dwave', version)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def clear_all_caches():
 *         """Removes all caches created by this class, up to and including the
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_clear_all_caches, __pyx_t_1) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":225
 *                 dirstack.pop()
 * 
 *     def get_clique_cache(self, sparse=False, compress=True):             # <<<<<<<<<<<<<<
 *         """Returns an object representing the cliques stored in this cache.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_11get_clique_cache, 0, __pyx_n_s_busgraph_cache_get_clique_cache, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__40);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_get_clique_cache, __pyx_t_1) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":254
 *         return cache
 * 
 *     def update_clique_cache(self, cache, write_to_disk = True):             # <<<<<<<<<<<<<<
 *         """Replaces or sets the cliques stored by this cache with the provided
 *         cache object.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_13update_clique_cache, 0, __pyx_n_s_busgraph_cache_update_clique_cac_2, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__43);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_update_clique_cache, __pyx_t_1) < 0) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":287
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def insert_clique_embedding(self, emb, write_to_disk=True, quality_function=None, direct=False):             # <<<<<<<<<<<<<<
 *         """Inserts a clique embedding into the cache.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_15insert_clique_embedding, 0, __pyx_n_s_busgraph_cache_insert_clique_emb_2, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__46);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_insert_clique_embedding, __pyx_t_1) < 0) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":341
 *             self.merge_clique_cache(cache, quality_function, write_to_disk)
 * 
 *     def merge_clique_cache(self, cache, quality_function=None, write_to_disk=True):             # <<<<<<<<<<<<<<
 *         """Merges the clique cache of another ``busgraph_cache`` object
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_17merge_clique_cache, 0, __pyx_n_s_busgraph_cache_merge_clique_cach_2, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__49);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_merge_clique_cache, __pyx_t_1) < 0) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":421
 *             self._fetch_cache('clique', lambda: self._cliques, force_write=True)
 * 
 *     def _fetch_cache(self, dirname, compute, force_write=False):             # <<<<<<<<<<<<<<
 *         """This is an ad-hoc implementation of a file-cache using a LRU strategy.
 *         It's intended to be platform independent, thread- and multiprocess-safe,
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_19_fetch_cache, 0, __pyx_n_s_busgraph_cache__fetch_cache, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__52);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_fetch_cache, __pyx_t_1) < 0) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":501
 *         return cache
 * 
 *     def largest_clique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_21largest_clique, 0, __pyx_n_s_busgraph_cache_largest_clique, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_largest_clique, __pyx_t_1) < 0) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":520
 *             return {}
 * 
 *     def largest_clique_by_chainlength(self, chainlength):             # <<<<<<<<<<<<<<
 *         """Returns the largest-found clique in the clique cache, with a specified
 *         maximum ``chainlength``.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_23largest_clique_by_chainlength, 0, __pyx_n_s_busgraph_cache_largest_clique_by, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_largest_clique_by_chainlength, __pyx_t_1) < 0) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":545
 *             return {}
 * 
 *     def find_clique_embedding(self, nn):             # <<<<<<<<<<<<<<
 *         """Returns a clique embedding, minimizing the maximum chainlength given
 *         its size.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_25find_clique_embedding, 0, __pyx_n_s_busgraph_cache_find_clique_embed, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_find_clique_embedding, __pyx_t_1) < 0) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":572
 *         return self._graph.relabel(emb)
 * 
 *     def largest_balanced_biclique(self):             # <<<<<<<<<<<<<<
 *         """Returns the largest-size biclique where both sides have equal size.
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_27largest_balanced_biclique, 0, __pyx_n_s_busgraph_cache_largest_balanced_2, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_largest_balanced_biclique, __pyx_t_1) < 0) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":599
 *         return self._graph.relabel(dict(enumerate(emb0 + emb1)))
 * 
 *     def find_biclique_embedding(self, nn, mm):             # <<<<<<<<<<<<<<
 *         """Returns a biclique embedding, minimizing the maximum chain length
 *         given its size.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_29find_biclique_embedding, 0, __pyx_n_s_busgraph_cache_find_biclique_emb, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_find_biclique_embedding, __pyx_t_1) < 0) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":658
 *         return {}
 * 
 *     def draw_fragment_embedding(self, emb, **kwargs):             # <<<<<<<<<<<<<<
 *         m, n, t, nodes, edges = self._graph.fragment_graph_spec()
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_14busgraph_cache_31draw_fragment_embedding, 0, __pyx_n_s_busgraph_cache_draw_fragment_emb, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_draw_fragment_embedding, __pyx_t_1) < 0) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":134
 *         return busgraph(g, seed=seed).find_clique_embedding(nodes)
 * 
 * class busgraph_cache:             # <<<<<<<<<<<<<<
 *     """A cache class for Chimera, Pegasus and Zephyr graphs, and their
 *     associated cliques and bicliques.
 */
  __pyx_t_1 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_busgraph_cache, __pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_busgraph_cache, __pyx_t_1) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":748
 *     return {'raw': raw, 'size': by_size, 'max_side': max_side}
 * 
 * def _trivial_relabeler(emb):             # <<<<<<<<<<<<<<
 *     """This doesn't relabel anything"""
 *     return {v: tuple(chain) for v, chain in emb.items()}
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_5_trivial_relabeler, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_trivial_relabeler, __pyx_t_2) < 0) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":752
 *     return {v: tuple(chain) for v, chain in emb.items()}
 * 
 * def _make_relabeler(f):             # <<<<<<<<<<<<<<
 *     """
 *     This returns an embedding-relabeling function, which applies `f` to each
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_7_make_relabeler, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_relabeler, __pyx_t_2) < 0) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1135
 *             return self.topo.topo.fragment_nodes(nodes)
 * 
 * def _default_quality_function(emb):             # <<<<<<<<<<<<<<
 *     """A function that returns a tuple corresponds to the "quality" of the embedding.
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_9_default_quality_function, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_default_quality_function, __pyx_t_2) < 0) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1155
 *     return -maxlen, sum(len(chain) == maxlen for chain in emb)
 * 
 * def _regularize_embedding(g, emb):             # <<<<<<<<<<<<<<
 *     """Best-effort augmented embedding, with hopefully uniform chainlegnths.
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_11_regularize_embedding, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_regularize_embedding, __pyx_t_2) < 0) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1199
 *     return {i: tuple(chain) for i, chain in emb.items()}
 * 
 * def mine_clique_embeddings(             # <<<<<<<<<<<<<<
 *         g,
 *         quality_function=None,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10minorminer_9busclique_13mine_clique_embeddings, NULL, __pyx_n_s_minorminer_busclique); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mine_clique_embeddings, __pyx_t_2) < 0) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1321
 *     return bgc
 * 
 * class copy_on_close_context:             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         self.files = {}
 */
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_copy_on_close_context, __pyx_n_s_copy_on_close_context, (PyObject *) NULL, __pyx_n_s_minorminer_busclique, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "minorminer/busclique.pyx":1322
 * 
 * class copy_on_close_context:
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.files = {}
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_21copy_on_close_context_1__init__, 0, __pyx_n_s_copy_on_close_context___init, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_1) < 0) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1325
 *         self.files = {}
 * 
 *     def open(self, file_name, mode):             # <<<<<<<<<<<<<<
 *         if file_name in self.files:
 *             self.files.clear()
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_21copy_on_close_context_3open, 0, __pyx_n_s_copy_on_close_context_open, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_open, __pyx_t_1) < 0) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1335
 *             return file_obj
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         for file_name, (temp_name, file_obj) in self.files.items():
 *             file_obj.close()
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10minorminer_9busclique_21copy_on_close_context_5close, 0, __pyx_n_s_copy_on_close_context_close, NULL, __pyx_n_s_minorminer_busclique, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_close, __pyx_t_1) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "minorminer/busclique.pyx":1321
 *     return bgc
 * 
 * class copy_on_close_context:             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         self.files = {}
 */
  __pyx_t_1 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_copy_on_close_context, __pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copy_on_close_context, __pyx_t_1) < 0) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "minorminer/busclique.pyx":1
 * # Copyright 2020 D-Wave Systems Inc.             # <<<<<<<<<<<<<<
 * #
 * #    Licensed under the Apache License, Version 2.0 (the "License");
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_uint8_t")
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init minorminer.busclique", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init minorminer.busclique");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       CYTHON_UNUSED int is_safe_type) {
    PyObject* value;
#if PY_VERSION_HEX >= 0x030400A0
    if ((1)) {
        value = PyDict_SetDefault(d, key, default_value);
        if (unlikely(!value)) return NULL;
        Py_INCREF(value);
#else
    if (is_safe_type == 1 || (is_safe_type == -1 &&
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
            (PyUnicode_CheckExact(key) || PyString_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItemWithError(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#else
            (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#endif
#endif
    } else {
        value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_setdefault, d, key, default_value);
    }
    return value;
}

/* PyIntCompare */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* None */
    static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* BytesEquals */
    static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
    static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
            if (unlikely(b == -1 && ((unsigned long)a) == 0-(unsigned long)a))
                return PyInt_Type.tp_as_number->nb_floor_divide(op1, op2);
            else {
                long q, r;
                q = a / b;
                r = a - q*b;
                q -= ((r != 0) & ((r ^ b) < 0));
                x = q;
            }
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_floor_divide(op1, op2);
            }
        }
                __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
                {
                    long q, r;
                    q = a / b;
                    r = a - q*b;
                    q -= ((r != 0) & ((r ^ b) < 0));
                    x = q;
                }
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                {
                    PY_LONG_LONG q, r;
                    q = lla / llb;
                    r = lla - q*llb;
                    q -= ((r != 0) & ((r ^ llb) < 0));
                    llx = q;
                }
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceFloorDivide : PyNumber_FloorDivide)(op1, op2);
}
#endif

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
            if (8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53))) {
                return PyFloat_FromDouble((double)a / (double)b);
            }
            return PyInt_Type.tp_as_number->nb_true_divide(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            }
        }
                __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
                if ((8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53)))
                        || __Pyx_sst_abs(size) <= 52 / PyLong_SHIFT) {
                    return PyFloat_FromDouble((double)a / (double)b);
                }
                return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            return PyLong_FromLong(x);
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            if (unlikely(zerodivision_check && b == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");
                return NULL;
            }
            PyFPE_START_PROTECT("divide", return NULL)
            result = ((double)a) / (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
}
#endif

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
    static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
    static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
    static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
    static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
    static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* SwapException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
    #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
    static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
    #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
    static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
    static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
